/*
  Copyright (C) 2009 - 2010 SWOAG Technology <www.swoag.com>

  This program is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free Software
  Foundation, version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  LIMOX entry point.

  REVISION:

    1) initial creation. ----------------------------------------- 2009-08-19 MG
*/

#include "limoa-log.h"
#include "limipc.h"
#include "limthread.h"
#include "limmemory.h"
#include "limqlist.h"
#include "OpenMAXAL.h"
#include "OMX_Core.h"
#include "ox-session.h"
#include "ox-main.h"

#undef LIMDBG
#ifndef LIMDBG
#define LIMDBG(fmt, arg...)
#endif

static int limox_loop = 1;


static void limox_log (
    const void * arg,
    int level,
    const void * log,
    int bytes)
{
    switch (level)
    {
    case LIM_LOG_FATAL :
    case LIM_LOG_ERROR :
    case LIM_LOG_WARNING :
        printf("%s", (char *)log);

    default :
        break;
    }
}

static void signal_handler (int signum)
{
    LIMDBG("signal [%d] caught, cleanup and exit.\n", signum);
    limox_loop = 0;
}

static inline int pre_setup (void)
{
    struct sigaction sa;

    lim_log_init(1024 * 1024 * 2, LIM_LOG_DEBUG, limox_log, NULL);

    if (OMX_ErrorNone != OMX_Init())
    {
        LIMERR("unable to initialize OMX_IL core.\n");
        return -1;
    }

    memset(&sa, 0, sizeof(struct sigaction));
    sa.sa_handler = signal_handler;
    sigaction(SIGINT, &sa, NULL);
    sigaction(SIGTERM, &sa, NULL);

    ox_omxil_debug();

    return 0;
}

static inline void post_cleanup (void)
{
    if (OMX_ErrorNone != OMX_Deinit())
    {
        LIMWARN("unable to unload OMX_IL core.\n");
    } else
    {
        LIMDBG("OMX IL core unloaded.\n");
    }
}

int main (int argc, char ** argv)
{
    struct sockaddr_un un;
    int sock_fd = -1;
    int max_fd;
    int client_fd;
    socklen_t len;
    fd_set rdfds, allfds;
    ox_session_t * ox;
    list_t * session_list;

    LIMDBG("limox startup.\n");

    // limox startup preprocessing
    if (pre_setup()) exit(-1);

    // limox session initialization, goes to listening mode.
    if (ox_session_init(NULL))
    {
        LIMERR("unable to initialize session.\n");
        goto bail1;
    }

    sock_fd = socket(AF_UNIX, SOCK_STREAM, 0);
    if (sock_fd <= 0)
    {
        LIMERR("unable to cretae limox server socket.\n");
        goto bail2;
    }

    memset(&un, 0, sizeof(un));
    un.sun_family = AF_UNIX;
    sprintf(un.sun_path, "%s", LIMOX_SOCKET_PATH);
    len = offsetof(struct sockaddr_un, sun_path) + strlen(un.sun_path);

    unlink(un.sun_path);
    if (bind(sock_fd, (struct sockaddr *)&un, len) < 0)
    {
        LIMERR("unable to bind limox socket to: %s\n", un.sun_path);
        goto bail2;
    }

    listen(sock_fd, 100);

    FD_ZERO(&allfds);
    FD_SET(sock_fd, &allfds);
    max_fd = sock_fd;

    while (limox_loop)
    {
        rdfds = allfds;
        if (select(max_fd + 1, &rdfds, NULL, NULL, NULL) < 0)
        {
            LIMWARN("unable to select fds.\n");
            continue;
        }

        if (FD_ISSET(sock_fd, &rdfds))
        {
            client_fd = accept(sock_fd, (struct sockaddr *)&un, &len);
            if (client_fd < 0)
            {
                LIMWARN("unable to accept client.\n");
            } else
            {
                FD_SET(client_fd, &allfds);
                if (client_fd > max_fd) max_fd = client_fd;
                ox_session_create(client_fd);
            }
            continue;
        }

        session_list = ox_session_list_head();
        while (session_list->next != ox_session_list_head())
        {
            ipc_packet_t pkt;
            int ret;

            ox = (ox_session_t *)session_list->next;
            session_list = session_list->next;

            if (!FD_ISSET(ox->ipc.sock_fd, &rdfds)) continue;

            lthread_mutex_lock(&ox->ipc.mutex);
            ret = ipc_read_packet(ox->ipc.sid, &pkt);
            if (0 > ret)
            {
                lthread_mutex_unlock(&ox->ipc.mutex);
                LIMERR("packet read error.\n");
                continue;
            }
            if (0 == ret)
            {
                lthread_mutex_unlock(&ox->ipc.mutex);
                LIMDBG("client socket closed.\n");
                FD_CLR(ox->ipc.sock_fd, &allfds);
                ipc_release_packet(ox->ipc.sid, &pkt);
                session_list = ox_session_list_head();
                ox_session_destroy(ox);

#if 1 //REMOVE: debug code to automatically shut limox down.
                limox_loop = 0;
                break;
#endif
                continue;
            }
            ox_session_parse_packet(ox, &pkt, NULL);
            lthread_mutex_unlock(&ox->ipc.mutex);

            ipc_release_packet(ox->ipc.sid, &pkt);
        }
    }

 bail2:
    if (sock_fd > 0) close(sock_fd);
 bail1:
    post_cleanup();
    LIMDBG("limox shuts down.\n");
    return 0;
}
