/*
  Copyright (C) 2009 - 2010 SWOAG Technology <www.swoag.com>

  This program is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free Software
  Foundation, version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  LIMOX output mix object.

  REVISION:

    1) initial creation. ----------------------------------------- 2009-09-20 MG
*/

#include "limoa-log.h"
#include "output-mix-be.h"

#undef LIMDBG
#ifndef LIMDBG
#define LIMDBG(fmt, arg...)
#endif

typedef struct
{
    list_t list;
    OMX_CALLBACKTYPE cb;
    OMX_PTR priv_data;
} ox_omix_cb_internal_t;

OMX_ERRORTYPE sink_event_handler (
    OMX_HANDLETYPE hComponent,
    OMX_PTR pAppData,
    OMX_EVENTTYPE eEvent,
    OMX_U32 Data1,
    OMX_U32 Data2,
    OMX_PTR pEventData)
{
    ox_omix_t * omix = (ox_omix_t *)pAppData;
    ox_omix_cb_internal_t * cb_i;
    list_t * list;

    LIMDBG("audio sink event handler.\n");

    lthread_mutex_lock(&omix->cb_mutex);
    list = omix->cb_list_head.next;
    while (list != &omix->cb_list_head)
    {
        cb_i = (ox_omix_cb_internal_t *)list;
        cb_i->cb.EventHandler(hComponent, cb_i->priv_data,
                              eEvent, Data1, Data2, pEventData);
        list = list->next;
    }
    lthread_mutex_unlock(&omix->cb_mutex);

    return OMX_ErrorNone;
}

OMX_ERRORTYPE sink_empty_buffer_done (
    OMX_HANDLETYPE hComponent,
    OMX_PTR pAppData,
    OMX_BUFFERHEADERTYPE* pBuffer)
{
    ox_omix_t * omix = (ox_omix_t *)pAppData;
    ox_omix_cb_internal_t * cb_i;
    list_t * list;

    LIMDBG("audio sink empty buffer done.\n");

    lthread_mutex_lock(&omix->cb_mutex);
    list = omix->cb_list_head.next;
    while (list != &omix->cb_list_head)
    {
        cb_i = (ox_omix_cb_internal_t *)list;
        cb_i->cb.EmptyBufferDone(hComponent, cb_i->priv_data, pBuffer);
        list = list->next;
    }
    lthread_mutex_unlock(&omix->cb_mutex);

    return OMX_ErrorNone;
}

OMX_CALLBACKTYPE sink_callbacks = {
    .EventHandler    = sink_event_handler,
    .EmptyBufferDone = sink_empty_buffer_done,
    .FillBufferDone  = NULL
};

static void release_callbacks (list_t * cb_list_head)
{
    list_t * list = cb_list_head->next;
    ox_omix_cb_internal_t * cb_i;

    while (list != cb_list_head)
    {
        cb_i = (ox_omix_cb_internal_t *)list;
        list = list->next;

        list_del(&cb_i->list);
        free(cb_i);
    }
}

XAresult ox_omix_realize (ox_omix_t * omix)
{
    XAresult res;

    if (OMX_ErrorNone !=
        OMX_GetHandle(&omix->sink_handle, "OMX.limoi.alsa_sink",
                      omix, &sink_callbacks))
    {
        res = XA_RESULT_RESOURCE_ERROR;
        LIMERR("unable to get audio sink handle.\n");
        goto bail;
    }

    list_init(&omix->cb_list_head);
    lthread_mutex_init(&omix->cb_mutex);
    lthread_mutex_init(&omix->mutex);
    lthread_cond_init(&omix->cond);

    LIMDBG("output mix realized.\n");
    res = XA_RESULT_SUCCESS;
 bail:
    return res;
}

int ox_omix_destroy (ox_omix_t * omix)
{
    OMX_FreeHandle(omix->sink_handle);

    lthread_mutex_lock(&omix->cb_mutex);
    release_callbacks(&omix->cb_list_head);
    lthread_mutex_unlock(&omix->cb_mutex);

    lthread_mutex_destroy(&omix->cb_mutex);
    lthread_mutex_destroy(&omix->mutex);
    lthread_cond_destroy(&omix->cond);

    return 0;
}

int ox_omix_register_callback (
    ox_omix_t * omix,
    ox_omix_cb_t * omix_cb)
{
    ox_omix_cb_internal_t * cb_i;

    cb_i = malloc(sizeof(ox_omix_cb_internal_t));
    if (NULL == cb_i)
    {
        LIMERR("out of memory.\n");
        return -1;
    }

    list_init(&cb_i->list);
    cb_i->cb = omix_cb->cb;
    cb_i->priv_data = omix_cb->priv_data;

    lthread_mutex_lock(&omix->cb_mutex);
    list_add_tail(&cb_i->list, &omix->cb_list_head);
    lthread_mutex_unlock(&omix->cb_mutex);

    return 0;
}

int ox_omix_unregister_callback (
    ox_omix_t * omix,
    ox_omix_cb_t * omix_cb)
{
    ox_omix_cb_internal_t * cb_i;
    list_t * list;

    lthread_mutex_lock(&omix->cb_mutex);
    list = omix->cb_list_head.next;

    while (list != &omix->cb_list_head)
    {
        cb_i = (ox_omix_cb_internal_t *)list;

        if (cb_i->cb.EventHandler == omix_cb->cb.EventHandler
            && cb_i->cb.EmptyBufferDone == omix_cb->cb.EmptyBufferDone
            && cb_i->cb.FillBufferDone == omix_cb->cb.FillBufferDone
            && cb_i->priv_data == omix_cb->priv_data)
        {
            list_del(&cb_i->list);
            free(cb_i);
            goto bail;
        } else list = list->next;
    }

    LIMDBG("unable to find callback.\n");

 bail:
    lthread_mutex_unlock(&omix->cb_mutex);
    return 0;
}
