/*
  Copyright (C) 2009 - 2010 SWOAG Technology <www.swoag.com>

  This program is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free Software
  Foundation, version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  LIMOX player object.

  REVISION:

    1) initial creation. ----------------------------------------- 2009-09-19 MG
*/

#include "limoa-log.h"
#include "player-be.h"
#include "common-be.h"
#include "output-mix-be.h"

//#undef LIMDBG
#ifndef LIMDBG
#define LIMDBG(fmt, arg...)
#endif

//TODO: enable this to synchronize the startup of audio and video,
//      both path shall be monitored and gracefully timeout if things
//      go wrong, so that either audio or video can play by itself.
//      until then, audio/video starts up on itself.
//#define AUDIO_VIDEO_STARTUP_SYNC

#define set_flags(flags) do {                 \
        lthread_mutex_lock(&player->mutex);   \
        flags;                                \
        lthread_mutex_unlock(&player->mutex); \
    } while(0)
#define broadcast_set_flags(flags) \
    lthread_event_broadcast(&player->cond, &player->mutex, flags)
#define wait_for_flags(flags, ret) \
    lthread_event_timedwait(&player->cond, &player->mutex, 5000, flags, ret)


//OpenMAX codec ID and role name mapping structure.
typedef struct
{
    int codec_id;
    const char * role_name;
} omx_codec_role_map_t;

//available demuxer roles.
static char * omx_demux_role_map[] =
{
    "container_demuxer.all",
    NULL
};

//audio decoder roles map.
static omx_codec_role_map_t omx_audio_decode_role_map[] =
{
    {OMX_AUDIO_CodingMP3, "audio_decoder.mp3"},
    {OMX_AUDIO_CodingAAC, "audio_decoder.aac"},
    {OMX_AUDIO_CodingAMR, "audio_decoder.amr"},
    {OMX_AUDIO_CodingVORBIS, "audio_decoder.ogg"},
    {OMX_AUDIO_CodingWMA, "audio_decoder.wma"},
    {OMX_AUDIO_CodingRA, "audio_decoder.ra"},
    {OMX_AUDIO_CodingADPCM, "audio_decoder.adpcm"},
    {OMX_AUDIO_CodingPCM, "audio_decoder.pcm"},
    {OMX_AUDIO_CodingMax, NULL}
};

//video decoder roles map.
static omx_codec_role_map_t omx_video_decode_role_map[] =
{
    {OMX_VIDEO_CodingMPEG2, "video_decoder.mpeg2"},
    {OMX_VIDEO_CodingH263, "video_decoder.h263"},
    {OMX_VIDEO_CodingMPEG4, "video_decoder.mpeg4"},
    {OMX_VIDEO_CodingWMV, "video_decoder.wmv"},
    {OMX_VIDEO_CodingRV, "video_decoder.rv"},
    {OMX_VIDEO_CodingAVC, "video_decoder.avc"},
    {OMX_VIDEO_CodingMJPEG, "video_decoder.mjpeg"},
    {OMX_VIDEO_CodingMax, NULL}
};

#if 0
static inline void print_player_flags (
    const char * caller,
    ox_player_t * player)
{
    LIMINFO("%s: player flags\n"
            "\tclock           | port : %08x | %08x\n"
            "\tdemux           | port : %08x | %08x\n"
            "\taudio decode    | port : %08x | %08x\n"
            "\taudio sink      | port : %08x | %08x\n"
            "\tvideo decode    | port : %08x | %08x\n"
            "\tvideo scheduler | port : %08x | %08x\n"
            "\tvideo sink      | port : %08x | %08x\n",
            caller,
            player->clock_flags, player->clock_port_flags,
            player->demux_flags, player->demux_port_flags,
            player->audio_decode_flags, player->audio_decode_port_flags,
            player->audio_sink_flags, player->audio_sink_port_flags,
            player->video_decode_flags, player->video_decode_port_flags,
            player->video_scheduler_flags, player->video_scheduler_port_flags,
            player->video_sink_flags, player->video_sink_port_flags);
}
#else
#define print_player_flags(c, p)
#endif

static OMX_STRING get_role_by_codec_id (
    omx_codec_role_map_t * map,
    int id)
{
    OMX_STRING role = NULL;

    while (map->codec_id != id && map->codec_id != OMX_AUDIO_CodingMax) map++;
    if (map->codec_id == id) role = (OMX_STRING)map->role_name;
    return role;
}

static OMX_ERRORTYPE audio_dec_event_handler (
    OMX_HANDLETYPE hComponent,
    OMX_PTR pAppData,
    OMX_EVENTTYPE eEvent,
    OMX_U32 Data1,
    OMX_U32 Data2,
    OMX_PTR pEventData)
{
    ox_player_t * player = (ox_player_t *)pAppData;

    if(eEvent == OMX_EventCmdComplete)
    {
        if (Data1 == OMX_CommandStateSet)
        {
            LIMDBG("audio decoder state changed to: %s\n",
                  omxil_state_name((int)Data2));

           switch (Data2)
            {
            case OMX_StateIdle :
                broadcast_set_flags(player->audio_decode_flags
                                    |= PSF_IDLE);
                break;

            case OMX_StateExecuting :
                broadcast_set_flags(player->audio_decode_flags
                                    |= PSF_EXECUTING);
                break;

            case OMX_StatePause :
                broadcast_set_flags(player->audio_decode_flags
                                    |= PSF_PAUSE);
                break;

            case OMX_StateLoaded :
                broadcast_set_flags(player->audio_decode_flags
                                    |= PSF_LOADED);
                break;

            default :
                LIMWARN("unhandled cmd event: param1 = %d param2 = %d\n",
                       (int)Data1, (int)Data2);
                break;
            }
        } else if (Data1 == OMX_CommandPortEnable)
        {
            LIMDBG("audio decoder port %u enabled.\n", Data2);
        } else if (Data1 == OMX_CommandPortDisable)
        {
            LIMDBG("audio decoder port %u disabled.\n", Data2);
            broadcast_set_flags(player->audio_decode_port_flags
                                |= PSF_PORT_DISABLED(Data2));
        } else if (Data1 == OMX_CommandFlush)
        {
            LIMDBG("audio decoder port %u flushed.\n", Data2);
            broadcast_set_flags(player->audio_decode_port_flags
                                |= PSF_PORT_FLUSHED(Data2));
        }
    } else if(eEvent == OMX_EventPortSettingsChanged)
    {
        LIMDBG("audio decoder port settings changed: %d %d.\n",
              (int)Data1, (int)Data2);
        if (Data1 == 1)
        {
            set_flags(player->audio_decode_port_flags &=
                      ~PSF_PORT_DISABLED(1);
                      player->audio_sink_port_flags &=
                      ~PSF_PORT_DISABLED(0);
                      player->sync_flags |=
                      PSR_AUDIO_DECODER_OUTPUT_PORTSETTINGSCHANGED);

            OMX_SendCommand(player->audio_sink_handle,
                            OMX_CommandPortDisable, 0, NULL);
            OMX_SendCommand(player->audio_decode_handle,
                            OMX_CommandPortDisable, 1, NULL);
        }
    } else if(eEvent == OMX_EventBufferFlag)
    {
        LIMDBG("audio decoder buffer flag event.\n");
    } else if (eEvent == OMX_EventError)
    {
        LIMDBG("error event: %x\n", (unsigned int)Data1);
        broadcast_set_flags(player->audio_decode_flags
                            |= PSF_INVALID);
    } else
    {
        LIMWARN("unhandled event: param1 = %d param2 = %d\n",
               (int)Data1, (int)Data2);
    }
    return OMX_ErrorNone;
}

OMX_CALLBACKTYPE audio_decode_callbacks = {
    .EventHandler    = audio_dec_event_handler,
    .EmptyBufferDone = NULL,
    .FillBufferDone  = NULL
};


static OMX_ERRORTYPE video_dec_event_handler (
    OMX_HANDLETYPE hComponent,
    OMX_PTR pAppData,
    OMX_EVENTTYPE eEvent,
    OMX_U32 Data1,
    OMX_U32 Data2,
    OMX_PTR pEventData)
{
    ox_player_t * player = (ox_player_t *)pAppData;

    if(eEvent == OMX_EventCmdComplete)
    {
        if (Data1 == OMX_CommandStateSet)
        {
            LIMDBG("video decoder state changed to: %s\n",
                  omxil_state_name((int)Data2));

           switch (Data2)
            {
            case OMX_StateIdle :
                broadcast_set_flags(player->video_decode_flags
                                    |= PSF_IDLE);
                break;

            case OMX_StateExecuting :
                broadcast_set_flags(player->video_decode_flags
                                    |= PSF_EXECUTING);
                break;

            case OMX_StatePause :
                broadcast_set_flags(player->video_decode_flags
                                    |= PSF_PAUSE);
                break;

            case OMX_StateLoaded :
                broadcast_set_flags(player->video_decode_flags
                                    |= PSF_LOADED);
                break;

            default :
                LIMWARN("unhandled cmd event: param1 = %d param2 = %d\n",
                       (int)Data1, (int)Data2);
                break;
            }
        } else if (Data1 == OMX_CommandPortEnable)
        {
            LIMDBG("video decoder port %u enabled.\n", Data2);
        } else if (Data1 == OMX_CommandPortDisable)
        {
            LIMDBG("video decoder port %u disabled.\n", Data2);
            broadcast_set_flags(player->video_decode_port_flags
                                |= PSF_PORT_DISABLED(Data2));
        } else if (Data1 == OMX_CommandFlush)
        {
            LIMDBG("video decoder port %u flushed.\n", Data2);
            broadcast_set_flags(player->video_decode_port_flags
                                |= PSF_PORT_FLUSHED(Data2));
        }
    } else if(eEvent == OMX_EventPortSettingsChanged)
    {
        LIMDBG("video decoder port settings changed: %d %d.\n",
              (int)Data1, (int)Data2);
        if (Data1 == 1)
        {
            set_flags(player->video_decode_port_flags &=
                      ~PSF_PORT_DISABLED(1);
                      player->video_scheduler_port_flags &=
                      ~PSF_PORT_DISABLED(0);
                      player->sync_flags |=
                      PSR_VIDEO_DECODER_OUTPUT_PORTSETTINGSCHANGED);

            //NOTE: disable down link component ports, just in case that
            //component implements buffer sharing, where a tight depency
            //would be formed across all tunneled ports.
            //this is where IL spec needs to be revised to present
            //one way for client to know the existance of this tight
            //dependency across ports.
            OMX_SendCommand(player->video_scheduler_handle,
                            OMX_CommandPortDisable, 1, NULL);
            OMX_SendCommand(player->video_sink_handle,
                            OMX_CommandPortDisable, 0, NULL);

            OMX_SendCommand(player->video_scheduler_handle,
                            OMX_CommandPortDisable, 0, NULL);
            OMX_SendCommand(player->video_decode_handle,
                            OMX_CommandPortDisable, 1, NULL);
        }
    } else if(eEvent == OMX_EventBufferFlag)
    {
        LIMDBG("video decoder buffer flag event.\n");
    } else if (eEvent == OMX_EventError)
    {
        LIMDBG("error event: %x\n", (unsigned int)Data1);
        broadcast_set_flags(player->video_decode_flags
                            |= PSF_INVALID);
    } else
    {
        LIMWARN("unhandled event: param1 = %d param2 = %d\n",
               (int)Data1, (int)Data2);
    }
    return OMX_ErrorNone;
}

OMX_CALLBACKTYPE video_decode_callbacks = {
    .EventHandler    = video_dec_event_handler,
    .EmptyBufferDone = NULL,
    .FillBufferDone  = NULL
};

static OMX_ERRORTYPE video_sch_event_handler (
    OMX_HANDLETYPE hComponent,
    OMX_PTR pAppData,
    OMX_EVENTTYPE eEvent,
    OMX_U32 Data1,
    OMX_U32 Data2,
    OMX_PTR pEventData)
{
    ox_player_t * player = (ox_player_t *)pAppData;

    if(eEvent == OMX_EventCmdComplete)
    {
        if (Data1 == OMX_CommandStateSet)
        {
            LIMDBG("video scheduler state changed to: %s\n",
                  omxil_state_name((int)Data2));

           switch (Data2)
            {
            case OMX_StateIdle :
                broadcast_set_flags(player->video_scheduler_flags
                                    |= PSF_IDLE);
                break;

            case OMX_StateExecuting :
                broadcast_set_flags(player->video_scheduler_flags
                                    |= PSF_EXECUTING);
                break;

            case OMX_StatePause :
                broadcast_set_flags(player->video_scheduler_flags
                                    |= PSF_PAUSE);
                break;

            case OMX_StateLoaded :
                broadcast_set_flags(player->video_scheduler_flags
                                    |= PSF_LOADED);
                break;

            default :
                LIMWARN("unhandled cmd event: param1 = %d param2 = %d\n",
                       (int)Data1, (int)Data2);
                break;
            }
        } else if (Data1 == OMX_CommandPortEnable)
        {
            LIMDBG("video scheduler port %u enabled.\n", Data2);
            broadcast_set_flags(player->video_scheduler_port_flags
                                |= PSF_PORT_ENABLED(Data2));
        } else if (Data1 == OMX_CommandPortDisable)
        {
            LIMDBG("video scheduler port %u disabled.\n", Data2);
            broadcast_set_flags(player->video_scheduler_port_flags
                                |= PSF_PORT_DISABLED(Data2));
        } else if (Data1 == OMX_CommandFlush)
        {
            LIMDBG("video scheduler port %u flushed.\n", Data2);
            broadcast_set_flags(player->video_scheduler_port_flags
                                |= PSF_PORT_FLUSHED(Data2));
        }
    } else if(eEvent == OMX_EventPortSettingsChanged)
    {
        LIMDBG("video scheduler port settings changed: %d %d.\n",
              (int)Data1, (int)Data2);
    } else if(eEvent == OMX_EventBufferFlag)
    {
        LIMDBG("video scheduler buffer flag event.\n");
    } else if (eEvent == OMX_EventError)
    {
        LIMDBG("error event: %x\n", (unsigned int)Data1);
        broadcast_set_flags(player->video_scheduler_flags
                            |= PSF_INVALID);
    } else
    {
        LIMWARN("unhandled event: param1 = %d param2 = %d\n",
               (int)Data1, (int)Data2);
    }
    return OMX_ErrorNone;
}

OMX_CALLBACKTYPE video_scheduler_callbacks = {
    .EventHandler    = video_sch_event_handler,
    .EmptyBufferDone = NULL,
    .FillBufferDone  = NULL
};

static OMX_ERRORTYPE video_sink_event_handler (
    OMX_HANDLETYPE hComponent,
    OMX_PTR pAppData,
    OMX_EVENTTYPE eEvent,
    OMX_U32 Data1,
    OMX_U32 Data2,
    OMX_PTR pEventData)
{
    ox_player_t * player = (ox_player_t *)pAppData;

    if(eEvent == OMX_EventCmdComplete)
    {
        if (Data1 == OMX_CommandStateSet)
        {
            LIMDBG("video sink state changed to: %s\n",
                  omxil_state_name((int)Data2));

           switch (Data2)
            {
            case OMX_StateIdle :
                broadcast_set_flags(player->video_sink_flags
                                    |= PSF_IDLE);
                break;

            case OMX_StateExecuting :
                broadcast_set_flags(player->video_sink_flags
                                    |= PSF_EXECUTING);
                break;

            case OMX_StatePause :
                broadcast_set_flags(player->video_sink_flags
                                    |= PSF_PAUSE);
                break;

            case OMX_StateLoaded :
                broadcast_set_flags(player->video_sink_flags
                                    |= PSF_LOADED);
                break;

            default :
                LIMWARN("unhandled cmd event: param1 = %d param2 = %d\n",
                       (int)Data1, (int)Data2);
                break;
            }
        } else if (Data1 == OMX_CommandPortEnable)
        {
            LIMDBG("video sink port %u enabled.\n", Data2);
            broadcast_set_flags(player->video_sink_port_flags
                                |= PSF_PORT_ENABLED(Data2));
        } else if (Data1 == OMX_CommandPortDisable)
        {
            LIMDBG("video sink port %d enabled.\n", Data2);
            broadcast_set_flags(player->video_sink_port_flags
                                |= PSF_PORT_DISABLED(Data2));
        } else if (Data1 == OMX_CommandFlush)
        {
            LIMDBG("video sink port %u flushed.\n", Data2);
            broadcast_set_flags(player->video_sink_port_flags
                                |= PSF_PORT_FLUSHED(Data2));
        }
    } else if(eEvent == OMX_EventPortSettingsChanged)
    {
        LIMDBG("video sink port settings changed: %d %d.\n",
              (int)Data1, (int)Data2);
    } else if(eEvent == OMX_EventBufferFlag)
    {
        LIMDBG("video sink buffer flag event.\n");
        broadcast_set_flags(player->video_sink_flags
                            |=  PSF_EOS_DETECTED);
    } else if (eEvent == OMX_EventError)
    {
        LIMDBG("error event: %x\n", (unsigned int)Data1);
        broadcast_set_flags(player->video_sink_flags
                            |= PSF_INVALID);
    } else
    {
        LIMWARN("unhandled event: param1 = %d param2 = %d\n",
               (int)Data1, (int)Data2);
    }
    return OMX_ErrorNone;
}

OMX_CALLBACKTYPE video_sink_callbacks = {
    .EventHandler    = video_sink_event_handler,
    .EmptyBufferDone = NULL,
    .FillBufferDone  = NULL
};

static OMX_ERRORTYPE demux_event_handler (
    OMX_HANDLETYPE hComponent,
    OMX_PTR pAppData,
    OMX_EVENTTYPE eEvent,
    OMX_U32 Data1,
    OMX_U32 Data2,
    OMX_PTR pEventData)
{
    ox_player_t * player = (ox_player_t *)pAppData;

    if(eEvent == OMX_EventCmdComplete)
    {
        if (Data1 == OMX_CommandStateSet)
        {
            LIMDBG("demuxer state changed to: %s\n",
                  omxil_state_name((int)Data2));

            switch (Data2)
            {
            case OMX_StateIdle :
                broadcast_set_flags(player->demux_flags |= PSF_IDLE);
                break;

            case OMX_StateExecuting :
                broadcast_set_flags(player->demux_flags |= PSF_EXECUTING);
                break;

            case OMX_StatePause :
                broadcast_set_flags(player->demux_flags |= PSF_PAUSE);
                break;

            case OMX_StateLoaded :
                broadcast_set_flags(player->demux_flags |= PSF_LOADED);
                break;

            default :
                LIMWARN("unhandled cmd event: param1 = %d param2 = %d\n",
                       (int)Data1, (int)Data2);
                break;
            }
        } else if (Data1 == OMX_CommandPortEnable)
        {
            LIMDBG("demuxer port %u enabled.\n", Data2);
            broadcast_set_flags(player->demux_port_flags
                                |= PSF_PORT_ENABLED(Data2));
        } else if (Data1 == OMX_CommandPortDisable)
        {
            LIMDBG("demuxer port %u disabled.\n", Data2);
            broadcast_set_flags(player->demux_port_flags
                                |= PSF_PORT_DISABLED(Data2));
        } else if (Data1 == OMX_CommandFlush)
        {
            LIMDBG("demuxer port %u flushed.\n", Data2);
            broadcast_set_flags(player->demux_port_flags
                                |= PSF_PORT_FLUSHED(Data2));
        }
    } else if (eEvent == OMX_EventPortFormatDetected)
    {
        LIMDBG("demuxer port %u format detected.\n", Data1);
    } else if (eEvent == OMX_EventPortSettingsChanged)
    {
        LIMDBG("demuxer port %u settings changed.\n", Data1);
        OMX_SendCommand(hComponent,
                        OMX_CommandPortDisable, Data1, NULL);
        broadcast_set_flags(player->demux_port_flags
                            |= PSF_PORT_SETTINGS_CHANGED(Data1));
    } else if (eEvent == OMX_EventBufferFlag)
    {
        LIMDBG("\ndemuxer EOS detected\n");
        if (!((!player->video_sink_handle
               || player->video_sink_flags & PSF_EOS_DETECTED)
              && (!player->audio_sink_handle
                  || player->audio_sink_flags & PSF_EOS_DETECTED)))
        {
            set_flags(player->sync_flags |= PSR_DEMUXER_EOS_DETECTED);
        }
    } else if (eEvent == OMX_EventError)
    {
        broadcast_set_flags(player->demux_flags |= PSF_INVALID);
    } else
    {
        LIMWARN("unhandled event: param1 = %d param2 = %d\n",
               (int)Data1, (int)Data2);
    }

    return OMX_ErrorNone;
}

static OMX_ERRORTYPE demux_fill_buffer_done (
    OMX_HANDLETYPE hComponent,
    OMX_PTR pAppData,
    OMX_BUFFERHEADERTYPE* pBuffer)
{
    //FIXME: assuming this is called upon port disabling, since normally
    //       port is in tuneling mode.
    LIMDBG("demuxer free port %u buffer %p.\n",
          pBuffer->nOutputPortIndex, pBuffer);
    OMX_FreeBuffer(hComponent, pBuffer->nOutputPortIndex, pBuffer);
    return OMX_ErrorNone;
}

static OMX_CALLBACKTYPE demux_callbacks = {
    .EventHandler    = demux_event_handler,
    .EmptyBufferDone  = NULL,
    .FillBufferDone  = demux_fill_buffer_done
};

static OMX_ERRORTYPE clock_event_handler (
    OMX_HANDLETYPE hComponent,
    OMX_PTR pAppData,
    OMX_EVENTTYPE eEvent,
    OMX_U32 Data1,
    OMX_U32 Data2,
    OMX_PTR pEventData)
{
    ox_player_t * player = (ox_player_t *)pAppData;

    if(eEvent == OMX_EventCmdComplete)
    {
        if (Data1 == OMX_CommandStateSet)
        {
            LIMDBG("clock state changed to: %s\n", omxil_state_name((int)Data2));

            switch (Data2)
            {
            case OMX_StateIdle :
                broadcast_set_flags(player->clock_flags |= PSF_IDLE);
                break;

            case OMX_StateExecuting :
                broadcast_set_flags(player->clock_flags |= PSF_EXECUTING);
                break;

            case OMX_StateLoaded :
                broadcast_set_flags(player->clock_flags |= PSF_LOADED);
                break;

            default :
                LIMWARN("unhandled cmd event: param1 = %d param2 = %d\n",
                       (int)Data1, (int)Data2);
                break;
            }
        } else if (Data1 == OMX_CommandPortEnable)
        {
            LIMDBG("clock port %u enabled.\n", Data2);
            broadcast_set_flags(player->clock_port_flags
                                |= PSF_PORT_ENABLED(Data2));
        } else if (Data1 == OMX_CommandPortDisable)
        {
            LIMDBG("clock port %u disabled.\n", Data2);
            broadcast_set_flags(player->clock_port_flags
                                |= PSF_PORT_DISABLED(Data2));
        }
    } else if (eEvent == OMX_EventPortFormatDetected)
    {
        LIMDBG("clock port %u format detected.\n", Data1);
    } else if (eEvent == OMX_EventPortSettingsChanged)
    {
        LIMDBG("clock port %u settings changed.\n", Data1);
        OMX_SendCommand(hComponent,
                        OMX_CommandPortDisable, Data1, NULL);
        broadcast_set_flags(player->clock_port_flags
                            |= PSF_PORT_SETTINGS_CHANGED(Data1));
    } else if (eEvent == OMX_EventBufferFlag)
    {
        LIMDBG("\nEOS detected\n");
    } else if (eEvent == OMX_EventError)
    {
        broadcast_set_flags(player->clock_flags |= PSF_INVALID);
    } else
    {
        LIMWARN("unhandled event: param1 = %d param2 = %d\n",
               (int)Data1, (int)Data2);
    }

    return OMX_ErrorNone;
}

static OMX_ERRORTYPE clock_fill_buffer_done (
    OMX_HANDLETYPE hComponent,
    OMX_PTR pAppData,
    OMX_BUFFERHEADERTYPE* pBuffer)
{
    LIMDBG("clock free port %u buffer.\n", pBuffer->nOutputPortIndex);
    OMX_FreeBuffer(hComponent, pBuffer->nOutputPortIndex, pBuffer);
    return OMX_ErrorNone;
}

static OMX_CALLBACKTYPE clock_callbacks = {
    .EventHandler    = clock_event_handler,
    .EmptyBufferDone  = NULL,
    .FillBufferDone  = clock_fill_buffer_done
};

static OMX_ERRORTYPE audio_sink_event_handler (
    OMX_HANDLETYPE hComponent,
    OMX_PTR pAppData,
    OMX_EVENTTYPE eEvent,
    OMX_U32 Data1,
    OMX_U32 Data2,
    OMX_PTR pEventData)
{
    ox_player_t * player = (ox_player_t *)pAppData;

    if(eEvent == OMX_EventCmdComplete)
    {
        if (Data1 == OMX_CommandStateSet)
        {
            LIMDBG("audio sink state changed to: %s\n",
                  omxil_state_name((int)Data2));

            switch (Data2)
            {
            case OMX_StateIdle :
                broadcast_set_flags(player->audio_sink_flags |= PSF_IDLE);
                break;

            case OMX_StateExecuting :
                broadcast_set_flags(player->audio_sink_flags |= PSF_EXECUTING);
                break;

            case OMX_StatePause :
                broadcast_set_flags(player->audio_sink_flags |= PSF_PAUSE);
                break;

            case OMX_StateLoaded :
                broadcast_set_flags(player->audio_sink_flags |= PSF_LOADED);
                break;

            default :
                LIMWARN("unhandled cmd event: param1 = %d param2 = %d\n",
                       (int)Data1, (int)Data2);
                break;
            }
        } else if (Data1 == OMX_CommandPortEnable)
        {
            LIMDBG("audio sink port %u enabled.\n", Data2);
            broadcast_set_flags(player->audio_sink_port_flags
                                |= PSF_PORT_ENABLED(Data2));
        } else if (Data1 == OMX_CommandPortDisable)
        {
            LIMDBG("audio sink port %u disabled.\n", Data2);
            broadcast_set_flags(player->audio_sink_port_flags
                                |= PSF_PORT_DISABLED(Data2));
        } else if (Data1 == OMX_CommandFlush)
        {
            LIMDBG("audio sink port %u flushed.\n", Data2);
            broadcast_set_flags(player->audio_sink_port_flags
                                |= PSF_PORT_FLUSHED(Data2));
        }
    } else if (eEvent == OMX_EventPortFormatDetected)
    {
        LIMDBG("audio sink port %u format detected.\n", Data1);
    } else if (eEvent == OMX_EventPortSettingsChanged)
    {
        LIMDBG("audio sink port %u settings changed.\n", Data1);
        OMX_SendCommand(hComponent,
                        OMX_CommandPortDisable, Data1, NULL);
        broadcast_set_flags(player->audio_sink_port_flags
                            |= PSF_PORT_SETTINGS_CHANGED(Data1));
    } else if (eEvent == OMX_EventBufferFlag)
    {
        LIMDBG("audio sink buffer flag event.\n");
        broadcast_set_flags(player->audio_sink_flags
                            |=  PSF_EOS_DETECTED);
    } else if (eEvent == OMX_EventError)
    {
        broadcast_set_flags(player->audio_sink_flags |= PSF_INVALID);
    } else
    {
        LIMWARN("unhandled event: param1 = %d param2 = %d\n",
               (int)Data1, (int)Data2);
    }

    return OMX_ErrorNone;
}

ox_omix_cb_t omix_cb = {
    .cb.EventHandler = audio_sink_event_handler,
    .cb.EmptyBufferDone  = NULL,
    .cb.FillBufferDone  = NULL
};

static int load_audio_decoder (
    ox_player_t * player,
    OMX_AUDIO_PARAM_PORTFORMATTYPE * format)
{
    char audio_component_name[OMX_MAX_STRINGNAME_SIZE];
    OMX_U32 num_comps;
    OMX_U8 * names[1];
    OMX_STRING role;
    OMX_PARAM_COMPONENTROLETYPE cur_role;
    OMX_PARAM_PORTDEFINITIONTYPE port;
    int ret;

    role = get_role_by_codec_id(omx_audio_decode_role_map, format->eEncoding);
    if (NULL == role)
    {
        LIMWARN("unsupported audio format: %d\n", format->eEncoding);
        goto bail;
    }

    //load the first component that supports given role.
    names[0] = (OMX_U8 *)audio_component_name;
    num_comps = 1;
    if (OMX_ErrorNone != OMX_GetComponentsOfRole(role, &num_comps, names))
    {
        LIMERR("no component found for role: %s\n", role);
        goto bail;
    }

    if (OMX_ErrorNone !=
        OMX_GetHandle(&player->audio_decode_handle, audio_component_name,
                      player, &audio_decode_callbacks))
    {
        LIMERR("unable to get handle for %s\n",  audio_component_name);
        goto bail;
    }

    //set decoder component current role.
    omxil_set_spec_version(&cur_role, sizeof(OMX_PARAM_COMPONENTROLETYPE));
    strncpy((char *)cur_role.cRole, role, OMX_MAX_STRINGNAME_SIZE);
    OMX_SetParameter(player->audio_decode_handle,
                     OMX_IndexParamStandardComponentRole,
                     &cur_role);

    //set decoder component port definition.
    omxil_set_spec_version(&port, sizeof(OMX_PARAM_PORTDEFINITIONTYPE));
    port.nPortIndex = 0; //TODO: APB+
    OMX_GetParameter(player->audio_decode_handle,
                     OMX_IndexParamPortDefinition,
                     &port);
    port.format.audio.eEncoding = format->eEncoding;
    OMX_SetParameter(player->audio_decode_handle,
                     OMX_IndexParamPortDefinition,
                     &port);

    //audio path looks good, now assign the sink.
    player->audio_sink_handle = ((ox_omix_t *)(player->ox->omix))->sink_handle;
    omix_cb.priv_data = player;
    ox_omix_register_callback((ox_omix_t *)(player->ox->omix), &omix_cb);

    //disable audio sink clock port
    OMX_SendCommand(player->audio_sink_handle,
                    OMX_CommandPortDisable, 1, NULL);
    wait_for_flags(player->audio_sink_port_flags & PSF_PORT_DISABLED(1), ret);
    if (ret)
    {
        LIMWARN("clock port disable timed out.\n");
    }
    return 0;

 bail:
    return -1;
}

static int load_video_render (ox_player_t * player)
{
    char component_name[OMX_MAX_STRINGNAME_SIZE];
    OMX_U32 num_comps;
    OMX_U8 * names[1];
    int ret;

    //load the first video scheduler component.
    names[0] = (OMX_U8 *)component_name;
    num_comps = 1;
    if (OMX_ErrorNone != OMX_GetComponentsOfRole
        ("video_scheduler.binary", &num_comps, names))
    {
        LIMERR("no video scheduler component found.\n");
        goto bail;
    }

    if (OMX_ErrorNone !=
        OMX_GetHandle(&player->video_scheduler_handle, component_name,
                      player, &video_scheduler_callbacks))
    {
        LIMERR("unable to get handle for %s\n",  component_name);
        goto bail;
    }

    //load the first video renderer component.
    names[0] = (OMX_U8 *)component_name;
    num_comps = 1;
    if (OMX_ErrorNone != OMX_GetComponentsOfRole
        ("iv_renderer.yuv.overlay", &num_comps, names))
    {
        LIMERR("no yuv overlay component found.\n");
        goto bail;
    }

    if (OMX_ErrorNone !=
        OMX_GetHandle(&player->video_sink_handle, component_name,
                      player, &video_sink_callbacks))
    {
        LIMERR("unable to get handle for %s\n",  component_name);
        goto bail;
    }

    //disable clock port
    OMX_SendCommand(player->video_scheduler_handle,
                    OMX_CommandPortDisable, 2, NULL);
    wait_for_flags(player->video_scheduler_port_flags & PSF_PORT_DISABLED(2),
                   ret);
    if (ret)
    {
        LIMWARN("clock port disable timed out.\n");
    }

    return 0;

 bail:
    return -1;
}

static int load_video_decoder (
    ox_player_t * player,
    OMX_VIDEO_PARAM_PORTFORMATTYPE * format)
{
    char video_component_name[OMX_MAX_STRINGNAME_SIZE];
    OMX_U32 num_comps;
    OMX_U8 * names[1];
    OMX_STRING role;
    OMX_PARAM_COMPONENTROLETYPE cur_role;
    OMX_PARAM_PORTDEFINITIONTYPE port;

    role = get_role_by_codec_id(omx_video_decode_role_map,
                                format->eCompressionFormat);
    if (NULL == role)
    {
        LIMWARN("unsupported video format: %d\n",
               format->eCompressionFormat);
        goto bail;
    }

    //load the first component that supports given role.
    names[0] = (OMX_U8 *)video_component_name;
    num_comps = 1;
    if (OMX_ErrorNone != OMX_GetComponentsOfRole(role, &num_comps, names))
    {
        LIMERR("no component found for role: %s\n", role);
        goto bail;
    }

    if (OMX_ErrorNone !=
        OMX_GetHandle(&player->video_decode_handle, video_component_name,
                      player, &video_decode_callbacks))
    {
        LIMERR("unable to get handle for %s\n",  video_component_name);
        goto bail;
    }

    //set decoder component current role.
    omxil_set_spec_version(&cur_role, sizeof(OMX_PARAM_COMPONENTROLETYPE));
    strncpy((char *)cur_role.cRole, role, OMX_MAX_STRINGNAME_SIZE);
    OMX_SetParameter(player->video_decode_handle,
                     OMX_IndexParamStandardComponentRole,
                     &cur_role);

    //set decoder component port definition.
    omxil_set_spec_version(&port, sizeof(OMX_PARAM_PORTDEFINITIONTYPE));
    port.nPortIndex = 0; //TODO: VPB+
    OMX_GetParameter(player->video_decode_handle,
                     OMX_IndexParamPortDefinition,
                     &port);
    port.format.video.eCompressionFormat = format->eCompressionFormat;
    OMX_SetParameter(player->video_decode_handle,
                     OMX_IndexParamPortDefinition,
                     &port);

    port.nPortIndex = 1;
    OMX_GetParameter(player->video_decode_handle,
                     OMX_IndexParamPortDefinition,
                     &port);
    //enable port autodetect
    port.format.video.nFrameWidth = 0;
    port.format.video.nFrameHeight = 0;
    OMX_SetParameter(player->video_decode_handle,
                     OMX_IndexParamPortDefinition,
                     &port);

    //video path looks good, now load scheduler and renderer.
    return(load_video_render(player));

 bail:
    return -1;
}

//examine port format and activate the first available stream
//TODO: interface to allow client to pick stream.
//
//demuxer component has reported port settings change.
static int examine_port_format (
    ox_player_t * player)
{
    OMX_PARAM_U32TYPE audio_num_streams;
    OMX_PARAM_U32TYPE video_num_streams;
    OMX_AUDIO_PARAM_PORTFORMATTYPE audio_port_format;
    OMX_VIDEO_PARAM_PORTFORMATTYPE video_port_format;
    int ii;
    int audio_ret = 0;
    int video_ret = 0;

    //examine audio port
    omxil_set_spec_version(&audio_num_streams, sizeof(OMX_PARAM_U32TYPE));
    audio_num_streams.nPortIndex = 0;
    if (OMX_ErrorNone !=
        OMX_GetParameter(player->demux_handle,
                         OMX_IndexParamNumAvailableStreams,
                         &audio_num_streams))
    {
        LIMDBG("unable to get available audio streams.\n");
        audio_num_streams.nU32 = 0;
        audio_ret = -1;
    } else if (audio_num_streams.nU32)
    {
        omxil_set_spec_version(&audio_port_format,
                               sizeof(OMX_AUDIO_PARAM_PORTFORMATTYPE));
        audio_port_format.nPortIndex = 0;
        for (ii = 0; ii < audio_num_streams.nU32; ii++)
        {
            audio_port_format.nIndex = ii;
            if (OMX_ErrorNone !=
                OMX_GetParameter(player->demux_handle,
                                 OMX_IndexParamAudioPortFormat,
                                 &audio_port_format))
            {
                LIMERR("unable to get port 0 format with index %d\n", ii);
                break;
            }
        }
    }

    omxil_set_spec_version(&video_num_streams, sizeof(OMX_PARAM_U32TYPE));
    video_num_streams.nPortIndex = 1;
    if (OMX_ErrorNone !=
        OMX_GetParameter(player->demux_handle,
                         OMX_IndexParamNumAvailableStreams,
                         &video_num_streams))
    {
        LIMDBG("unable to get available video streams.\n");
        video_num_streams.nU32 = 0;
        video_ret = -1;
    } else if (video_num_streams.nU32)
    {
        omxil_set_spec_version(&video_port_format,
                               sizeof(OMX_VIDEO_PARAM_PORTFORMATTYPE));
        video_port_format.nPortIndex = 1;
        for (ii = 0; ii < video_num_streams.nU32; ii++)
        {
            video_port_format.nIndex = ii;
            if (OMX_ErrorNone !=
                OMX_GetParameter(player->demux_handle,
                                 OMX_IndexParamVideoPortFormat,
                                 &video_port_format))
            {
                LIMERR("unable to get port 1 format with index %d\n", ii);
                break;
            }
        }
    }

    if (!audio_num_streams.nU32 && !video_num_streams.nU32)
    {
        LIMDBG("neither audio nor video detected.\n");
    } else
    {
        if (audio_num_streams.nU32)
        {
            OMX_PARAM_U32TYPE active_stream;

            omxil_set_spec_version(&active_stream, sizeof(OMX_PARAM_U32TYPE));
            active_stream.nPortIndex = 0;
            active_stream.nU32 = 0;
            if (OMX_ErrorNone !=
                OMX_SetParameter(player->demux_handle,
                                 OMX_IndexParamActiveStream,
                                 &active_stream))
            {
                LIMERR("unable to set active audio stream to %u.\n",
                      active_stream.nU32);
            }

            omxil_set_spec_version(&audio_port_format,
                                   sizeof(OMX_AUDIO_PARAM_PORTFORMATTYPE));
            audio_port_format.nPortIndex = 0;
            audio_port_format.nIndex = 0;
            OMX_GetParameter(player->demux_handle,
                             OMX_IndexParamAudioPortFormat,
                             &audio_port_format);

            if (load_audio_decoder(player, &audio_port_format))
            {
                LIMERR("unable to find audio decoder.\n");
                audio_ret = -1;
            }
        }
        if (video_num_streams.nU32)
        {
            OMX_PARAM_U32TYPE active_stream;

            omxil_set_spec_version(&active_stream, sizeof(OMX_PARAM_U32TYPE));
            active_stream.nPortIndex = 1;
            active_stream.nU32 = 0;
            if (OMX_ErrorNone !=
                OMX_SetParameter(player->demux_handle,
                                 OMX_IndexParamActiveStream,
                                 &active_stream))
            {
                LIMERR("unable to set active video stream to %u.\n",
                      active_stream.nU32);
            }

            omxil_set_spec_version(&video_port_format,
                                   sizeof(OMX_VIDEO_PARAM_PORTFORMATTYPE));
            video_port_format.nPortIndex = 1;
            video_port_format.nIndex = 0;
            OMX_GetParameter(player->demux_handle,
                             OMX_IndexParamVideoPortFormat,
                             &video_port_format);

            if (load_video_decoder(player, &video_port_format))
            {
                LIMERR("unable to find video decoder.\n");
                video_ret = -1;
            }
        }
    }

    return (audio_ret == -1 && video_ret == -1)? -1 : 0;
}

//player->demux_handle has the demuxer candidate component.
static int demux_try (
    ox_player_t * player)
{
    int ret;
    char * uri;
    int bytes;
    OMX_PARAM_CONTENTURITYPE * uri_param;
    OMX_PARAM_PORTDEFINITIONTYPE port0;
    OMX_PARAM_PORTDEFINITIONTYPE port1;
    OMX_BUFFERHEADERTYPE ** dummy;
    int ii;

    uri = (char *)(((XADataLocator_URI *)player->data_src.pLocator)->pURI);
    bytes = strlen(uri);
    uri_param = (OMX_PARAM_CONTENTURITYPE *)malloc
        (sizeof(OMX_PARAM_CONTENTURITYPE) + bytes);
    omxil_set_spec_version(uri_param, sizeof(OMX_PARAM_CONTENTURITYPE) + bytes);
    strcpy((char *)uri_param->contentURI, uri);
    OMX_SetParameter(player->demux_handle, OMX_IndexParamContentURI, uri_param);
    free(uri_param);

    if (OMX_ErrorNone !=
        OMX_SendCommand(player->demux_handle,
                        OMX_CommandStateSet, OMX_StateIdle, NULL))
    {
        LIMERR("unable to set demuxer to idle.\n");
        goto bail;
    }

    omxil_set_spec_version(&port0, sizeof(OMX_PARAM_PORTDEFINITIONTYPE));
    port0.nPortIndex = 0;
    OMX_GetParameter(player->demux_handle,
                     OMX_IndexParamPortDefinition, &port0);

    omxil_set_spec_version(&port1, sizeof(OMX_PARAM_PORTDEFINITIONTYPE));
    port1.nPortIndex = 1;
    OMX_GetParameter(player->demux_handle,
                     OMX_IndexParamPortDefinition, &port1);

    dummy = malloc((port0.nBufferCountActual + port1.nBufferCountActual)
                   * sizeof(OMX_BUFFERHEADERTYPE *));
    for (ii = 0; ii < port0.nBufferCountActual; ii++)
    {
        OMX_UseBuffer(player->demux_handle, &dummy[ii],
                      0, NULL, 1, player->dummy_buf);
    }
    for (ii = 0; ii < port1.nBufferCountActual; ii++)
    {
        OMX_UseBuffer(player->demux_handle,
                      &dummy[ii + port0.nBufferCountActual],
                      1, NULL, 1, player->dummy_buf);
    }

    wait_for_flags(player->demux_flags & (PSF_INVALID | PSF_IDLE), ret);
    if (ret)
    {
        LIMERR("transition demuxer to IDLE timedout.\n");
        goto bail;
    }

    lthread_mutex_lock(&player->mutex);
    if (player->demux_flags & PSF_INVALID)
    {
        //TODO:
        LIMERR("unable to transition demuxer to IDLE.\n");
        lthread_mutex_unlock(&player->mutex);
        goto bail;
    }
    lthread_mutex_unlock(&player->mutex);

    for (ii = 0; ii < port0.nBufferCountActual + port1.nBufferCountActual; ii++)
    {
        OMX_FillThisBuffer(player->demux_handle, dummy[ii]);
    }
    free(dummy);

    if (OMX_ErrorNone !=
        OMX_SendCommand(player->demux_handle,
                        OMX_CommandStateSet, OMX_StateExecuting, NULL))
    {
        LIMERR("unable to execute demuxer.\n");
        goto bail;
    }

    wait_for_flags(player->demux_flags
                   & (PSF_INVALID | PSF_EXECUTING), ret);
    if (ret)
    {
        LIMERR("transition demuxer to EXECUTING timedout.\n");
        goto bail;
    }

    lthread_mutex_lock(&player->mutex);
    if (player->demux_flags & PSF_INVALID)
    {
        //TODO:
        LIMERR("unable to transition demuxer to EXECUTING.\n");
        lthread_mutex_unlock(&player->mutex);
        goto bail;
    }
    lthread_mutex_unlock(&player->mutex);

    //now wait for porting settings change event.
    wait_for_flags((player->demux_flags & PSF_INVALID)
                   || (player->demux_port_flags
                       & (PSF_PORT_SETTINGS_CHANGED(0)
                          | PSF_PORT_SETTINGS_CHANGED(1))), ret);
    if (ret)
    {
        LIMERR("waiting for PORT_SETTINGS_CHANGED timedout.\n");
        goto bail;
    }

    lthread_mutex_lock(&player->mutex);
    if (player->demux_flags & PSF_INVALID)
    {
        //TODO:
        LIMERR("unable to dectect content format.\n");
        lthread_mutex_unlock(&player->mutex);
        goto bail;
    }
    lthread_mutex_unlock(&player->mutex);

    //now wait for port disable to complete.
    //port setting changes could come in any order.
    if (player->demux_port_flags & PSF_PORT_SETTINGS_CHANGED(0))
    {
        wait_for_flags(player->demux_port_flags & PSF_PORT_DISABLED(0), ret);
        if (player->demux_port_flags & PSF_PORT_SETTINGS_CHANGED(1))
        {
            wait_for_flags(player->demux_port_flags & PSF_PORT_DISABLED(1), ret);
        }
    }
    if (player->demux_port_flags & PSF_PORT_SETTINGS_CHANGED(1))
    {
        wait_for_flags(player->demux_port_flags & PSF_PORT_DISABLED(1), ret);
        if (player->demux_port_flags & PSF_PORT_SETTINGS_CHANGED(0))
        {
            wait_for_flags(player->demux_port_flags & PSF_PORT_DISABLED(0), ret);
        }
    }

    if (examine_port_format(player))
    {
        LIMERR("unable to find decoder for content.\n");
        goto bail;
    }
    return 0;

 bail:
    return -1;
}

static int content_discovery (
    ox_player_t * player)
{
    int ii, ret;
    int role_index;
    char * name_buf = NULL;
    char * name_buf_ptr;
    OMX_U8 ** names = NULL;
    OMX_U32 num_comps;

    LIMDBG("content discovery in process.\n");

    role_index = 0;
 try_next_role:
    if (name_buf) free(name_buf);
    if (names) free(names);
    name_buf = NULL;
    names = NULL;

    if (NULL == omx_demux_role_map[role_index])
    {
        LIMERR("omx role map table exhausted.\n");
        goto bail;
    }

    if (OMX_ErrorNone !=
        OMX_GetComponentsOfRole((OMX_STRING)omx_demux_role_map[role_index],
                                &num_comps, NULL))
    {
        LIMERR("unable to find demuxer component.\n");
        goto bail;
    }

    if (!num_comps)
    {
        LIMDBG("no component found for role: %s\n",
              omx_demux_role_map[role_index]);
        role_index++;
        goto try_next_role;
    }

    name_buf = malloc(OMX_MAX_STRINGNAME_SIZE * num_comps);
    names = malloc(sizeof(OMX_U8 *) * num_comps);
    if (NULL == name_buf || NULL == names)
    {
        LIMERR("out of memory.\n");
        goto bail;
    }

    name_buf_ptr = name_buf;
    for (ii = 0; ii < num_comps; ii++)
    {
        names[ii] = (OMX_U8 *)name_buf_ptr;
        name_buf_ptr += OMX_MAX_STRINGNAME_SIZE;
    }

    if (OMX_ErrorNone !=
        OMX_GetComponentsOfRole((OMX_STRING)omx_demux_role_map[role_index],
                                &num_comps, names))
    {
        LIMERR("unable to find demuxer component.\n");
        goto bail;
    }

    name_buf_ptr = name_buf;
 try_next_comp:
    if (!num_comps)
    {
        role_index++;
        goto try_next_role;
    }

    if (OMX_ErrorNone !=
        OMX_GetHandle(&player->demux_handle, name_buf_ptr,
                      player, &demux_callbacks))
    {
        LIMERR("unable to get %s demuxer handle.\n", name_buf_ptr);
        name_buf_ptr += OMX_MAX_STRINGNAME_SIZE;
        num_comps--;
        goto try_next_comp;
    } else
    {
        //disable demuxer clock port
        LIMDBG("disabling demux port 2.\n");
        set_flags(player->demux_port_flags = 0;
                  player->demux_flags = 0);
        OMX_SendCommand(player->demux_handle,
                        OMX_CommandPortDisable, 2, NULL);
        wait_for_flags(player->demux_port_flags & PSF_PORT_DISABLED(2), ret);
        if (ret)
        {
            LIMWARN("clock port disable timed out.\n");
        }
    }

    LIMDBG("trying %s demuxer.\n", name_buf_ptr);
    if (demux_try(player))
    {
        OMX_FreeHandle(player->demux_handle);
        name_buf_ptr += OMX_MAX_STRINGNAME_SIZE;
        num_comps--;
        goto try_next_comp;
    }

    return 0;

 bail:
    if (name_buf) free(name_buf);
    if (names) free(names);
    return -1;
}

//clock port mapping:
// 0 : demuxer
// 1 : audio
// 2 : video
static int clock_init (ox_player_t * player)
{
    char component_name[OMX_MAX_STRINGNAME_SIZE];
    OMX_U32 num_comps;
    OMX_U8 * names[1];
    int ii;
    OMX_PORT_PARAM_TYPE port_param;

    names[0] = (OMX_U8 *)component_name;
    num_comps = 1;
    if (OMX_ErrorNone !=
        OMX_GetComponentsOfRole("clock.binary", &num_comps, names))
    {
        LIMERR("no clock component found for role: %s\n", "clock.binary");
        goto bail;
    }

    if (OMX_ErrorNone !=
        OMX_GetHandle(&player->clock_handle, component_name,
                      player, &clock_callbacks))
    {
        LIMERR("unable to get handle for %s\n",  component_name);
        goto bail;
    }

    //disable all clock ports.
    omxil_set_spec_version(&port_param, sizeof(OMX_PORT_PARAM_TYPE));
    OMX_GetParameter(player->clock_handle,
                     OMX_IndexParamOtherInit, &port_param);
    LIMDBG("number of clock ports: %u\n", port_param.nPorts);

    for (ii = 0; ii < port_param.nPorts; ii++)
    {
        OMX_SendCommand(player->clock_handle,
                        OMX_CommandPortDisable,
                        ii + port_param.nStartPortNumber,
                        NULL);
    }

    return 0;

 bail:
    return -1;
}

static int clock_config (ox_player_t * player)
{
    int ret;
    OMX_TIME_CONFIG_ACTIVEREFCLOCKTYPE ref;
    OMX_TIME_CONFIG_CLOCKSTATETYPE st;
    ltime_val_t tv;

    //instantiate clock component.
    OMX_SendCommand(player->clock_handle,
                    OMX_CommandStateSet, OMX_StateIdle, NULL);

    wait_for_flags(player->clock_flags & (PSF_INVALID | PSF_IDLE), ret);
    if (ret)
    {
        LIMERR("waiting for clock setup timedout.\n");
        goto bail;
    }

    LIMDBG("tunneling demuxer with clock.\n");
    if (OMX_ErrorNone !=
        OMX_SetupTunnel(player->clock_handle, 0, player->demux_handle, 2))
    {
        LIMERR("unable to tunnel clock and demuxer component.\n");
        goto bail;
    }

    LIMDBG("tunneling clock with audio renderer.\n");
    if (player->audio_sink_handle)
    {
        if (OMX_ErrorNone !=
            OMX_SetupTunnel(player->clock_handle, 1,
                            player->audio_sink_handle, 1))
        {
            LIMERR("unable to tunnel clock and audio sink component.\n");
            goto bail;
        }
    }

    LIMDBG("tunneling clock with video scheduler.\n");
    if (player->video_scheduler_handle)
    {
        if (OMX_ErrorNone !=
            OMX_SetupTunnel(player->clock_handle, 2,
                            player->video_scheduler_handle, 2))
        {
            LIMERR("unable to tunnel clock and video scheduler component.\n");
            goto bail;
        }
    }

    //enable clock ports
    OMX_SendCommand(player->clock_handle,
                    OMX_CommandPortEnable, 0, NULL);
    OMX_SendCommand(player->demux_handle,
                    OMX_CommandPortEnable, 2, NULL);
    wait_for_flags
        (((player->demux_flags & PSF_INVALID)
          || player->demux_port_flags & PSF_PORT_ENABLED(2))
         && ((player->clock_flags & PSF_INVALID)
             || player->clock_port_flags & PSF_PORT_ENABLED(0)), ret);
    if (ret)
    {
        LIMERR("waiting for clock port 0 enable timed out.\n");
        goto bail;
    }

    if (player->audio_sink_handle)
    {
        OMX_SendCommand(player->clock_handle,
                        OMX_CommandPortEnable, 1, NULL);
        OMX_SendCommand(player->audio_sink_handle,
                        OMX_CommandPortEnable, 1, NULL);
        wait_for_flags
            (((player->audio_sink_flags & PSF_INVALID)
              || player->audio_sink_port_flags & PSF_PORT_ENABLED(1))
             && ((player->clock_flags & PSF_INVALID)
                 || player->clock_port_flags & PSF_PORT_ENABLED(1)), ret);

        if (ret)
        {
            LIMERR("waiting for clock port 1 enable timed out.\n");
            goto bail;
        }
    }

    if (player->video_scheduler_handle)
    {
        OMX_SendCommand(player->clock_handle,
                        OMX_CommandPortEnable, 2, NULL);
        OMX_SendCommand(player->video_scheduler_handle,
                        OMX_CommandPortEnable, 2, NULL);
        wait_for_flags
            (((player->video_scheduler_flags & PSF_INVALID)
              || player->video_scheduler_port_flags & PSF_PORT_ENABLED(2))
             && ((player->clock_flags & PSF_INVALID)
                 || player->clock_port_flags & PSF_PORT_ENABLED(2)), ret);

        if (ret)
        {
            LIMERR("waiting for clock port 2 enable timed out.\n");
            goto bail;
        }
    }

    //audio as reference clock if present.
    omxil_set_spec_version(&ref, sizeof(OMX_TIME_CONFIG_ACTIVEREFCLOCKTYPE));
    ref.eClock = (player->audio_decode_handle)?
        OMX_TIME_RefClockAudio : OMX_TIME_RefClockVideo;
    if (OMX_ErrorNone !=
        OMX_SetConfig(player->clock_handle,
                      OMX_IndexConfigTimeActiveRefClock,
                      &ref))
    {
        LIMERR("unable to specifiy reference clock source.\n");
        goto bail;
    }

    //set clock state
    omxil_set_spec_version(&st, sizeof(OMX_TIME_CONFIG_CLOCKSTATETYPE));
    st.eState = OMX_TIME_ClockStateRunning;
    ltime_gettimeofday(&tv);
    st.nStartTime = ltime_tv2us(&tv);
    st.nOffset = 0;
    st.nWaitMask = 0;
    if (OMX_ErrorNone !=
        OMX_SetConfig(player->clock_handle,
                      OMX_IndexConfigTimeClockState,
                      &st))
    {
        LIMERR("unable to set clock state.\n");
        goto bail;
    }

    LIMDBG("clock config done.\n");
    return 0;

 bail:
    return -1;
}

static void * sync_thread (void * arg)
{
    ox_player_t * player = (ox_player_t *)arg;

    lthread_mutex_lock(&player->mutex);
    player->sync_state = OPSS_IDLE;
    lthread_cond_broadcast(&player->cond);
    lthread_mutex_unlock(&player->mutex);

    while (1)
    {
        lthread_mutex_lock(&player->mutex);
        if (player->sync_state == OPSS_STOP)
        {
            lthread_mutex_unlock(&player->mutex);
            break;
        }

        if (player->sync_flags & PSR_AUDIO_DECODER_OUTPUT_PORTSETTINGSCHANGED)
        {
            if (player->audio_decode_port_flags & PSF_PORT_DISABLED(1)
                && player->audio_sink_port_flags & PSF_PORT_DISABLED(0))
            {
                OMX_AUDIO_PARAM_PCMMODETYPE pcm;

                omxil_set_spec_version
                    (&pcm, sizeof(OMX_AUDIO_PARAM_PCMMODETYPE));

                pcm.nPortIndex = 1;
                OMX_GetParameter(player->audio_decode_handle,
                                 OMX_IndexParamAudioPcm, &pcm);

                pcm.nPortIndex=0;
                OMX_SetParameter(player->audio_sink_handle,
                                 OMX_IndexParamAudioPcm, &pcm);

#ifdef AUDIO_VIDEO_STARTUP_SYNC
                //if video is present, wait till video path is ready before
                //enabling audio.
                if (player->video_sink_handle)
                {
                    player->sync_flags |= PSR_DECODER_OUTPUT_ENABLE;
                } else
                {
                    OMX_SendCommand(player->audio_sink_handle,
                                    OMX_CommandPortEnable, 0, NULL);
                    OMX_SendCommand(player->audio_decode_handle,
                                    OMX_CommandPortEnable, 1, NULL);
                }
#else
                OMX_SendCommand(player->audio_sink_handle,
                                OMX_CommandPortEnable, 0, NULL);
                OMX_SendCommand(player->audio_decode_handle,
                                OMX_CommandPortEnable, 1, NULL);
#endif
                player->sync_flags &=
                    ~PSR_AUDIO_DECODER_OUTPUT_PORTSETTINGSCHANGED;
            }
        }

        if (player->sync_flags & PSR_VIDEO_DECODER_OUTPUT_PORTSETTINGSCHANGED)
        {
            if (player->video_decode_port_flags & PSF_PORT_DISABLED(1)
                && player->video_scheduler_port_flags & PSF_PORT_DISABLED(0))
            {
                OMX_PARAM_PORTDEFINITIONTYPE port;
                OMX_PARAM_PORTDEFINITIONTYPE port2;
                int w, h;

                omxil_set_spec_version
                    (&port, sizeof(OMX_PARAM_PORTDEFINITIONTYPE));

                port.nPortIndex = 1;
                OMX_GetParameter(player->video_decode_handle,
                                 OMX_IndexParamPortDefinition, &port);

                //420 plannar and aligned on 16 pixel boundary
                w = (port.format.video.nFrameWidth + 15) & ~15;
                h = (port.format.video.nFrameHeight + 15) & ~15;

                port.nBufferSize = (w + w / 2) * h;
                OMX_SetParameter(player->video_decode_handle,
                                 OMX_IndexParamPortDefinition, &port);


                omxil_set_spec_version
                    (&port2, sizeof(OMX_PARAM_PORTDEFINITIONTYPE));
                port2.nPortIndex = 0;
                OMX_GetParameter(player->video_scheduler_handle,
                                 OMX_IndexParamPortDefinition, &port2);
                port2.format.video = port.format.video;
                port2.nBufferSize = port.nBufferSize;
                if (OMX_ErrorNone !=
                    OMX_SetParameter(player->video_scheduler_handle,
                                     OMX_IndexParamPortDefinition, &port2))
                {
                    LIMERR("unable to set parameter.\n");
                }

                //propagate info to video sink
                player->video_scheduler_port_flags &= ~PSF_PORT_DISABLED(1);
                player->video_sink_port_flags &= ~PSF_PORT_DISABLED(0);
                player->sync_flags |=
                    PSR_VIDEO_SCHEDULER_INPUT_PORTSETTINGSCHANGED;

                OMX_SendCommand(player->video_sink_handle,
                                OMX_CommandPortDisable, 0, NULL);
                OMX_SendCommand(player->video_scheduler_handle,
                                OMX_CommandPortDisable, 1, NULL);

#ifdef AUDIO_VIDEO_STARTUP_SYNC
                //if audio is present, synchronize with it.
                if (player->audio_sink_handle)
                {
                } else
                {
                    OMX_SendCommand(player->video_scheduler_handle,
                                OMX_CommandPortEnable, 0, NULL);
                    OMX_SendCommand(player->video_decode_handle,
                                    OMX_CommandPortEnable, 1, NULL);
                }
#else
                OMX_SendCommand(player->video_scheduler_handle,
                                OMX_CommandPortEnable, 0, NULL);
                OMX_SendCommand(player->video_decode_handle,
                                OMX_CommandPortEnable, 1, NULL);
#endif
                player->sync_flags &=
                    ~PSR_VIDEO_DECODER_OUTPUT_PORTSETTINGSCHANGED;
                print_player_flags("VIDEO DECODER SETTING CHANGED", player);
            }
        }

        if (player->sync_flags & PSR_VIDEO_SCHEDULER_INPUT_PORTSETTINGSCHANGED)
        {
            if (player->video_scheduler_port_flags & PSF_PORT_DISABLED(1)
                && player->video_sink_port_flags & PSF_PORT_DISABLED(0))
            {
                OMX_PARAM_PORTDEFINITIONTYPE port;
                OMX_PARAM_PORTDEFINITIONTYPE port2;
                OMX_CONFIG_SCALEFACTORTYPE scale;

                omxil_set_spec_version
                    (&port, sizeof(OMX_PARAM_PORTDEFINITIONTYPE));

                port.nPortIndex = 0;
                OMX_GetParameter(player->video_scheduler_handle,
                                 OMX_IndexParamPortDefinition, &port);

                omxil_set_spec_version
                    (&port2, sizeof(OMX_PARAM_PORTDEFINITIONTYPE));
                port2.nPortIndex = 1;
                OMX_GetParameter(player->video_scheduler_handle,
                                 OMX_IndexParamPortDefinition, &port2);
                port2.format.video = port.format.video;
                port2.nBufferSize = port.nBufferSize;
                OMX_SetParameter(player->video_scheduler_handle,
                                 OMX_IndexParamPortDefinition, &port2);

                port2.nPortIndex = 0;
                OMX_GetParameter(player->video_sink_handle,
                                 OMX_IndexParamPortDefinition, &port2);
                port2.format.video = port.format.video;
                port2.nBufferSize = port.nBufferSize;
                OMX_SetParameter(player->video_sink_handle,
                                 OMX_IndexParamPortDefinition, &port2);

                omxil_set_spec_version
                    (&scale, sizeof(OMX_CONFIG_SCALEFACTORTYPE));
                scale.nPortIndex = 1;
                if (OMX_ErrorNone ==
                    OMX_GetConfig(player->video_decode_handle,
                                  OMX_IndexConfigCommonScale, &scale))
                {
                    scale.nPortIndex = 0;
                    OMX_SetConfig(player->video_sink_handle,
                                  OMX_IndexConfigCommonScale, &scale);
                } else
                {
                    LIMINFO("video aspect ratio info not available.\n");
                }
                LIMDBG("video resolution: %d x %d, aspect ratio: %d x %d\n",
                       port.format.video.nFrameWidth,
                       port.format.video.nFrameHeight,
                       scale.xWidth, scale.xHeight);
                print_player_flags("VIDEO SETUP DONE", player);

                if (player->audio_sink_handle)
                {
                    player->video_sink_port_flags &= ~PSF_PORT_ENABLED(0);
                }

                OMX_SendCommand(player->video_sink_handle,
                                OMX_CommandPortEnable, 0, NULL);

                OMX_SendCommand(player->video_scheduler_handle,
                                OMX_CommandPortEnable, 1, NULL);

                player->sync_flags &=
                    ~PSR_VIDEO_SCHEDULER_INPUT_PORTSETTINGSCHANGED;
            }
        }
#ifdef AUDIO_VIDEO_STARTUP_SYNC
        if (player->sync_flags & PSR_DECODER_OUTPUT_ENABLE)
        {
            LIMDBG("wait to enable decoder output port.\n");
            if (player->video_sink_port_flags & PSF_PORT_ENABLED(0))
            {
                LIMDBG("now to enable decoder output port.\n");
                OMX_SendCommand(player->audio_sink_handle,
                                OMX_CommandPortEnable, 0, NULL);
                OMX_SendCommand(player->audio_decode_handle,
                                OMX_CommandPortEnable, 1, NULL);

                OMX_SendCommand(player->video_scheduler_handle,
                                OMX_CommandPortEnable, 0, NULL);
                OMX_SendCommand(player->video_decode_handle,
                                OMX_CommandPortEnable, 1, NULL);
                player->sync_flags &= ~PSR_DECODER_OUTPUT_ENABLE;
            }
        }
#endif
        if (player->sync_flags & PSR_DEMUXER_EOS_DETECTED)
        {
            if ((!player->video_sink_handle
                || player->video_sink_flags & PSF_EOS_DETECTED)
                && (!player->audio_sink_handle
                    || player->audio_sink_flags & PSF_EOS_DETECTED))
            {
                //callback to client
                ipc_packet_t pkt;

                player->callback.event = XA_PLAYEVENT_HEADATEND;

                pkt.header.id.value = OX_CALLBACK_PLAY;
                pkt.header.bytes = sizeof(ox_callback_play_t);
                pkt.dptr = &player->callback;
                ox_session_send_packet(player->ox, &pkt);

                player->sync_flags &= ~PSR_DEMUXER_EOS_DETECTED;
            }
        }

        lthread_cond_wait(&player->cond, &player->mutex);
        lthread_mutex_unlock(&player->mutex);
    }

    lthread_mutex_lock(&player->mutex);
    player->sync_state = OPSS_INVALID;
    lthread_cond_broadcast(&player->cond);
    lthread_mutex_unlock(&player->mutex);
    return NULL;
}

static void sync_destroy (ox_player_t * player)
{
    lthread_mutex_lock(&player->mutex);
    while (player->sync_state != OPSS_INVALID)
    {
        player->sync_state = OPSS_STOP;
        lthread_cond_broadcast(&player->cond);
        if (lthread_cond_timedwait(&player->cond, &player->mutex, 5000))
        {
            LIMERR("sync thread state transition timeout.\n");
            break;
        }
    }
    lthread_mutex_unlock(&player->mutex);
}

static void obj_destroy (ox_player_t * player)
{
    ox_omix_unregister_callback((ox_omix_t *)(player->ox->omix), &omix_cb);
    lthread_mutex_destroy(&player->mutex);
    lthread_cond_destroy(&player->cond);
}

XAresult ox_player_realize (ox_player_t * player)
{
    int ret;
    XAresult res = XA_RESULT_RESOURCE_ERROR;

    lthread_mutex_init(&player->mutex);
    lthread_cond_init(&player->cond);
    player->sync_state = OPSS_INVALID;
    if (lthread_create(&player->sync_thread, sync_thread, player, 1))
    {
        LIMERR("unable to create sync thread.\n");
        return res;
    }

    lthread_mutex_lock(&player->mutex);
    while (player->sync_state != OPSS_IDLE)
    {
        if (lthread_cond_timedwait(&player->cond, &player->mutex, 5000))
        {
            LIMERR("sync thread state transition timeout.\n");
            lthread_mutex_unlock(&player->mutex);
            goto bail;
        }
    }
    lthread_mutex_unlock(&player->mutex);

    //content automatic discovery, populates the following omx handles if any.
    //-demux_handle
    //-audio_decode_handle
    //-video_decode_handle
    if (content_discovery(player))
    {
        LIMERR("content not found or not supported: %s\n",
              ((XADataLocator_URI *)player->data_src.pLocator)->pURI);
        res = XA_RESULT_CONTENT_UNSUPPORTED;
        goto bail;
    }

    if (clock_init(player))
    {
        LIMERR("unable to initialize clock.\n");
        goto bail;
    }

    if (player->video_decode_handle)
    {
        LIMDBG("tunneling demuxer with video decoder.\n");
        if (OMX_ErrorNone !=
            OMX_SetupTunnel(player->demux_handle, 1,
                            player->video_decode_handle, 0))
        {
            LIMERR("unable to tunnel demuxer and video decoder component.\n");
            goto bail;
        }

        LIMDBG("tunneling video decoder with scheduler.\n");
        if (OMX_ErrorNone !=
            OMX_SetupTunnel(player->video_decode_handle, 1,
                            player->video_scheduler_handle, 0))
        {
            LIMERR("unable to tunnel video decoder and scheduler component.\n");
            goto bail;
        }

        LIMDBG("tunneling video scheduler with sink.\n");
        if (OMX_ErrorNone !=
            OMX_SetupTunnel(player->video_scheduler_handle, 1,
                            player->video_sink_handle, 0))
        {
            LIMERR("unable to tunnel decoder and audio sink component.\n");
            goto bail;
        }

        LIMDBG("reenabling demuxer video port.\n");
        if (OMX_ErrorNone !=
            OMX_SendCommand(player->demux_handle,
                            OMX_CommandPortEnable, 1, NULL))
        {
            LIMERR("unable to enable demuxer video port.\n");
            goto bail;
        }

        LIMDBG("sending video decoder to idle.\n");
        if (OMX_ErrorNone !=
            OMX_SendCommand(player->video_decode_handle,
                            OMX_CommandStateSet, OMX_StateIdle, NULL))
        {
            LIMERR("unable to set decoder to idle.\n");
            goto bail;
        }

        LIMDBG("sending scheduler to idle.\n");
        if (OMX_ErrorNone !=
            OMX_SendCommand(player->video_scheduler_handle,
                            OMX_CommandStateSet, OMX_StateIdle, NULL))
        {
            LIMERR("unable to set scheduler to idle.\n");
            goto bail;
        }

        LIMDBG("sending video sink to idle.\n");
        if (OMX_ErrorNone !=
            OMX_SendCommand(player->video_sink_handle,
                            OMX_CommandStateSet, OMX_StateIdle, NULL))
        {
            LIMERR("unable to set video sink to idle.\n");
            goto bail;
        }

        wait_for_flags
            (((player->demux_flags & PSF_INVALID)
              || (player->demux_port_flags & PSF_PORT_ENABLED(1)))
             && (player->video_scheduler_flags & (PSF_INVALID | PSF_IDLE))
             && (player->video_sink_flags & (PSF_INVALID | PSF_IDLE))
             && (player->video_decode_flags & (PSF_INVALID | PSF_IDLE)), ret);
        if (ret)
        {
            LIMERR("waiting for video setup timedout.\n");
            goto bail;
        }
        print_player_flags("VIDEO IDLE", player);
    }

    if (player->audio_decode_handle)
    {
        LIMDBG("tunneling demuxer with audio decoder.\n");
        //TODO: check to validate audio port index.
        if (OMX_ErrorNone !=
            OMX_SetupTunnel(player->demux_handle, 0,
                            player->audio_decode_handle, 0))
        {
            LIMERR("unable to tunnel demuxer and audio decoder component.\n");
            goto bail;
        }

        LIMDBG("tunneling audio decoder with sink.\n");
        if (OMX_ErrorNone !=
            OMX_SetupTunnel(player->audio_decode_handle, 1,
                            player->audio_sink_handle, 0))
        {
            LIMERR("unable to tunnel decoder and audio sink component.\n");
            goto bail;
        }

        LIMDBG("reenabling demuxer audio port.\n");
        if (OMX_ErrorNone !=
            OMX_SendCommand(player->demux_handle,
                            OMX_CommandPortEnable, 0, NULL))
        {
            LIMERR("unable to enable demuxer audio port.\n");
            goto bail;
        }

        LIMDBG("sending decoder to idle.\n");
        if (OMX_ErrorNone !=
            OMX_SendCommand(player->audio_decode_handle,
                            OMX_CommandStateSet, OMX_StateIdle, NULL))
        {
            LIMERR("unable to set decoder to idle.\n");
            goto bail;
        }

        LIMDBG("sending audio sink to idle.\n");
        if (OMX_ErrorNone !=
            OMX_SendCommand(player->audio_sink_handle,
                            OMX_CommandStateSet, OMX_StateIdle, NULL))
        {
            LIMERR("unable to set audio sink to idle.\n");
            goto bail;
        }

        wait_for_flags
            (((player->demux_flags & PSF_INVALID)
              || (player->demux_port_flags & PSF_PORT_ENABLED(0)))
             && (player->audio_sink_flags & (PSF_INVALID | PSF_IDLE))
             && (player->audio_decode_flags & (PSF_INVALID | PSF_IDLE)), ret);
        if (ret)
        {
            LIMERR("waiting for audio setup timedout.\n");
            goto bail;
        }
        //TODO: error handling if any of the PSF_INVALID is detected.

        print_player_flags("AUDIO IDLE", player);
    }

    clock_config(player);

    //kick off clock
    OMX_SendCommand(player->clock_handle,
                    OMX_CommandStateSet, OMX_StateExecuting, NULL);

    wait_for_flags(player->clock_flags & (PSF_INVALID | PSF_EXECUTING), ret);
    if (ret)
    {
        LIMERR("waiting for clock EXECUTING timedout.\n");
        goto bail;
    }

    LIMDBG("player realized.\n");
    return XA_RESULT_SUCCESS;

 bail:
    sync_destroy(player);
    obj_destroy(player);
    return res;
}

XAresult ox_player_play (ox_player_t * player)
{
    int ret;
    XAresult res = XA_RESULT_RESOURCE_ERROR;

    if (player->audio_decode_handle)
    {
        if (OMX_ErrorNone !=
            OMX_SendCommand(player->audio_decode_handle,
                            OMX_CommandStateSet, OMX_StateExecuting, NULL))
        {
            LIMERR("unable to execute decoder.\n");
            goto bail;
        }

        if (OMX_ErrorNone !=
            OMX_SendCommand(player->audio_sink_handle,
                            OMX_CommandStateSet, OMX_StateExecuting, NULL))
        {
            LIMERR("unable to execute audio sink.\n");
            goto bail;
        }

        wait_for_flags
            (player->audio_decode_flags & (PSF_INVALID | PSF_EXECUTING)
             && player->audio_sink_flags & (PSF_INVALID | PSF_EXECUTING), ret);
        if (ret)
        {
            LIMERR("waiting for audio EXECUTING timedout.\n");
            goto bail;
        }
        print_player_flags("AUDIO PLAYING", player);
    }

    if (player->video_decode_handle)
    {
        if (OMX_ErrorNone !=
            OMX_SendCommand(player->video_decode_handle,
                            OMX_CommandStateSet, OMX_StateExecuting, NULL))
        {
            LIMERR("unable to execute decoder.\n");
            goto bail;
        }

        if (OMX_ErrorNone !=
            OMX_SendCommand(player->video_scheduler_handle,
                            OMX_CommandStateSet, OMX_StateExecuting, NULL))
        {
            LIMERR("unable to execute audio sink.\n");
            goto bail;
        }

        if (OMX_ErrorNone !=
            OMX_SendCommand(player->video_sink_handle,
                            OMX_CommandStateSet, OMX_StateExecuting, NULL))
        {
            LIMERR("unable to execute audio sink.\n");
            goto bail;
        }

        wait_for_flags
            (player->video_decode_flags & (PSF_INVALID | PSF_EXECUTING)
             && player->video_scheduler_flags & (PSF_INVALID | PSF_EXECUTING)
             && player->video_sink_flags & (PSF_INVALID | PSF_EXECUTING), ret);
        if (ret)
        {
            LIMERR("waiting for video EXECUTING timedout.\n");
            goto bail;
        }
        print_player_flags("VIDEO PLAYING", player);
    }

    res = XA_RESULT_SUCCESS;
 bail:
    return res;
}

//NOTE: it is up to the fontend to make sure player has been stopped before
//      this API is called.
int ox_player_destroy (ox_player_t * player)
{
    int ret;
    XAresult res = XA_RESULT_PERMISSION_DENIED;

    sync_destroy(player);

    set_flags(player->demux_flags &= ~PSF_LOADED);
    if (OMX_ErrorNone !=
        OMX_SendCommand(player->demux_handle,
                        OMX_CommandStateSet, OMX_StateLoaded, NULL))
    {
        LIMERR("unable to send demuxer to loaded state.\n");
        goto bail;
    }

    if (OMX_ErrorNone !=
        OMX_SendCommand(player->clock_handle,
                        OMX_CommandStateSet, OMX_StateLoaded, NULL))
    {
        LIMERR("unable to send clock to loaded state.\n");
        goto bail;
    }

    if (player->audio_decode_handle)
    {
        set_flags(player->audio_decode_flags &= ~PSF_LOADED;
                  player->audio_sink_flags &= ~PSF_LOADED);

        if (OMX_ErrorNone !=
            OMX_SendCommand(player->audio_decode_handle,
                            OMX_CommandStateSet, OMX_StateLoaded, NULL))
        {
        LIMERR("unable to send audio decoder to loaded state.\n");
            goto bail;
        }

        if (OMX_ErrorNone !=
            OMX_SendCommand(player->audio_sink_handle,
                            OMX_CommandStateSet, OMX_StateLoaded, NULL))
        {
            LIMERR("unable to send audio sink to loaded state.\n");
            goto bail;
        }

        wait_for_flags((player->audio_decode_flags & (PSF_INVALID | PSF_LOADED))
                       && (player->audio_sink_flags & (PSF_INVALID | PSF_LOADED)),
                       ret);
        if (ret)
        {
            LIMERR("waiting for audio LOADED timedout.\n");
            goto bail;
        }
    }

    if (player->video_decode_handle)
    {
        set_flags(player->video_decode_flags &= ~PSF_LOADED;
                  player->video_scheduler_flags &= ~PSF_LOADED;
                  player->video_sink_flags &= ~PSF_LOADED);

        if (OMX_ErrorNone !=
            OMX_SendCommand(player->video_decode_handle,
                            OMX_CommandStateSet, OMX_StateLoaded, NULL))
        {
            LIMERR("unable to send video decoder to loaded state.\n");
            goto bail;
        }

        if (OMX_ErrorNone !=
            OMX_SendCommand(player->video_scheduler_handle,
                            OMX_CommandStateSet, OMX_StateLoaded, NULL))
        {
            LIMERR("unable to send video scheduler to loaded state.\n");
            goto bail;
        }

        if (OMX_ErrorNone !=
            OMX_SendCommand(player->video_sink_handle,
                            OMX_CommandStateSet, OMX_StateLoaded, NULL))
        {
            LIMERR("unable to send video sink to loaded state.\n");
            goto bail;
        }

        wait_for_flags
            ((player->video_decode_flags & (PSF_INVALID | PSF_LOADED))
             && (player->video_scheduler_flags & (PSF_INVALID | PSF_LOADED))
             && (player->video_sink_flags & (PSF_INVALID | PSF_LOADED)),
             ret);
        if (ret)
        {
            LIMERR("waiting for video LOADED timedout.\n");
            goto bail;
        }
    }

    wait_for_flags(player->demux_flags & (PSF_INVALID | PSF_LOADED), ret);
    if (ret)
    {
        LIMERR("waiting for demuxer LOADED timedout.\n");
        goto bail;
    }

    OMX_FreeHandle(player->demux_handle);
    if (player->audio_decode_handle)
    {
        OMX_FreeHandle(player->audio_decode_handle);
    }

    if (player->video_decode_handle)
    {
        OMX_FreeHandle(player->video_decode_handle);
        OMX_FreeHandle(player->video_scheduler_handle);
        OMX_FreeHandle(player->video_sink_handle);
    }

    res = XA_RESULT_SUCCESS;

 bail:
    obj_destroy(player);

    return res;
}

XAresult ox_player_pause (ox_player_t * player)
{
    int ret;
    XAresult res = XA_RESULT_PERMISSION_DENIED;

    if (player->audio_decode_handle)
    {
        set_flags(player->audio_sink_flags &= ~PSF_PAUSE);

        if (OMX_ErrorNone !=
            OMX_SendCommand(player->audio_sink_handle,
                            OMX_CommandStateSet, OMX_StatePause, NULL))
        {
            LIMERR("unable to pause audio sink.\n");
            goto bail;
        }

        wait_for_flags(player->audio_sink_flags & (PSF_INVALID | PSF_PAUSE),
                       ret);
        if (ret)
        {
            LIMERR("waiting for audio PAUSE timedout.\n");
            goto bail;
        }
    }

    if (player->video_decode_handle)
    {
        set_flags(player->video_scheduler_flags &= ~PSF_PAUSE);

        if (OMX_ErrorNone !=
            OMX_SendCommand(player->video_scheduler_handle,
                            OMX_CommandStateSet, OMX_StatePause, NULL))
        {
            LIMERR("unable to pause video scheduler.\n");
            goto bail;
        }

        wait_for_flags(player->video_scheduler_flags & (PSF_INVALID | PSF_PAUSE),
                       ret);
        if (ret)
        {
            LIMERR("waiting for video scheduler PAUSE timedout.\n");
            goto bail;
        }
    }

    res = XA_RESULT_SUCCESS;
 bail:
    return res;
}

XAresult ox_player_stop (ox_player_t * player)
{
    int ret;
    XAresult res = XA_RESULT_PERMISSION_DENIED;

    set_flags(player->demux_flags &= ~(PSF_IDLE | PSF_INVALID));
    if (OMX_ErrorNone !=
        OMX_SendCommand(player->demux_handle,
                        OMX_CommandStateSet, OMX_StateIdle, NULL))
    {
        LIMERR("unable to idle demuxer.\n");
        goto bail;
    }

    set_flags(player->clock_flags &= ~(PSF_IDLE | PSF_INVALID));
    if (OMX_ErrorNone !=
        OMX_SendCommand(player->clock_handle,
                        OMX_CommandStateSet, OMX_StateIdle, NULL))
    {
        LIMERR("unable to idle clock.\n");
        goto bail;
    }

    if (player->audio_decode_handle)
    {
        set_flags(player->audio_decode_flags &= ~(PSF_IDLE | PSF_INVALID);
                  player->audio_sink_flags &= ~(PSF_IDLE | PSF_INVALID));

        if (OMX_ErrorNone !=
            OMX_SendCommand(player->audio_decode_handle,
                            OMX_CommandStateSet, OMX_StateIdle, NULL))
        {
            LIMERR("unable to idle audio decoder.\n");
            goto bail;
        }

        if (OMX_ErrorNone !=
            OMX_SendCommand(player->audio_sink_handle,
                            OMX_CommandStateSet, OMX_StateIdle, NULL))
        {
            LIMERR("unable to idle audio sink.\n");
            goto bail;
        }

        wait_for_flags((player->audio_decode_flags & (PSF_INVALID | PSF_IDLE))
                       && (player->audio_sink_flags & (PSF_INVALID | PSF_IDLE)),
                       ret);
        if (ret)
        {
            LIMERR("waiting for audio IDLE timedout.\n");
            goto bail;
        }
    }

    if (player->video_decode_handle)
    {
        set_flags(player->video_decode_flags &= ~(PSF_IDLE | PSF_INVALID);
                  player->video_scheduler_flags &= ~(PSF_IDLE | PSF_INVALID);
                  player->video_sink_flags &= ~(PSF_IDLE | PSF_INVALID));

        if (OMX_ErrorNone !=
            OMX_SendCommand(player->video_decode_handle,
                            OMX_CommandStateSet, OMX_StateIdle, NULL))
        {
            LIMERR("unable to idle video decoder.\n");
            goto bail;
        }

        if (OMX_ErrorNone !=
            OMX_SendCommand(player->video_scheduler_handle,
                            OMX_CommandStateSet, OMX_StateIdle, NULL))
        {
            LIMERR("unable to idle video scheduler.\n");
            goto bail;
        }

        if (OMX_ErrorNone !=
            OMX_SendCommand(player->video_sink_handle,
                            OMX_CommandStateSet, OMX_StateIdle, NULL))
        {
            LIMERR("unable to idle video sink.\n");
            goto bail;
        }

        wait_for_flags
            ((player->video_decode_flags & (PSF_INVALID | PSF_IDLE))
             && (player->video_scheduler_flags & (PSF_INVALID | PSF_IDLE))
             && (player->video_sink_flags & (PSF_INVALID | PSF_IDLE)),
             ret);
        if (ret)
        {
            LIMERR("waiting for video IDLE timedout.\n");
            goto bail;
        }
    }

    wait_for_flags(player->demux_flags & (PSF_INVALID | PSF_IDLE), ret);
    if (ret)
    {
        LIMERR("waiting for demuxer IDLE timedout.\n");
        goto bail;
    }
    res = XA_RESULT_SUCCESS;

 bail:
    return res;
}

XAresult ox_player_get_position (ox_player_t * player, XAmillisecond * pos)
{
    OMX_TIME_CONFIG_TIMESTAMPTYPE ts;

    omxil_set_spec_version(&ts, sizeof(OMX_TIME_CONFIG_TIMESTAMPTYPE));
    if (OMX_ErrorNone == OMX_GetConfig
        (player->clock_handle, OMX_IndexConfigTimeCurrentMediaTime, &ts))
    {
        if (ts.nTimestamp < 0) ts.nTimestamp = 0;
        *pos = ts.nTimestamp / 1000;
    } else
    {
        LIMERR("unable to get current media time.\n");
        return XA_RESULT_PARAMETER_INVALID;
    }

    LIMDBG("position got as : %d\n", *pos);
    return XA_RESULT_SUCCESS;
}

XAresult ox_player_set_position (
    ox_player_t * player,
    XAmillisecond pos,
    XAuint32 seek_mode)
{
    int ret;
    XAresult res = XA_RESULT_PERMISSION_DENIED;
    OMX_TIME_CONFIG_CLOCKSTATETYPE st;
    OMX_TIME_CONFIG_TIMESTAMPTYPE ts;
    OMX_TIME_CONFIG_SEEKMODETYPE mode;

    //pause components.
    set_flags(player->demux_flags &= ~PSF_PAUSE);
    if (OMX_ErrorNone !=
        OMX_SendCommand(player->demux_handle,
                        OMX_CommandStateSet, OMX_StatePause, NULL))
    {
        LIMERR("unable to pause demuxer.\n");
        goto bail;
    }

    wait_for_flags(player->demux_flags & (PSF_INVALID | PSF_PAUSE),
                   ret);
    if (ret)
    {
        LIMERR("waiting for demux PAUSE timedout.\n");
        goto bail;
    }

    if (player->audio_decode_handle)
    {
        set_flags(player->audio_sink_flags &= ~PSF_PAUSE;
                  player->audio_decode_flags &= ~PSF_PAUSE);
        if (OMX_ErrorNone !=
            OMX_SendCommand(player->audio_sink_handle,
                            OMX_CommandStateSet, OMX_StatePause, NULL))
        {
            LIMERR("unable to pause audio sink.\n");
            goto bail;
        }

        if (OMX_ErrorNone !=
            OMX_SendCommand(player->audio_decode_handle,
                            OMX_CommandStateSet, OMX_StatePause, NULL))
        {
            LIMERR("unable to pause audio decoder.\n");
            goto bail;
        }

        wait_for_flags(player->audio_sink_flags & (PSF_INVALID | PSF_PAUSE)
                       && player->audio_decode_flags & (PSF_INVALID | PSF_PAUSE),
                       ret);
        if (ret)
        {
            LIMERR("waiting for audio PAUSE timedout.\n");
            goto bail;
        }
    }

    if (player->video_decode_handle)
    {
        set_flags(player->video_scheduler_flags &= ~PSF_PAUSE;
                  player->video_decode_flags &= ~PSF_PAUSE);

        if (OMX_ErrorNone !=
            OMX_SendCommand(player->video_scheduler_handle,
                            OMX_CommandStateSet, OMX_StatePause, NULL))
        {
            LIMERR("unable to pause video scheduler.\n");
            goto bail;
        }

        if (OMX_ErrorNone !=
            OMX_SendCommand(player->video_decode_handle,
                            OMX_CommandStateSet, OMX_StatePause, NULL))
        {
            LIMERR("unable to pause video decoder.\n");
            goto bail;
        }

        wait_for_flags(player->video_scheduler_flags & (PSF_INVALID | PSF_PAUSE)
                       && player->video_decode_flags & (PSF_INVALID | PSF_PAUSE),
                       ret);
        if (ret)
        {
            LIMERR("waiting for video PAUSE timedout.\n");
            goto bail;
        }
    }

    //stop clock component
    omxil_set_spec_version(&st, sizeof(OMX_TIME_CONFIG_CLOCKSTATETYPE));
    if (OMX_ErrorNone != OMX_GetConfig
        (player->clock_handle, OMX_IndexConfigTimeClockState, &st))
    {
        LIMERR("unable to get clock state.\n");
        goto bail;
    }
    st.eState = OMX_TIME_ClockStateStopped;
    if (OMX_ErrorNone != OMX_SetConfig
        (player->clock_handle, OMX_IndexConfigTimeClockState, &st))
    {
        LIMERR("unable to set clock state.\n");
        goto bail;
    }

    //seek to desired position
    omxil_set_spec_version(&mode, sizeof(OMX_TIME_SEEKMODETYPE));
    mode.eType = (seek_mode == XA_SEEKMODE_ACCURATE)?
        OMX_TIME_SeekModeAccurate : OMX_TIME_SeekModeFast;
    if (OMX_ErrorNone != OMX_SetConfig
        (player->demux_handle, OMX_IndexConfigTimeSeekMode, &mode))
    {
        LIMERR("unable to set seek mode\n");
        goto bail;
    }

    omxil_set_spec_version(&ts, sizeof(OMX_TIME_CONFIG_TIMESTAMPTYPE));
    ts.nTimestamp = (OMX_TICKS)pos * 1000;
    LIMDBG("seeking to %"PRId64"\n", ts.nTimestamp);
    if (OMX_ErrorNone != OMX_SetConfig
        (player->demux_handle, OMX_IndexConfigTimePosition, &ts))
    {
        LIMERR("unable to seek to position %"PRId64" in demuxer.\n",
               (OMX_TICKS)pos * 1000);
        goto continue_to_play;
    }

    //flush all components
    if (player->audio_decode_handle)
    {
        set_flags(player->demux_port_flags &= ~PSF_PORT_FLUSHED(0);
                  player->audio_sink_port_flags &= ~PSF_PORT_FLUSHED(0);
                  player->audio_decode_port_flags &=
                  ~(PSF_PORT_FLUSHED(0) | PSF_PORT_FLUSHED(1)));

        if (OMX_ErrorNone !=
            OMX_SendCommand(player->demux_handle,
                        OMX_CommandFlush, 0, NULL))
        {
            LIMERR("unable to flush demuxer audio port.\n");
            goto bail;
        }

        if (OMX_ErrorNone !=
            OMX_SendCommand(player->audio_sink_handle,
                            OMX_CommandFlush, 0, NULL))
        {
            LIMERR("unable to flush audio sink.\n");
            goto bail;
        }

        if (OMX_ErrorNone !=
            OMX_SendCommand(player->audio_decode_handle,
                            OMX_CommandFlush, OMX_ALL, NULL))
        {
            LIMERR("unable to flush audio decoder input.\n");
            goto bail;
        }
        wait_for_flags(player->demux_port_flags & PSF_PORT_FLUSHED(0)
                       && player->audio_sink_port_flags & PSF_PORT_FLUSHED(0)
                       && player->audio_decode_port_flags &
                       (PSF_PORT_FLUSHED(0) | PSF_PORT_FLUSHED(1)),
                       ret);
        if (ret)
        {
            LIMERR("waiting for audio flush timedout.\n");
            goto bail;
        }
    }

    if (player->video_decode_handle)
    {
        set_flags(player->demux_port_flags &= ~PSF_PORT_FLUSHED(1);
                  player->video_scheduler_port_flags &= ~PSF_PORT_FLUSHED(0);
                  player->video_decode_port_flags &=
                  ~(PSF_PORT_FLUSHED(0) | PSF_PORT_FLUSHED(1)));

        if (OMX_ErrorNone !=
            OMX_SendCommand(player->demux_handle,
                            OMX_CommandFlush, 1, NULL))
        {
            LIMERR("unable to flush demuxer video port.\n");
            goto bail;
        }

        if (OMX_ErrorNone !=
            OMX_SendCommand(player->video_scheduler_handle,
                            OMX_CommandFlush, 0, NULL))
        {
            LIMERR("unable to flush video scheduler.\n");
            goto bail;
        }

        if (OMX_ErrorNone !=
            OMX_SendCommand(player->video_decode_handle,
                            OMX_CommandFlush, OMX_ALL, NULL))
        {
            LIMERR("unable to flush video decoder.\n");
            goto bail;
        }
        wait_for_flags(player->demux_port_flags & PSF_PORT_FLUSHED(1)
                       && player->video_scheduler_port_flags
                       & PSF_PORT_FLUSHED(0)
                       && player->video_decode_port_flags &
                       (PSF_PORT_FLUSHED(0) | PSF_PORT_FLUSHED(1)),
                       ret);
        if (ret)
        {
            LIMERR("waiting for audio flush timedout.\n");
            goto bail;
        }
    }

 continue_to_play:
    //restart clock
    st.eState = OMX_TIME_ClockStateRunning;
    if (OMX_ErrorNone != OMX_SetConfig
        (player->clock_handle, OMX_IndexConfigTimeClockState, &st))
    {
        LIMERR("unable to set clock state.\n");
        goto bail;
    }

    //unpause components
    set_flags(player->demux_flags &= ~PSF_EXECUTING);
    if (OMX_ErrorNone !=
        OMX_SendCommand(player->demux_handle,
                        OMX_CommandStateSet, OMX_StateExecuting, NULL))
    {
        LIMERR("unable to unpause demuxer.\n");
        goto bail;
    }

    wait_for_flags(player->demux_flags & (PSF_INVALID | PSF_EXECUTING),
                   ret);
    if (ret)
    {
        LIMERR("waiting for demux unpause timedout.\n");
        goto bail;
    }

    if (player->audio_decode_handle)
    {
        set_flags(player->audio_sink_flags &= ~PSF_EXECUTING;
                  player->audio_decode_flags &= ~PSF_EXECUTING);

        if (OMX_ErrorNone !=
            OMX_SendCommand(player->audio_sink_handle,
                            OMX_CommandStateSet, OMX_StateExecuting, NULL))
        {
            LIMERR("unable to unpause audio sink.\n");
            goto bail;
        }

        if (OMX_ErrorNone !=
            OMX_SendCommand(player->audio_decode_handle,
                            OMX_CommandStateSet, OMX_StateExecuting, NULL))
        {
            LIMERR("unable to unpause audio decoder.\n");
            goto bail;
        }


        wait_for_flags(player->audio_sink_flags & (PSF_INVALID | PSF_EXECUTING)
                       && player->audio_decode_flags
                       & (PSF_INVALID | PSF_EXECUTING),
                       ret);
        if (ret)
        {
            LIMERR("waiting for audio unpause timedout.\n");
            goto bail;
        }
    }

    if (player->video_decode_handle)
    {
        set_flags(player->video_scheduler_flags &= ~PSF_EXECUTING;
                  player->video_decode_flags &= ~PSF_EXECUTING);

        if (OMX_ErrorNone !=
            OMX_SendCommand(player->video_scheduler_handle,
                            OMX_CommandStateSet, OMX_StateExecuting, NULL))
        {
            LIMERR("unable to unpause video scheduler.\n");
            goto bail;
        }

        if (OMX_ErrorNone !=
            OMX_SendCommand(player->video_decode_handle,
                            OMX_CommandStateSet, OMX_StateExecuting, NULL))
        {
            LIMERR("unable to unpause video decoder.\n");
            goto bail;
        }

        wait_for_flags(player->video_scheduler_flags
                       & (PSF_INVALID | PSF_EXECUTING)
                       && player->video_decode_flags
                       & (PSF_INVALID | PSF_EXECUTING),
                       ret);
        if (ret)
        {
            LIMERR("waiting for video unpause timedout.\n");
            goto bail;
        }
    }

    res = XA_RESULT_SUCCESS;

 bail:
    return res;
}
