/*
  Copyright (C) 2009 - 2010 SWOAG Technology <www.swoag.com>

  This program is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free Software
  Foundation, version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  LIMOX IPC packet parser routine for player sub-system.

  REVISION:

    1) initial creation. ----------------------------------------- 2009-09-17 MG
*/

#include "limoa-log.h"
#include "limipc.h"
#include "limthread.h"
#include "limmemory.h"
#include "limqlist.h"
#include "OpenMAXAL.h"
#include "ox-session.h"
#include "common-be.h"
#include "player-be.h"

static inline ox_player_t * player_init (ox_session_t * ox)
{
    ox_player_t * player = NULL;

    player = lim_mem_pool_get_buf(ox->pool, sizeof(ox_player_t));
    if (player) memset(player, 0, sizeof(ox_player_t));

    ox->player = player;
    player->ox = ox;

    return player;
}

static inline int player_create (
    ox_session_t * ox,
    ipc_packet_t * rqst,
    ipc_packet_t * rsps)
{
    ox_rqst_player_create_t * player_rqst;
    ipc_standard_rsps_t * player_rsps;
    void * ptrdiff;
    void * dptr;
    ox_player_t * player;
    int bytes;

    player_rsps = ipc_init_standard_response(rsps);

    player = player_init(ox);
    if (NULL == player)
    {
        LIMERR("unable to initialize player.\n");
        player_rsps->result = XA_RESULT_MEMORY_FAILURE;
        goto bail;
    }

    dptr = player->source_sink_dptr = lim_mem_pool_get_buf
        (ox->pool, rqst->header.bytes);
    if (NULL == player->source_sink_dptr)
    {
        LIMERR("unable to allocate memory for player controls.\n");
        player_rsps->result = XA_RESULT_MEMORY_FAILURE;
        goto bail;
    }

    player_rqst = (ox_rqst_player_create_t *)rqst->dptr;
    ptrdiff = NULL + (rqst->dptr - player_rqst->ptrbase);

    dptr += deep_cp_data_source(&player->data_src, &player_rqst->data_src,
                                dptr, ptrdiff);
    dptr += deep_cp_data_source(&player->bank_src, &player_rqst->bank_src,
                                dptr, ptrdiff);
    dptr += deep_cp_data_sink(&player->audio_snk, &player_rqst->audio_snk,
                              dptr, ptrdiff);
    dptr += deep_cp_data_sink(&player->image_video_snk,
                              &player_rqst->image_video_snk,
                              dptr, ptrdiff);
    dptr += deep_cp_data_sink(&player->vibra_snk, &player_rqst->vibra_snk,
                              dptr, ptrdiff);
    dptr += deep_cp_data_sink(&player->LED_array_snk,
                              &player_rqst->LED_array_snk,
                              dptr, ptrdiff);

    bytes = dptr - player->source_sink_dptr;
    ASSERT(bytes <= rqst->header.bytes);

 bail:
    return 0;
}

static inline int player_realize (
    ox_session_t * ox,
    ipc_packet_t * rqst,
    ipc_packet_t * rsps)
{
    ox_player_t * player = (ox_player_t *)ox->player;
    ox_rqst_player_realize_t * player_rqst;
    ipc_standard_rsps_t * player_rsps;

    player_rsps = ipc_init_standard_response(rsps);

    if (NULL == player)
    {
        LIMERR("player is not available.\n");
        player_rsps->result = XA_RESULT_PRECONDITIONS_VIOLATED;
        goto bail;
    }

    player_rqst = rqst->dptr;
    if (player_rqst->async == XA_BOOLEAN_TRUE)
    {
        //TODO: async realization, spawn off a thread here ...
        LIMWARN("async realization of player not supported.\n");
    }
    player_rsps->result = ox_player_realize(player);
 bail:
    return 0;
}

static inline int player_destroy (
    ox_session_t * ox,
    ipc_packet_t * rqst,
    ipc_packet_t * rsps)
{
    ox_player_t * player = (ox_player_t *)ox->player;
    int ret = -1;

    rsps->header.bytes = 0;
    if (NULL == player)
    {
        LIMERR("player is not available.\n");
        goto bail;
    }
    ret = ox_player_destroy(player);
 bail:
    return ret;
}

static inline int player_play (
    ox_session_t * ox,
    ipc_packet_t * rqst,
    ipc_packet_t * rsps)
{
    ox_player_t * player = (ox_player_t *)ox->player;
    ipc_standard_rsps_t * player_rsps;

    player_rsps = ipc_init_standard_response(rsps);

    if (NULL == player)
    {
        LIMERR("player is not available.\n");
        player_rsps->result = XA_RESULT_PERMISSION_DENIED;
        goto bail;
    }

    player_rsps->result = ox_player_play(player);
 bail:
    return 0;
}

static inline int player_pause (
    ox_session_t * ox,
    ipc_packet_t * rqst,
    ipc_packet_t * rsps)
{
    ox_player_t * player = (ox_player_t *)ox->player;
    ipc_standard_rsps_t * player_rsps;

    player_rsps = ipc_init_standard_response(rsps);

    if (NULL == player)
    {
        LIMERR("player is not available.\n");
        player_rsps->result = XA_RESULT_PERMISSION_DENIED;
        goto bail;
    }

    player_rsps->result = ox_player_pause(player);
 bail:
    return 0;
}

static inline int player_stop (
    ox_session_t * ox,
    ipc_packet_t * rqst,
    ipc_packet_t * rsps)
{
    ox_player_t * player = (ox_player_t *)ox->player;
    ipc_standard_rsps_t * player_rsps;

    player_rsps = ipc_init_standard_response(rsps);

    if (NULL == player)
    {
        LIMERR("player is not available.\n");
        player_rsps->result = XA_RESULT_PERMISSION_DENIED;
        goto bail;
    }

    player_rsps->result = ox_player_stop(player);
 bail:
    return 0;
}

static inline int player_get_position (
    ox_session_t * ox,
    ipc_packet_t * rqst,
    ipc_packet_t * rsps)
{
    ox_player_t * player = (ox_player_t *)ox->player;
    ox_rsps_player_get_position_t * player_rsps;

    player_rsps = ipc_init_response(rsps, sizeof(ox_rsps_player_get_position_t));
    if (NULL == player)
    {
        LIMERR("player is not available.\n");
        player_rsps->result = XA_RESULT_PERMISSION_DENIED;
        goto bail;
    }

    player_rsps->result = ox_player_get_position(player, &player_rsps->pos);

 bail:
    return 0;
}

static inline int player_set_position (
    ox_session_t * ox,
    ipc_packet_t * rqst,
    ipc_packet_t * rsps)
{
    ox_player_t * player = (ox_player_t *)ox->player;
    ox_rqst_player_set_position_t * player_rqst;
    ipc_standard_rsps_t * player_rsps;

    player_rsps = ipc_init_standard_response(rsps);
    if (NULL == player)
    {
        LIMERR("player is not available.\n");
        player_rsps->result = XA_RESULT_PERMISSION_DENIED;
        goto bail;
    }

    player_rqst = rqst->dptr;
    player_rsps->result = ox_player_set_position
        (player, player_rqst->pos, player_rqst->seek_mode);

 bail:
    return 0;
}

static inline int player_register_callback (
    ox_session_t * ox,
    ipc_packet_t * rqst,
    ipc_packet_t * rsps)
{
    ox_player_t * player = (ox_player_t *)ox->player;
    ox_rqst_player_register_callback_t * player_rqst;
    int ret = -1;

    rsps->header.bytes = 0;
    if (NULL == player)
    {
        LIMERR("player is not available.\n");
        goto bail;
    }
    player_rqst = (ox_rqst_player_register_callback_t *)rqst->dptr;
    player->callback.caller = player_rqst->caller;
    player->callback.func = player_rqst->func;
    player->callback.context = player_rqst->context;

    ret = 0;
 bail:
    return ret;
}

static inline int player_set_callback_events_mask (
    ox_session_t * ox,
    ipc_packet_t * rqst,
    ipc_packet_t * rsps)
{
    ox_player_t * player = (ox_player_t *)ox->player;
    ox_rqst_player_set_callback_events_mask_t * player_rqst;
    int ret = -1;

    rsps->header.bytes = 0;
    if (NULL == player)
    {
        LIMERR("player is not available.\n");
        goto bail;
    }
    player_rqst = (ox_rqst_player_set_callback_events_mask_t *)rqst->dptr;
    player->callback_events_mask = player_rqst->mask;

    ret = 0;
 bail:
    return ret;
}

int player_parse_packet (
    ox_session_t * ox,
    ipc_packet_t * rqst,
    ipc_packet_t * rsps)
{
    switch (rqst->header.id.value)
    {
    case OX_RQST_PLAYER_CREATE :
        return player_create(ox, rqst, rsps);

    case OX_RQST_PLAYER_REALIZE :
        return player_realize(ox, rqst, rsps);

    case OX_RQST_PLAYER_DESTROY :
        return player_destroy(ox, rqst, rsps);

    case OX_RQST_PLAYER_PLAY :
        return player_play(ox, rqst, rsps);

    case OX_RQST_PLAYER_STOP :
        return player_stop(ox, rqst, rsps);

    case OX_RQST_PLAYER_PAUSE :
        return player_pause(ox, rqst, rsps);

    case OX_RQST_PLAYER_GET_POSITION :
        return player_get_position(ox, rqst, rsps);

    case OX_RQST_PLAYER_SET_POSITION :
        return player_set_position(ox, rqst, rsps);

    case OX_RQST_PLAYER_REGISTER_CALLBACK :
        return player_register_callback(ox, rqst, rsps);

    case OX_RQST_PLAYER_SET_CALLBACK_EVENTS_MASK :
        return player_set_callback_events_mask(ox, rqst, rsps);

    default :
        LIMWARN("unkown player message, segment: %d, id: %d\n",
               MSG_SEGMENT(rqst->header.id.value),
               MSG_ID(rqst->header.id.value));
        rsps->header.bytes = 0;
        break;
    }
    return 0;
}
