/*
  Copyright (C) 2010 SWOAG Technology <www.swoag.com>

  This program is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free Software
  Foundation, version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  LIMOX session stub for thread communication.

  REVISION:

    1) initial creation. ----------------------------------------- 2010-04-06 MG
*/

#include "limoa-log.h"
#include "limipc.h"
#include "limthread.h"
#include "limmemory.h"
#include "OpenMAXAL.h"
#include "ox-session.h"
#include "ox-main.h"
#include "engine-fe.h"
#include "OpenMAXAL.h"
#include "OMX_Core.h"

#define DEFAULT_SESSION_POOL_BYTES (1024 * 32)

typedef struct
{
    int fd;
    lim_mem_pool_t pool;
} ipc_session_internal_t;

static ox_session_t backend_ox;
static engine_ipc_t * frontend_ipc;

//this stub is called from frontend at engine initialization.
int ox_session_init (void * arg)
{
    frontend_ipc = (engine_ipc_t *)arg;

    memset(&backend_ox, 0, sizeof(ox_session_t));
    if (OMX_ErrorNone != OMX_Init())
    {
        LIMERR("unable to initialize OMX_IL core.\n");
        return -1;
    }

    backend_ox.pool = lim_mem_pool_create(DEFAULT_SESSION_POOL_BYTES);
    if (NULL == backend_ox.pool)
    {
        LIMERR("unable to allocate session memory.\n");
        OMX_Deinit();
        return -1;
    }

    return 0;
}

//this stub is called from frontend at engine shutdown.
void ox_session_destroy (ox_session_t * ox)
{
    if (backend_ox.pool)
    {
        lim_mem_pool_destroy(backend_ox.pool);
    }

    if (OMX_ErrorNone != OMX_Deinit())
    {
        LIMWARN("unable to unload OMX_IL core.\n");
    } else
    {
        LIMDBG("OMX IL core unloaded.\n");
    }
}

//call to frontend packet parser without holding any mutex.
int ox_session_send_packet (ox_session_t * ox, ipc_packet_t * pkt)
{
    engine_callback_parser(pkt);
    return 0;
}

//this stub is called from frontend packet sender, with ox = NULL.
//caller must have held ipc->mutex
int ox_session_parse_packet (
    ox_session_t * ox,
    ipc_packet_t * pkt,
    ipc_packet_t * rsps_pkt)
{
    ipc_packet_t rsps;
#define _MAXIMUM_RESPONSE_BYTES    (1024 * 4)
    // hard-coded max reponse size for both coding ase and efficiency
    // subsequent call to fill the response buffer must check to make
    // sure buffer is big enough for that particular response, if not,
    // buffer can be 'malloc'ed and automatically gets 'free'ed here.
    void * buf[_MAXIMUM_RESPONSE_BYTES / sizeof(void *)];
    int ret = 0;

    rsps.header.bytes = _MAXIMUM_RESPONSE_BYTES;
    rsps.header.id.direction = LIMOA_IPC_RESPONSE_TO_CLIENT;
    rsps.dptr = (void *)&buf[0];

    ret = ox_parse_packet(&backend_ox, pkt, &rsps);

    if (ret)
    {
        LIMDBG("unable to parse packet, NAK it.\n");
        rsps.header.id.value = LIMOA_IPC_NAK;
        goto bail;
    } else
    {
        rsps.header.id.value = pkt->header.id.value;
    }

    if (rsps_pkt)
    {
        ipc_session_internal_t * hdl;
        hdl = (ipc_session_internal_t *)frontend_ipc->sid;

        *rsps_pkt = rsps;
        if (rsps.header.bytes)
        {
            rsps_pkt->dptr = lim_mem_pool_get_buf(hdl->pool, rsps.header.bytes);
            if (NULL == rsps_pkt->dptr)
            {
                LIMERR("unable to allocate IPC packet memory.\n");
                ret = -1;
                goto bail;
            }
            memcpy(rsps_pkt->dptr, rsps.dptr, rsps.header.bytes);
        }
    }

    // see above notes, buffer must have been 'malloc'ed by parser
    if (rsps.header.bytes > _MAXIMUM_RESPONSE_BYTES)
    {
        free(rsps.dptr);
    }

 bail:
    return ret;
}


