/*
  Copyright (C) 2009 - 2010 SWOAG Technology <www.swoag.com>

  This program is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free Software
  Foundation, version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  LIMOX session management routines.

  REVISION:

    1) initial creation. ----------------------------------------- 2009-08-30 MG
*/

#include "limoa-log.h"
#include "limipc.h"
#include "limthread.h"
#include "limmemory.h"
#include "OpenMAXAL.h"
#include "ox-session.h"
#include "ox-main.h"

#define DEFAULT_SESSION_POOL_BYTES (1024 * 32)
static list_t session_list_head;

int ox_session_init (void * arg)
{
    list_init(&session_list_head);
    return 0;
}

list_t * ox_session_list_head (void)
{
    return &session_list_head;
}

int ox_session_create (int fd)
{
    ox_session_t * ox;

    ox = (ox_session_t *)malloc(sizeof(ox_session_t));
    if (NULL == ox)
    {
        LIMERR("unable to create session.\n");
        return -1;
    }

    ox->pool = lim_mem_pool_create(DEFAULT_SESSION_POOL_BYTES);
    if (NULL == ox->pool)
    {
        LIMERR("unable to allocate session memory.\n");
        free(ox);
        return -1;
    }

    ox->ipc.sid = ipc_session_create(fd);
    if (NULL == ox->ipc.sid)
    {
        LIMERR("unable to open IPC session.\n");
        lim_mem_pool_destroy(ox->pool);
        free(ox);
        return -1;
    }
    ox->ipc.sock_fd = fd;
    lthread_mutex_init(&ox->ipc.mutex);

    list_init(&ox->list);
    list_add(&ox->list, &session_list_head);

    return 0;
}

void ox_session_destroy (ox_session_t * ox)
{
    if (NULL == ox) return;

    if (ox->ipc.sid)
    {
        ipc_session_destroy(ox->ipc.sid);
        lthread_mutex_destroy(&ox->ipc.mutex);
    }

    if (ox->pool)
    {
        lim_mem_pool_destroy(ox->pool);
    }

    list_del(&ox->list);
    free(ox);
}

int ox_session_send_packet (ox_session_t * ox, ipc_packet_t * pkt)
{
    int ret;

    lthread_mutex_lock(&ox->ipc.mutex);
    pkt->header.id.direction = LIMOA_IPC_CALLBACK_TO_CLIENT;
    ret = ipc_write_packet(ox->ipc.sid, pkt);
    lthread_mutex_unlock(&ox->ipc.mutex);

    return ret;
}

int ox_session_parse_packet (
    ox_session_t * ox,
    ipc_packet_t * pkt,
    ipc_packet_t * rsps_pkt)
{
    ipc_packet_t rsps;
#define _MAXIMUM_RESPONSE_BYTES    (1024 * 4)
    // hard-coded max reponse size for both coding ase and efficiency
    // subsequent call to fill the response buffer must check to make
    // sure buffer is big enough for that particular response, if not,
    // buffer can be 'malloc'ed and automatically gets 'free'ed here.
    void * buf[_MAXIMUM_RESPONSE_BYTES / sizeof(void *)];
    int ret = 0;

    if (pkt->header.id.direction != LIMOA_IPC_REQUEST_TO_SERVER)
    {
        LIMWARN("invalid IPC packet detected and discarded.\n");
        goto bail;
    }

    rsps.header.bytes = _MAXIMUM_RESPONSE_BYTES;
    rsps.header.id.direction = LIMOA_IPC_RESPONSE_TO_CLIENT;
    rsps.dptr = (void *)&buf[0];

    ret = ox_parse_packet(ox, pkt, &rsps);

    if (ret)
    {
        LIMDBG("unable to parse packet, NAK it.\n");
        rsps.header.id.value = LIMOA_IPC_NAK;
        goto bail;
    } else
    {
        rsps.header.id.value = pkt->header.id.value;
    }

    ipc_write_packet(ox->ipc.sid, &rsps);

    // see above notes, buffer must have been 'malloc'ed by parser
    if (rsps.header.bytes > _MAXIMUM_RESPONSE_BYTES)
    {
        free(rsps.dptr);
    }

 bail:
    return ret;
}
