/*
  Copyright (C) 2009 - 2010 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX AL ipc implementation.

  REVISION:

    1) initial creation. ----------------------------------------- 2009-08-20 MG
*/

#include "limlog.h"
#include "limipc.h"
#include "limthread.h"
#include "limmemory.h"

typedef struct
{
    int fd;
    lim_mem_pool_t pool;
} ipc_session_internal_t;

static int read_bytes (int fd, void * buf, int bytes)
{
    int rd_bytes = bytes;
    int ret;

    while (rd_bytes)
    {
        ret = read(fd, buf, rd_bytes);
        if (0 > ret) return ret;
        if (0 == ret) break;
        rd_bytes -= ret;
        buf += ret;
    }
    return (bytes - rd_bytes);
}

static void discard_bytes (int fd, int bytes)
{
    char tmp;
    while(bytes--) if(read(fd, &tmp, 1)){};
}

static int write_bytes (int fd, void * buf, int bytes)
{
    int wr_bytes = bytes;
    int ret;

    while (wr_bytes)
    {
        ret = write(fd, buf, wr_bytes);
        if (0 > ret) return ret;
        if (0 == ret) break;
        wr_bytes -= ret;
        buf += ret;
    }
    return (bytes - wr_bytes);
}

ipc_session_t ipc_session_create (int fd)
{
    ipc_session_internal_t * hdl;

    hdl = (ipc_session_internal_t *)malloc(sizeof(ipc_session_internal_t));
    if (hdl)
    {
        hdl->fd = fd;
        hdl->pool = lim_mem_pool_create(DEFAULT_IPC_PACKET_POOL_BYTES);
        if (NULL == hdl->pool)
        {
            LIMERR("unable to allocate session memory pool.\n");
            free(hdl);
            hdl = NULL;
        }
    } else
    {
        LIMERR("unabled to allocate session.\n");
    }

    if (NULL == hdl && -1 != fd)
    {
        close(fd);
    }
    return (ipc_session_t)hdl;
}

void ipc_session_destroy (ipc_session_t sid)
{
    ipc_session_internal_t * hdl = (ipc_session_internal_t *)sid;

    if (hdl)
    {
        if (hdl->pool) lim_mem_pool_destroy(hdl->pool);
        if (-1 != hdl->fd) close(hdl->fd);
        free(hdl);
    }
}

int ipc_read_packet (ipc_session_t sid, ipc_packet_t * pkt)
{
    int ret;
    ipc_session_internal_t * hdl = (ipc_session_internal_t *)sid;

    pkt->dptr = NULL;
    ret = read_bytes(hdl->fd, &pkt->header, sizeof(ipc_packet_header_t));
    if ( 0 >= ret)
    {
        //LIMERR("unable to read IPC header.\n");
        return ret;
    }
    ASSERT(ret == sizeof(ipc_packet_header_t));

    if (pkt->header.bytes)
    {
        pkt->dptr = lim_mem_pool_get_buf(hdl->pool, pkt->header.bytes);
        if (NULL == pkt->dptr)
        {
            LIMERR("unable to allocate IPC packet memory.\n");
            discard_bytes(hdl->fd, pkt->header.bytes);
            return -1;
        }
        ret = read_bytes(hdl->fd, pkt->dptr, pkt->header.bytes);
        if ( 0 >= ret)
        {
            LIMERR("unable to read IPC body.\n");
            return ret;
        }
        ASSERT(ret == pkt->header.bytes);
    }

    return ret;
}

int ipc_write_packet (ipc_session_t sid, ipc_packet_t * pkt)
{
    int ret;
    ipc_session_internal_t * hdl = (ipc_session_internal_t *)sid;

    ret = write_bytes(hdl->fd, &pkt->header, sizeof(ipc_packet_header_t));

    if ( 0 >= ret)
    {
        LIMERR("unable to write IPC header.\n");
        return ret;
    }
    ASSERT(ret == sizeof(ipc_packet_header_t));

    if (pkt->header.bytes)
    {
        ret = write_bytes(hdl->fd, pkt->dptr, pkt->header.bytes);
        if ( 0 >= ret)
        {
            LIMERR("unable to write IPC body.\n");
            return ret;
        }
        ASSERT(ret == pkt->header.bytes);
    }
    return ret;
}

void ipc_release_packet (ipc_session_t sid, ipc_packet_t * pkt)
{
    ipc_session_internal_t * hdl = (ipc_session_internal_t *)sid;
    if (pkt) if (pkt->dptr) lim_mem_pool_put_buf(hdl->pool, pkt->dptr);
}
