/*
  Copyright (C) 2009 - 2010 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX AL ENGINE: ipc implementation.

  REVISION:

    1) initial creation. ----------------------------------------- 2009-09-03 MG
*/

#include "limoa-log.h"
#include "limipc.h"
#include "limthread.h"
#include "OpenMAXAL.h"
#include "engine-fe.h"

#undef LIMDBG
#ifndef LIMDBG
#define LIMDBG(fmt, arg...)
#endif

#define ENGINE_MAX_QUEUED_PACKETS 64

static int connect2limox (void)
{
    int ret;
    int fd = -1;
    struct sockaddr_un un;
    socklen_t len;

    fd = socket(AF_UNIX, SOCK_STREAM, 0);
    if (fd <= 0)
    {
        LIMERR("unable to create socket.\n");
        ret = -1;
        goto bail;
    }

    memset(&un, 0, sizeof(un));
    un.sun_family = AF_UNIX;
    sprintf(un.sun_path, "%s%05d", LIMOA_SOCKET_PATH, getpid());
    len = offsetof(struct sockaddr_un, sun_path) + strlen(un.sun_path);

    unlink(un.sun_path);
    if (bind(fd, (struct sockaddr *)&un, len) < 0)
    {
        LIMERR("unable to bind socket.\n");
        ret = -2;
        goto bail;
    }

    memset(&un, 0, sizeof(un));
    un.sun_family = AF_UNIX;
    strcpy(un.sun_path, LIMOX_SOCKET_PATH);
    len = offsetof(struct sockaddr_un, sun_path) + strlen(un.sun_path);

    if (connect(fd, (struct sockaddr *)&un, len) < 0)
    {
        LIMERR("unable to connect to limox.\n");
        ret = -3;
        goto bail;
    }

    return fd;

 bail:
    if (fd > 0) close(fd);
    return ret;
}

static void * callback (void * arg)
{
    fd_set rfds, rfds_save;
    engine_t * engine = (engine_t *)arg;
    engine_ipc_t * ipc = &engine->ipc;

    FD_ZERO(&rfds_save);
    FD_SET(ipc->sock_fd, &rfds_save);

    while (1)
    {
        ipc_packet_t * pkt;
        struct timeval tv;
        int ret;

        lthread_mutex_lock(&ipc->mutex);
        if (ipc->destroy)
        {
            ipc->destroy = 0;
            lthread_cond_broadcast(&ipc->cond);
            lthread_mutex_unlock(&ipc->mutex);
            break;
        }
        lthread_mutex_unlock(&ipc->mutex);

        tv.tv_sec = 0;
        tv.tv_usec = 100000;
        rfds = rfds_save;
        ret = select(ipc->sock_fd + 1, &rfds, NULL, NULL, &tv);
        if (-1 == ret)
        {
            LIMWARN("oops, unable to monitor socket fd.\n");
            continue;
        } else if (0 == ret) continue;

        lthread_mutex_lock(&ipc->mutex);
        if (ipc->sock_read_completed)
        {
            LIMDBG("socket reading done already.\n");
            ipc->sock_read_completed = 0;
        } else
        {
            pkt = limq_enqueue_get(ipc->pkt_q);
            ret = ipc_read_packet(ipc->sid, pkt);
            if (0 > ret)
            {
                LIMERR("packet read error.\n");
                lthread_mutex_unlock(&ipc->mutex);
                break;
            } else if (0 == ret)
            {
                LIMERR("limox shut down?\n");
                lthread_mutex_unlock(&ipc->mutex);
                //TODO: gracefully shut down
                exit(-1);
                break;
            }
            limq_enqueue_put(ipc->pkt_q, pkt);
        }

        // examine callback request queue
        while ((pkt = limq_dequeue_get(ipc->pkt_q)))
        {
            if (pkt->header.id.direction != LIMOA_IPC_CALLBACK_TO_CLIENT)
            {
                LIMWARN("unexpected packet.\n");
            } else
            {
                lthread_mutex_unlock(&ipc->mutex);
                engine_callback_parser(pkt);
                lthread_mutex_lock(&ipc->mutex);
            }
            limq_dequeue_put(ipc->pkt_q, pkt);
            ipc_release_packet(ipc->sid, pkt);
        }

        lthread_mutex_unlock(&ipc->mutex);
    }
    return NULL;
}

int engine_ipc_init (engine_t * engine)
{
    int ret = -1;
    engine_ipc_t * ipc = &engine->ipc;

    lthread_mutex_init(&ipc->mutex);
    lthread_cond_init(&ipc->cond);
    ipc->sock_read_completed = 0;

    ipc->sock_fd = connect2limox();
    if (ipc->sock_fd <= 0)
    {
        LIMERR("unable to reach limox server.\n");
        goto bail1;
    }

    ipc->sid = ipc_session_create(ipc->sock_fd);
    if (NULL == ipc->sid)
    {
        LIMERR("unable to initialize session.\n");
        goto bail1;
    }

    ipc->pkt_q = limq_create(sizeof(ipc_packet_t),
                             ENGINE_MAX_QUEUED_PACKETS,
                             "LIMOA IPC_PACKET QUEUE");
    if (NULL == ipc->pkt_q)
    {
        LIMERR("unable to initialize session.\n");
        goto bail2;
    }

    ipc->destroy = 0;
    if (lthread_create(&ipc->cb_thread, callback, engine, 1))
    {
        LIMERR("unable to create server callback thread.\n");
        goto bail3;
    }

    return 0;

 bail3:
    limq_destroy(ipc->pkt_q);
 bail2:
    ipc_session_destroy(ipc->sid);
 bail1:
    return ret;
}

int engine_ipc_destroy (engine_t * engine)
{
    engine_ipc_t * ipc = &engine->ipc;

    lthread_mutex_lock(&ipc->mutex);
    ipc->destroy = 1;
    while(1 == ipc->destroy)
    {
        lthread_cond_wait(&ipc->cond, &ipc->mutex);
    }
    lthread_mutex_unlock(&ipc->mutex);

    lthread_mutex_destroy(&ipc->mutex);
    lthread_cond_destroy(&ipc->cond);

    ipc_session_destroy(ipc->sid);
    limq_destroy(ipc->pkt_q);

    return 0;
}

int engine_ipc_send_packet (
    engine_t * engine,
    ipc_packet_t * rqst,
    ipc_packet_t * rsps)
{
    engine_ipc_t * ipc = &engine->ipc;
    ipc_packet_t pkt;
    int ret = -1;

    lthread_mutex_lock(&ipc->mutex);

    rqst->header.id.direction = LIMOA_IPC_REQUEST_TO_SERVER;

    if (0 > ipc_write_packet(ipc->sid, rqst)) goto bail;

 retry:
    if (0 > ipc_read_packet(ipc->sid, &pkt)) goto bail;

    if (LIMOA_IPC_RESPONSE_TO_CLIENT == pkt.header.id.direction)
    {
        if (rqst->header.id.value == pkt.header.id.value)
        {
            if (rsps) *rsps = pkt;
            else ipc_release_packet(ipc->sid, &pkt);
            ret = 0;
            goto bail;
        }
        LIMWARN("message NAKed.\n");
    } else if (LIMOA_IPC_CALLBACK_TO_CLIENT != pkt.header.id.direction)
    {
        LIMERR("unknown packet, interface out of sequence?\n");
    } else
    {
        ipc_packet_t * qpkt;
        qpkt = limq_enqueue_get(ipc->pkt_q);
        *qpkt = pkt;
        limq_enqueue_put(ipc->pkt_q, qpkt);
        goto retry;
    }
    ipc_release_packet(ipc->sid, &pkt);

 bail:
    ipc->sock_read_completed = 1;
    lthread_mutex_unlock(&ipc->mutex);
    return ret;
}

int engine_ipc_ping (engine_t * engine)
{
    ipc_packet_t pkt;

    pkt.header.id.value = OX_RQST_ENGINE_PING;
    pkt.header.bytes = 0;
    return engine_ipc_send_packet(engine, &pkt, NULL);
}
