/*
  Copyright (C) 2009 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX AL ENGINE: object interface implementation.

  REVISION:

    1) initial creation. ----------------------------------------- 2009-08-15 MG
*/

#include "limoa-log.h"
#include "limipc.h"
#include "limthread.h"
#include "OpenMAXAL.h"
#include "engine-fe.h"

#undef LIMDBG
#ifndef LIMDBG
#define LIMDBG(fmt, arg...)
#endif

static void engine_log (
    const void * arg,
    int level,
    const void * log,
    int bytes)
{
    switch (level)
    {
    case LIM_LOG_FATAL :
    case LIM_LOG_ERROR :
    case LIM_LOG_WARNING :
        printf("%s", (char *)log);

    default :
        break;
    }
}

static XAresult Realize (
    XAObjectItf self,
    XAboolean async)
{
    XAresult res = XA_RESULT_UNKNOWN_ERROR;
    engine_t * engine = (engine_t *)self;

    if (engine->objstate != OBJ_STATE_UNREALIZED)
    {
        res = XA_RESULT_PRECONDITIONS_VIOLATED;
        goto bail;
    }

    engine->objstate = OBJ_STATE_REALIZED;
    res = XA_RESULT_SUCCESS;
 bail:
    return XA_RESULT_SUCCESS;
}

static XAresult Resume (
    XAObjectItf self,
    XAboolean async)
{
    LIMWARN("API not implemented.\n");
    return XA_RESULT_SUCCESS;
}

static XAresult GetState (
    XAObjectItf self,
    XAuint32 * pState)
{
    LIMWARN("API not implemented.\n");
    return XA_RESULT_SUCCESS;
}

static XAresult GetInterface (
    XAObjectItf self,
    const XAInterfaceID iid,
    void * pInterface)
{
    XAresult res = XA_RESULT_SUCCESS;
    engine_t * engine = (engine_t *)self;

    if (engine->objstate != OBJ_STATE_REALIZED)
    {
        res = XA_RESULT_PRECONDITIONS_VIOLATED;
        LIMWARN("%s\n", xa_error_string(res));
        goto bail;
    }

    if (XA_IID_ENGINE == iid)
    {
        *((XAEngineItf *)pInterface) = (XAEngineItf)&engine->engine_itf;
    } else
    {
        res = XA_RESULT_FEATURE_UNSUPPORTED;
        LIMWARN("%s\n", xa_error_string(res));
    }

 bail:
    return res;
}

static XAresult RegisterCallback (
    XAObjectItf self,
    xaObjectCallback callback,
    void * pContext)
{
    LIMWARN("API not implemented.\n");
    return XA_RESULT_SUCCESS;
}

static void AbortAsyncOperation (
    XAObjectItf self)
{
    LIMWARN("API not implemented.\n");
}

static void Destroy (
    XAObjectItf self)
{
    engine_t * engine = (engine_t *)self;

    engine_ipc_destroy(engine);
    free(engine);

    LIMDBG("engine destroyed.\n");
}

static XAresult SetPriority (
    XAObjectItf self,
    XAuint32 priority)
{
    LIMWARN("API not implemented.\n");
    return XA_RESULT_SUCCESS;
}

static XAresult GetPriority (
    XAObjectItf self,
    XAuint32 * pPriority)
{
    LIMWARN("API not implemented.\n");
    return XA_RESULT_SUCCESS;
}

static XAresult SetLossOfControlInterfaces (
    XAObjectItf self,
    XAuint16 numInterfaces,
    const XAInterfaceID * pInterfaceIDs,
    XAboolean enabled)
{
    LIMWARN("API not implemented.\n");
    return XA_RESULT_SUCCESS;
}

XAresult engine_obj_create (XAObjectItf * pEngine,
    XAuint32 numOptions,
    const XAEngineOption * pEngineOptions,
    XAuint32 numInterfaces,
    const XAInterfaceID * pInterfaceIds,
    const XAboolean * pInterfaceRequired)
{
    XAresult res;
    engine_t * engine;

    lim_log_init(1024 * 1024 * 2, LIM_LOG_DEBUG, engine_log, NULL);

    engine = (engine_t *)calloc(1, sizeof(engine_t));

    if (NULL == engine)
    {
        LIMERR("unable to allocate memory for engine object.\n");
        res = XA_RESULT_MEMORY_FAILURE;
        goto bail1;
    }
    engine->objstate = OBJ_STATE_UNREALIZED;
    engine->obj_itf = &engine->objItf;

    engine->obj_itf->Realize = Realize;
    engine->obj_itf->Resume = Resume;
    engine->obj_itf->GetState = GetState;
    engine->obj_itf->GetInterface = GetInterface;
    engine->obj_itf->RegisterCallback = RegisterCallback;
    engine->obj_itf->AbortAsyncOperation = AbortAsyncOperation;
    engine->obj_itf->Destroy = Destroy;
    engine->obj_itf->SetPriority = SetPriority;
    engine->obj_itf->GetPriority = GetPriority;
    engine->obj_itf->SetLossOfControlInterfaces = SetLossOfControlInterfaces;

    engine_interface_init(engine);

    if (engine_ipc_init(engine))
    {
        res = XA_RESULT_RESOURCE_ERROR;
        LIMWARN("%s\n", xa_error_string(res));
        goto bail2;
    }

    *pEngine = (XAObjectItf)&engine->obj_itf;
    return XA_RESULT_SUCCESS;

 bail2:
    free(engine);
 bail1:
    return res;
}
