/*
  Copyright (C) 2009 - 2010 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX AL OUTPUT-MIX: object interface implementation.

  REVISION:

    1) initial creation. ----------------------------------------- 2009-09-03 MG
*/

#include "limoa-log.h"
#include "limipc.h"
#include "limthread.h"
#include "OpenMAXAL.h"
#include "output-mix-fe.h"

//#undef LIMDBG
#ifndef LIMDBG
#define LIMDBG(fmt, arg...)
#endif

static inline int ox_omix_create (engine_t * engine)
{
    ipc_packet_t rqst;
    int ret = -1;

    ipc_init_request(&rqst, OX_RQST_OUTPUT_MIX_CREATE, 0, NULL);
    if (engine_ipc_send_packet(engine, &rqst, NULL))
    {
        LIMERR("unable to send output mix creation request to limox.\n");
        goto bail;
    }
    ret = 0;
 bail:
    return ret;
}

static inline int ox_omix_realize (output_mix_t * omx)
{
    engine_t * engine = omx->engine;
    ipc_packet_t rqst;
    int ret = -1;

    ipc_init_request(&rqst, OX_RQST_OUTPUT_MIX_REALIZE, 0, NULL);
    if (engine_ipc_send_packet(engine, &rqst, NULL))
    {
        LIMERR("unable to send output mix realize request to limox.\n");
        goto bail;
    }
    ret = 0;
 bail:
    return ret;
}

static inline int ox_omix_destroy (output_mix_t * omx)
{
    engine_t * engine = omx->engine;
    ipc_packet_t rqst;
    int ret = -1;

    ipc_init_request(&rqst, OX_RQST_OUTPUT_MIX_DESTROY, 0, NULL);
    if (engine_ipc_send_packet(engine, &rqst, NULL))
    {
        LIMERR("unable to send output mix destroy request to limox.\n");
        goto bail;
    }
    ret = 0;
 bail:
    return ret;
}

static XAresult Realize (
    XAObjectItf self,
    XAboolean async)
{
    XAresult res = XA_RESULT_UNKNOWN_ERROR;
    output_mix_t * omix = (output_mix_t *)self;

    if (omix->objstate != OBJ_STATE_UNREALIZED)
    {
        res = XA_RESULT_PRECONDITIONS_VIOLATED;
        LIMWARN("%s\n", xa_error_string(res));
        goto bail;
    }

    if (ox_omix_realize(omix))
    {
        LIMERR("unable to realize output mix.\n");
        goto bail;
    }

    omix->objstate = OBJ_STATE_REALIZED;
    res = XA_RESULT_SUCCESS;
 bail:
    return res;
}

static XAresult Resume (
    XAObjectItf self,
    XAboolean async)
{
    LIMWARN("API not implemented.\n");
    return XA_RESULT_SUCCESS;
}

static XAresult GetState (
    XAObjectItf self,
    XAuint32 * pState)
{
    LIMWARN("API not implemented.\n");
    return XA_RESULT_SUCCESS;
}

static XAresult GetInterface (
    XAObjectItf self,
    const XAInterfaceID iid,
    void * pInterface)
{
    XAresult res = XA_RESULT_SUCCESS;
    output_mix_t * omix = (output_mix_t *)self;

    if (omix->objstate != OBJ_STATE_REALIZED)
    {
        res = XA_RESULT_PRECONDITIONS_VIOLATED;
        LIMWARN("%s\n", xa_error_string(res));
        goto bail;
    }

    if (XA_IID_VOLUME == iid)
    {
    } else if (XA_IID_EQUALIZER == iid)
    {

    } else
    {
        LIMWARN("Feature is not supported.\n");
        res = XA_RESULT_FEATURE_UNSUPPORTED;
    }

 bail:
    return res;
}

static XAresult RegisterCallback (
    XAObjectItf self,
    xaObjectCallback callback,
    void * pContext)
{
    LIMWARN("API not implemented.\n");
    return XA_RESULT_SUCCESS;
}

static void AbortAsyncOperation (
    XAObjectItf self)
{
    LIMWARN("API not implemented.\n");
}

static void Destroy (
    XAObjectItf self)
{
    output_mix_t * omix = (output_mix_t *)self;
    if (ox_omix_destroy(omix))
    {
        LIMERR("unknown error upon omix destroy.\n");
    }
    free(omix);
    LIMDBG("output-mix destroyed.\n");
}

static XAresult SetPriority (
    XAObjectItf self,
    XAuint32 priority)
{
    LIMWARN("API not implemented.\n");
    return XA_RESULT_SUCCESS;
}

static XAresult GetPriority (
    XAObjectItf self,
    XAuint32 * pPriority)
{
    LIMWARN("API not implemented.\n");
    return XA_RESULT_SUCCESS;
}

static XAresult SetLossOfControlInterfaces (
    XAObjectItf self,
    XAuint16 numInterfaces,
    const XAInterfaceID * pInterfaceIDs,
    XAboolean enabled)
{
    LIMWARN("API not implemented.\n");
    return XA_RESULT_SUCCESS;
}

XAresult output_mix_obj_create (
    XAEngineItf self,
    XAObjectItf * pMix,
    XAuint32 numInterfaces,
    const XAInterfaceID * pInterfaceIds,
    const XAboolean * pInterfaceRequired)
{
    XAresult res;
    output_mix_t * omix;
    engine_t * engine = ((engine_itf_t *)(*self))->engine;

    omix = (output_mix_t *)calloc(1, sizeof(output_mix_t));
    if (NULL == omix)
    {
        LIMERR("unable to allocate memory for output mix object.\n");
        res = XA_RESULT_MEMORY_FAILURE;
        goto bail1;
    }
    omix->engine = engine;
    omix->objstate = OBJ_STATE_UNREALIZED;
    omix->obj_itf = &omix->objItf;

    omix->obj_itf->Realize = Realize;
    omix->obj_itf->Resume = Resume;
    omix->obj_itf->GetState = GetState;
    omix->obj_itf->GetInterface = GetInterface;
    omix->obj_itf->RegisterCallback = RegisterCallback;
    omix->obj_itf->AbortAsyncOperation = AbortAsyncOperation;
    omix->obj_itf->Destroy = Destroy;
    omix->obj_itf->SetPriority = SetPriority;
    omix->obj_itf->GetPriority = GetPriority;
    omix->obj_itf->SetLossOfControlInterfaces = SetLossOfControlInterfaces;

    //TODO: validate interfaces.

    // create backend object
    if (ox_omix_create(engine))
    {
        res = XA_RESULT_RESOURCE_ERROR;
        LIMWARN("%s\n", xa_error_string(res));
        goto bail2;
    }

    *pMix = (XAObjectItf)&omix->obj_itf;
    return XA_RESULT_SUCCESS;

 bail2:
    free(omix);
 bail1:
    return res;
}
