/*
  Copyright (C) 2009 - 2010 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX AL PLAYER: play interface implementation.

  REVISION:

    1) initial creation. ----------------------------------------- 2009-09-04 MG
*/

#include "limoa-log.h"
#include "limipc.h"
#include "limthread.h"
#include "OpenMAXAL.h"
#include "engine-fe.h"
#include "player-fe.h"

//#undef LIMDBG
#ifndef LIMDBG
#define LIMDBG(fmt, arg...)
#endif

static XAresult player_start(player_t * player)
{
    engine_t * engine = player->engine;
    ipc_packet_t rqst;
    ipc_packet_t rsps;
    ipc_standard_rsps_t * player_rsps;
    XAresult result = XA_RESULT_PERMISSION_DENIED;

    ipc_init_request(&rqst, OX_RQST_PLAYER_PLAY, 0, NULL);
    if (engine_ipc_send_packet(engine, &rqst, &rsps))
    {
        LIMERR("unable to send player play request.\n");
        goto bail;
    }
    player_rsps = rsps.dptr;
    result = player_rsps->result;
    ipc_release_packet(engine->ipc.sid, &rsps);

 bail:
    return result;
}

static XAresult player_stop(player_t * player)
{
    engine_t * engine = player->engine;
    ipc_packet_t rqst;
    ipc_packet_t rsps;
    ipc_standard_rsps_t * player_rsps;
    XAresult result = XA_RESULT_PERMISSION_DENIED;

    ipc_init_request(&rqst, OX_RQST_PLAYER_STOP, 0, NULL);
    if (engine_ipc_send_packet(engine, &rqst, &rsps))
    {
        LIMERR("unable to send player stop request.\n");
        goto bail;
    }
    player_rsps = rsps.dptr;
    result = player_rsps->result;
    ipc_release_packet(engine->ipc.sid, &rsps);

 bail:
    return result;
}

static XAresult player_pause(player_t * player)
{
    engine_t * engine = player->engine;
    ipc_packet_t rqst;
    ipc_packet_t rsps;
    ipc_standard_rsps_t * player_rsps;
    XAresult result = XA_RESULT_PERMISSION_DENIED;

    ipc_init_request(&rqst, OX_RQST_PLAYER_PAUSE, 0, NULL);
    if (engine_ipc_send_packet(engine, &rqst, &rsps))
    {
        LIMERR("unable to send player pause request.\n");
        goto bail;
    }
    player_rsps = rsps.dptr;
    result = player_rsps->result;
    ipc_release_packet(engine->ipc.sid, &rsps);

 bail:
    return result;
}

static XAresult SetPlayState (
    XAPlayItf self,
    XAuint32 state)
{
    XAresult res = XA_RESULT_SUCCESS;
    player_t * player = ((play_itf_t *)(*self))->player;

    switch (state)
    {
    case XA_PLAYSTATE_PLAYING :
        res = player_start(player);
        break;

    case XA_PLAYSTATE_STOPPED :
        res = player_stop(player);
        break;

    case XA_PLAYSTATE_PAUSED :
        res = player_pause(player);
        break;

    default :
        res = XA_RESULT_PARAMETER_INVALID;
        LIMWARN("%s\n", xa_error_string(res));
        goto bail;
    }

    if (XA_RESULT_SUCCESS != res)
    {
        player->playstate = XA_PLAYSTATE_STOPPED;
    } else
    {
        player->playstate = state;
    }

 bail:
    return res;
}

static XAresult GetPlayState (
    XAPlayItf self,
    XAuint32 * pState)
{
    XAresult res = XA_RESULT_SUCCESS;
    player_t * player = ((play_itf_t *)(*self))->player;

    if (NULL == pState)
    {
        res = XA_RESULT_PARAMETER_INVALID;
        LIMWARN("%s\n", xa_error_string(res));
        goto bail;
    }

    *pState = player->playstate;

 bail:
    return res;
}

static XAresult GetPosition (
    XAPlayItf self,
    XAmillisecond * pMsec)
{
    XAresult result = XA_RESULT_PARAMETER_INVALID;
    player_t * player = ((play_itf_t *)(*self))->player;
    engine_t * engine = player->engine;
    ipc_packet_t rqst;
    ipc_packet_t rsps;
    ox_rsps_player_get_position_t * player_rsps;

    ipc_init_request(&rqst, OX_RQST_PLAYER_GET_POSITION, 0, NULL);
    if (engine_ipc_send_packet(engine, &rqst, &rsps))
    {
        LIMERR("unable to send player get position request.\n");
        goto bail;
    }
    player_rsps = rsps.dptr;

    result = player_rsps->result;
    if (XA_RESULT_SUCCESS == result)
    {
        *pMsec = player_rsps->pos;
    }

 bail:
    return result;
}

static XAresult RegisterCallback (
    XAPlayItf self,
    xaPlayCallback callback,
    void * pContext)
{
    XAresult result = XA_RESULT_PARAMETER_INVALID;
    player_t * player = ((play_itf_t *)(*self))->player;
    engine_t * engine = player->engine;
    ipc_packet_t rqst;
    ox_rqst_player_register_callback_t player_rqst;

    ipc_init_request(&rqst, OX_RQST_PLAYER_REGISTER_CALLBACK,
                     sizeof(ox_rqst_player_register_callback_t),
                     &player_rqst);

    player_rqst.caller = self;
    player_rqst.func = callback;
    player_rqst.context = pContext;

    if (engine_ipc_send_packet(engine, &rqst, NULL))
    {
        LIMERR("unable to send player register callback request.\n");
        goto bail;
    }
    result = XA_RESULT_SUCCESS;
 bail:
    return result;
}

static XAresult SetCallbackEventsMask (
    XAPlayItf self,
    XAuint32 eventFlags)
{
    XAresult result = XA_RESULT_PARAMETER_INVALID;
    player_t * player = ((play_itf_t *)(*self))->player;
    engine_t * engine = player->engine;
    ipc_packet_t rqst;
    ox_rqst_player_set_callback_events_mask_t player_rqst;

    ipc_init_request(&rqst, OX_RQST_PLAYER_SET_CALLBACK_EVENTS_MASK,
                     sizeof(ox_rqst_player_set_callback_events_mask_t),
                     &player_rqst);

    player_rqst.mask = eventFlags;

    if (engine_ipc_send_packet(engine, &rqst, NULL))
    {
        LIMERR("unable to send player set callback events mask request.\n");
        goto bail;
    }
    result = XA_RESULT_SUCCESS;
 bail:
    return result;
}

void play_interface_init (player_t * player)
{
    struct XAPlayItf_ * itf;

    player->playItf.player = player;

    itf =
    player->play_itf = &player->playItf.itf;

    itf->SetPlayState = SetPlayState;
    itf->GetPlayState = GetPlayState;
    itf->GetPosition = GetPosition;
    itf->RegisterCallback = RegisterCallback;
    itf->SetCallbackEventsMask = SetCallbackEventsMask;
}
