/*
  Copyright (C) 2009 - 2010 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX AL PLAYER: object interface implementation.

  REVISION:

    1) initial creation. ----------------------------------------- 2009-09-03 MG
*/

#include "limoa-log.h"
#include "limipc.h"
#include "limthread.h"
#include "OpenMAXAL.h"
#include "player-fe.h"

//#undef LIMDBG
#ifndef LIMDBG
#define LIMDBG(fmt, arg...)
#endif

#define PLAYER_IPC_SCRATCH_BUFFER_BYTES 1024

static inline int ox_player_create (
    engine_t * engine,
    const XADataSource * pDataSrc,
    const XADataSource * pBankSrc,
    const XADataSink * pAudioSnk,
    const XADataSink * pImageVideoSnk,
    const XADataSink * pVibra,
    const XADataSink * pLEDArray)
{
    ipc_packet_t rqst;
    ipc_packet_t rsps;
    ox_rqst_player_create_t * player_rqst;
    ipc_standard_rsps_t * player_rsps;
    void * scratch_buf[BUF_BYTES(PLAYER_IPC_SCRATCH_BUFFER_BYTES, void *)];
    void * ptrbase;
    int bytes = 0;
    int ret = -1;

    ptrbase = (void *)&scratch_buf[0];
    player_rqst = (ox_rqst_player_create_t *)ptrbase;
    memset(player_rqst, 0, sizeof(ox_rqst_player_create_t));
    player_rqst->ptrbase = ptrbase;
    ptrbase += sizeof(ox_rqst_player_create_t);

    ptrbase += deep_cp_data_source(&player_rqst->data_src, pDataSrc, ptrbase);
    ptrbase += deep_cp_data_source(&player_rqst->bank_src, pBankSrc, ptrbase);

    ptrbase += deep_cp_data_sink(&player_rqst->audio_snk,
                                 pAudioSnk, ptrbase);
    ptrbase += deep_cp_data_sink(&player_rqst->image_video_snk,
                                 pImageVideoSnk, ptrbase);
    ptrbase += deep_cp_data_sink(&player_rqst->image_video_snk,
                                 pImageVideoSnk, ptrbase);
    ptrbase += deep_cp_data_sink(&player_rqst->vibra_snk,
                                 pVibra, ptrbase);
    ptrbase += deep_cp_data_sink(&player_rqst->LED_array_snk,
                                 pLEDArray, ptrbase);

    bytes = ptrbase - player_rqst->ptrbase;
    ASSERT(bytes <= PLAYER_IPC_SCRATCH_BUFFER_BYTES);

    ipc_init_request(&rqst, OX_RQST_PLAYER_CREATE, bytes, player_rqst);
    if (engine_ipc_send_packet(engine, &rqst, &rsps))
    {
        LIMERR("unable to send player creation request to limox\n");
        goto bail1;
    }

    player_rsps = rsps.dptr;
    if (player_rsps->result != XA_RESULT_SUCCESS)
    {
        LIMERR("limox unable to create player object, result = %d\n",
              (int)player_rsps->result);
        goto bail2;
    }
    ret = 0;

 bail2:
    ipc_release_packet(engine->ipc.sid, &rsps);
 bail1:
    return ret;
}

static inline XAresult ox_player_realize (player_t * player, XAboolean async)
{
    engine_t * engine = player->engine;
    ipc_packet_t rqst;
    ipc_packet_t rsps;
    ox_rqst_player_realize_t player_rqst;
    ipc_standard_rsps_t * player_rsps;
    XAresult result = XA_RESULT_IO_ERROR;

    player_rqst.async = async;
    ipc_init_request(&rqst, OX_RQST_PLAYER_REALIZE,
                     sizeof(ox_rqst_player_realize_t), &player_rqst);
    if (engine_ipc_send_packet(engine, &rqst, &rsps))
    {
        LIMERR("unable to send player realize request to limox\n");
        goto bail;
    }

    player_rsps = rsps.dptr;
    result = player_rsps->result;
    ipc_release_packet(engine->ipc.sid, &rsps);

 bail:
    return result;
}

static XAresult Realize (
    XAObjectItf self,
    XAboolean async)
{
    XAresult res = XA_RESULT_UNKNOWN_ERROR;
    player_t * player = (player_t *)self;

    if (player->objstate != OBJ_STATE_UNREALIZED)
    {
        res = XA_RESULT_PRECONDITIONS_VIOLATED;
        LIMWARN("%s\n", xa_error_string(res));
        goto bail;
    }

    res = ox_player_realize(player, async);
    if (XA_RESULT_SUCCESS != res)
    {
        LIMERR("unable to realize player.\n");
        goto bail;
    }

    player->objstate = OBJ_STATE_REALIZED;
    res = XA_RESULT_SUCCESS;
 bail:
    return res;
}

static XAresult Resume (
    XAObjectItf self,
    XAboolean async)
{
    LIMWARN("API not implemented.\n");
    return XA_RESULT_SUCCESS;
}

static XAresult GetState (
    XAObjectItf self,
    XAuint32 * pState)
{
    LIMWARN("API not implemented.\n");
    return XA_RESULT_SUCCESS;
}

static XAresult GetInterface (
    XAObjectItf self,
    const XAInterfaceID iid,
    void * pInterface)
{
    XAresult res = XA_RESULT_SUCCESS;
    player_t * player = (player_t *)self;

    if (player->objstate != OBJ_STATE_REALIZED)
    {
        res = XA_RESULT_PRECONDITIONS_VIOLATED;
        LIMWARN("%s\n", xa_error_string(res));
        goto bail;
    }

    if (XA_IID_PLAY == iid)
    {
        *((XAPlayItf *)pInterface) = (XAPlayItf)&player->play_itf;
    } else if (XA_IID_SEEK == iid)
    {
        *((XASeekItf *)pInterface) = (XASeekItf)&player->seek_itf;
    } else
    {
        res = XA_RESULT_FEATURE_UNSUPPORTED;
        LIMWARN("%s\n", xa_error_string(res));
    }

 bail:
    return res;
}

static XAresult RegisterCallback (
    XAObjectItf self,
    xaObjectCallback callback,
    void * pContext)
{
    LIMWARN("API not implemented.\n");
    return XA_RESULT_SUCCESS;
}

static void AbortAsyncOperation (
    XAObjectItf self)
{
    LIMWARN("API not implemented.\n");
}

static void Destroy (
    XAObjectItf self)
{
    player_t * player = (player_t *)self;
    ipc_packet_t rqst;

    //TODO: automatically query playItf and stop player before destroying.

    ipc_init_request(&rqst, OX_RQST_PLAYER_DESTROY, 0, NULL);
    if (engine_ipc_send_packet(player->engine, &rqst, NULL))
    {
        LIMWARN("unable to send player destroy request.\n");
    }

    free(player);
    LIMDBG("player destroyed.\n");
}

static XAresult SetPriority (
    XAObjectItf self,
    XAuint32 priority)
{
    LIMWARN("API not implemented.\n");
    return XA_RESULT_SUCCESS;
}

static XAresult GetPriority (
    XAObjectItf self,
    XAuint32 * pPriority)
{
    LIMWARN("API not implemented.\n");
    return XA_RESULT_SUCCESS;
}

static XAresult SetLossOfControlInterfaces (
    XAObjectItf self,
    XAuint16 numInterfaces,
    const XAInterfaceID * pInterfaceIDs,
    XAboolean enabled)
{
    LIMWARN("API not implemented.\n");
    return XA_RESULT_SUCCESS;
}

XAresult player_obj_create (
    XAEngineItf self,
    XAObjectItf * pPlayer,
    const XADataSource * pDataSrc,
    const XADataSource * pBankSrc,
    const XADataSink * pAudioSnk,
    const XADataSink * pImageVideoSnk,
    const XADataSink * pVibra,
    const XADataSink * pLEDArray,
    XAuint32 numInterfaces,
    const XAInterfaceID * pInterfaceIds,
    const XAboolean * pInterfaceRequired)
{
    XAresult res;
    player_t * player;
    engine_t * engine = ((engine_itf_t *)(*self))->engine;

    player = (player_t *)calloc(1, sizeof(player_t));
    if (NULL == player)
    {
        LIMERR("unable to allocate memory for output mix object.\n");
        res = XA_RESULT_MEMORY_FAILURE;
        goto bail1;
    }
    player->engine = engine;
    player->objstate = OBJ_STATE_UNREALIZED;
    player->obj_itf = &player->objItf;

    player->obj_itf->Realize = Realize;
    player->obj_itf->Resume = Resume;
    player->obj_itf->GetState = GetState;
    player->obj_itf->GetInterface = GetInterface;
    player->obj_itf->RegisterCallback = RegisterCallback;
    player->obj_itf->AbortAsyncOperation = AbortAsyncOperation;
    player->obj_itf->Destroy = Destroy;
    player->obj_itf->SetPriority = SetPriority;
    player->obj_itf->GetPriority = GetPriority;
    player->obj_itf->SetLossOfControlInterfaces = SetLossOfControlInterfaces;

    //TODO: validate interfaces.

    // create backend object
    if (ox_player_create(engine, pDataSrc, pBankSrc,
                         pAudioSnk, pImageVideoSnk, pVibra, pLEDArray))
    {
        res = XA_RESULT_RESOURCE_ERROR;
        LIMWARN("%s\n", xa_error_string(res));
        goto bail2;
    }

    play_interface_init(player);
    seek_interface_init(player);

    *pPlayer = (XAObjectItf)&player->obj_itf;
    return XA_RESULT_SUCCESS;

 bail2:
    free(player);
 bail1:
    return res;
}
