/*
  Copyright (C) 2009-2010 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX AL player demo code.
      - based on sample code from OpenMAX spec.

  REVISION:

    1) initial creation. ----------------------------------------- 2009-08-19 MG
*/

#include "unistd.h"
#include "linux/limits.h"
#include "stdio.h"
#include "stdlib.h"
#include "string.h"
#include "termios.h"
#include "OpenMAXAL.h"

#define MAX_NUMBER_INTERFACES 5

//#define DBG_ON
#ifdef DBG_ON
#define DBG(fmt, arg...) do {\
		printf("%s:%s: %d\n", __BASE_FILE__, __FUNCTION__, __LINE__); \
		printf(fmt, ##arg); } while(0)
#else
#define DBG(fmt, arg...)
#endif

/* Global variables. (Should be local in real application.) */
static XAObjectItf engine; /* OpenMAX AL Engine */
static XAObjectItf player;
static XAObjectItf outputMix;
static XAPlayItf playItf;
static XASeekItf seekItf;
static XAEqualizerItf equalizerItf;
static int eof;
static int is_image;

/* Checks for error. If any errors exit the application! */
void CheckErr (XAresult res)
{
    if (res != XA_RESULT_SUCCESS)
    {
        /* Debug printing to be placed here */
        printf("OMX error code: %u\n", res);
        exit(1);
    }
}

/*
 * Draws single EQ band to the screen. Called by drawEQDisplay
 */
void drawEQBand (int minFreq, int maxFreq, int level)
{
    /* Insert drawing routines here for single EQ band. (Use
     * GetBandLevelRange and screen height to map the level to screen
     * y-coordinate.) */
}

/*
 * Called when the display is repainted.
 */
void drawEQDisplay (void)
{
    XAuint16 numBands;
    XAmillibel bandLevel;
    XAmillibel minLevel;

    XAmillibel maxLevel;
    XAmilliHertz minFreq;
    XAmilliHertz maxFreq;
    int band;
    XAresult res;

    res = (*equalizerItf)->GetNumberOfBands(equalizerItf,
                                            &numBands);
    CheckErr(res);

    res = (*equalizerItf)->GetBandLevelRange(equalizerItf,
                                             &minLevel, &maxLevel);
    CheckErr(res);

    for (band = 0; band<numBands; band++)
    {
        res = (*equalizerItf)->GetBandFreqRange(equalizerItf,
                                                (XAint16) band,
                                                &minFreq, &maxFreq);
        CheckErr(res);
        res = (*equalizerItf)->GetBandLevel(equalizerItf,
                                            (XAint16)band, &bandLevel);
        CheckErr(res);
        drawEQBand(minFreq, maxFreq, bandLevel);
    }
}

/* registered player callback. */
void my_callback (XAPlayItf caller, void * pContext, XAuint32 event)
{
    switch (event)
    {
    case XA_PLAYEVENT_HEADATEND :
        printf("playback hit the end, goodbye.\n");
        if (!is_image) eof = 1;
        break;

    case XA_PLAYEVENT_HEADATMARKER :
        printf("playback hit the marker.\n");
        break;

    case XA_PLAYEVENT_HEADATNEWPOS :
        printf("playback at new position.\n");
        break;

    case XA_PLAYEVENT_HEADMOVING :
        printf("playback is active.\n");
        break;

    case XA_PLAYEVENT_HEADSTALLED :
        printf("playback stalled.\n");
        break;

    default :
        printf("bogus callback event : %d\n", (int)event);
        break;
    }
}

static int image_extension_detect (const char * filename)
{
    char * pc = strrchr(filename, '.');

    if (pc++)
    {
        if (!strcasecmp(pc, "jpg") || !strcasecmp(pc, "jpeg")
            || !strcasecmp(pc, "bmp") || !strcasecmp(pc, "png")
            || !strcasecmp(pc, "gif") || !strcasecmp(pc, "tiff")
            || !strcasecmp(pc, "ppm") || !strcasecmp(pc, "pbm")
            || !strcasecmp(pc, "pnm"))
        {
            return 1;
        }
    }
    return 0;
}


/*
 * Initializes the OpenMAX AL engine, starts the playback of some
 * music from a file and draws the graphical equalizer
 */
void init (const char * filename)
{
    XAEngineItf EngineItf;
    XADataSource audioSource;
    XADataLocator_URI uri;
    XADataFormat_MIME mime;
    XADataSink audioSink;
    XADataLocator_OutputMix locator_outputmix;
    XAVolumeItf volumeItf;
    XAresult res;
    int ii;
    XAboolean required[MAX_NUMBER_INTERFACES];
    XAInterfaceID iidArray[MAX_NUMBER_INTERFACES];
    XAEngineOption EngineOption[] = {
        (XAuint32) XA_ENGINEOPTION_THREADSAFE,
        (XAuint32) XA_BOOLEAN_TRUE};
    XAchar uri_path[512];

    is_image = image_extension_detect(filename);

    //sprintf(uri_path, "file:///%s", filename);
    sprintf(uri_path, "%s", filename);

    /* Create OpenMAX AL */
	DBG("creating engine.\n");
    res = xaCreateEngine(&engine, 1, EngineOption, 0, NULL, NULL);
    CheckErr(res);

    /* Realizing the XA Engine in synchronous mode. */
	DBG("realizing engine.\n");
    res = (*engine)->Realize(engine, XA_BOOLEAN_FALSE);
    CheckErr(res);

    /* Get the XA Engine Interface which is implicit */
    res = (*engine)->GetInterface(engine, XA_IID_ENGINE, (void*)&EngineItf);
    CheckErr(res);

    /* Initialize arrays required[] and iidArray[] */
    for (ii = 0; ii < MAX_NUMBER_INTERFACES; ii++)
    {
        required[ii] = XA_BOOLEAN_FALSE;
        iidArray[ii] = XA_IID_NULL;
    }
    /* Set arrays  required[] and iidArray[] for VOLUME and EQUALIZER
     * interfaces */
    required[0] =  XA_BOOLEAN_TRUE;
    iidArray[0] =  XA_IID_VOLUME;
    //required[1] =  XA_BOOLEAN_TRUE;
    //iidArray[1] =  XA_IID_EQUALIZER;

    /* Create Output Mix object to be used by player */
    res = (*EngineItf)->CreateOutputMix(EngineItf, &outputMix, 2,
                                        iidArray, required);
    CheckErr(res);

    /* Realizing the Output Mix object in synchronous mode. */
    res = (*outputMix)->Realize(outputMix, XA_BOOLEAN_FALSE);
    CheckErr(res);

    /* Get play and equalizer interfaces */
    res = (*outputMix)->GetInterface(outputMix, XA_IID_VOLUME,
                                     (void*) &volumeItf);
    CheckErr(res);
    res = (*outputMix)->GetInterface(outputMix, XA_IID_EQUALIZER,
                                     (void*) &equalizerItf);
    CheckErr(res);

    /* Setup the data source structure */
    uri.locatorType = XA_DATALOCATOR_URI;
    uri.pURI = uri_path;
    mime.formatType = XA_DATAFORMAT_MIME;
    mime.pMimeType = (XAchar *) "audio/ogg";
    mime.containerType = XA_CONTAINERTYPE_WAV;
    audioSource.pLocator = (void*) &uri;
    audioSource.pFormat = (void*) &mime;

    /* Setup the data sink structure */
    locator_outputmix.locatorType = XA_DATALOCATOR_OUTPUTMIX;
    locator_outputmix.outputMix = outputMix;
    audioSink.pLocator = (void*) &locator_outputmix;
    audioSink.pFormat = NULL;

    /* Set arrays required[] and iidArray[] for no interfaces (PlayItf
     * is implicit) */
    required[0] = XA_BOOLEAN_FALSE;
    iidArray[0] = XA_IID_NULL;
    required[1] = XA_BOOLEAN_FALSE;
    iidArray[1] = XA_IID_NULL;

    /* Create the music player */
	DBG("creating media player.\n");
    res = (*EngineItf)->CreateMediaPlayer(EngineItf, &player,
                                          &audioSource, NULL,
                                          &audioSink, NULL, NULL, NULL,
                                          0, iidArray, required);
    CheckErr(res);

    /* Realizing the player in synchronous mode. */
	DBG("realizing player.\n");
    res = (*player)->Realize(player, XA_BOOLEAN_FALSE);
    CheckErr(res);

    /* Get the play interface */
	DBG("get play interface.\n");
    res = (*player)->GetInterface(player, XA_IID_PLAY, (void*) &playItf);
    CheckErr(res);

    /* Register callback. */
	DBG("register callback.\n");
    res = (*playItf)->RegisterCallback(playItf, my_callback, NULL);
    CheckErr(res);

    /* only interested in EOF. */
	DBG("set callback event mask.\n");
    res = (*playItf)->SetCallbackEventsMask(playItf, XA_PLAYEVENT_HEADATEND);
    CheckErr(res);

    /* Get the play interface */
	DBG("get seek interface.\n");
    res = (*player)->GetInterface(player, XA_IID_SEEK, (void*) &seekItf);
    CheckErr(res);

#if 0
    /* Before we start set volume to -3dB (-300mB) and enable equalizer */
    res = (*volumeItf)->SetVolumeLevel(volumeItf, -300);
    CheckErr(res);

    res = (*equalizerItf)->SetEnabled(equalizerItf, XA_BOOLEAN_TRUE);
    CheckErr(res);
#endif
    /* Play the music */
	DBG("set PLAYSTATE_PLAYING.\n");
    res = (*playItf)->SetPlayState(playItf, XA_PLAYSTATE_PLAYING);
    CheckErr(res);
#if 0
    /* Draw the graphical EQ */
    drawEQDisplay();
#endif
}

/*
 * Shuts down the OpenMAX AL engine.
 */
void destroy (void)
{
    XAresult res;

    /* Stop the music */
    res = (*playItf)->SetPlayState(playItf, XA_PLAYSTATE_STOPPED);
    CheckErr(res);

    /* Destroy the player */
    (*player)->Destroy(player);

    /* Destroy Output Mix object */
    (*outputMix)->Destroy(outputMix);

    /* Shutdown OpenMAX AL */
    (*engine)->Destroy(engine);
}

/*
 * Called by UI when user increases or decreases a band level.
 */
void setBandLevel (XAint16 band, XAboolean increase)
{
    XAuint16 numBands;
    XAmillibel bandLevel;
    XAmillibel minLevel;
    XAmillibel maxLevel;
    XAresult res;

    res = (*equalizerItf)->GetNumberOfBands(equalizerItf, &numBands);
    CheckErr(res);

    res = (*equalizerItf)->GetBandLevelRange(equalizerItf,
                                             &minLevel, &maxLevel);
    CheckErr(res);
    if (band >= numBands)
    {
        /* Error. Insert debug print here. */
        exit(0);
    }
    res = (*equalizerItf)->GetBandLevel(equalizerItf,
                                        band, &bandLevel);
    CheckErr(res);

    if (increase==XA_BOOLEAN_TRUE)
    {
        /* increase the level by 1 dB (100mB) if the max supported level
         * is not exceeded */
        bandLevel = bandLevel + 100;
        if(bandLevel < maxLevel)
        {
            res = (*equalizerItf)->SetBandLevel(equalizerItf,
                                                band, bandLevel);
            CheckErr(res);
            drawEQDisplay();
        }
    }
    else /* increase==false */
    {
        /* decrease the level by 1 dB (100mB) if the min supported level
         * is not crossed */
        bandLevel = bandLevel - 100;
        if( bandLevel > minLevel )
        {
            res = (*equalizerItf)->SetBandLevel(equalizerItf,
                                                band, bandLevel);
            CheckErr(res);
            drawEQDisplay();
        }
    }
}

static struct termios orig_termios;
static void restore_termios (int status, void * arg)
{
    tcsetattr(STDIN_FILENO, TCSANOW, &orig_termios);
}

int main (int argc, char ** argv)
{
    XAuint32 play_state;
    fd_set rdfds;
    char fname[PATH_MAX];
    struct termios new_termios;
    XAmillisecond cur_pos;

    if (argc != 2)
    {
        printf("\nusage: ./player <file-name>\n\n");
        exit(-1);
    }

    tcgetattr(STDIN_FILENO, &orig_termios);
    new_termios = orig_termios;
    new_termios.c_lflag &= ~(ICANON | ECHO | ECHOCTL | ECHONL);
    new_termios.c_cflag |= HUPCL;
    new_termios.c_cc[VMIN] = 0;
    tcsetattr(STDIN_FILENO, TCSANOW, &new_termios);
    on_exit(restore_termios, &orig_termios);

    if (argv[1][0] != '/')
    {
        if (getcwd(fname, PATH_MAX))
        {
            strcat(fname, "/");
            strcat(fname, argv[1]);
            init(fname);
        } else
        {
            printf("ERROR: unable to get current working path.\n");
        }
    } else init(argv[1]);

    eof = 0;
    while (!eof)
    {
        struct timeval tv;
        int ret;
        int ch[8];
        int chnum = 0;

        FD_ZERO(&rdfds);
        FD_SET(0, &rdfds);

        tv.tv_sec = 0;
        tv.tv_usec = 250000;
        ret = select(1, &rdfds, NULL, NULL, &tv);
        if (-1 == ret) break;
        if (0 == ret) continue;

        while((ch[chnum] = getchar()) != EOF) chnum++;
        if (chnum > 1) ch[0] = ch[chnum - 1] | (ch[chnum - 2] << 8);

        switch (ch[0])
        {
        case 'p' :
        case ' ' :
            (*playItf)->GetPlayState(playItf, &play_state);
            if (XA_PLAYSTATE_PLAYING == play_state)
            {
                printf("pause\n");
                (*playItf)->SetPlayState(playItf, XA_PLAYSTATE_PAUSED);
            } else if (XA_PLAYSTATE_PAUSED == play_state)
            {
                printf("unpause\n");
                (*playItf)->SetPlayState(playItf, XA_PLAYSTATE_PLAYING);
            } else
            {
                printf("playback stopped already?\n");
                goto bail;
            }
            break;

        case 'q' :
            goto bail;

        case 0x5b42 : //KEY_DOWN
            (*playItf)->GetPosition(playItf, &cur_pos);
            (*seekItf)->SetPosition(seekItf, (cur_pos + 60000 < 0)?
                                    0 : cur_pos + 60000,
                                    XA_SEEKMODE_FAST);
            break;

        case 0x5b41 : //KEY_UP
            (*playItf)->GetPosition(playItf, &cur_pos);
            (*seekItf)->SetPosition(seekItf, (cur_pos > 60000)?
                                    cur_pos - 60000 : 0,
                                    XA_SEEKMODE_FAST);
            break;
        case 0x5b43 : //KEY_RIGHT
            (*playItf)->GetPosition(playItf, &cur_pos);
            DBG(">> current position: %d\n", cur_pos);
            (*seekItf)->SetPosition(seekItf, (cur_pos + 10000 < 0)?
                                    0 : cur_pos + 10000,
                                    XA_SEEKMODE_FAST);
            break;

        case 0x5b44 : //KEY_LEFT
            (*playItf)->GetPosition(playItf, &cur_pos);
            DBG("<< current position: %d\n", cur_pos);
            (*seekItf)->SetPosition(seekItf, (cur_pos > 10000)?
                                    cur_pos - 10000 : 0,
                                    XA_SEEKMODE_FAST);
            break;

        case 'L' :
            lim_log_dump();
            break;

        default :
            printf("unkown command [%x], valid commands are :\n"
                   "\t-----------------------------------------\n"
                   "\tp     : pause/unpause playback.\n"
                   "\tq     : stop playback and quit program.\n"
                   "\tLEFT  : seek backward 10 seconds.\n"
                   "\tRIGHT : seek forward 10 seconds.\n"
                   "\tUP    : seek backward 60 seconds.\n"
                   "\tDOWN  : seek forward 60 seconds.\n"
                   "\t-----------------------------------------\n",
                   ch[0]);
            break;
        }
    }

 bail:
    destroy();
    printf("OpenMAXAL engine destroyed, exit.\n");
}
