/*
  Copyright (C) 2010 - 2011 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX IL standard audio encoder component.

  REVISION:

    1) initial creation. ----------------------------------------- 2010-05-19 MG
*/
#define _GNU_SOURCE
#include "limoi.h"
#include "audio-priv.h"

//#undef LIMDBG
#ifndef LIMDBG
#define LIMDBG(fmt, arg...)
#endif

static OMX_ERRORTYPE main_handle (
    OMX_HANDLETYPE hComp,
    limoi_port_t ** audio_port,
    limoi_port_t ** video_port,
    limoi_port_t ** other_port,
    int * preloaded)
{
    limoi_audio_encoder_t * encoder = LIMOI_AUDIO_ENCODER(hComp);
    limoi_audio_encoder_priv_t * pencoder = encoder->priv;
    limoi_port_t * audio_port_in = *audio_port;
    limoi_port_t * audio_port_out = *(audio_port + 1);
    OMX_BUFFERHEADERTYPE ** in = NULL;
    OMX_BUFFERHEADERTYPE ** out = NULL;
    OMX_ERRORTYPE ret = OMX_ErrorNone;

    if (*preloaded == 0)
    {
        //do not requrie preload.
        *preloaded = LIMOI_PRELOAD_BUFFER_COUNT;
        goto bail1;
    }

    LIMDBG("%s encoder get input port buffer.\n", LIMOI_COMPONENT_PRIV(hComp)->name);
    in = limoi_component_get_port_buffer(hComp, audio_port_in, -1);
    if (NULL == in)
    {
        LIMDBG("%s unable to get input port buffer.\n",
               LIMOI_COMPONENT_PRIV(hComp)->name);
        ret = OMX_ErrorNotReady;
        goto bail1;
    } else if (!(*in)->nFilledLen && !((*in)->nFlags & OMX_BUFFERFLAG_EOS))
    {
        LIMDBG("%s put input port buffer back: %p.\n",
               LIMOI_COMPONENT_PRIV(hComp)->name, (*in)->pBuffer);
        if (-1 == limoi_component_put_port_buffer(hComp, audio_port_in, in))
        {
            ret = OMX_ErrorNotReady;
        }
        goto bail1;
    }
    LIMDBG("%s input buffer len: %u\n",
           LIMOI_COMPONENT_PRIV(hComp)->name, (*in)->nFilledLen);

    LIMDBG("%s encoder get output port buffer.\n",
           LIMOI_COMPONENT_PRIV(hComp)->name);
    out = limoi_component_get_port_buffer(hComp, audio_port_out, -1);
    if (NULL == out)
    {
        LIMDBG("%s unable to get output port buffer.\n",
               LIMOI_COMPONENT_PRIV(hComp)->name);
        ret = OMX_ErrorNotReady;
        goto bail2;
    } else if ((*out)->nFilledLen == 0)
    {
        LIMDBG("%s output buffer len: %u\n",
               LIMOI_COMPONENT_PRIV(hComp)->name, (*out)->nAllocLen);
        if ((*in)->nFilledLen)
        {
            ret = pencoder->callbacks.encode(hComp, *in, *out);
            LIMDBG("%s encoding bytes in: %u bytes out: %u offset: %u\n",
                   LIMOI_COMPONENT_PRIV(hComp)->name, (*in)->nFilledLen,
                   (*out)->nFilledLen, (*out)->nOffset);
        }

        if ((*in)->nFlags & OMX_BUFFERFLAG_EOS)
        {
            (*out)->nFlags |= OMX_BUFFERFLAG_EOS;
            limoi_component_deliver_event
                (hComp, OMX_EventBufferFlag, 0, (*in)->nFlags, NULL);
            limoi_component_deliver_event
                (hComp, OMX_EventBufferFlag, 1, (*out)->nFlags, NULL);
        }

        if (!(*in)->nFilledLen)
        {
            int tmp;

            LIMDBG("%s put input port buffer back: %p.\n",
                   LIMOI_COMPONENT_PRIV(hComp)->name, (*in)->pBuffer);
            tmp = limoi_component_put_port_buffer(hComp, audio_port_in, in);
            if (tmp)
            {
                if (tmp == -1) ret = OMX_ErrorNotReady;
                else
                {
                    //client or tunnel component rejeted this buffer.
                    LIMDBG("%s unable to put back port buffer\n",
                           LIMOI_COMPONENT_PRIV(hComp)->name);
                }
            }
            in = NULL;
        }
    }

    if ((*out)->nFilledLen || (*out)->nFlags & OMX_BUFFERFLAG_EOS)
    {
        int tmp;

        LIMDBG("%s put output port buffer back: %p.\n",
               LIMOI_COMPONENT_PRIV(hComp)->name, (*out)->pBuffer);
        tmp = limoi_component_put_port_buffer(hComp, audio_port_out, out);
        if (tmp)
        {
            if (tmp == -1)
            {
                ret = OMX_ErrorNotReady;
            } else
            {
                //client or tunnel component rejeted this buffer.
                LIMDBG("%s unable to put back port buffer\n",
                       LIMOI_COMPONENT_PRIV(hComp)->name);
            }
        }
        out = NULL;
    }

 bail2:
    if (in) limoi_component_release_port_buffer(hComp, audio_port_in, in);
    if (out) limoi_component_release_port_buffer(hComp, audio_port_out, out);
 bail1:
    return ret;
}

static OMX_ERRORTYPE prep_cmd_handle (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    limoi_audio_encoder_t * encoder = LIMOI_AUDIO_ENCODER(cmd->hComponent);
    limoi_audio_encoder_priv_t * pencoder = encoder->priv;

    if (pencoder->callbacks.command_prep)
    {
        ret = pencoder->callbacks.command_prep(cmd);
        if (OMX_ErrorNone != ret)
        {
            LIMERR("%s command callback returned error %s\n",
                   LIMOI_COMPONENT_PRIV(cmd->hComponent)->name,
                   limoi_omx_error_string(ret));
        }
    }
    return ret;
}

static OMX_ERRORTYPE post_cmd_handle (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    limoi_audio_encoder_t * encoder = LIMOI_AUDIO_ENCODER(cmd->hComponent);
    limoi_audio_encoder_priv_t * pencoder = encoder->priv;

    if (pencoder->callbacks.command_post)
    {
        ret = pencoder->callbacks.command_post(cmd);
        if (OMX_ErrorNone != ret)
        {
            LIMERR("%s command callback returned error %s\n",
                   LIMOI_COMPONENT_PRIV(cmd->hComponent)->name,
                   limoi_omx_error_string(ret));
        }
    }
    return ret;
}

static int add_port (
    OMX_HANDLETYPE hComp,
    OMX_AUDIO_CODINGTYPE type)
{
	OMX_PARAM_PORTDEFINITIONTYPE port;
    OMX_PORT_PARAM_TYPE port_param;

    limoi_set_spec_version(&port_param, sizeof(OMX_PORT_PARAM_TYPE));
    port_param.nPorts = 2;
    port_param.nStartPortNumber = 0; //APB = 0
    limoi_component_set_parameter(hComp, OMX_IndexParamAudioInit, &port_param);

    limoi_set_spec_version(&port, sizeof(OMX_PARAM_PORTDEFINITIONTYPE));

    port.nPortIndex = 0;
    port.eDir = OMX_DirInput;
    port.nBufferCountActual = LIMOI_AUDIO_ENCODER_INPUT_BUFFER_NUM;
    port.nBufferCountMin = LIMOI_AUDIO_ENCODER_INPUT_BUFFER_NUM;
    port.nBufferSize = LIMOI_AUDIO_ENCODER_INPUT_BUFFER_BYTES;
    port.bEnabled = OMX_TRUE;
    port.bPopulated = OMX_FALSE;
    port.eDomain = OMX_PortDomainAudio;
    port.bBuffersContiguous = OMX_FALSE;
    port.nBufferAlignment = 0;
    port.format.audio.cMIMEType = NULL;
    port.format.audio.pNativeRender = 0;
    port.format.audio.bFlagErrorConcealment = OMX_FALSE;
    port.format.audio.eEncoding = OMX_AUDIO_CodingPCM;
    limoi_component_add_port(hComp, &port);

    port.nPortIndex = 1;
    port.eDir = OMX_DirOutput;
    port.nBufferCountActual = LIMOI_AUDIO_ENCODER_OUTPUT_BUFFER_NUM;
    port.nBufferCountMin = LIMOI_AUDIO_ENCODER_OUTPUT_BUFFER_NUM;
    port.nBufferSize = LIMOI_AUDIO_ENCODER_OUTPUT_BUFFER_BYTES;
    port.format.audio.eEncoding = type;
    limoi_component_add_port(hComp, &port);

    return 0;
}

int limoi_audio_encoder_set_callbacks (
    OMX_HANDLETYPE hComp,
    limoi_audio_encoder_callbacks_t * callbacks)
{
    limoi_audio_encoder_t * encoder = LIMOI_AUDIO_ENCODER(hComp);
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComp);

    lthread_mutex_lock(&base->mutex);
    encoder->priv->callbacks = *callbacks;
    lthread_mutex_unlock(&base->mutex);

    return 0;
}

OMX_ERRORTYPE limoi_audio_encoder_deinit (
    OMX_HANDLETYPE hComponent)
{
    limoi_audio_encoder_t * encoder = LIMOI_AUDIO_ENCODER(hComponent);

    free(encoder->priv);

    // parent destructor.
    return limoi_component_deinit(hComponent);
}

static OMX_ERRORTYPE audio_encoder_init (
    OMX_HANDLETYPE * hComp,
	int bytes,
	const char * comp_name,
	const char * comp_version,
    OMX_AUDIO_CODINGTYPE type,
    int synchronous)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    OMX_COMPONENTTYPE * hdl;
    limoi_component_callbacks_t callbacks;
    limoi_audio_encoder_t * encoder;
    limoi_audio_encoder_priv_t * priv = NULL;

    LIMDBG("init component %s, version %s.\n", comp_name, comp_version);

    priv = calloc(1, sizeof(limoi_audio_encoder_priv_t));
    if (NULL == priv)
    {
        LIMERR("out of memory upon audio encoder init.\n");
        ret = OMX_ErrorInsufficientResources;
        goto bail;
    }

    ret = synchronous?
        limoi_component_init_sync(hComp, bytes) :
        limoi_component_init(hComp, bytes);

    if (OMX_ErrorNone != ret) goto bail;
    hdl = (OMX_COMPONENTTYPE *)*hComp;

    encoder = LIMOI_AUDIO_ENCODER(*hComp);
    encoder->priv = priv;

    limoi_component_set_name_version(*hComp, comp_name, comp_version);
    add_port(*hComp, type);

    callbacks.command_prep = prep_cmd_handle;
    callbacks.command_post = post_cmd_handle;
    callbacks.main = main_handle;
    callbacks.sub = NULL;
    limoi_component_set_callbacks(*hComp, &callbacks);

    LIMDBG("component %s initialized.\n", comp_name);
    hdl->ComponentDeInit = limoi_audio_encoder_deinit;

 bail:
    if (OMX_ErrorNone != ret && priv) free(priv);
    return ret;
}

OMX_ERRORTYPE limoi_audio_encoder_init (
    OMX_HANDLETYPE * hComp,
	int bytes,
	const char * comp_name,
	const char * comp_version,
    OMX_AUDIO_CODINGTYPE type)
{
    return audio_encoder_init(hComp, bytes, comp_name, comp_version, type, 0);
}

OMX_ERRORTYPE limoi_audio_encoder_init_sync (
    OMX_HANDLETYPE * hComp,
	int bytes,
	const char * comp_name,
	const char * comp_version,
    OMX_AUDIO_CODINGTYPE type)
{
    return audio_encoder_init(hComp, bytes, comp_name, comp_version, type, 1);
}
