/*
  Copyright (C) 2009 - 2011 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX IL standard audio renderer component.

  REVISION:

    1) initial creation. ----------------------------------------- 2009-12-06 MG
*/
#define _GNU_SOURCE
#include "limoi.h"
#include "audio-priv.h"
#include <sys/stat.h>
#include <fcntl.h>

//#undef LIMDBG
#ifndef LIMDBG
#define LIMDBG(fmt, arg...)
#endif

//DEBUG: define this to write pcm to 'renderer-debug.pcm' while rendering.
//#define DEBUG_SAVE2FILE (1024 * 1024 * 8)

#ifdef DEBUG_SAVE2FILE
static void debug_save2file (void * buf, int len)
{
    static int fd = -1;
    static int bytes;

    if (-1 == fd)
    {
        fd = open("renderer-debug.pcm", O_WRONLY | O_CREAT | O_TRUNC,
                  S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
        if (fd < 0)
        {
            LIMERR("unable to create pcm debug file.\n");
            fd = -2;
        }
    }
    if (fd > 0)
    {
        if (bytes < DEBUG_SAVE2FILE)
        {
            bytes += write(fd, buf, len);
        } else
        {
            LIMINFO("yuv debug file closed.\n");
            close(fd);
            fd = -2;
        }
    }
}
#else
#define debug_save2file(x, y) ;
#endif

static OMX_ERRORTYPE sub_handle (
    OMX_HANDLETYPE hComp,
    limoi_port_t ** audio_port,
    limoi_port_t ** video_port,
    limoi_port_t ** other_port)
{
    limoi_audio_renderer_t * renderer = LIMOI_AUDIO_RENDERER(hComp);
    limoi_audio_renderer_priv_t * prenderer = renderer->priv;
    limoi_port_t * clock_port = *other_port;
    OMX_BUFFERHEADERTYPE ** clock_buf = NULL;
    OMX_ERRORTYPE ret = OMX_ErrorNone;

    if ((!clock_port->tunnel_comp
         && LIMOI_PORTDEF(clock_port)->bEnabled == OMX_TRUE)
        || !prenderer->callbacks.clock)
    {
        ret = OMX_ErrorNoMore;
        goto bail;
    }

    LIMDBG("%s renderer get clock port buffer.\n",
           LIMOI_COMPONENT_PRIV(hComp)->name);
    clock_buf = limoi_component_get_port_buffer(hComp, clock_port, 0);

    if (NULL == clock_buf)
    {
        LIMDBG("%s renderer null clock port buffer.\n",
               LIMOI_COMPONENT_PRIV(hComp)->name);
        ret = OMX_ErrorNotReady;
        goto bail;
    }

    if (!(*clock_buf)->nFilledLen)
    {
        LIMDBG("%s renderer clock port buffer zero length.\n",
               LIMOI_COMPONENT_PRIV(hComp)->name);
        if (-1 == limoi_component_put_port_buffer(hComp, clock_port, clock_buf))
        {
            ret = OMX_ErrorNotReady;
        }
        goto bail;
    }

    prenderer->callbacks.clock(hComp, clock_port->tunnel_comp, *clock_buf);

    if (!(*clock_buf)->nFilledLen)
    {
        int tmp;

        tmp = limoi_component_put_port_buffer(hComp, clock_port, clock_buf);
        if (tmp)
        {
            if (tmp == -1) ret = OMX_ErrorNotReady;
            else
            {
                LIMDBG("%s unable to put back port buffer.\n",
                       LIMOI_COMPONENT_PRIV(hComp)->name);
            }
        }
    } else
    {
        LIMDBG("%s renderer clock port buffer not consumed.\n",
               LIMOI_COMPONENT_PRIV(hComp)->name);
        limoi_component_release_port_buffer(hComp, clock_port, clock_buf);
    }

 bail:
    return ret;
}

static OMX_ERRORTYPE main_handle (
    OMX_HANDLETYPE hComp,
    limoi_port_t ** audio_port,
    limoi_port_t ** video_port,
    limoi_port_t ** other_port,
    int * preloaded)
{
    limoi_audio_renderer_t * renderer = LIMOI_AUDIO_RENDERER(hComp);
    limoi_audio_renderer_priv_t * prenderer = renderer->priv;
    limoi_port_t * pcm_port = *audio_port;
    OMX_BUFFERHEADERTYPE ** buf;
    limoi_port_t * clock_port = *other_port;
    OMX_ERRORTYPE ret = OMX_ErrorNone;

    if (*preloaded == 0)
    {
        //do not requrie preload.
        *preloaded = LIMOI_PRELOAD_BUFFER_COUNT;
        goto bail;
    }

    LIMDBG("%s renderer get port buffer.\n",
           LIMOI_COMPONENT_PRIV(hComp)->name);
    buf = limoi_component_get_port_buffer(hComp, pcm_port, -1);
    if (NULL == buf)
    {
        LIMDBG("%s unable to get port buffer.\n",
               LIMOI_COMPONENT_PRIV(hComp)->name);
        ret = OMX_ErrorNotReady;
        goto bail;
    } else if ((*buf)->nFilledLen || (*buf)->nFlags & OMX_BUFFERFLAG_EOS)
    {
        if ((*buf)->nFilledLen)
        {
            LIMDBG("%s renderer render buffer bytes: %u.\n",
                   LIMOI_COMPONENT_PRIV(hComp)->name, (*buf)->nFilledLen);
            debug_save2file((*buf)->pBuffer + (*buf)->nOffset,
                            (*buf)->nFilledLen);
            prenderer->callbacks.render
                (hComp, *buf,
                 clock_port->tunnel_comp, clock_port->tunnel_port,
                 clock_port->index);
        }

        if ((*buf)->nFlags & OMX_BUFFERFLAG_EOS)
        {
            limoi_component_deliver_event
                (hComp, OMX_EventBufferFlag, 0, (*buf)->nFlags, NULL);
        }

        if (!(*buf)->nFilledLen)
        {
            if (-1 == limoi_component_put_port_buffer(hComp, pcm_port, buf))
            {
                LIMDBG("%s unable to put back port buffer.\n",
                       LIMOI_COMPONENT_PRIV(hComp)->name);
                //not an error, tunnel port may be in the wrong state.
                ret = OMX_ErrorNotReady;
            }
            buf = NULL;
        }
    }

    if (buf)
    {
        if ((*buf)->nFilledLen)
            limoi_component_release_port_buffer(hComp, pcm_port, buf);
        else
            limoi_component_put_port_buffer(hComp, pcm_port, buf);
    }

 bail:
    return ret;
}

static OMX_ERRORTYPE prep_cmd_handle (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    limoi_audio_renderer_t * renderer = LIMOI_AUDIO_RENDERER(cmd->hComponent);
    limoi_audio_renderer_priv_t * prenderer = renderer->priv;

    //renderer callback
    if (prenderer->callbacks.command_prep)
    {
        ret = prenderer->callbacks.command_prep(cmd);
        if (OMX_ErrorNone != ret)
        {
            LIMERR("%s command callback returned error %s\n",
                   LIMOI_COMPONENT_PRIV(cmd->hComponent)->name,
                   limoi_omx_error_string(ret));
        }
    }
    return ret;
}

static OMX_ERRORTYPE post_cmd_handle (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    limoi_audio_renderer_t * renderer = LIMOI_AUDIO_RENDERER(cmd->hComponent);
    limoi_audio_renderer_priv_t * prenderer = renderer->priv;

    if (prenderer->callbacks.command_post)
    {
        ret = prenderer->callbacks.command_post(cmd);
        if (OMX_ErrorNone != ret)
        {
            LIMERR("%s command callback returned error %s\n",
                   LIMOI_COMPONENT_PRIV(cmd->hComponent)->name,
                   limoi_omx_error_string(ret));
        }
    }
    return ret;
}

static int add_port (
    OMX_HANDLETYPE hComp)
{
	OMX_PARAM_PORTDEFINITIONTYPE port;
    OMX_PORT_PARAM_TYPE port_param;

    limoi_set_spec_version(&port_param, sizeof(OMX_PORT_PARAM_TYPE));
    port_param.nPorts = 1;
    port_param.nStartPortNumber = 0; //APB = 0
    limoi_component_set_parameter(hComp, OMX_IndexParamAudioInit, &port_param);

    port_param.nStartPortNumber = 1; //OPB = 1
    limoi_component_set_parameter(hComp, OMX_IndexParamOtherInit, &port_param);

    limoi_set_spec_version(&port, sizeof(OMX_PARAM_PORTDEFINITIONTYPE));

    port.nPortIndex = 0;
    port.eDir = OMX_DirInput;
    port.nBufferCountActual = 2;
    port.nBufferCountMin = 2;
    port.nBufferSize = 1024 * 32;
    port.bEnabled = OMX_TRUE;
    port.bPopulated = OMX_FALSE;
    port.eDomain = OMX_PortDomainAudio;
    port.bBuffersContiguous = OMX_FALSE;
    port.nBufferAlignment = 0;
    port.format.audio.cMIMEType = NULL;
    port.format.audio.pNativeRender = 0;
    port.format.audio.bFlagErrorConcealment = OMX_FALSE;
    port.format.audio.eEncoding = OMX_AUDIO_CodingPCM;
    limoi_component_add_port(hComp, &port);

    port.nPortIndex = 1;
    port.nBufferCountActual = LIMOI_MAX_CLOCK_PORT_BUFFERS;
    port.nBufferCountMin = LIMOI_MAX_CLOCK_PORT_BUFFERS;
    port.nBufferSize = sizeof(OMX_TIME_MEDIATIMETYPE);
    port.eDomain = OMX_PortDomainOther;
    port.format.other.eFormat = OMX_OTHER_FormatTime;
    limoi_component_add_port(hComp, &port);

    return 0;
}

int limoi_audio_renderer_set_callbacks (
    OMX_HANDLETYPE hComp,
    limoi_audio_renderer_callbacks_t * callbacks)
{
    limoi_audio_renderer_t * renderer = LIMOI_AUDIO_RENDERER(hComp);
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComp);

    lthread_mutex_lock(&base->mutex);
    renderer->priv->callbacks = *callbacks;
    lthread_mutex_unlock(&base->mutex);

    return 0;
}

OMX_ERRORTYPE limoi_audio_renderer_deinit (
    OMX_HANDLETYPE hComponent)
{
    limoi_audio_renderer_t * renderer = LIMOI_AUDIO_RENDERER(hComponent);

    free(renderer->priv);

    // parent destructor.
    return limoi_component_deinit(hComponent);
}

OMX_ERRORTYPE limoi_audio_renderer_init (
    OMX_HANDLETYPE * hComp,
	int bytes,
	const char * comp_name,
	const char * comp_version)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    OMX_COMPONENTTYPE * hdl;
    limoi_component_callbacks_t callbacks;
    limoi_audio_renderer_t * renderer;
    limoi_audio_renderer_priv_t * priv = NULL;

    LIMDBG("init component %s, version %s.\n", comp_name, comp_version);

    priv = calloc(1, sizeof(limoi_audio_renderer_priv_t));
    if (NULL == priv)
    {
        LIMERR("out of memory upon audio renderer init.\n");
        ret = OMX_ErrorInsufficientResources;
        goto bail;
    }

    ret = limoi_component_init(hComp, bytes);
    if (OMX_ErrorNone != ret) goto bail;
    hdl = (OMX_COMPONENTTYPE *)*hComp;

    renderer = LIMOI_AUDIO_RENDERER(*hComp);
    renderer->priv = priv;

    limoi_component_set_name_version(*hComp, comp_name, comp_version);
    limoi_component_add_role(*hComp, LIMOI_AUDIO_RENDERER_ROLE);
    add_port(*hComp);

    callbacks.command_prep = prep_cmd_handle;
    callbacks.command_post = post_cmd_handle;
    callbacks.main = main_handle;
    callbacks.sub = sub_handle;
    limoi_component_set_callbacks(*hComp, &callbacks);

    LIMDBG("component %s initialized.\n", comp_name);
    hdl->ComponentDeInit = limoi_audio_renderer_deinit;

 bail:
    if (OMX_ErrorNone != ret && priv) free(priv);
    return ret;
}
