/*
  Copyright (C) 2011 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX IL base component: buffer sharing implementation subroutines.

  REVISION:

    1) initial creation. ----------------------------------------- 2011-07-20 MG
*/
#define _GNU_SOURCE
#include "limoi.h"
#include "component.h"
#include "string.h"
#include <stdarg.h>

//#undef LIMDBG
#ifndef LIMDBG
#define LIMDBG(fmt, arg...)
#endif

typedef struct
{
    list_t list;
    limoi_port_t * in_port;
    limoi_port_t * out_port;
    OMX_BUFFERHEADERTYPE ** in_buf_header;
    OMX_BUFFERHEADERTYPE * out_buf_header;
    int ref_count;
} share_request_t;

typedef struct
{
    share_request_t rqst[LIMOI_BUFFER_SHARING_POOL_MAX];
    list_t rqst_head;
    int count;
} share_pool_t;

//allocate buffer share request pool if not already.
static int allocate_buffer_share_pool (OMX_HANDLETYPE hComp)
{
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComp);
    int ii;

    if (base->buffer_share_pool) return 0;

    //allocate buffer share pool
    base->buffer_share_pool = malloc(sizeof(share_pool_t));
    if (NULL == base->buffer_share_pool)
    {
        LIMERR("%s unable to allocate buffer share pool.\n", base->name);
        return -1;
    } else
    {
        share_pool_t * pool = base->buffer_share_pool;
        memset(pool, 0, sizeof(share_pool_t));
        list_init(&pool->rqst_head);
        for (ii = 0; ii < LIMOI_BUFFER_SHARING_POOL_MAX; ii++)
        {
            list_init(&pool->rqst[ii].list);
            list_add(&pool->rqst[ii].list, &pool->rqst_head);
        }
    }

    return 0;
}

int buffer_share_validate_policy (OMX_HANDLETYPE hComp, OMX_U32 nPortIndex)
{
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComp);
    limoi_port_t * limoi_port;
    OMX_PARAM_PORTDEFINITIONTYPE * port;
    limoi_port_t * limoi_port2;
    OMX_PARAM_PORTDEFINITIONTYPE * port2;
    int ii, jj;
    int valid_share_policy_found = 0;

    LIMDBG("%s port %u validating buffer share policy.\n",
           base->name, nPortIndex);

    lthread_mutex_lock(&base->mutex);
    if (!base->buffer_share_policy) goto bail;

    for (ii = 0; ii < base->buffer_share_policy->num_share_pairs; ii++)
    {
        if (base->buffer_share_policy->ports[ii * 2 + 1] != nPortIndex)
            continue;

        jj = ii + 1;
        while (jj < base->buffer_share_policy->num_share_pairs)
        {
            if (base->buffer_share_policy->ports[jj * 2 + 1]
                == nPortIndex)
            {
                LIMERR("%s port %u has more than one supplier.\n",
                       base->name, nPortIndex);
                goto bail;
            }
            jj++;
        }

        limoi_port2 = limoi_component_get_port(hComp, nPortIndex);
        ASSERT(limoi_port2);
        port2 = LIMOI_PORTDEF(limoi_port2);

        limoi_port = limoi_component_get_port
            (hComp, base->buffer_share_policy->ports[ii * 2]);
        ASSERT(limoi_port);
        port = LIMOI_PORTDEF(limoi_port);

        //skip if port has already been validated.
        for (jj = 0; jj < limoi_port->buffer_share.share_num; jj++)
        {
            if (limoi_port->buffer_share.ports[jj] == limoi_port2->index)
                goto bail;
        }

        if (allocate_buffer_share_pool(hComp)) break;

        if (!tunnel_is_supplier(limoi_port2))
        {
            LIMDBG("%s sharing port %u must be tunneled supplier.\n",
                   base->name, limoi_port2->index);
            //limoi_omx_port_status(hComp, limoi_port2->index);
            break;
        }

        if (port->eDir == OMX_DirOutput)
        {
            if (port2->eDir != OMX_DirInput)
            {
                LIMERR("%s buffer can't be shared among output ports.\n",
                       base->name);
                break;
            }

            ASSERT(!limoi_port->buffer_share.share_num);
            //output port supplier can only have one share
            jj = ii + 1;
            while (jj < base->buffer_share_policy->num_share_pairs)
            {
                if (base->buffer_share_policy->ports[jj * 2]
                    == limoi_port->index)
                {
                    LIMERR("%s multiple sharing of output buffer.\n",
                           base->name);
                    goto bail;
                }
                jj++;
            }
        } else
        {
            if (port2->eDir != OMX_DirOutput)
            {
                LIMERR("%s buffer can't be shared among input ports.\n",
                       base->name);
                break;
            }
        }

        //valid share pair found.
        //check to populate valid consumer
        limoi_port2->buffer_share.type = LIMOX_BufferShareConsumer;
        limoi_port2->buffer_share.ports[0] = limoi_port->index;
        limoi_port2->buffer_share.share_num++;

        //check to populate valid supplier
        limoi_port->buffer_share.type = LIMOX_BufferShareSupplier;
        limoi_port->buffer_share.ports
            [limoi_port->buffer_share.share_num] = limoi_port2->index;
        limoi_port->buffer_share.share_num++;

        //order ports to make sure suppliers are at top, so they
        //are allocated first.
        list_del_init(&limoi_port->list);
        list_add(&limoi_port->list, &base->ports_list_head);

        LIMDBG("%s buffer share enabled %u --> %u.\n",
               base->name, nPortIndex, limoi_port2->index);
        //limoi_omx_port_status(hComp, limoi_port->index);
        //limoi_omx_port_status(hComp, limoi_port2->index);

        valid_share_policy_found = 1;
        break;
    }

bail:
    lthread_mutex_unlock(&base->mutex);

    LIMDBG("%s port %u validated buffer share policy: %s.\n",
           base->name, nPortIndex, valid_share_policy_found? "YES" : "NO");

    return valid_share_policy_found? 0 : -1;
}

//don't care about mutex protection, as buffer share config is validated
//at tunnel setup, and remains readonly.
int buffer_share_is_supplier (
    OMX_HANDLETYPE hComp,
    limoi_port_t * limoi_port)
{
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComp);

    if (!base->buffer_share_policy || !base->buffer_share_pool) return 0;

    return (limoi_port->buffer_share.type == LIMOX_BufferShareSupplier)? 1 : 0;
}

void * buffer_share_get_supplier_buffer (
    OMX_HANDLETYPE hComp,
    limoi_port_t * limoi_port,
    int index)
{
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComp);
    limoi_port_t * supplier_port;
    limoi_port_buffer_t * port_buf;
    list_t * list;
    void * pBuffer = NULL;

    lthread_mutex_unlock(&limoi_port->mutex);

    supplier_port = limoi_component_get_port
        (hComp, limoi_port->buffer_share.ports[0]);
    if (NULL == supplier_port)
    {
        LIMERR("%s unable to find supplier port %u\n", base->name,
               limoi_port->buffer_share.ports[0]);
        goto bail;
    }

    lthread_mutex_lock(&supplier_port->mutex);
    list = supplier_port->buffers_list_head.next;
    while (index-- && list) list = list->next;
    if (NULL == list)
    {
        LIMERR("%s supplier port %u not enabled?\n", base->name,
               limoi_port->buffer_share.ports[0]);
        goto bail2;
    }
    port_buf = (limoi_port_buffer_t *)list;
    pBuffer = port_buf->header->pBuffer;

 bail2:
    lthread_mutex_unlock(&supplier_port->mutex);

 bail:
    lthread_mutex_lock(&limoi_port->mutex);

    return pBuffer;
}

int buffer_share_get_buffer_count_size (
    OMX_HANDLETYPE hComp,
    limoi_port_t * limoi_port,
    OMX_U32 * count,
    OMX_U32 * bytes)
{
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComp);
    limoi_port_t * share_port;
    OMX_PARAM_PORTDEFINITIONTYPE * port;
    OMX_PARAM_PORTDEFINITIONTYPE tunneled_port;
    int ii;
    int ret = 0;

    LIMDBG("%s port %u get actual buffer count and size.\n",
           base->name, limoi_port->index);

    lthread_mutex_lock(&limoi_port->mutex);
    port = LIMOI_PORTDEF(limoi_port);
    *bytes = port->nBufferSize;
    *count = port->nBufferCountActual;
    lthread_mutex_unlock(&limoi_port->mutex);

    if (tunnel_is_supplier(limoi_port))
    {
        limoi_set_spec_version(&tunneled_port,
                               sizeof(OMX_PARAM_PORTDEFINITIONTYPE));
        tunneled_port.nPortIndex = limoi_port->tunnel_port;
        if (OMX_ErrorNone !=
            OMX_GetParameter(limoi_port->tunnel_comp,
                             OMX_IndexParamPortDefinition,
                             &tunneled_port))
        {
            LIMERR("%s unable to get tunneled port %d params.\n",
                   base->name, limoi_port->tunnel_port);
            ret = -1;
            goto bail;
        }
        if (*bytes < tunneled_port.nBufferSize)
            *bytes = tunneled_port.nBufferSize;
        if (*count < tunneled_port.nBufferCountActual)
            *count = tunneled_port.nBufferCountActual;
    }

    if (!buffer_share_is_supplier(hComp, limoi_port)) goto bail;

    for (ii = 0; ii < limoi_port->buffer_share.share_num; ii++)
    {
        OMX_U32 count2;
        OMX_U32 bytes2;

        share_port = limoi_component_get_port
            (hComp, limoi_port->buffer_share.ports[ii]);

        if (buffer_share_get_buffer_count_size
            (hComp, share_port, &count2, &bytes2))
        {
            LIMERR("%s port %u unable to get buffer count.\n",
                   base->name, share_port->index);
            ret = -1;
            goto bail;
        }

        if (*bytes < bytes2) *bytes = bytes2;
        if (*count < count2) *count = count2;
    }

 bail:
    LIMDBG("%s port %u actual buffer count: %u, size: %u\n", base->name,
           limoi_port->index, *count, *bytes);

    return ret;
}

int buffer_share_set_buffer_count_size (
    OMX_HANDLETYPE hComp,
    limoi_port_t * limoi_port,
    OMX_U32 count,
    OMX_U32 bytes)
{
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComp);
    limoi_port_t * share_port;
    OMX_PARAM_PORTDEFINITIONTYPE * port;
    OMX_PARAM_PORTDEFINITIONTYPE tunneled_port;
    int ii;
    int ret = 0;

    LIMDBG("%s port %u set actual buffer count %d and size %d.\n",
           base->name, limoi_port->index, count, bytes);

    lthread_mutex_lock(&limoi_port->mutex);
    port = LIMOI_PORTDEF(limoi_port);
    port->nBufferSize = bytes;
    port->nBufferCountActual = count;
    lthread_mutex_unlock(&limoi_port->mutex);

    if (tunnel_is_supplier(limoi_port))
    {
        limoi_set_spec_version(&tunneled_port,
                               sizeof(OMX_PARAM_PORTDEFINITIONTYPE));
        tunneled_port.nPortIndex = limoi_port->tunnel_port;
        if (OMX_ErrorNone !=
            OMX_GetParameter(limoi_port->tunnel_comp,
                             OMX_IndexParamPortDefinition,
                             &tunneled_port))
        {
            LIMERR("%s unable to get tunneled port %d params.\n",
                   base->name, limoi_port->tunnel_port);
            ret = -1;
            goto bail;
        } else
        {
            tunneled_port.nBufferSize = bytes;
            tunneled_port.nBufferCountActual = count;
            if (OMX_ErrorNone !=
                OMX_SetParameter(limoi_port->tunnel_comp,
                                 OMX_IndexParamPortDefinition,
                                 &tunneled_port))
            {
                LIMERR("%s unable to set tunneled port %d params.\n",
                       base->name, limoi_port->tunnel_port);
                ret = -1;
                goto bail;
            }
        }
    }

    if (!buffer_share_is_supplier(hComp, limoi_port)) goto bail;

    for (ii = 0; ii < limoi_port->buffer_share.share_num; ii++)
    {
        share_port = limoi_component_get_port
            (hComp, limoi_port->buffer_share.ports[ii]);

        if (buffer_share_set_buffer_count_size
            (hComp, share_port, count, bytes))
        {
            LIMERR("%s port %u unable to set buffer count size.\n",
                   base->name, share_port->index);
            ret = -1;
            goto bail;
        }
    }

 bail:
    return ret;
}

int buffer_share_pool_in (
    OMX_HANDLETYPE hComp,
    limoi_port_t * in_port,
    limoi_port_t * out_port,
    OMX_BUFFERHEADERTYPE ** in_buf_header,
    OMX_BUFFERHEADERTYPE * out_buf_header)
{
    list_t * list;
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComp);
    share_pool_t * pool;
    share_request_t * rqst;
    int ret = 0;

    ASSERT(base->buffer_share_pool);
    ASSERT(in_port && in_buf_header && out_buf_header);

    pool = base->buffer_share_pool;
    if (pool->count >= LIMOI_BUFFER_SHARING_POOL_MAX)
    {
        LIMERR("%s buffer sharing pool full.\n", base->name);
        ret = -1; goto bail;
    }

    LIMDBG("%s port %u buffer share request in: %p\n"
           "\trequests in pool  : %d\n",
           base->name, in_port->index,
           in_buf_header, pool->count);

    list = pool->rqst_head.next;
    while (list != &pool->rqst_head)
    {
        rqst = (share_request_t *)list;
        if (in_port == rqst->in_port
            && out_port == rqst->out_port
            && in_buf_header == rqst->in_buf_header
            && out_buf_header == rqst->out_buf_header)
        {
            rqst->ref_count++;
            goto bail;
        }

        if (NULL == rqst->in_port)
        {
            rqst->in_port = in_port;
            rqst->out_port = out_port;
            rqst->in_buf_header = in_buf_header;
            rqst->out_buf_header = out_buf_header;
            rqst->ref_count = 1;
            pool->count++;
            goto bail;
        }
        list = list->next;
    }

    LIMERR("%s unexpected error, pool overflow.\n", base->name);
    ASSERT(0);

 bail:
    return ret;
}

int buffer_share_pool_out (
    OMX_HANDLETYPE hComp,
    limoi_port_t * out_port,
    OMX_BUFFERHEADERTYPE * out_buf_header)
{
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComp);
    share_pool_t * pool;
    share_request_t * rqst;
    list_t * list;
    int ret = 1;

    ASSERT(out_port && out_buf_header);

    pool = base->buffer_share_pool;
    if (pool->count <= 0)
    {
        LIMDBG("%s buffer sharing pool empty.\n", base->name);
        goto bail;
    }

    list = pool->rqst_head.next;
    while (list != &pool->rqst_head)
    {
        rqst = (share_request_t *)list;
        if (rqst->in_port && out_buf_header == rqst->out_buf_header)
        {
            share_request_t rqst_copy = *rqst;
            rqst->ref_count--;

            LIMDBG("%s port %u buffer share request out: %p\n"
                   "\trequest ref_count : %d\n"
                   "\trequests in pool  : %d\n",
                   base->name, rqst->in_port->index,
                   rqst->in_buf_header, rqst->ref_count, pool->count);

            if (rqst->ref_count == 0)
            {
                pool->count--;
                rqst->in_port = NULL;
                list_del_init(&rqst->list);
                list_add_tail(&rqst->list, &pool->rqst_head);
            }

            lthread_mutex_unlock(&base->mutex);

            limoi_component_release_port_buffer
                (hComp, rqst_copy.in_port, rqst_copy.in_buf_header);
            lthread_mutex_lock(&base->mutex);
            ret = 0;
            goto bail;
        }

        if (NULL == rqst->in_port)
        {
            LIMDBG("%s buffer not in share request pool.\n", base->name);
            goto bail;
        }
        list = list->next;
    }

 bail:
    return ret;
}

OMX_ERRORTYPE limoi_component_set_buffer_share_policy (
    OMX_HANDLETYPE hComp,
    const limoi_buffer_share_policy_t * policy)
{
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComp);
    OMX_ERRORTYPE ret = OMX_ErrorNone;

    lthread_mutex_lock(&base->mutex);
    if (base->state != OMX_StateLoaded)
    {
        LIMERR("%s buffer sharing policy can only be set in loaded state\n",
               base->name);
        ret = OMX_ErrorInvalidState;
        goto bail;
    }

    if (NULL == base->buffer_share_policy)
        base->buffer_share_policy = malloc(sizeof(limoi_buffer_share_policy_t));

    if (NULL == base->buffer_share_policy)
    {
        LIMERR("%s out of memory.\n", base->name);
        ret = OMX_ErrorInsufficientResources;
        goto bail;
    }

    *(base->buffer_share_policy) = *policy;

 bail:
    lthread_mutex_unlock(&base->mutex);

    return ret;
}

int limoi_component_port_buffer_share_enabled (limoi_port_t * port)
{
    return (LIMOX_BufferShareDisabled == port->buffer_share.type)? 0 : 1;
}
