/*
  Copyright (C) 2011 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX IL component command interface subroutines.

  REVISION:

    1) initial creation. ----------------------------------------- 2011-07-22 MG
*/
#define _GNU_SOURCE
#include "limoi.h"
#include "component.h"
#include "string.h"
#include <stdarg.h>

//#undef LIMDBG
#ifndef LIMDBG
#define LIMDBG(fmt, arg...)
#endif

static void set_port_transition_state (
    OMX_HANDLETYPE hComp,
    OMX_U32 nPortIndex,
    int transition_state)
{
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComp);
    list_t * list;
    limoi_port_t * limoi_port;

    list = base->ports_list_head.next;
    while (list != &base->ports_list_head)
    {
        limoi_port = (limoi_port_t *)list;
        lthread_mutex_lock(&limoi_port->mutex);
        if (nPortIndex == limoi_port->index || nPortIndex == OMX_ALL)
        {
            if (transition_state) limoi_port->bufq_state_transition++;
            else limoi_port->bufq_state_transition--;
            lthread_cond_broadcast(&limoi_port->cond);
        }
        if (nPortIndex == limoi_port->index)
        {
            lthread_mutex_unlock(&limoi_port->mutex);
            break;
        }
        lthread_mutex_unlock(&limoi_port->mutex);
        list = list->next;
    }
}

static void process_cmd (
    limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(cmd->hComponent);

    switch (cmd->Cmd)
    {
    case OMX_CommandStateSet :
        ret = state_set(cmd);
        break;

    case OMX_CommandPortDisable :
        port_disable(cmd);
        goto bail;

    case OMX_CommandPortEnable :
        port_enable(cmd);
        goto bail;

    case OMX_CommandFlush :
        port_flush(cmd);
        goto bail;

    case OMX_CommandMarkBuffer :

    default :
        LIMDBG("unhandled command : %d\n", cmd->Cmd);
        break;
    }

    // now send return code back to client through callback.
    if (OMX_ErrorNone == ret)
    {
        event_deliver(cmd->hComponent, OMX_EventCmdComplete,
                      cmd->Cmd, cmd->nParam1, cmd->pCmdData);
    } else
    {
        event_deliver(cmd->hComponent, OMX_EventError,
                      ret, 0, NULL);
    }

 bail:
    lthread_mutex_lock(&base->mutex);
    if (cmd->Cmd == OMX_CommandStateSet || cmd->Cmd == OMX_CommandFlush)
    {
        set_port_transition_state(cmd->hComponent, OMX_ALL, 0);
        lthread_cond_broadcast(&base->cond);
    }
    lthread_mutex_unlock(&base->mutex);

    event_deliver_real(cmd->hComponent);

    if (LIMOI_COMPONENT_PRIV(cmd->hComponent)->synchronous)
    {
        if (OMX_CommandPortEnable == cmd->Cmd
            || (OMX_CommandStateSet == cmd->Cmd
                && OMX_StateExecuting == cmd->nParam1))
        {
            sync_consume_buffers(cmd->hComponent);
        }
    }
}

static void * cmd_thread (
    void * arg)
{
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(arg);
    limoi_component_cmd_t * cmd;

    lthread_mutex_lock(&base->cmd_mutex);
    base->cmd_received = 0;
    lthread_cond_broadcast(&base->cmd_cond);
    lthread_mutex_unlock(&base->cmd_mutex);

    while (1)
    {
        lthread_mutex_lock(&base->cmd_mutex);

        while (base->cmd_received == 0)
        {
            lthread_cond_wait(&base->cmd_cond, &base->cmd_mutex);
        }

        if (base->cmd_received < 0)
        {
            lthread_mutex_unlock(&base->cmd_mutex);
            break;
        }

        cmd = limq_dequeue_get(base->cmd_q);
        if (!cmd)
        {
            lthread_mutex_unlock(&base->cmd_mutex);
            continue;
        }
        base->cmd_received--;
        lthread_mutex_unlock(&base->cmd_mutex);

        LIMDBG("%s processing command: %s arg: %u\n", base->name,
               limoi_omx_command_string(cmd->Cmd), cmd->nParam1);
        process_cmd(cmd);
        LIMDBG("%s processed command: %s arg: %u\n", base->name,
               limoi_omx_command_string(cmd->Cmd), cmd->nParam1);

        lthread_mutex_lock(&base->cmd_mutex);
        limq_dequeue_put(base->cmd_q, cmd);
        lthread_mutex_unlock(&base->cmd_mutex);
    }

    lthread_mutex_lock(&base->cmd_mutex);
    base->cmd_received = -1;
    lthread_cond_broadcast(&base->cmd_cond);
    lthread_mutex_unlock(&base->cmd_mutex);

    return NULL;
}

int command_interface_init (OMX_HANDLETYPE hComp)
{
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComp);

    // command interface intialization.
    lthread_mutex_init(&base->cmd_mutex);
    lthread_cond_init(&base->cmd_cond);
    base->cmd_received = -1;

    base->cmd_q = limq_create(sizeof(limoi_component_cmd_t),
                              LIMOI_COMPONENT_CMD_Q_DEPTH_MAX,
                              "component cmd queue");
    if (NULL == base->cmd_q)
    {
        LIMERR("unable to allocate command queue.\n");
        goto bail;
    }
    LIMDBG("created cmd queue.\n");
    if (lthread_create(&base->cmd_thread, cmd_thread, hComp, 0))
    {
        LIMERR("unable to create command thread.\n");
        goto bail;
    }
    // wait till command thread is up running.
    lthread_mutex_lock(&base->cmd_mutex);
    while (base->cmd_received == -1)
    {
        lthread_cond_wait(&base->cmd_cond, &base->cmd_mutex);
    }
    lthread_mutex_unlock(&base->cmd_mutex);

    return 0;

 bail:
    lthread_mutex_destroy(&base->cmd_mutex);
    lthread_cond_destroy(&base->cmd_cond);
    if (base->cmd_q) limq_destroy(base->cmd_q);

    return -1;
}

void command_interface_deinit (OMX_HANDLETYPE hComp)
{
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComp);

    // shuts down the command thread.
    lthread_mutex_lock(&base->cmd_mutex);
    base->cmd_received = -LIMOI_COMPONENT_CMD_Q_DEPTH_MAX;
    lthread_cond_broadcast(&base->cmd_cond);

    // wait till command thread is shut down.
    while (base->cmd_received != -1)
    {
        lthread_cond_wait(&base->cmd_cond, &base->cmd_mutex);
    }
    lthread_mutex_unlock(&base->cmd_mutex);
    lthread_join(base->cmd_thread);
    LIMDBG("%s deinit: cmd thread shut down.\n", base->name);

    lthread_mutex_destroy(&base->cmd_mutex);
    lthread_cond_destroy(&base->cmd_cond);
    LIMDBG("%s deinit: destroy cmd_q.\n", base->name);
    limq_destroy(base->cmd_q);
    LIMDBG("%s deinit: destroyed cmd_q.\n", base->name);
}

OMX_ERRORTYPE command_callback (
    limoi_component_priv_t * base,
    limoi_component_cmd_t * cmd,
    int prep)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;

    if ((prep && !base->limoi_callbacks.command_prep)
        || (!prep && !base->limoi_callbacks.command_post))
    {
        LIMDBG("%s command handle not set, prep | post: %p | %p.\n",
               base->name, base->limoi_callbacks.command_prep,
               base->limoi_callbacks.command_post);
        goto bail;
    } else
    {
        LIMDBG("%s limoi component command callback.\n", base->name);
        ret = prep?
            base->limoi_callbacks.command_prep(cmd) :
            base->limoi_callbacks.command_post(cmd);

        if (OMX_ErrorNone != ret)
        {
            LIMERR("%s command callback returned error %s\n",
                   base->name, limoi_omx_error_string(ret));
        }
    }

bail:
    return ret;
}

OMX_ERRORTYPE omx_send_command (
    OMX_HANDLETYPE hComponent,
    OMX_COMMANDTYPE Cmd,
    OMX_U32 nParam1,
    OMX_PTR pCmdData)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    limoi_component_cmd_t * cmd;
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComponent);

    lthread_mutex_lock(&base->mutex);
    //block waiting on 'putting_buffer' flag only, instead of
    //'base->busy', this gives component better response, as long
    //as downstream component is IL compliant, the waiting here
    //should not be longer than 5ms.
    while (base->putting_buffer)
    {
        if (lthread_cond_timedwait(&base->cond, &base->mutex, 100))
        {
            LIMINFO("%s timeout, downstream component blocking too long?\n",
                    base->name);
            break;
        }
    }

    //port is immediately disabled, so no buffer gets leaked
    //out once this call returns.
    //port enabling takes sometime, have to wait till all buffers
    //are allocated and populated.
    if (Cmd == OMX_CommandPortDisable)
    {
        port_disable_now(hComponent, nParam1);
    } else if (Cmd == OMX_CommandStateSet || Cmd == OMX_CommandFlush)
    {
        //immediately disable buffer access for all ports, buffer can still
        //be 'put back/released', but 'get' is not allowed, so that state
        //transition can complete without block waiting.
        set_port_transition_state(hComponent, OMX_ALL, 1);
    }
    lthread_mutex_unlock(&base->mutex);

    //queue command
    lthread_mutex_lock(&base->cmd_mutex);

    cmd = limq_enqueue_get(base->cmd_q);
    if (NULL == cmd)
    {
        LIMERR("%s command event queue full.\n", base->name);
        ret = OMX_ErrorInsufficientResources;
        goto bail;
    }
    cmd->hComponent = hComponent;
    cmd->Cmd = Cmd;
    cmd->nParam1 = nParam1;
    cmd->pCmdData = pCmdData;

    limq_enqueue_put(base->cmd_q, cmd);
    base->cmd_received++;
    LIMDBG("%s received command: %s arg: %u\n", base->name,
           limoi_omx_command_string(Cmd), nParam1);
    lthread_cond_broadcast(&base->cmd_cond);
 bail:
    lthread_mutex_unlock(&base->cmd_mutex);

    return ret;
}
