/*
  Copyright (C) 2009 - 2011 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX IL component common code.

  REVISION:

    2) synchronous component support. ---------------------------- 2011-06-01 MG
    1) initial creation. ----------------------------------------- 2009-11-30 MG
*/
#define _GNU_SOURCE
#include "limoi.h"
#include "component.h"
#include "string.h"
#include <stdarg.h>

//#undef LIMDBG
#ifndef LIMDBG
#define LIMDBG(fmt, arg...)
#endif

static OMX_ERRORTYPE omx_get_extension_index (
    OMX_HANDLETYPE hComponent,
    OMX_STRING cParameterName,
    OMX_INDEXTYPE* pIndexType)
{
    return OMX_ErrorNotImplemented;
}

static OMX_ERRORTYPE omx_use_EGL_image (
    OMX_HANDLETYPE hComponent,
    OMX_BUFFERHEADERTYPE** ppBufferHdr,
    OMX_U32 nPortIndex,
    OMX_PTR pAppPrivate,
    void* eglImage)
{
    return OMX_ErrorNotImplemented;
}

#define omx_component_deinit limoi_component_deinit

static OMX_ERRORTYPE component_init (
    OMX_HANDLETYPE * hComp,
    int bytes,
    int synchronous)
{
    void * priv;
    limoi_component_priv_t * base;
    OMX_COMPONENTTYPE * comp;

    *hComp = calloc(1, sizeof(OMX_COMPONENTTYPE));
    if (NULL == *hComp)
    {
        LIMERR("unable to allocate component.\n");
        return OMX_ErrorInsufficientResources;
    }

    comp = (OMX_COMPONENTTYPE *)(*hComp);
    priv = calloc(1, bytes);
    if (NULL == priv)
    {
        LIMERR("unable to allocate component private data: %d.\n", bytes);
        free(*hComp);
        return OMX_ErrorInsufficientResources;
    }

    comp->pComponentPrivate = priv;

    // allocate base component private data
    base = calloc(1, sizeof(limoi_component_priv_t));
    if (NULL == base)
    {
        free(priv);
        free(*hComp);
        LIMERR("unable to allocate base component private data.\n");
        return OMX_ErrorInsufficientResources;
    } else *((limoi_component_t *)priv) = base;

    base->synchronous = synchronous;
    base->busy = 0;
    base->state_transition = 0;

    if (command_interface_init(*hComp))
    {
        free(base);
        free(priv);
        free(*hComp);
        return OMX_ErrorInsufficientResources;
    }

    // client event set intialization.
    lthread_mutex_init(&base->event_mutex);
    base->event_set = limset_create
        (sizeof(limoi_component_client_event_t),
         LIMOI_COMPONENT_CLIENT_EVENT_SET_DEPTH_MAX,
         "component client event set");
    if (NULL == base->event_set)
    {
        LIMERR("unable to allocate client event set.\n");
        command_interface_deinit(*hComp);
        free(base);
        free(priv);
        free(*hComp);
        return OMX_ErrorInsufficientResources;
    }

    // other generic intialization.
    base->state = OMX_StateLoaded;
    lthread_mutex_init2(&base->mutex, LTHREAD_MUTEX_RECURSIVE);
    lthread_cond_init(&base->cond);
    base->lock_count = 0;

    // other
    list_init(&base->roles_list_head);
    list_init(&base->ports_list_head);
    list_init(&base->params_list_head);

    // construct base component.
    limoi_set_spec_version(comp, sizeof(OMX_COMPONENTTYPE));

    comp->GetComponentVersion = omx_get_component_version;
    comp->SendCommand = omx_send_command;
    comp->GetParameter = omx_get_parameter;
    comp->SetParameter = omx_set_parameter;
    comp->GetConfig = omx_get_config;
    comp->SetConfig = omx_set_config;
    comp->GetExtensionIndex = omx_get_extension_index;
    comp->GetState = omx_get_state;
    comp->ComponentTunnelRequest = omx_component_tunnel_request;
    comp->UseBuffer = omx_use_buffer;
    comp->AllocateBuffer = omx_allocate_buffer;
    comp->FreeBuffer = omx_free_buffer;
    comp->EmptyThisBuffer = omx_empty_this_buffer;
    comp->FillThisBuffer = omx_fill_this_buffer;
    comp->SetCallbacks = omx_set_callbacks;
    comp->ComponentDeInit = omx_component_deinit;
    comp->UseEGLImage = omx_use_EGL_image;
    comp->ComponentRoleEnum = omx_component_role_enum;

    return OMX_ErrorNone;
}

OMX_ERRORTYPE limoi_component_init (OMX_HANDLETYPE * hComp, int bytes)
{
    return component_init(hComp, bytes, 0);
}

OMX_ERRORTYPE limoi_component_init_sync (OMX_HANDLETYPE * hComp, int bytes)
{
    return component_init(hComp, bytes, 1);
}

OMX_ERRORTYPE limoi_component_deinit (OMX_HANDLETYPE hComponent)
{
    OMX_COMPONENTTYPE * comp = (OMX_COMPONENTTYPE *)hComponent;
    limoi_component_priv_t * base;
    if (NULL == comp) goto bail;

    if (NULL == comp->pComponentPrivate) goto bail;
    else base = *((limoi_component_t *)comp->pComponentPrivate);
    ASSERT(base);
    LIMDBG("deiniting component: %s\n", base->name);

    command_interface_deinit(hComponent);

    if (!base->synchronous)
    {
        lthread_mutex_lock(&base->mutex);
        // last resort to cleanly destroy component if in wrong state.
        port_disable_now(hComponent, OMX_ALL);
        if (OMX_StateLoaded != base->state && OMX_StateInvalid != base->state)
        {
            base->state = OMX_StateInvalid;
            while (-1 != base->state_transition)
            {
                if (lthread_cond_timedwait(&base->cond, &base->mutex, 5000))
                {
                    LIMERR("%s main thread shutdown timedout.\n", base->name);
                    break;
                }
            }
        }
        lthread_mutex_unlock(&base->mutex);
    }
    lthread_mutex_destroy(&base->mutex);

    limset_destroy(base->event_set);
    lthread_mutex_destroy(&base->event_mutex);

    // other
    LIMDBG("%s deinit: releasing roles and ports.\n", base->name);
    role_release(&base->roles_list_head);
    port_release(&base->ports_list_head, hComponent);
    param_release(&base->params_list_head);

    if (base->buffer_share_policy) free(base->buffer_share_policy);
    if (base->buffer_share_pool) free(base->buffer_share_pool);

    if (base->name) free(base->name);

    if (base->audio_port) free(base->audio_port);
    if (base->video_port) free(base->video_port);
    if (base->other_port) free(base->other_port);

    free(base);
    free(comp->pComponentPrivate);
    free(hComponent);

 bail:
    return OMX_ErrorNone;
}
