/*
  Copyright (C) 2010 - 2011 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX IL standard container demuxer component.

  REVISION:

    1) initial creation. ----------------------------------------- 2010-01-27 MG
*/
#define _GNU_SOURCE
#include "limoi.h"
#include "other-priv.h"

//#undef LIMDBG
#ifndef LIMDBG
#define LIMDBG(fmt, arg...)
#endif

static OMX_ERRORTYPE sub_handle (
    OMX_HANDLETYPE hComp,
    limoi_port_t ** audio_port,
    limoi_port_t ** video_port,
    limoi_port_t ** other_port)
{
    limoi_demuxer_t * demuxer = LIMOI_DEMUXER(hComp);
    limoi_port_t * clock_port = *other_port;
    OMX_BUFFERHEADERTYPE ** clock_buf = NULL;
    OMX_ERRORTYPE ret = OMX_ErrorNone;

    if ((!clock_port->tunnel_comp
         && LIMOI_PORTDEF(clock_port)->bEnabled == OMX_TRUE)
        || !demuxer->priv->callbacks.clock)
    {
        ret = OMX_ErrorNoMore;
        goto bail;
    }

    LIMDBG("%s demuxer get clock port buffer.\n",
           LIMOI_COMPONENT_PRIV(hComp)->name);
    clock_buf = limoi_component_get_port_buffer(hComp, clock_port, 0);

    if (NULL == clock_buf)
    {
        LIMDBG("%s demuxer null clock port buffer.\n",
               LIMOI_COMPONENT_PRIV(hComp)->name);
        ret = OMX_ErrorNotReady;
        goto bail;
    }

    if (!(*clock_buf)->nFilledLen)
    {
        LIMDBG("%s demuxer clock port buffer zero length.\n",
               LIMOI_COMPONENT_PRIV(hComp)->name);
        if (-1 == limoi_component_put_port_buffer(hComp, clock_port, clock_buf))
        {
            ret = OMX_ErrorNotReady;
        }
        goto bail;
    }

    demuxer->priv->callbacks.clock(hComp, clock_port->tunnel_comp, *clock_buf);

    if (!(*clock_buf)->nFilledLen)
    {
        int tmp;

        tmp = limoi_component_put_port_buffer(hComp, clock_port, clock_buf);
        if (tmp)
        {
            if (tmp == -1) ret = OMX_ErrorNotReady;
            else
            {
                LIMDBG("%s unable to put back port buffer.\n",
                       LIMOI_COMPONENT_PRIV(hComp)->name);
            }
        }
    } else
    {
        LIMDBG("%s demuxer clock port buffer not consumed.\n",
               LIMOI_COMPONENT_PRIV(hComp)->name);
        limoi_component_release_port_buffer(hComp, clock_port, clock_buf);
    }

 bail:
    return ret;
}

static OMX_ERRORTYPE main_handle (
    OMX_HANDLETYPE hComp,
    limoi_port_t ** audio_port,
    limoi_port_t ** video_port,
    limoi_port_t ** other_port,
    int * preloaded)
{
    OMX_ERRORTYPE  res = OMX_ErrorNone;
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComp);
    limoi_demuxer_t * demuxer = LIMOI_DEMUXER(hComp);
    limoi_demuxer_priv_t * pdemuxer = demuxer->priv;
    limoi_port_t * audio_port_out = *audio_port;
    limoi_port_t * video_port_out = *video_port;
    OMX_BUFFERHEADERTYPE ** audio_out = NULL;
    OMX_BUFFERHEADERTYPE ** video_out = NULL;
    OMX_U32 audio_flags;
    OMX_U32 video_flags;
    int fatal_error = 0;

    if (*preloaded == 0)
    {
        //do not requrie preload.
        *preloaded = LIMOI_PRELOAD_BUFFER_COUNT;
        goto bail;
    }

    LIMDBG("%s demuxer get audio output port buffer.\n", base->name);
    audio_out = limoi_component_get_port_buffer(hComp, audio_port_out, -1);
    if (NULL == audio_out)
    {
        LIMDBG("%s unable to get audio output port buffer.\n", base->name);
        res = OMX_ErrorNotReady;
        goto bail;
    } else
    {
        audio_flags = (*audio_out)->nFlags;
    }
    LIMDBG("%s audio output buffer len: %u %p\n",
           base->name, (*audio_out)->nAllocLen, (*audio_out)->pBuffer);

    LIMDBG("%s demuxer get video output port buffer.\n", base->name);
    video_out = limoi_component_get_port_buffer(hComp, video_port_out, -1);
    if (NULL == video_out)
    {
        LIMDBG("%s unable to get video output port buffer.\n", base->name);
        res = OMX_ErrorNotReady;
        goto bail;
    } else
    {
        video_flags = (*video_out)->nFlags;
    }
    LIMDBG("%s video output buffer len: %u/%u %p\n",
           base->name, (*video_out)->nFilledLen, (*video_out)->nAllocLen,
           (*video_out)->pBuffer);

    //continue demuxing only if output has been fully consumed.
    if ((*audio_out)->nFilledLen == 0 && (*video_out)->nFilledLen == 0)
    {
        res = pdemuxer->callbacks.demux
            (hComp, pdemuxer->pipe, pdemuxer->content, *audio_out, *video_out);
    }

    if (OMX_ErrorNone == res)
    {
        if (audio_flags & OMX_BUFFERFLAG_EOS)
        {
            limoi_component_deliver_event
                (hComp, OMX_EventBufferFlag, 0, audio_flags, NULL);
            res = OMX_ErrorNoMore;
        }

        if (video_flags & OMX_BUFFERFLAG_EOS)
        {
            limoi_component_deliver_event
                (hComp, OMX_EventBufferFlag, 1, video_flags, NULL);
            res = OMX_ErrorNoMore;
        }

        if ((*audio_out)->nFilledLen || audio_flags & OMX_BUFFERFLAG_EOS)
        {
            LIMDBG("%s put audio output port buffer back: %p.\n",
               base->name, (*audio_out)->pBuffer);
            if (-1 == limoi_component_put_port_buffer
                (hComp, audio_port_out, audio_out))
            {
                LIMDBG("%s unable to put back audio output port buffer.\n"
                       "\tnFilledLen: %u, flags: %u\n",
                       base->name, (*audio_out)->nFilledLen, audio_flags);
                res = OMX_ErrorNotReady;
            }
            audio_out = NULL;
        }

        if ((*video_out)->nFilledLen || video_flags & OMX_BUFFERFLAG_EOS)
        {
            LIMDBG("%s put video output port buffer back:\n"
                   "\taddr       : %p\n"
                   "\tnFilledLen : %u\n"
                   "\tflags      : %x\n",
                   base->name, (*video_out)->pBuffer,
                   (*video_out)->nFilledLen, video_flags);
            if (-1 == limoi_component_put_port_buffer
                (hComp, video_port_out, video_out))
            {
                LIMDBG("%s unable to put back video output port buffer.\n"
                        "\tnFilledLen: %u, flags: %u\n",
                        base->name, (*video_out)->nFilledLen, video_flags);
                res = OMX_ErrorNotReady;
            }
            video_out = NULL;
        }
    } else fatal_error = 1;

 bail :
    if (audio_out)
        limoi_component_release_port_buffer(hComp, audio_port_out, audio_out);
    if (video_out)
        limoi_component_release_port_buffer(hComp, video_port_out, video_out);

    if (fatal_error)
    {
        limoi_component_goto_invalid_state(hComp);
    }

    return res;
}

static OMX_ERRORTYPE set_state (
    limoi_component_cmd_t * cmd,
    int * cmd_handle_called)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    OMX_STATETYPE next_state = cmd->nParam1;
    limoi_demuxer_t * demuxer = LIMOI_DEMUXER(cmd->hComponent);
    limoi_demuxer_priv_t * pdemuxer = demuxer->priv;
    limoi_component_priv_t * base =  LIMOI_COMPONENT_PRIV(cmd->hComponent);
    void * buf[(sizeof(OMX_PARAM_CONTENTURITYPE) + LIMOI_URI_MAX_LENGTH) /
               sizeof(void *)];
    OMX_PARAM_CONTENTURITYPE * uri_param;

    lthread_mutex_lock(&base->mutex);
    switch (base->state)
    {
    case OMX_StateLoaded :

        // transition to StateIdle
        pdemuxer->pipe = NULL;
        pdemuxer->content = NULL;

        uri_param = (OMX_PARAM_CONTENTURITYPE *)&buf[0];
        limoi_set_spec_version(uri_param, sizeof(OMX_PARAM_CONTENTURITYPE)
                               + LIMOI_URI_MAX_LENGTH);
        ret = limoi_component_get_parameter(cmd->hComponent,
                                            OMX_IndexParamContentURI,
                                            uri_param);
        if (OMX_ErrorNone == ret)
        {
            ret = OMX_GetContentPipe((OMX_HANDLETYPE *)&pdemuxer->pipe,
                                     (OMX_STRING)uri_param->contentURI);
            if (OMX_ErrorNone != ret)
            {
                LIMERR("%s unable to open content pipe at %s\n",
                       base->name, uri_param->contentURI);
            } else
            {
                if (pdemuxer->pipe->Open(&pdemuxer->content,
                                         (OMX_STRING)uri_param->contentURI,
                                         CP_AccessRead))
                {
                    LIMERR("%s unable to access content: %s\n",
                           base->name, uri_param->contentURI);
                    ret = OMX_ErrorContentPipeOpenFailed;
                }
            }
        } else
        {
            LIMERR("%s unable to get URI parameter.\n", base->name);
        }
        break;

    case OMX_StateIdle :
        if (OMX_StateLoaded == next_state)
        {
            if (pdemuxer->callbacks.command_post)
            {
                *cmd_handle_called = 1;
                ret = pdemuxer->callbacks.command_post(cmd);
                if (OMX_ErrorNone != ret)
                {
                    LIMERR("%s command callback returned error %s\n",
                           LIMOI_COMPONENT_PRIV(cmd->hComponent)->name,
                           limoi_omx_error_string(ret));
                }
            }

            if (pdemuxer->pipe && pdemuxer->content)
            {
                pdemuxer->pipe->Close(pdemuxer->content);
            }
        }
        break;

    default :
        break;
    }
    lthread_mutex_unlock(&base->mutex);
    return ret;
}

static OMX_ERRORTYPE prep_cmd_handle (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    limoi_demuxer_t * demuxer = LIMOI_DEMUXER(cmd->hComponent);
    limoi_demuxer_priv_t * pdemuxer = demuxer->priv;

    if (pdemuxer->callbacks.command_prep)
    {
        ret = pdemuxer->callbacks.command_prep(cmd);
        if (OMX_ErrorNone != ret)
        {
            LIMERR("%s command callback returned error %s\n",
                   LIMOI_COMPONENT_PRIV(cmd->hComponent)->name,
                   limoi_omx_error_string(ret));
        }
    }
    return ret;
}

static OMX_ERRORTYPE post_cmd_handle (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    limoi_demuxer_t * demuxer = LIMOI_DEMUXER(cmd->hComponent);
    limoi_demuxer_priv_t * pdemuxer = demuxer->priv;
    int post_cmd_handle_called = 0;

    switch (cmd->Cmd)
    {
    case OMX_CommandStateSet :
        ret = set_state(cmd, &post_cmd_handle_called);
        if (OMX_ErrorNone != ret)
        {
            LIMERR("%s post state set handle returned error %s\n",
                   LIMOI_COMPONENT_PRIV(cmd->hComponent)->name,
                   limoi_omx_error_string(ret));
            goto bail;
        }
        break;

    default :
        break;
    }

    if (pdemuxer->callbacks.command_post && !post_cmd_handle_called)
    {
        ret = pdemuxer->callbacks.command_post(cmd);
        if (OMX_ErrorNone != ret)
        {
            LIMERR("%s command callback returned error %s\n",
                   LIMOI_COMPONENT_PRIV(cmd->hComponent)->name,
                   limoi_omx_error_string(ret));
            goto bail;
        }
    }

 bail:
    return ret;
}

static int add_port (
    OMX_HANDLETYPE hComp)
{
	OMX_PARAM_PORTDEFINITIONTYPE port;
    OMX_PORT_PARAM_TYPE port_param;

    limoi_set_spec_version(&port_param, sizeof(OMX_PORT_PARAM_TYPE));
    port_param.nPorts = 1;
    port_param.nStartPortNumber = 0; //APB = 0
    limoi_component_set_parameter(hComp, OMX_IndexParamAudioInit, &port_param);

    port_param.nStartPortNumber = 1; //VPB = 1
    limoi_component_set_parameter(hComp, OMX_IndexParamVideoInit, &port_param);

    port_param.nStartPortNumber = 2; //OPB = 2
    limoi_component_set_parameter(hComp, OMX_IndexParamOtherInit, &port_param);

    limoi_set_spec_version(&port, sizeof(OMX_PARAM_PORTDEFINITIONTYPE));

    port.nPortIndex = 0;
    port.eDir = OMX_DirOutput;
    port.nBufferCountActual = LIMOI_DEMUXER_AUDIO_OUTPUT_BUFFER_NUM;
    port.nBufferCountMin = LIMOI_DEMUXER_AUDIO_OUTPUT_BUFFER_NUM;
    port.nBufferSize = LIMOI_DEMUXER_AUDIO_OUTPUT_BUFFER_BYTES;
    port.bEnabled = OMX_TRUE;
    port.bPopulated = OMX_FALSE;
    port.eDomain = OMX_PortDomainAudio;
    port.bBuffersContiguous = OMX_FALSE;
    port.nBufferAlignment = 0;
    port.format.audio.cMIMEType = NULL;
    port.format.audio.pNativeRender = 0;
    port.format.audio.bFlagErrorConcealment = OMX_FALSE;
    port.format.audio.eEncoding = OMX_AUDIO_CodingAutoDetect;
    limoi_component_add_port(hComp, &port);

    port.nPortIndex = 1;
    port.nBufferCountActual = LIMOI_DEMUXER_VIDEO_OUTPUT_BUFFER_NUM;
    port.nBufferCountMin = LIMOI_DEMUXER_VIDEO_OUTPUT_BUFFER_NUM;
    port.nBufferSize = LIMOI_DEMUXER_VIDEO_OUTPUT_BUFFER_BYTES;
    port.eDomain = OMX_PortDomainVideo;
    port.format.video.eCompressionFormat = OMX_VIDEO_CodingAutoDetect;
    port.format.video.eColorFormat = OMX_COLOR_FormatUnused;
    limoi_component_add_port(hComp, &port);

    port.nPortIndex = 2;
    port.eDir = OMX_DirInput;
    port.nBufferCountActual = LIMOI_MAX_CLOCK_PORT_BUFFERS;
    port.nBufferCountMin = LIMOI_MAX_CLOCK_PORT_BUFFERS;
    port.eDomain = OMX_PortDomainOther;
    port.format.other.eFormat = OMX_OTHER_FormatTime;
    limoi_component_add_port(hComp, &port);

    return 0;
}

static int add_metadata_filter (OMX_HANDLETYPE hComp)
{
    OMX_PARAM_METADATAFILTERTYPE filter;

    limoi_set_spec_version(&filter, sizeof(OMX_PARAM_METADATAFILTERTYPE));

    filter.bAllKeys = OMX_TRUE;
    filter.nLanguageCountrySizeUsed = 0;
    filter.bEnabled = OMX_TRUE;

    limoi_component_set_parameter
        (hComp, OMX_IndexParamMetadataFilterType, &filter);

    return 0;
}

int limoi_demuxer_set_port_params (
    OMX_HANDLETYPE hComp,
    int num_audio_streams,
    limoi_audio_format_t * audio_format,
    int num_video_streams,
    limoi_video_format_t * video_format)
{
    OMX_PARAM_U32TYPE num_streams;
    OMX_AUDIO_PARAM_PORTFORMATTYPE audio_port_format;
    OMX_VIDEO_PARAM_PORTFORMATTYPE video_port_format;
    int audio_port_format_detected = 0;
    int audio_port_settings_changed = 0;
    int video_port_format_detected = 0;
    int video_port_settings_changed = 0;
    int ii;

    LIMDBG("%s number of audio streams: %d, number of video streams: %d\n",
           LIMOI_COMPONENT_PRIV(hComp)->name, num_audio_streams, num_video_streams);

    limoi_set_spec_version(&num_streams, sizeof(OMX_PARAM_U32TYPE));
    num_streams.nPortIndex = 0;
    if (OMX_ErrorNone ==
        limoi_component_get_parameter(hComp, OMX_IndexParamNumAvailableStreams,
                                      &num_streams))
    {
        if (num_streams.nU32 != num_audio_streams)
            audio_port_settings_changed = 1;
    } else if (num_audio_streams) audio_port_settings_changed = 1;

    if (audio_port_settings_changed)
    {
        num_streams.nU32 = num_audio_streams;
        limoi_component_set_parameter(hComp, OMX_IndexParamNumAvailableStreams,
                                      &num_streams);

        if (num_audio_streams)
        {
            audio_port_format_detected = 1;
            limoi_set_spec_version(&audio_port_format,
                                   sizeof(OMX_AUDIO_PARAM_PORTFORMATTYPE));
            audio_port_format.nPortIndex = 0;
            for (ii = 0; ii < num_audio_streams; ii++)
            {
                audio_port_format.nIndex = ii;
                audio_port_format.eEncoding = audio_format->coding_type;

                limoi_component_set_parameter
                    (hComp, OMX_IndexParamAudioPortFormat,
                     &audio_port_format);
                audio_format++;
            }

            //lazy, reuse num_streams to default active stream
            num_streams.nU32 = 0;
            limoi_component_set_parameter(hComp, OMX_IndexParamActiveStream,
                                          &num_streams);
        }
    }

    num_streams.nPortIndex = 1;
    if (OMX_ErrorNone ==
        limoi_component_get_parameter(hComp, OMX_IndexParamNumAvailableStreams,
                                      &num_streams))
    {
        if (num_streams.nU32 != num_video_streams)
            video_port_settings_changed = 1;
    } else if (num_video_streams) video_port_settings_changed = 1;

    if (video_port_settings_changed)
    {
        num_streams.nU32 = num_video_streams;
        limoi_component_set_parameter(hComp, OMX_IndexParamNumAvailableStreams,
                                      &num_streams);

        if (num_video_streams)
        {
            video_port_format_detected = 1;
            limoi_set_spec_version(&video_port_format,
                                   sizeof(OMX_VIDEO_PARAM_PORTFORMATTYPE));
            video_port_format.nPortIndex = 1;
            for (ii = 0; ii < num_video_streams; ii++)
            {
                video_port_format.nIndex = ii;
                video_port_format.eCompressionFormat =
                    video_format->coding_type;
                video_port_format.eColorFormat = video_format->color_format;
                video_port_format.xFramerate = video_format->framerate;
                limoi_component_set_parameter
                    (hComp, OMX_IndexParamVideoPortFormat,
                     &video_port_format);
                video_format++;
            }

            //lazy, reuse num_streams to default active stream
            num_streams.nU32 = 0;
            limoi_component_set_parameter(hComp, OMX_IndexParamActiveStream,
                                          &num_streams);
        }
    }

    if (audio_port_settings_changed)
    {
        if (audio_port_format_detected)
            limoi_component_deliver_event
                (hComp, OMX_EventPortFormatDetected, 0, 0, NULL);

        limoi_component_deliver_event
            (hComp, OMX_EventPortSettingsChanged, 0, 0, NULL);
    }

    if (video_port_settings_changed)
    {
        if (video_port_format_detected)
            limoi_component_deliver_event
                (hComp, OMX_EventPortFormatDetected, 1, 0, NULL);

        limoi_component_deliver_event
            (hComp, OMX_EventPortSettingsChanged, 1, 0, NULL);
    }

    return 0;
}

int limoi_demuxer_set_callbacks (
    OMX_HANDLETYPE hComp,
    limoi_demuxer_callbacks_t * callbacks)
{
    limoi_demuxer_t * demuxer = LIMOI_DEMUXER(hComp);
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComp);

    lthread_mutex_lock(&base->mutex);
    demuxer->priv->callbacks = *callbacks;
    lthread_mutex_unlock(&base->mutex);

    return 0;
}

static OMX_ERRORTYPE limoi_demuxer_get_parameter (
    OMX_HANDLETYPE hComp,
    OMX_INDEXTYPE nIndex,
    OMX_PTR pCompConfigStruct)
{
    limoi_component_priv_t * base =  LIMOI_COMPONENT_PRIV(hComp);
    limoi_demuxer_t * demuxer = LIMOI_DEMUXER(hComp);
    limoi_demuxer_priv_t * pdemuxer = demuxer->priv;
    OMX_ERRORTYPE ret = OMX_ErrorNone;

    lthread_mutex_lock(&base->mutex);
    switch (nIndex)
    {
    case OMX_IndexConfigMetadataItemCount :
    {
        OMX_CONFIG_METADATAITEMCOUNTTYPE * count;
        count = (OMX_CONFIG_METADATAITEMCOUNTTYPE *)pCompConfigStruct;
        if (NULL == pdemuxer->callbacks.get_metadata_count)
        {
            ret = OMX_ErrorNotImplemented;
            break;
        }

        //TODO: fetch language country specifier from current filter
        count->nMetadataItemCount = pdemuxer->callbacks.get_metadata_count
            (hComp, count->eScopeMode, count->nScopeSpecifier, NULL);
        break;
    }
    case OMX_IndexConfigMetadataItem :
    {
        OMX_CONFIG_METADATAITEMTYPE * item;
        item = (OMX_CONFIG_METADATAITEMTYPE *)pCompConfigStruct;

        if (NULL == pdemuxer->callbacks.get_metadata)
        {
            ret = OMX_ErrorNotImplemented;
            break;
        }

        switch (item->eSearchMode)
        {
        case OMX_MetadataSearchValueSizeByIndex :
        {
            void * buf[(sizeof(OMX_CONFIG_METADATAITEMTYPE) + 1024)
                       / sizeof(void*)];
            OMX_CONFIG_METADATAITEMTYPE * item_tmp;
            item_tmp = (OMX_CONFIG_METADATAITEMTYPE *)buf;

            item_tmp->nKeySizeUsed = 128;
            item_tmp->nValueSizeUsed = 1024;

            ret = pdemuxer->callbacks.get_metadata
                (hComp, item_tmp->eScopeMode, item_tmp->nScopeSpecifier, NULL,
                 item_tmp->nMetadataItemIndex,
                 &item_tmp->nKeySizeUsed, item_tmp->nKey,
                 &item_tmp->eKeyCharset,
                 &item_tmp->nValueSizeUsed, item_tmp->nValue,
                 &item_tmp->eValueCharset);
            if (OMX_ErrorNone == ret)
            {
                item->nValueMaxSize = item_tmp->nValueSizeUsed;
            }
            break;
        }
        case OMX_MetadataSearchItemByIndex :
            item->nKeySizeUsed = 128;
            item->nValueSizeUsed = item->nValueMaxSize;
            ret = pdemuxer->callbacks.get_metadata
                (hComp, item->eScopeMode, item->nScopeSpecifier, NULL,
                 item->nMetadataItemIndex,
                 &item->nKeySizeUsed, item->nKey, &item->eKeyCharset,
                 &item->nValueSizeUsed, item->nValue, &item->eValueCharset);
            break;

        case OMX_MetadataSearchNextItemByKey :
            ret = OMX_ErrorNotImplemented;
            break;

        default :
            ret = OMX_ErrorBadParameter;
            break;
        }

        break;
    }
    default:
        lthread_mutex_unlock(&base->mutex);
        return limoi_component_get_parameter(hComp, nIndex, pCompConfigStruct);
    }
    lthread_mutex_unlock(&base->mutex);
    return ret;
}

OMX_ERRORTYPE limoi_demuxer_deinit (
    OMX_HANDLETYPE hComponent)
{
    limoi_demuxer_t * demuxer = LIMOI_DEMUXER(hComponent);

    free(demuxer->priv);

    // parent destructor.
    return limoi_component_deinit(hComponent);
}

OMX_ERRORTYPE limoi_demuxer_init (
    OMX_HANDLETYPE * hComp,
	int bytes,
	const char * comp_name,
	const char * comp_version)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    OMX_COMPONENTTYPE * hdl;
    limoi_component_callbacks_t callbacks;
    limoi_demuxer_t * demuxer;
    limoi_demuxer_priv_t * priv = NULL;

    LIMDBG("init component %s, version %s.\n", comp_name, comp_version);

    priv = calloc(1, sizeof(limoi_demuxer_priv_t));
    if (NULL == priv)
    {
        LIMERR("out of memory upon content demuxer init.\n");
        ret = OMX_ErrorInsufficientResources;
        goto bail;
    }

    ret = limoi_component_init(hComp, bytes);
    if (OMX_ErrorNone != ret) goto bail;
    hdl = (OMX_COMPONENTTYPE *)*hComp;

    demuxer = LIMOI_DEMUXER(*hComp);
    demuxer->priv = priv;

    limoi_component_set_name_version(*hComp, comp_name, comp_version);
    add_port(*hComp);
    add_metadata_filter(*hComp);

    callbacks.command_prep = prep_cmd_handle;
    callbacks.command_post = post_cmd_handle;
    callbacks.main = main_handle;
    callbacks.sub = sub_handle;
    limoi_component_set_callbacks(*hComp, &callbacks);

    LIMDBG("component %s initialized.\n", comp_name);
    hdl->ComponentDeInit = limoi_demuxer_deinit;
    hdl->GetConfig = limoi_demuxer_get_parameter;

 bail:
    if (OMX_ErrorNone != ret && priv) free(priv);
    return ret;
}
