/*
  Copyright (C) 2010 - 2011 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX IL standard container muxer component.

  REVISION:

    1) initial creation. ----------------------------------------- 2010-05-14 MG
*/
#define _GNU_SOURCE
#include "limoi.h"
#include "other-priv.h"

//#undef LIMDBG
#ifndef LIMDBG
#define LIMDBG(fmt, arg...)
#endif


static OMX_ERRORTYPE main_handle (
    OMX_HANDLETYPE hComp,
    limoi_port_t ** audio_port,
    limoi_port_t ** video_port,
    limoi_port_t ** other_port,
    int * preloaded)
{
    OMX_ERRORTYPE  res = OMX_ErrorNone;
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComp);
    limoi_muxer_t * muxer = LIMOI_MUXER(hComp);
    limoi_muxer_priv_t * pmuxer = muxer->priv;
    limoi_port_t * audio_port_in = *audio_port;
    limoi_port_t * video_port_in = *video_port;
    OMX_BUFFERHEADERTYPE ** audio_in = NULL;
    OMX_BUFFERHEADERTYPE ** video_in = NULL;
    OMX_BUFFERHEADERTYPE * audio_buf = NULL;
    OMX_BUFFERHEADERTYPE * video_buf = NULL;
    OMX_U32 audio_flags = 0;
    OMX_U32 video_flags = 0;
    OMX_ERRORTYPE ret = 0;

    if (*preloaded == 0)
    {
        //do not requrie preload.
        *preloaded = LIMOI_PRELOAD_BUFFER_COUNT;
        goto bail;
    }

    LIMDBG("%s muxer get audio input port buffer.\n", base->name);
    audio_in = limoi_component_get_port_buffer(hComp, audio_port_in, 0);
    if (audio_in)
    {
        audio_flags = (*audio_in)->nFlags;
        LIMDBG("%s audio input buffer len: %u %p\n",
               base->name, (*audio_in)->nAllocLen, (*audio_in)->pBuffer);
        audio_buf = *audio_in;
    }

    LIMDBG("%s muxer get video input port buffer.\n", base->name);
    video_in = limoi_component_get_port_buffer(hComp, video_port_in, 0);
    if (video_in)
    {
        video_flags = (*video_in)->nFlags;
        LIMDBG("%s video input buffer len: %u/%u %p\n",
               base->name, (*video_in)->nFilledLen, (*video_in)->nAllocLen,
               (*video_in)->pBuffer);
        video_buf = *video_in;
    }

    if (NULL == audio_buf && NULL == video_buf)
    {
        LIMDBG("%s unable to get audio/video input port buffer.\n", base->name);
        //set return code to signal component main thread to
        //wait a bit before retrying.
        res = OMX_ErrorNotReady;
        goto bail;
    }

    //continue muxing if input is available or EOS
    if ((audio_buf && (audio_buf->nFilledLen
                       || audio_flags & OMX_BUFFERFLAG_EOS))
        || (video_buf && (video_buf->nFilledLen
                          || video_flags & OMX_BUFFERFLAG_EOS)))
    {
        ret = pmuxer->callbacks.mux
            (hComp, pmuxer->pipe, pmuxer->content, audio_buf, video_buf);
    }

    if (0 == ret)
    {
        if (audio_buf && !audio_buf->nFilledLen)
        {
            LIMDBG("%s put audio input port buffer back: %p.\n",
                   base->name, audio_buf->pBuffer);
            if (-1 == limoi_component_put_port_buffer
                (hComp, audio_port_in, audio_in))
            {
                LIMDBG("%s unable to put back audio input port buffer.\n"
                       "\tnFilledLen: %u, flags: %u\n",
                       base->name, audio_buf->nFilledLen, audio_flags);
                ret = OMX_ErrorNotReady;
            }
            audio_in = NULL;
        }

        if (video_buf && !video_buf->nFilledLen)
        {
            LIMDBG("%s put video input port buffer back:\n"
                   "\taddr       : %p\n"
                   "\tnFilledLen : %u\n"
                   "\tflags      : %x\n",
                   base->name, video_buf->pBuffer,
                   video_buf->nFilledLen, video_flags);
            if (-1 == limoi_component_put_port_buffer
                (hComp, video_port_in, video_in))
            {
                LIMDBG("%s unable to put back video input port buffer.\n"
                       "\tnFilledLen: %u, flags: %u\n",
                       base->name, video_buf->nFilledLen, video_flags);
                ret = OMX_ErrorNotReady;
            }
            video_in = NULL;
        }
    }

 bail :
    if (audio_in)
    {
        if (!audio_buf->nFilledLen)
        {
            limoi_component_put_port_buffer(hComp, audio_port_in, audio_in);
        } else
        {
            limoi_component_release_port_buffer(hComp, audio_port_in, audio_in);
        }
    }
    if (video_in)
    {
        if (!video_buf->nFilledLen)
        {
            limoi_component_put_port_buffer(hComp, video_port_in, video_in);
        } else
        {
            limoi_component_release_port_buffer(hComp, video_port_in, video_in);
        }
    }

    return res;
}

static OMX_ERRORTYPE set_state (
    limoi_component_cmd_t * cmd,
    int * cmd_handle_called)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    OMX_STATETYPE next_state = cmd->nParam1;
    limoi_muxer_t * muxer = LIMOI_MUXER(cmd->hComponent);
    limoi_muxer_priv_t * pmuxer = muxer->priv;
    limoi_component_priv_t * base =  LIMOI_COMPONENT_PRIV(cmd->hComponent);
    void * buf[(sizeof(OMX_PARAM_CONTENTURITYPE) + LIMOI_URI_MAX_LENGTH) /
               sizeof(void *)];
    OMX_PARAM_CONTENTURITYPE * uri_param;

    lthread_mutex_lock(&base->mutex);
    switch (base->state)
    {
    case OMX_StateLoaded :

        // transition to StateIdle
        pmuxer->pipe = NULL;
        pmuxer->content = NULL;

        uri_param = (OMX_PARAM_CONTENTURITYPE *)&buf[0];
        limoi_set_spec_version(uri_param, sizeof(OMX_PARAM_CONTENTURITYPE)
                               + LIMOI_URI_MAX_LENGTH);
        ret = limoi_component_get_parameter(cmd->hComponent,
                                            OMX_IndexParamContentURI,
                                            uri_param);
        if (OMX_ErrorNone == ret)
        {
            ret = OMX_GetContentPipe((OMX_HANDLETYPE *)&pmuxer->pipe,
                                     (OMX_STRING)uri_param->contentURI);
            if (OMX_ErrorNone != ret)
            {
                LIMERR("%s unable to open content pipe at %s\n",
                       base->name, uri_param->contentURI);
            } else
            {
                if (pmuxer->pipe->Open(&pmuxer->content,
                                      (OMX_STRING)uri_param->contentURI,
                                      CP_AccessReadWrite))
                {
                    LIMERR("%s unable to create content: %s\n",
                           base->name, uri_param->contentURI);
                    ret = OMX_ErrorContentPipeOpenFailed;
                }
            }
        } else
        {
            LIMERR("%s unable to get URI parameter.\n", base->name);
        }
        break;

    case OMX_StateIdle :
        if (OMX_StateLoaded == next_state)
        {
            if (pmuxer->callbacks.command_post)
            {
                *cmd_handle_called = 1;
                ret = pmuxer->callbacks.command_post(cmd);
                if (OMX_ErrorNone != ret)
                {
                    LIMERR("%s command callback returned error %s\n",
                           LIMOI_COMPONENT_PRIV(cmd->hComponent)->name,
                           limoi_omx_error_string(ret));
                }
            }

            if (pmuxer->pipe && pmuxer->content)
            {
                pmuxer->pipe->Close(pmuxer->content);
            }
        }
        break;

    default :
        break;
    }
    lthread_mutex_unlock(&base->mutex);
    return ret;
}

static OMX_ERRORTYPE prep_cmd_handle (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    limoi_muxer_t * muxer = LIMOI_MUXER(cmd->hComponent);
    limoi_muxer_priv_t * pmuxer = muxer->priv;

    if (pmuxer->callbacks.command_prep)
    {
        ret = pmuxer->callbacks.command_prep(cmd);
        if (OMX_ErrorNone != ret)
        {
            LIMERR("%s command callback returned error %s\n",
                   LIMOI_COMPONENT_PRIV(cmd->hComponent)->name,
                   limoi_omx_error_string(ret));
        }
    }
    return ret;
}

static OMX_ERRORTYPE post_cmd_handle (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    limoi_muxer_t * muxer = LIMOI_MUXER(cmd->hComponent);
    limoi_muxer_priv_t * pmuxer = muxer->priv;
    int post_cmd_handle_called = 0;

    switch (cmd->Cmd)
    {
    case OMX_CommandStateSet :
        ret = set_state(cmd, &post_cmd_handle_called);
        if (OMX_ErrorNone != ret)
        {
            LIMERR("%s post state set handle returned error %s\n",
                   LIMOI_COMPONENT_PRIV(cmd->hComponent)->name,
                   limoi_omx_error_string(ret));
            goto bail;
        }
         break;

    default :
        break;
    }

    if (pmuxer->callbacks.command_post && !post_cmd_handle_called)
    {
        ret = pmuxer->callbacks.command_post(cmd);
        if (OMX_ErrorNone != ret)
        {
            LIMERR("%s command callback returned error %s\n",
                   LIMOI_COMPONENT_PRIV(cmd->hComponent)->name,
                   limoi_omx_error_string(ret));
            goto bail;
        }
    }

 bail:
    return ret;
}

static int add_port (
    OMX_HANDLETYPE hComp)
{
	OMX_PARAM_PORTDEFINITIONTYPE port;
    OMX_PORT_PARAM_TYPE port_param;

    limoi_set_spec_version(&port_param, sizeof(OMX_PORT_PARAM_TYPE));
    port_param.nPorts = 1;
    port_param.nStartPortNumber = 0; //APB = 0
    limoi_component_set_parameter(hComp, OMX_IndexParamAudioInit, &port_param);

    port_param.nStartPortNumber = 1; //VPB = 1
    limoi_component_set_parameter(hComp, OMX_IndexParamVideoInit, &port_param);

    limoi_set_spec_version(&port, sizeof(OMX_PARAM_PORTDEFINITIONTYPE));

    port.nPortIndex = 0;
    port.eDir = OMX_DirInput;
    port.nBufferCountActual = LIMOI_MUXER_AUDIO_INPUT_BUFFER_NUM;
    port.nBufferCountMin = LIMOI_MUXER_AUDIO_INPUT_BUFFER_NUM;
    port.nBufferSize = LIMOI_MUXER_AUDIO_INPUT_BUFFER_BYTES;
    port.bEnabled = OMX_TRUE;
    port.bPopulated = OMX_FALSE;
    port.eDomain = OMX_PortDomainAudio;
    port.bBuffersContiguous = OMX_FALSE;
    port.nBufferAlignment = 0;
    port.format.audio.cMIMEType = NULL;
    port.format.audio.pNativeRender = 0;
    port.format.audio.bFlagErrorConcealment = OMX_FALSE;
    port.format.audio.eEncoding = OMX_AUDIO_CodingAutoDetect;
    limoi_component_add_port(hComp, &port);

    port.nPortIndex = 1;
    port.nBufferCountActual = LIMOI_MUXER_VIDEO_INPUT_BUFFER_NUM;
    port.nBufferCountMin = LIMOI_MUXER_VIDEO_INPUT_BUFFER_NUM;
    port.nBufferSize = LIMOI_MUXER_VIDEO_INPUT_BUFFER_BYTES;
    port.eDomain = OMX_PortDomainVideo;
    port.format.video.eCompressionFormat = OMX_VIDEO_CodingAutoDetect;
    port.format.video.eColorFormat = OMX_COLOR_FormatUnused;
    limoi_component_add_port(hComp, &port);

    return 0;
}

int limoi_muxer_set_callbacks (
    OMX_HANDLETYPE hComp,
    limoi_muxer_callbacks_t * callbacks)
{
    limoi_muxer_t * muxer = LIMOI_MUXER(hComp);
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComp);

    lthread_mutex_lock(&base->mutex);
    muxer->priv->callbacks = *callbacks;
    lthread_mutex_unlock(&base->mutex);

    return 0;
}

OMX_ERRORTYPE limoi_muxer_deinit (
    OMX_HANDLETYPE hComponent)
{
    limoi_muxer_t * muxer = LIMOI_MUXER(hComponent);

    free(muxer->priv);

    // parent destructor.
    return limoi_component_deinit(hComponent);
}

OMX_ERRORTYPE limoi_muxer_init (
    OMX_HANDLETYPE * hComp,
	int bytes,
	const char * comp_name,
	const char * comp_version)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    OMX_COMPONENTTYPE * hdl;
    limoi_component_callbacks_t callbacks;
    limoi_muxer_t * muxer;
    limoi_muxer_priv_t * priv = NULL;

    LIMDBG("init component %s, version %s.\n", comp_name, comp_version);

    priv = calloc(1, sizeof(limoi_muxer_priv_t));
    if (NULL == priv)
    {
        LIMERR("out of memory upon content muxer init.\n");
        ret = OMX_ErrorInsufficientResources;
        goto bail;
    }

    ret = limoi_component_init(hComp, bytes);
    if (OMX_ErrorNone != ret) goto bail;
    hdl = (OMX_COMPONENTTYPE *)*hComp;

    muxer = LIMOI_MUXER(*hComp);
    muxer->priv = priv;

    limoi_component_set_name_version(*hComp, comp_name, comp_version);
    add_port(*hComp);

    callbacks.command_prep = prep_cmd_handle;
    callbacks.command_post = post_cmd_handle;
    callbacks.main = main_handle;
    callbacks.sub = NULL;
    limoi_component_set_callbacks(*hComp, &callbacks);

    LIMDBG("component %s initialized.\n", comp_name);
    hdl->ComponentDeInit = limoi_muxer_deinit;

 bail:
    if (OMX_ErrorNone != ret && priv) free(priv);
    return ret;
}
