/*
  Copyright (C) 2011 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX IL component parameter set/get subroutines.

  REVISION:

    1) initial creation. ----------------------------------------- 2011-07-22 MG
*/
#define _GNU_SOURCE
#include "limoi.h"
#include "component.h"
#include "string.h"
#include <stdarg.h>

//#undef LIMDBG
#ifndef LIMDBG
#define LIMDBG(fmt, arg...)
#endif

/** all port param structures start with this structure. */
typedef struct
{
    OMX_U32 nSize;
    OMX_VERSIONTYPE nVersion;
    OMX_U32 nPortIndex;
    OMX_U32 nIndex;
} limoi_port_param_t;

//param_config_type
#define PARAM_CONFIG_TYPE_COMPONENT_                   0
#define PARAM_CONFIG_TYPE_PORT_                        1
#define PARAM_CONFIG_TYPE_PORT_WITH_ENUMERATION_INDEX_ 2
static int get_param_config_type (
    OMX_INDEXTYPE nIndex)
{
    if (nIndex == OMX_IndexParamNumAvailableStreams
        || nIndex == OMX_IndexParamActiveStream
        || (nIndex >= OMX_IndexParamPortDefinition
            && nIndex <= OMX_IndexParamCompBufferSupplier)
        || (nIndex >= OMX_IndexParamAudioPortFormat
            && nIndex <= OMX_IndexConfigAudioChannelVolume)
        || (nIndex >= OMX_IndexParamImagePortFormat
            && nIndex <= OMX_IndexConfigFlashControl)
        || (nIndex >= OMX_IndexParamVideoPortFormat
            && nIndex <= OMX_IndexConfigVideoNalSize)
        || (nIndex >= OMX_IndexParamCommonDeblocking
            && nIndex <= OMX_IndexConfigCommonTransitionEffect)
        || nIndex == OMX_IndexParamOtherPortFormat
        )
    {
        if (nIndex == OMX_IndexParamAudioPortFormat
            || nIndex == OMX_IndexParamImagePortFormat
            || nIndex == OMX_IndexParamVideoPortFormat
            || nIndex == OMX_IndexParamOtherPortFormat)
            return PARAM_CONFIG_TYPE_PORT_WITH_ENUMERATION_INDEX_;
        else
            return PARAM_CONFIG_TYPE_PORT_;
    }
    return PARAM_CONFIG_TYPE_COMPONENT_;
}

void param_release (list_t * params_list_head)
{
    list_t * list = params_list_head->next;
    limoi_param_config_t * reg;

    while (list != params_list_head)
    {
        reg = (limoi_param_config_t *)list;

        LIMDBG("release param/config: %08x\n", reg->index);

        free((void *)reg->param_config);

        list = list->next;

        list_del(&reg->list);
        free(reg);
    }
}

OMX_ERRORTYPE param_allocate_and_set (
    OMX_INDEXTYPE nIndex,
    void * param,
    int bytes,
    list_t * params_list_head)
{
    int ret = OMX_ErrorNone;
    limoi_param_config_t * reg;

    reg = malloc(sizeof(limoi_param_config_t));
    if (NULL == reg)
    {
        LIMERR("out of memory.\n");
        ret = OMX_ErrorInsufficientResources;
        goto bail;
    }

    reg->param_config = malloc(bytes);
    if (NULL == reg->param_config)
    {
        LIMERR("out of memory.\n");
        free(reg);
        ret = OMX_ErrorInsufficientResources;
        goto bail;
    }

    reg->index = nIndex;
    memcpy(reg->param_config, param, bytes);

    list_init(&reg->list);
    list_add(&reg->list, params_list_head);

 bail:
    return ret;
}

OMX_ERRORTYPE omx_get_parameter (
    OMX_HANDLETYPE hComponent,
    OMX_INDEXTYPE nParamIndex,
    OMX_PTR pComponentParameterStructure)
{
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComponent);
    OMX_ERRORTYPE ret = OMX_ErrorNone;

    lthread_mutex_lock(&base->mutex);
    if (OMX_StateInvalid == base->state)
    {
        LIMERR("%s invalid state.\n", base->name);
        ret = OMX_ErrorInvalidState;
        goto bail;
    }
    lthread_mutex_unlock(&base->mutex);

    return limoi_component_get_parameter(hComponent, nParamIndex,
                                         pComponentParameterStructure);

 bail:
    lthread_mutex_unlock(&base->mutex);
    return ret;
}

OMX_ERRORTYPE omx_set_parameter (
    OMX_HANDLETYPE hComponent,
    OMX_INDEXTYPE nIndex,
    OMX_PTR pComponentParameterStructure)
{
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComponent);
    OMX_ERRORTYPE ret = OMX_ErrorNone;

    lthread_mutex_lock(&base->mutex);
    if (OMX_StateLoaded != base->state)
    {
        if (get_param_config_type(nIndex) != PARAM_CONFIG_TYPE_COMPONENT_)
        {
            limoi_port_t * limoi_port;
            limoi_port_param_t * port_param;

            port_param = (limoi_port_param_t *)pComponentParameterStructure;
            limoi_port = limoi_component_get_port
                (hComponent, port_param->nPortIndex);
            if (NULL == limoi_port)
            {
                LIMERR("unable to find port: %u of %s.\n",
                       port_param->nPortIndex, base->name);
                ret = OMX_ErrorBadParameter;
                goto bail;
            }

            lthread_mutex_lock(&limoi_port->mutex);
            if (LIMOI_PORTDEF(limoi_port)->bEnabled == OMX_TRUE
                && LIMOI_PORTDEF(limoi_port)->bPopulated == OMX_TRUE)
            {
                lthread_mutex_unlock(&limoi_port->mutex);
                LIMERR("%s port %u enabled, unable to set parameter.\n",
                       base->name, port_param->nPortIndex);
                ret = OMX_ErrorInvalidState;
                goto bail;
            }
            lthread_mutex_unlock(&limoi_port->mutex);
        } else
        {
            LIMERR("%s invalid state.\n", base->name);
            ret = OMX_ErrorInvalidState;
            goto bail;
        }
    }
    lthread_mutex_unlock(&base->mutex);

    return limoi_component_set_parameter(hComponent, nIndex,
                                         pComponentParameterStructure);

 bail:
    lthread_mutex_unlock(&base->mutex);
    return ret;
}

OMX_ERRORTYPE omx_get_config (
    OMX_HANDLETYPE hComponent,
    OMX_INDEXTYPE nIndex,
    OMX_PTR pComponentConfigStructure)
{
    return omx_get_parameter(hComponent, nIndex, pComponentConfigStructure);
}

OMX_ERRORTYPE omx_set_config (
    OMX_HANDLETYPE hComponent,
    OMX_INDEXTYPE nIndex,
    OMX_PTR pComponentConfigStructure)
{
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComponent);
    OMX_ERRORTYPE ret = OMX_ErrorInvalidState;

    lthread_mutex_lock(&base->mutex);
    if (OMX_StateInvalid != base->state)
    {
        ret = limoi_component_set_parameter(hComponent, nIndex,
                                            pComponentConfigStructure);
    }
    lthread_mutex_unlock(&base->mutex);
    return ret;
}

OMX_ERRORTYPE omx_get_component_version (
    OMX_HANDLETYPE hComponent,
    OMX_STRING pComponentName,
    OMX_VERSIONTYPE* pComponentVersion,
    OMX_VERSIONTYPE* pSpecVersion,
    OMX_UUIDTYPE* pComponentUUID)
{
    OMX_COMPONENTTYPE * comp = (OMX_COMPONENTTYPE *)hComponent;
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(comp);

    if (pComponentName)
    {
        strcpy(pComponentName, base->name);
    }

    if (pComponentVersion)
    {
        *pComponentVersion = base->version;
    }

    if (pSpecVersion)
    {
        *pSpecVersion = comp->nVersion;
    }

    if (pComponentUUID)
    {
        //TODO: assign component UUID.
    }

    return 0;
}

OMX_ERRORTYPE omx_set_callbacks (
    OMX_HANDLETYPE hComponent,
    OMX_CALLBACKTYPE* pCallbacks,
    OMX_PTR pAppData)
{
    OMX_COMPONENTTYPE * comp = (OMX_COMPONENTTYPE *)hComponent;
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComponent);
    OMX_ERRORTYPE ret = OMX_ErrorNone;

    lthread_mutex_lock(&base->mutex);
    if (OMX_StateLoaded != base->state)
    {
        LIMERR("%s invalid state.\n", base->name);
        ret = OMX_ErrorInvalidState;
    } else
    {
        base->callbacks = *pCallbacks;
        comp->pApplicationPrivate = pAppData;
        LIMDBG("%s set callbacks: %p/%p\n", base->name,
               base->callbacks.EventHandler, pCallbacks->EventHandler);
    }
    lthread_mutex_unlock(&base->mutex);

    return ret;
}

OMX_ERRORTYPE limoi_component_get_parameter (
    OMX_HANDLETYPE hComponent,
    OMX_INDEXTYPE nParamIndex,
    OMX_PTR pComponentParameterStructure)
{
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComponent);
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    limoi_port_param_t * port_param;
    list_t * params_list_head;
    list_t * list;
    limoi_param_config_t * reg;
    int port_lock = 0;
    limoi_port_t * limoi_port = NULL;
    OMX_U32 enum_index = OMX_IndexMax;
    int param_config_type;

    //support of per-port param.
    port_param = (limoi_port_param_t *)pComponentParameterStructure;
    if (!limoi_is_version_compatible(port_param->nVersion))
        return OMX_ErrorVersionMismatch;

    lthread_mutex_lock(&base->mutex);

    param_config_type = get_param_config_type(nParamIndex);
    if (param_config_type != PARAM_CONFIG_TYPE_COMPONENT_)
    {
        limoi_port = limoi_component_get_port
            (hComponent, port_param->nPortIndex);
        if (NULL == limoi_port)
        {
            LIMERR("unable to find port: %u of %s.\n",
                   port_param->nPortIndex, base->name);
            ret = OMX_ErrorBadParameter;
            goto bail;
        }
        lthread_mutex_unlock(&base->mutex);

        if (param_config_type != PARAM_CONFIG_TYPE_PORT_)
        {
            enum_index = port_param->nIndex;
        }

        port_lock = 1;
        lthread_mutex_lock(&limoi_port->mutex);
        params_list_head = &limoi_port->params_list_head;
    } else // per-component param
    {
        params_list_head = &base->params_list_head;
    }

    // search for param/config
    list = params_list_head->next;
    while (list != params_list_head)
    {
        reg = (limoi_param_config_t *)list;
        if (reg->index == nParamIndex)
        {
            if (enum_index == OMX_IndexMax
                || enum_index ==
                ((limoi_port_param_t *)(reg->param_config))->nIndex)
            {
                memcpy(pComponentParameterStructure, reg->param_config,
                       port_param->nSize);
                goto bail;
            }
        }
        list = list->next;
    }
    LIMDBG("unable to find parameter %x of %s.\n",
           nParamIndex, base->name);
    ret = OMX_ErrorBadParameter;

 bail:
    if (port_lock)
    {
        lthread_mutex_unlock(&limoi_port->mutex);
    } else
    {
        lthread_mutex_unlock(&base->mutex);
    }

    if (OMX_IndexParamPortDefinition == nParamIndex && OMX_ErrorNone == ret)
    {
        OMX_U32 count, bytes;
        OMX_PARAM_PORTDEFINITIONTYPE * port = pComponentParameterStructure;

        LIMDBG("%s port %u get parameter.\n", base->name, limoi_port->index);

        lthread_mutex_lock(&base->mutex);
        if (!buffer_share_get_buffer_count_size
            (hComponent, limoi_port, &count, &bytes))
        {
            lthread_mutex_lock(&limoi_port->mutex);
            port->nBufferCountActual = count;
            port->nBufferSize = bytes;
            lthread_mutex_unlock(&limoi_port->mutex);
        }
        lthread_mutex_unlock(&base->mutex);
    }
    return ret;
}

OMX_ERRORTYPE limoi_component_set_parameter (
    OMX_HANDLETYPE hComponent,
    OMX_INDEXTYPE nIndex,
    OMX_PTR pComponentParameterStructure)
{
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComponent);
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    limoi_port_param_t * port_param;
    list_t * params_list_head;
    list_t * list;
    limoi_param_config_t * reg;
    int port_lock = 0;
    limoi_port_t * limoi_port = NULL;
    OMX_U32 enum_index = OMX_IndexMax;
    int param_config_type;
    int port_buffer_count_size_change = 0;

    //support of per-port param.
    port_param = (limoi_port_param_t *)pComponentParameterStructure;
    if (!limoi_is_version_compatible(port_param->nVersion))
        return OMX_ErrorVersionMismatch;

    lthread_mutex_lock(&base->mutex);

    param_config_type = get_param_config_type(nIndex);
    if (param_config_type != PARAM_CONFIG_TYPE_COMPONENT_)
    {
        limoi_port = limoi_component_get_port
            (hComponent, port_param->nPortIndex);
        if (NULL == limoi_port)
        {
            LIMERR("%s unable to find port %u.\n",
                   base->name, port_param->nPortIndex);
            ret = OMX_ErrorBadParameter;
            goto bail;
        }
        lthread_mutex_unlock(&base->mutex);

        if (param_config_type != PARAM_CONFIG_TYPE_PORT_)
        {
            enum_index = port_param->nIndex;
        }

        port_lock = 1;
        lthread_mutex_lock(&limoi_port->mutex);
        params_list_head = &limoi_port->params_list_head;

        if (nIndex == OMX_IndexParamPortDefinition)
        {
            OMX_PARAM_PORTDEFINITIONTYPE * port = LIMOI_PORTDEF(limoi_port);
            OMX_PARAM_PORTDEFINITIONTYPE * port2 = pComponentParameterStructure;
            if (port->nBufferCountActual != port2->nBufferCountActual
                || port->nBufferSize != port2->nBufferSize)
            {
                port_buffer_count_size_change = 1;
                LIMDBG("%s port %u buffer count size change\n"
                       "\tcound : %d -> %d\n"
                       "\tsize  : %d -> %d\n", base->name, port->nPortIndex,
                       port->nBufferCountActual, port2->nBufferCountActual,
                       port->nBufferSize, port2->nBufferSize);
            }
        }
    } else // per-component param
    {
        params_list_head = &base->params_list_head;
    }

    // search for param/config
    list = params_list_head->next;
    while (list != params_list_head)
    {
        reg = (limoi_param_config_t *)list;
        if (reg->index == nIndex)
        {
            if (enum_index == OMX_IndexMax
                || enum_index ==
                ((limoi_port_param_t *)(reg->param_config))->nIndex)
            {
                memcpy(reg->param_config, pComponentParameterStructure,
                       port_param->nSize);
                goto bail;
            }
        }
        list = list->next;
    }

    // unable to find param/config, allocate and set.
    ret = param_allocate_and_set(nIndex, pComponentParameterStructure,
                                 port_param->nSize, params_list_head);

 bail:
    if (port_lock)
    {
        //walk through enumerated portformat parameters and populate
        //port settings.
        if (OMX_ErrorNone == ret && nIndex == OMX_IndexParamActiveStream)
        {
            OMX_PARAM_U32TYPE * active_stream =
                (OMX_PARAM_U32TYPE *)pComponentParameterStructure;
            OMX_PARAM_PORTDEFINITIONTYPE * port = LIMOI_PORTDEF(limoi_port);

            switch (port->eDomain)
            {
            case OMX_PortDomainAudio :
                {
                    OMX_AUDIO_PARAM_PORTFORMATTYPE audio_port_format;

                    limoi_set_spec_version
                        (&audio_port_format,
                         sizeof(OMX_AUDIO_PARAM_PORTFORMATTYPE));

                    audio_port_format.nPortIndex = limoi_port->index;
                    audio_port_format.nIndex = active_stream->nU32;
                    ret = limoi_component_get_parameter
                        (hComponent, OMX_IndexParamAudioPortFormat,
                         &audio_port_format);
                    if (OMX_ErrorNone == ret)
                    {
                        port->format.audio.eEncoding =
                            audio_port_format.eEncoding;
                    }
                }
                break;

            case OMX_PortDomainImage :
                {
                    OMX_IMAGE_PARAM_PORTFORMATTYPE image_port_format;

                    limoi_set_spec_version
                        (&image_port_format,
                         sizeof(OMX_IMAGE_PARAM_PORTFORMATTYPE));

                    image_port_format.nPortIndex = limoi_port->index;
                    image_port_format.nIndex = active_stream->nU32;
                    ret = limoi_component_get_parameter
                        (hComponent, OMX_IndexParamImagePortFormat,
                         &image_port_format);
                    if (OMX_ErrorNone == ret)
                    {
                        port->format.image.eCompressionFormat =
                            image_port_format.eCompressionFormat;
                        port->format.image.eColorFormat =
                            image_port_format.eColorFormat;
                    }
                }
                break;

            case OMX_PortDomainVideo :
                {
                    OMX_VIDEO_PARAM_PORTFORMATTYPE video_port_format;

                    limoi_set_spec_version
                        (&video_port_format,
                         sizeof(OMX_VIDEO_PARAM_PORTFORMATTYPE));

                    video_port_format.nPortIndex = limoi_port->index;
                    video_port_format.nIndex = active_stream->nU32;
                    ret = limoi_component_get_parameter
                        (hComponent, OMX_IndexParamVideoPortFormat,
                         &video_port_format);
                    if (OMX_ErrorNone == ret)
                    {
                        port->format.video.eCompressionFormat =
                            video_port_format.eCompressionFormat;
                        port->format.video.eColorFormat =
                            video_port_format.eColorFormat;
                        port->format.video.xFramerate =
                            video_port_format.xFramerate;
                    }
                }
                break;

            case OMX_PortDomainOther :
                {
                    OMX_OTHER_PARAM_PORTFORMATTYPE other_port_format;

                    limoi_set_spec_version
                        (&other_port_format,
                         sizeof(OMX_OTHER_PARAM_PORTFORMATTYPE));

                    other_port_format.nPortIndex = limoi_port->index;
                    other_port_format.nIndex = active_stream->nU32;
                    ret = limoi_component_get_parameter
                        (hComponent, OMX_IndexParamOtherPortFormat,
                         &other_port_format);
                    if (OMX_ErrorNone == ret)
                    {
                        port->format.other.eFormat =
                            other_port_format.eFormat;
                    }
                }
                break;

            default :
                LIMWARN("%s bogus port domain: %d\n",
                        base->name, port->eDomain);
                break;
            }
        }
        lthread_mutex_unlock(&limoi_port->mutex);
    } else
    {
        lthread_mutex_unlock(&base->mutex);
    }

    if (OMX_IndexParamPortDefinition == nIndex && OMX_ErrorNone == ret
        && port_buffer_count_size_change)
    {
        OMX_PARAM_PORTDEFINITIONTYPE * port = pComponentParameterStructure;
        lthread_mutex_lock(&base->mutex);
        buffer_share_set_buffer_count_size
            (hComponent, limoi_port, port->nBufferCountActual,
             port->nBufferSize);
        lthread_mutex_unlock(&base->mutex);
    }
    return ret;
}

OMX_PARAM_PORTDEFINITIONTYPE * limoi_component_get_port_definition (
     limoi_port_t * port)
{
    return ((OMX_PARAM_PORTDEFINITIONTYPE *)
            (((limoi_param_config_t *)
              (port->params_list_head.prev))->param_config));
}

OMX_PARAM_BUFFERSUPPLIERTYPE * limoi_component_get_port_buffer_supplier (
     limoi_port_t * port)
{
    return ((OMX_PARAM_BUFFERSUPPLIERTYPE *)
            (((limoi_param_config_t *)
              (port->params_list_head.prev->prev))->param_config));
}

int limoi_component_set_name_version (
    OMX_HANDLETYPE hComp,
    const char * name,
    const char * version)
{
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComp);

    if (NULL == name) return -1;

    lthread_mutex_lock(&base->mutex);
    base->name = strndup(name, OMX_MAX_STRINGNAME_SIZE - 1);

    if (version)
    {
        char * ver = strndup(version, OMX_MAX_STRINGNAME_SIZE - 1);
        char * pc;

        pc = strtok(ver, ".");
        if (pc)
        {
            base->version.s.nVersionMajor = atoi(pc);
        }

        pc = strtok(NULL, ".");
        if (pc)
        {
            base->version.s.nVersionMinor = atoi(pc);
        }

        pc = strtok(NULL, ".");
        if (pc)
        {
            base->version.s.nRevision = atoi(pc);
        }

        pc = strtok(NULL, ".");
        if (pc)
        {
            base->version.s.nStep = atoi(pc);
        }
        free(ver);

        LIMDBG("%d.%d.%d.%d - %s\n",
               base->version.s.nVersionMajor,
               base->version.s.nVersionMinor,
               base->version.s.nRevision,
               base->version.s.nStep,
               version);
    }
    lthread_mutex_unlock(&base->mutex);

    return 0;
}

int limoi_component_set_callbacks (
    OMX_HANDLETYPE hComp,
    limoi_component_callbacks_t * callbacks)
{
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComp);

    lthread_mutex_lock(&base->mutex);
    base->limoi_callbacks = *callbacks;
    lthread_mutex_unlock(&base->mutex);
    return 0;
}
