/*
  Copyright (C) 2011 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX IL component resource manager.

  REVISION:

    1) initial creation. ----------------------------------------- 2011-04-03 MG
*/
#define _GNU_SOURCE
#include "limoi.h"
#include "string.h"
#include <stdarg.h>
#include "limoi-resource-manager.h"

#define RM_HASH_ENTRIES 256
#define RMHASH(id) (((id)^((id)>>24)^((id)>>16)^((id)>>8)) & 0xff)

static list_t rm_list_head[RM_HASH_ENTRIES];
static int rm_inited;
static lthread_mutex_t rm_mutex = LTHREAD_MUTEX_INITIALIZER;

typedef struct
{
    list_t list;
    int manager_id;
    int * resource_id;
    int resource_num;
    limoi_resource_get_cb_t get;
    limoi_resource_put_cb_t put;
    void * arg;
    list_t wait_list_head;
    list_t usage_list_head;
} lim_rm_reg_t;

typedef struct
{
    list_t list;
    int resource_id;
    int priority;
    limoi_resource_notify_cb_t notify;
    OMX_HANDLETYPE hComp;
} lim_rsc_wait_reg_t;

typedef struct
{
    list_t list;
    int resource_id;
    limoi_resource_preempt_cb_t preempt;
    OMX_HANDLETYPE hComp;
} lim_rsc_usage_reg_t;

static lim_rm_reg_t * locate_rm (int manager_id, int resource_id)
{
    list_t * list;

    list = rm_list_head[RMHASH(manager_id)].next;
    while (list != &rm_list_head[RMHASH(manager_id)])
    {
        lim_rm_reg_t * rm = (lim_rm_reg_t *)list;
        if (resource_id >= rm->resource_id[0]
            && resource_id <= rm->resource_id[rm->resource_num - 1])
            return rm;
        list = list->next;
    }
    return NULL;
}

static void rm_sort (int * resource_id, int resource_num)
{
    int ii = 0;
    int tmp;
    int sorted;

    if (resource_num < 2) return;
    LIMDBG("sorting id: %d\n", resource_num);
    do
    {
        sorted = 1;
        for (ii = 0; ii < resource_num - 1; ii++)
        {
            if (resource_id[ii] > resource_id[ii + 1])
            {
                tmp = resource_id[ii];
                resource_id[ii] = resource_id[ii + 1];
                resource_id[ii + 1] = tmp;
                sorted = 0;
            }
        }
    } while (sorted == 0);
}

//return -1: error
//        0: found and updated
//        1: no match found
static int find_and_update_entry (
    int manager_id,
    int resource_id[],
    limoi_resource_get_cb_t get,
    limoi_resource_put_cb_t put,
    void * arg)
{
    int ii;
    list_t * list = rm_list_head[RMHASH(manager_id)].next;

    while (list != &rm_list_head[RMHASH(manager_id)])
    {
        lim_rm_reg_t * rm = (lim_rm_reg_t *)list;
        list = list->next;
        if (rm->manager_id != manager_id) continue;

        ii = 0;
        while (resource_id[ii] != -1)
        {
            if (resource_id[ii] >= rm->resource_id[0]
                && resource_id[ii] <= rm->resource_id[rm->resource_num - 1])
            {
                int jj = 0;
                while (jj < rm->resource_num)
                {
                    if (resource_id[ii] == rm->resource_id[jj])
                    {
                        LIMERR("duplicate resource id: %d manager: %d\n",
                               resource_id[ii], manager_id);
                        return -1;
                    }
                    jj++;
                }
            }
            ii++;
        }

        //no duplicate resource id found.
        if (rm->get == get && rm->put == put && rm->arg == arg)
        {
            //match entry found, update it.
            int num_rsc = 0;
            int * new_resource_id;

            while (resource_id[num_rsc] != -1) num_rsc++;
            ASSERT(num_rsc);

            new_resource_id =
                malloc((num_rsc + rm->resource_num) * sizeof(int));
            if (NULL == new_resource_id)
            {
                LIMERR("out of memory, unable to add resource: %d\n",
                       manager_id);
                return -1;
            }

            memcpy(new_resource_id, rm->resource_id,
                   rm->resource_num * sizeof(int));
            free(rm->resource_id);
            memcpy(new_resource_id + rm->resource_num * sizeof(int),
                   resource_id, num_rsc * sizeof(int));

            rm->resource_id = new_resource_id;
            rm->resource_num += num_rsc;

            rm_sort(rm->resource_id, rm->resource_num);

            return 0;
        }
    }

    return 1;
}

int limoi_resource_register (
    int manager_id,
    int resource_id[],
    limoi_resource_get_cb_t get,
    limoi_resource_put_cb_t put,
    void * arg)
{
    int ret;
    lim_rm_reg_t * rm = NULL;
    int num_rsc = 0;

    while (resource_id[num_rsc] != -1 && num_rsc < 10000) num_rsc++;
    if (num_rsc <= 0 || num_rsc >= 10000)
    {
        LIMERR("unable to register %d resource for manager: %X\n",
               num_rsc, manager_id);
        return -1;
    }

    lthread_mutex_lock(&rm_mutex);
    if (!rm_inited)
    {
        int ii;
        for (ii = 0; ii < RM_HASH_ENTRIES; ii++) list_init(&rm_list_head[ii]);
        rm_inited = 1;
    }

    LIMDBG("find and update entry: %X\n", manager_id);
    ret = find_and_update_entry(manager_id, resource_id, get, put, arg);
    if (ret > 0)
    {
        rm = malloc(sizeof(lim_rm_reg_t));
        if (NULL == rm)
        {
            LIMERR("out of memory, unable to register resource manager: %d\n",
                   manager_id);
            ret = -1;
            goto bail;
        }

        rm->resource_id = malloc(num_rsc * sizeof(int));
        if (NULL == rm->resource_id)
        {
            LIMERR("out of memory, unable to register resource manager: %d\n",
                   manager_id);
            ret = -1;
            goto bail;
        }

        rm->manager_id = manager_id;
        rm->get = get;
        rm->put = put;
        rm->arg = arg;
        memcpy(rm->resource_id, resource_id, num_rsc);
        rm->resource_num = num_rsc;

        rm_sort(rm->resource_id, rm->resource_num);

        list_init(&rm->wait_list_head);
        list_init(&rm->usage_list_head);

        list_init(&rm->list);
        list_add(&rm->list, &rm_list_head[RMHASH(manager_id)]);

        ret = 0;
    }

 bail:
    if (ret)
    {
        if (rm)
        {
            if (rm->resource_id) free(rm->resource_id);
            free(rm);
        }
    }
    lthread_mutex_unlock(&rm_mutex);

    return ret;
}

void limoi_resource_unregister (int manager_id)
{
    list_t * list;

    lthread_mutex_lock(&rm_mutex);
    list = rm_list_head[RMHASH(manager_id)].next;
    while (list != &rm_list_head[RMHASH(manager_id)])
    {
        lim_rm_reg_t * rm = (lim_rm_reg_t *)list;
        list = list->next;
        if (rm->manager_id == manager_id)
        {
            ASSERT(rm->wait_list_head.next == &rm->wait_list_head);
            ASSERT(rm->usage_list_head.next == &rm->usage_list_head);

            free(rm->resource_id);
            list_del(&rm->list);
            free(rm);
        }
    }
    lthread_mutex_unlock(&rm_mutex);
}

void limoi_resource_post (int manager_id, int resource_id)
{
    lim_rm_reg_t * rm;
    list_t * list;
    limoi_resource_notify_cb_t notify = NULL;
    OMX_HANDLETYPE hComp;

    lthread_mutex_lock(&rm_mutex);
    rm = locate_rm(manager_id, resource_id);
    list = rm->wait_list_head.next;
    while (list != &rm->wait_list_head)
    {
        lim_rsc_wait_reg_t * rw = (lim_rsc_wait_reg_t *)list;
        if (rw->resource_id == resource_id)
        {
            hComp = rw->hComp;
            notify = rw->notify;
            list_del(list);
            free(rw);
            break;
        } else list = list->next;
    }
    lthread_mutex_unlock(&rm_mutex);

    if (notify) notify(manager_id, resource_id, 0, hComp);
}

//------------------------------------------------------------------------------
//start of IL component resource manager APIs.
void * limoi_resource_acquire (
    int manager_id,
    int resource_id,
    limoi_resource_preempt_cb_t preempt,
    OMX_HANDLETYPE hComp)
{
    list_t * list;
    lim_rm_reg_t * rm;
    void * rsc = NULL;

    lthread_mutex_lock(&rm_mutex);
    rm = locate_rm(manager_id, resource_id);
    lthread_mutex_unlock(&rm_mutex);

    rsc = rm->get(manager_id, resource_id, rm->arg);

    //preempt lower priority component and retry
    if (NULL == rsc)
    {
        lthread_mutex_lock(&rm_mutex);
        list = rm->usage_list_head.next;
        while (list != &rm->usage_list_head && NULL == rsc)
        {
            limoi_resource_preempt_cb_t preempt;
            OMX_HANDLETYPE comp;
            lim_rsc_usage_reg_t * ru = (lim_rsc_usage_reg_t *)list;
            if (ru->resource_id == resource_id)
            {
                preempt = ru->preempt;
                comp = ru->preempt;
                lthread_mutex_unlock(&rm_mutex);
                if (!preempt(manager_id, resource_id, comp))
                {
                    rsc = rm->get(manager_id, resource_id, rm->arg);
                    if (rsc) goto reg_rsc;
                }
                lthread_mutex_lock(&rm_mutex);
            }
            list = list->next;
        }
        lthread_mutex_unlock(&rm_mutex);
    }

 reg_rsc:
    if (rsc)
    {
        lim_rsc_usage_reg_t * ru;
        lthread_mutex_lock(&rm_mutex);
        ru = malloc(sizeof(lim_rsc_usage_reg_t));
        if (NULL == ru)
        {
            lthread_mutex_unlock(&rm_mutex);
            LIMERR("out of memory, unable to acquire resource.\n");
            rm->put(manager_id, resource_id, rm->arg);
            return NULL;
        }
        ru->resource_id = resource_id;
        ru->preempt = preempt;
        ru->hComp = hComp;

        list_init(&ru->list);
        list_add(&ru->list, &rm->usage_list_head);
        lthread_mutex_unlock(&rm_mutex);
    }

    return rsc;
}

int limoi_resource_release (
    int manager_id,
    int resource_id,
    OMX_HANDLETYPE hComp)
{
    lim_rm_reg_t * rm;
    list_t * list;
    limoi_resource_notify_cb_t notify = NULL;
    limoi_resource_put_cb_t put = NULL;
    void * arg = NULL;

    lthread_mutex_lock(&rm_mutex);
    rm = locate_rm(manager_id, resource_id);
    if (rm)
    {
        put = rm->put;
        arg = rm->arg;

        list = rm->wait_list_head.next;
        while (list != &rm->wait_list_head)
        {
            lim_rsc_wait_reg_t * rw = (lim_rsc_wait_reg_t *)list;
            if (rw->resource_id == resource_id)
            {
                hComp = rw->hComp;
                notify = rw->notify;
                list_del(list);
                free(rw);
                //only the one on top of the waiting list gets notified.
                break;
            } else list = list->next;
        }

        list = rm->usage_list_head.next;
        while (list != &rm->usage_list_head)
        {
            lim_rsc_usage_reg_t * ru = (lim_rsc_usage_reg_t *)list;
            if (ru->resource_id == resource_id
                && ru->hComp == hComp)
            {
                list_del(list);
                free(ru);
                break;
            } else list = list->next;
        }
    }
    lthread_mutex_unlock(&rm_mutex);

    if (put) put(manager_id, resource_id, arg);
    if (notify) notify(manager_id, resource_id, 0, hComp);

    return 0;
}

static OMX_U32 get_component_priority (OMX_HANDLETYPE hComp)
{
    OMX_PRIORITYMGMTTYPE pm;

    if (OMX_ErrorNone == limoi_component_get_parameter
        (hComp, OMX_IndexParamPriorityMgmt, &pm))
    {
        return pm.nGroupPriority;
    }
    //default priority if not set.
    return 100;
}

int limoi_resource_wait (
    int manager_id,
    int resource_id,
    OMX_HANDLETYPE hComp,
    limoi_resource_notify_cb_t notify)
{
    lim_rm_reg_t * rm;
    list_t * list;
    lim_rsc_wait_reg_t * rw;
    int found = 0;
    list_t tmp_head;

    lthread_mutex_lock(&rm_mutex);
    rm = locate_rm(manager_id, resource_id);
    list = rm->wait_list_head.next;
    //already in wait list?
    while (list != &rm->wait_list_head)
    {
        rw = (lim_rsc_wait_reg_t *)list;
        if (rw->resource_id == resource_id && rw->hComp == hComp)
        {
            rw->notify = notify;
            found = 1;
            break;
        } else list = list->next;
    }

    if (!found)
    {
        rw = malloc(sizeof(lim_rsc_wait_reg_t));
        if (NULL == rw)
        {
            LIMERR("out of memory, unable to wait on resource.\n");
            lthread_mutex_unlock(&rm_mutex);
            return -1;
        }
        list_init(&rw->list);
        rw->resource_id = resource_id;
        rw->notify = notify;
        rw->hComp = hComp;
        list_add(&rw->list, &rm->wait_list_head);
    }

    //sort wait list based on current component priority
    list = rm->wait_list_head.next;
    while (list != &rm->wait_list_head)
    {
        rw = (lim_rsc_wait_reg_t *)list;
        list = list->next;
        rw->priority = get_component_priority(hComp);
    }

    tmp_head = rm->wait_list_head;
    list_init(&rm->wait_list_head);

    list = tmp_head.next;
    list_del(list);
    list_add(list, &rm->wait_list_head);

    list = tmp_head.next;
    while (list != &tmp_head)
    {
        list_t * list2;
        lim_rsc_wait_reg_t * rw2;

        rw = (lim_rsc_wait_reg_t *)list;
        list = list->next;

        list2 = rm->wait_list_head.prev;
        rw2 = (lim_rsc_wait_reg_t *)list2;
        while (list2 != &rm->wait_list_head)
        {
            if (rw->priority >= rw2->priority) break;
            list2 = list2->prev;
            rw2 = (lim_rsc_wait_reg_t *)list2;
        }
        list_del(&rw->list);
        list_add(&rw->list, &rw2->list);
    }
    lthread_mutex_unlock(&rm_mutex);

    return 0;
}

int limoi_resource_cancel_wait (
    int manager_id,
    int resource_id,
    OMX_HANDLETYPE hComp)
{
    lim_rm_reg_t * rm;
    list_t * list;

    lthread_mutex_lock(&rm_mutex);
    rm = locate_rm(manager_id, resource_id);
    list = rm->wait_list_head.next;
    while (list != &rm->wait_list_head)
    {
        lim_rsc_wait_reg_t * rw = (lim_rsc_wait_reg_t *)list;
        if (rw->resource_id == resource_id && rw->hComp == hComp)
        {
            list_del(list);
            free(rw);
            break;
        } else list = list->next;
    }
    lthread_mutex_unlock(&rm_mutex);

    return 0;
}
