/*
  Copyright (C) 2011 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX IL component state support subroutines.

  REVISION:

    1) initial creation. ----------------------------------------- 2011-07-22 MG
*/
#define _GNU_SOURCE
#include "limoi.h"
#include "component.h"
#include "string.h"
#include <stdarg.h>

//#undef LIMDBG
#ifndef LIMDBG
#define LIMDBG(fmt, arg...)
#endif

OMX_ERRORTYPE state_set (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    OMX_HANDLETYPE hComp = cmd->hComponent;
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComp);
    list_t * port = base->ports_list_head.next;
    OMX_STATETYPE next_state = cmd->nParam1;
    int post_command_callback_called = 0;

    lthread_mutex_lock(&base->mutex);
    LIMDBG("%s setting state from %s to %s\n",
           base->name, limoi_omx_state_string(base->state),
           limoi_omx_state_string(next_state));
    if (base->state == next_state)
    {
        LIMDBG("%s set to same state : %s\n", base->name,
               limoi_omx_state_string(next_state));
        ret = OMX_ErrorSameState;
        goto bail_out;
    }

    //'state_transition' flag essentially puts data processing threads on hold.
    //buffer can still be put 'back/released', but 'get' will not be allowed.
    base->state_transition = 1;
    lthread_cond_broadcast(&base->cond);

    switch (base->state)
    {
    case OMX_StateLoaded :
        if (OMX_StateIdle != next_state)
        {
            LIMERR("%s incorrect state transition from %s to %s\n",
                   base->name, limoi_omx_state_string(base->state),
                   limoi_omx_state_string(next_state));
            ret = OMX_ErrorIncorrectStateTransition;
            goto bail_transition_done;
        }

        lthread_mutex_unlock(&base->mutex);
        ret = command_callback(base, cmd, 1);
        lthread_mutex_lock(&base->mutex);
        if (OMX_ErrorNone != ret)
        {
            base->state = OMX_StateInvalid;
            goto bail_transition_done;
        }

        // loop through ports to wait for buffers to be allocated
        LIMDBG("%s allocating port buffers\n", base->name);
        while (port != &base->ports_list_head)
        {
            lthread_mutex_unlock(&base->mutex);
            //base->mutex is unlocked, port buffer allocatin is locked
            //on port granularity, let's try to avoid the big global
            //base->mutex as much as possible.
            //plus, allocating port buffer may require to call API of
            //tunneled component, all mutex must be relinquished before
            //doing that.

            if (port_allocate_buffer(hComp, (limoi_port_t *)port))
            {
                lthread_mutex_lock(&base->mutex);
                LIMERR("unable to allocate buffer for port: %u.\n",
                       ((limoi_port_t *)port)->index);
                ret = OMX_ErrorUndefined;
                base->state = OMX_StateInvalid;
                goto bail_transition_done;
            }
            lthread_mutex_lock(&base->mutex);
            port = port->next;
            LIMDBG("%s current port %p, port head: %p\n",
                   base->name, port, &base->ports_list_head);
        }
        LIMDBG("%s allocating port buffers done.\n", base->name);

        if (OMX_ErrorNone == ret)
        {
            post_command_callback_called = 1;

            lthread_mutex_unlock(&base->mutex);
            ret = command_callback(base, cmd, 0);
            lthread_mutex_lock(&base->mutex);
            if (OMX_ErrorNone != ret)
            {
                base->state = OMX_StateInvalid;
                goto bail_transition_done;
            }
        }

        //check to create component main thread.
        if (!base->synchronous)
        {
            ASSERT(base->limoi_callbacks.main);
            lthread_mutex_unlock(&base->mutex);
            if (lthread_create(&base->main_thread, thread_main_handle, hComp, 1))
            {
                lthread_mutex_lock(&base->mutex);
                LIMERR("%s unable to create main thread.\n", base->name);
                ret = OMX_ErrorInsufficientResources;
                next_state = OMX_StateInvalid;
                goto bail_wait_for_transition;
            }
            lthread_mutex_lock(&base->mutex);
        }
        break;

    case OMX_StateIdle :
        lthread_mutex_unlock(&base->mutex);
        ret = command_callback(base, cmd, 1);
        lthread_mutex_lock(&base->mutex);
        if (OMX_ErrorNone != ret) next_state = OMX_StateInvalid;
        else if (OMX_StateExecuting == next_state
                 || OMX_StatePause == next_state)
        {
            base->preloaded = 0;
            ret = port_activate_buffer_all(hComp);
            if (OMX_ErrorNone != ret) next_state = OMX_StateInvalid;
        } else if (OMX_StateLoaded == next_state)
        {
            while (base->busy)
            {
                if (lthread_cond_timedwait(&base->cond, &base->mutex, 5000))
                {
                    LIMERR("%s unable to transition to LOADED.\n", base->name);
                    ret = OMX_ErrorUndefined;
                    goto bail_transition_done;
                }
            }
            port_free_buffer_all(hComp);
        } else
        {
            LIMERR("%s incorrect state transition from Idle to %s\n",
                   base->name, limoi_omx_state_string(next_state));
            ret = OMX_ErrorIncorrectStateTransition;
            goto bail_transition_done;
        }
        break;

    case OMX_StateExecuting :
        lthread_mutex_unlock(&base->mutex);
        ret = command_callback(base, cmd, 1);
        lthread_mutex_lock(&base->mutex);
        if (OMX_ErrorNone != ret) next_state = OMX_StateInvalid;
        else if (OMX_StatePause == next_state)
        {
            //do not wait for state transition, so that entire data path
            //can be paused by pausing just one component.
            //base->state = next_state;
            //goto bail2;
        } else if (OMX_StateIdle == next_state)
        {
            while (base->busy)
            {
                if (lthread_cond_timedwait(&base->cond, &base->mutex, 5000))
                {
                    LIMERR("%s unable to transition to IDLE.\n", base->name);
                    ret = OMX_ErrorUndefined;
                    goto bail_transition_done;
                }
            }
            port_deactivate_buffer_all(hComp);
        } else
        {
            LIMERR("%s incorrect state transition %s -> %s\n",
                   base->name, limoi_omx_state_string(base->state),
                   limoi_omx_state_string(next_state));
            ret = OMX_ErrorIncorrectStateTransition;
            goto bail_transition_done;
        }
        break;

    case OMX_StatePause :
        lthread_mutex_unlock(&base->mutex);
        ret = command_callback(base, cmd, 1);
        lthread_mutex_lock(&base->mutex);
        if (OMX_ErrorNone != ret) next_state = OMX_StateInvalid;
        else if (OMX_StateExecuting == next_state)
        {
            //do nothing
        } else if (OMX_StateIdle == next_state)
        {
            while (base->busy)
            {
                if (lthread_cond_timedwait(&base->cond, &base->mutex, 5000))
                {
                    LIMERR("%s unable to transition to IDLE.\n", base->name);
                    ret = OMX_ErrorUndefined;
                    goto bail_transition_done;
                }
            }
            port_deactivate_buffer_all(hComp);
        } else
        {
            LIMERR("%s incorrect state transition %s -> %s\n",
                   base->name, limoi_omx_state_string(base->state),
                   limoi_omx_state_string(next_state));
            ret = OMX_ErrorIncorrectStateTransition;
            goto bail_transition_done;
        }
        break;

    default :
        LIMERR("%s bogus state detected: %s -> %s\n", base->name,
               limoi_omx_state_string(base->state),
               limoi_omx_state_string(next_state));
        ret = OMX_ErrorIncorrectStateTransition;
        goto bail_transition_done;
    }

  bail_wait_for_transition:
    while (base->busy)
    {
        if (lthread_cond_timedwait(&base->cond, &base->mutex, 5000))
        {
            LIMERR("%s state transition %s -> %s timed out: %x.\n",
                   base->name, limoi_omx_state_string(base->state),
                   limoi_omx_state_string(next_state),
                   base->busy);
            ret = OMX_ErrorUndefined;
            goto bail_transition_done;
        }
    }
    base->state = next_state;

  bail_transition_done:
    base->state_transition = 0;

  bail_out:
    lthread_cond_broadcast(&base->cond);
    lthread_mutex_unlock(&base->mutex);

    if (OMX_ErrorNone == ret && !post_command_callback_called)
    {
        ret = command_callback(base, cmd, 0);
        if (OMX_ErrorNone != ret)
        {
            lthread_mutex_lock(&base->mutex);
            base->state = OMX_StateInvalid;
            lthread_cond_broadcast(&base->cond);
            lthread_mutex_unlock(&base->mutex);
        }
    }

    LIMDBG("%s set state to %s done\n",
           base->name, limoi_omx_state_string(base->state));
    return ret;
}

OMX_ERRORTYPE omx_get_state (
    OMX_HANDLETYPE hComponent,
    OMX_STATETYPE * pState)
{
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComponent);

    lthread_mutex_lock(&base->mutex);
    *pState = base->state;
    lthread_mutex_unlock(&base->mutex);

    return OMX_ErrorNone;
}

void limoi_component_goto_invalid_state (OMX_HANDLETYPE hComp)
{
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComp);

    lthread_mutex_lock(&base->mutex);
    port_disable_now(hComp, OMX_ALL);
    base->state = OMX_StateInvalid;
    lthread_mutex_unlock(&base->mutex);

    event_deliver(hComp, OMX_EventError, OMX_ErrorInvalidState, 0, NULL);
}

OMX_STATETYPE limoi_component_state (OMX_HANDLETYPE hComp)
{
    OMX_STATETYPE state;
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComp);

    lthread_mutex_lock(&base->mutex);
    state = base->state;
    lthread_mutex_unlock(&base->mutex);
    return state;
}
