/*
  Copyright (C) 2011 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX IL component thread and mutex conditional subroutines.

  REVISION:

    1) initial creation. ----------------------------------------- 2011-07-22 MG
*/
#define _GNU_SOURCE
#include "limoi.h"
#include "component.h"
#include "string.h"
#include <stdarg.h>

//#undef LIMDBG
#ifndef LIMDBG
#define LIMDBG(fmt, arg...)
#endif

/** sub thread private data structure. */
typedef struct
{
    OMX_HANDLETYPE hComp;
    limoi_port_t ** audio_port;
    limoi_port_t ** video_port;
    limoi_port_t ** other_port;
    int stop; //-1 stopping, 0 running, 1, stopped
} sub_thread_priv_t;

static void * sub_thread (void * arg)
{
    sub_thread_priv_t * sub_priv = (sub_thread_priv_t *)arg;
    OMX_HANDLETYPE hComp = sub_priv->hComp;
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComp);
    limoi_port_t ** audio_port = sub_priv->audio_port;
    limoi_port_t ** video_port = sub_priv->video_port;
    limoi_port_t ** other_port = sub_priv->other_port;
    OMX_ERRORTYPE  res = OMX_ErrorNone;

    lthread_mutex_lock(&base->mutex);
    sub_priv->stop = 0;
    lthread_cond_broadcast(&base->cond);
    lthread_mutex_unlock(&base->mutex);

    while (1)
    {
        lthread_mutex_lock(&base->mutex);
        while (base->state_transition
               || (OMX_StateExecuting != base->state
                   && OMX_StatePause != base->state
                   && OMX_StateInvalid != base->state
                   && !sub_priv->stop)
               || OMX_ErrorNone != res)
        {
            LIMDBG("%s sub thread timed waiting on:\n"
                   "\t state     : %s\n",
                   base->name, limoi_omx_state_string(base->state));
            lthread_cond_timedwait(&base->cond, &base->mutex, 50);
            LIMDBG("%s sub thread running.\n", base->name);
            res = OMX_ErrorNone;
        }

        if (OMX_StateInvalid == base->state || sub_priv->stop)
        {
            lthread_mutex_unlock(&base->mutex);
            LIMDBG("%s sub thread stopped.\n", base->name);
            break;
        }
        base->busy |= COMP_SUB_THREAD_BUSY;
        lthread_mutex_unlock(&base->mutex);

        LIMDBG("%s sub callback in.\n", base->name);
        res = base->limoi_callbacks.sub
            (hComp, audio_port, video_port, other_port);
        LIMDBG("%s sub callback out.\n", base->name);

        lthread_mutex_lock(&base->mutex);
        base->busy &= ~COMP_SUB_THREAD_BUSY;
        lthread_cond_broadcast(&base->cond);
        lthread_mutex_unlock(&base->mutex);

        event_deliver_real(hComp);

        if (OMX_ErrorNoMore == res)
        {
            LIMDBG("%s sub thread no more callback.\n", base->name);
            break;
        }
    }

    lthread_mutex_lock(&base->mutex);
    sub_priv->stop = 1;
    lthread_cond_broadcast(&base->cond);
    lthread_mutex_unlock(&base->mutex);

    return NULL;
}

static int enumerate_ports (OMX_HANDLETYPE hComp)
{
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComp);
    int aports = 0;
    int apb = 0;
    int vports = 0;
    int vpb = 0;
    int oports = 0;
    int opb = 0;
    list_t * list;
    limoi_port_t * limoi_port;
    OMX_PORT_PARAM_TYPE port_param;

    lthread_mutex_lock(&base->mutex);

    if (base->audio_port || base->audio_port || base->audio_port) goto bail;

    limoi_set_spec_version(&port_param, sizeof(OMX_PORT_PARAM_TYPE));
    if (OMX_ErrorNone == limoi_component_get_parameter
        (hComp, OMX_IndexParamAudioInit, &port_param))
    {
        aports = port_param.nPorts;
        apb = port_param.nStartPortNumber;
    }

    if (OMX_ErrorNone == limoi_component_get_parameter
        (hComp, OMX_IndexParamVideoInit, &port_param))
    {
        vports = port_param.nPorts;
        vpb = port_param.nStartPortNumber;
    }

    if (OMX_ErrorNone == limoi_component_get_parameter
        (hComp, OMX_IndexParamOtherInit, &port_param))
    {
        oports = port_param.nPorts;
        opb = port_param.nStartPortNumber;
    }

    if (aports)
        base->audio_port = malloc(sizeof(limoi_port_t *) * aports);
    if (vports)
        base->video_port = malloc(sizeof(limoi_port_t *) * vports);
    if (oports)
        base->other_port = malloc(sizeof(limoi_port_t *) * oports);

    list = base->ports_list_head.next;
    while (list != &base->ports_list_head)
    {
        limoi_port = (limoi_port_t *)list;
        switch (LIMOI_PORTDEF(limoi_port)->eDomain)
        {
        case OMX_PortDomainAudio :
            aports--;
            *(base->audio_port + limoi_port->index - apb) = limoi_port;
            break;

        case OMX_PortDomainVideo :
            vports--;
            *(base->video_port + limoi_port->index - vpb) = limoi_port;
            break;

        default :
            oports--;
            *(base->other_port + limoi_port->index - opb) = limoi_port;
        }
        list = list->next;
    }
    ASSERT(!aports && !vports && !oports);

 bail:
    lthread_mutex_unlock(&base->mutex);

    return 0;
}

void * thread_main_handle (void * arg)
{
    OMX_HANDLETYPE hComp = (OMX_HANDLETYPE)arg;
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComp);
    OMX_ERRORTYPE  res = OMX_ErrorNone;
    sub_thread_priv_t sub_priv;

    enumerate_ports(hComp);

    //check to create sub thread for other domain
    if (base->other_port && base->limoi_callbacks.sub)
    {
        sub_priv.stop = 1;
        sub_priv.hComp = hComp;
        sub_priv.audio_port = base->audio_port;
        sub_priv.video_port = base->video_port;
        sub_priv.other_port = base->other_port;
        if (lthread_create(&base->sub_thread, sub_thread, &sub_priv, 0))
        {
            LIMERR("%s unable to create sub thread.\n", base->name);
            //error is ignored, it is up to implementor to decided
            //what to do to fix.
        }

        //wait till sub thread is up running
        lthread_mutex_lock(&base->mutex);
        while (sub_priv.stop != 0)
        {
            if (lthread_cond_timedwait(&base->cond, &base->mutex, 5000))
            {
                LIMERR("%s unable to start sub thread.\n", base->name);
                break;
            }
        }
        lthread_mutex_unlock(&base->mutex);
    }

    //continues to main thread.
    while (1)
    {
        lthread_mutex_lock(&base->mutex);

        if (base->preloaded > LIMOI_PRELOAD_BUFFER_COUNT)
            base->preloaded = LIMOI_PRELOAD_BUFFER_COUNT + 1;

        while (base->state_transition
               || OMX_StateIdle == base->state
               || (OMX_StatePause == base->state
                   && LIMOI_PRELOAD_BUFFER_COUNT <= base->preloaded)
               || OMX_ErrorNone != res)
        {
            LIMDBG("%s main thread timed waiting on:\n"
                   "\t state     : %s\n"
                   "\t preloaded : %d\n",
                   base->name, limoi_omx_state_string(base->state),
                   base->preloaded);
            lthread_cond_timedwait(&base->cond, &base->mutex, 50);
            LIMDBG("%s main thread running.\n", base->name);
            res = OMX_ErrorNone;
        }

        if (OMX_StateLoaded == base->state || OMX_StateInvalid == base->state)
        {
            lthread_mutex_unlock(&base->mutex);
            break;
        }
        base->busy |= COMP_MAIN_THREAD_BUSY;
        lthread_mutex_unlock(&base->mutex);

        LIMDBG("%s main callback in.\n", base->name);
        res = base->limoi_callbacks.main
            (hComp, base->audio_port, base->video_port,
             base->other_port, &base->preloaded);
        LIMDBG("%s main callback out.\n", base->name);

        lthread_mutex_lock(&base->mutex);
        base->busy &= ~COMP_MAIN_THREAD_BUSY;
        lthread_cond_broadcast(&base->cond);
        lthread_mutex_unlock(&base->mutex);

        event_deliver_real(hComp);
    }

    //check to shut sub thread down.
    if (base->other_port && base->limoi_callbacks.sub)
    {
        lthread_mutex_lock(&base->mutex);
        while (sub_priv.stop != 1)
        {
            sub_priv.stop = -1;
            lthread_cond_broadcast(&base->cond);
            if (lthread_cond_timedwait(&base->cond, &base->mutex, 5000))
            {
                LIMERR("%s unable to stop sub thread.\n", base->name);
                break;
            }
        }
        lthread_mutex_unlock(&base->mutex);
        lthread_join(base->sub_thread);
    }

    //broadcast main thread shutdown.
    lthread_mutex_lock(&base->mutex);
    //this indicates that main thread has been completely shut down.
    base->state_transition = -1;
    lthread_cond_broadcast(&base->cond);
    lthread_mutex_unlock(&base->mutex);

    LIMDBG("%s main thread exits.\n", base->name);
    return NULL;
}

//caller should hold neither base->mutex nor port->mutex
//synchronous component must guarantee that all buffers can be consumed
//within 5 milliseconds, per IL spec.
OMX_ERRORTYPE sync_consume_buffers (OMX_HANDLETYPE hComp)
{
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComp);
    OMX_ERRORTYPE  res = OMX_ErrorNone;

    lthread_mutex_lock(&base->mutex);
    if (base->busy & COMP_BUFFER_CONSUMING_BUSY) goto bail;
    base->busy |= COMP_BUFFER_CONSUMING_BUSY;
    lthread_mutex_unlock(&base->mutex);

    enumerate_ports (hComp);

 consume_more:
    LIMDBG("%s consume buffers.\n", base->name);
    lthread_mutex_lock(&base->mutex);

    if (base->preloaded > LIMOI_PRELOAD_BUFFER_COUNT)
        base->preloaded = LIMOI_PRELOAD_BUFFER_COUNT + 1;

    if (base->state_transition
        || OMX_StateIdle == base->state
        || (OMX_StatePause == base->state
            && LIMOI_PRELOAD_BUFFER_COUNT <= base->preloaded)
        || OMX_ErrorNone != res
        || OMX_StateLoaded == base->state
        || OMX_StateInvalid == base->state)
    {
        LIMDBG("%s buffer consuming bailed because:\n"
               "\t state            : %s\n"
               "\t preloaded        : %d\n"
               "\t state transition : %d\n",
               base->name, limoi_omx_state_string(base->state),
               base->preloaded, base->state_transition);
        lthread_mutex_unlock(&base->mutex);
    } else
    {
        lthread_mutex_unlock(&base->mutex);

        LIMDBG("%s main callback in.\n", base->name);
        res = base->limoi_callbacks.main
            (hComp, base->audio_port, base->video_port,
             base->other_port, &base->preloaded);
        LIMDBG("%s main callback out.\n", base->name);

        event_deliver_real(hComp);

        if (OMX_ErrorNone == res)
        {
            LIMDBG("%s main callback continue consuming.\n", base->name);
            goto consume_more;
        }
    }

    lthread_mutex_lock(&base->mutex);
    base->busy &= ~COMP_BUFFER_CONSUMING_BUSY;
 bail:
    lthread_cond_broadcast(&base->cond);
    lthread_mutex_unlock(&base->mutex);
    return OMX_ErrorNone;
}

void limoi_component_lock (OMX_HANDLETYPE hComp)
{
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComp);
    lthread_mutex_lock(&base->mutex);
    base->lock_count++;
}

void limoi_component_unlock (OMX_HANDLETYPE hComp)
{
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComp);
    base->lock_count--;
    lthread_mutex_unlock(&base->mutex);
}

void limoi_component_cond_broadcast (OMX_HANDLETYPE hComp)
{
    lthread_cond_broadcast(&(LIMOI_COMPONENT_PRIV(hComp)->cond));
}

int limoi_component_cond_timedwait (
    OMX_HANDLETYPE hComp,
    int miliseconds)
{
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComp);
    int ret;

    base->lock_count--;
    ret = lthread_cond_timedwait(&base->cond, &base->mutex, miliseconds);
    base->lock_count++;
    return ret;
}
