/*
  Copyright (C) 2011 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX IL component tunnel support subroutines.

  REVISION:

    1) initial creation. ----------------------------------------- 2011-07-22 MG
*/
#define _GNU_SOURCE
#include "limoi.h"
#include "component.h"
#include "string.h"
#include <stdarg.h>

//#undef LIMDBG
#ifndef LIMDBG
#define LIMDBG(fmt, arg...)
#endif

OMX_ERRORTYPE omx_component_tunnel_request (
    OMX_HANDLETYPE hComp,
    OMX_U32 nPort,
    OMX_HANDLETYPE hTunneledComp,
    OMX_U32 nTunneledPort,
    OMX_TUNNELSETUPTYPE* pTunnelSetup)
{
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComp);
    limoi_port_t * limoi_port;
    OMX_PARAM_PORTDEFINITIONTYPE * port;
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    int retry_count = 0;

 retry:
    lthread_mutex_lock(&base->mutex);

    limoi_port = limoi_component_get_port(hComp, nPort);
    if (NULL == limoi_port)
    {
        LIMERR("unable to find port: %u of %s\n", nPort, base->name);
        ret = OMX_ErrorBadParameter;
        goto bail;
    }

    lthread_mutex_lock(&limoi_port->mutex);
    port = LIMOI_PORTDEF(limoi_port);

    if (OMX_StateLoaded != base->state && port->bEnabled == OMX_TRUE)
    {
        LIMERR("%s invalid state %d and port %u enabled %d.\n", base->name,
               base->state, port->nPortIndex, port->bEnabled);
        ret = OMX_ErrorInvalidState;
        goto bail;
    }

    if (port->bPopulated == OMX_TRUE)
    {
        LIMWARN("%s port %u already populated?\n",
                base->name, limoi_port->index);
        //client may not wait for the port disabling callback before
        //trying to set up the tunnel, wait a bit then retry before bailing.
        if (retry_count++ < 60)
        {
            lthread_mutex_unlock(&limoi_port->mutex);
            lthread_mutex_unlock(&base->mutex);
            usleep(50000);
            goto retry;
        } else
        {
            LIMERR("%s invalid state %d and port %u\n"
                   "\tenabled   : %d\n"
                   "\tpopulated : %d\n", base->name,
                   base->state, port->nPortIndex,
                   port->bEnabled, port->bPopulated);
            ret = OMX_ErrorInvalidState;
            goto bail;
        }
    }

    if (NULL == pTunnelSetup)
    {
        ret = OMX_ErrorBadParameter;
        goto bail;
    }

    limoi_port->tunnel_comp = NULL;
    if (NULL == hTunneledComp)
    {
        goto bail;
    }

	if (OMX_DirOutput == port->eDir)
	{
        if (pTunnelSetup->eSupplier != OMX_BufferSupplyMax)
        {
            LIMERR("both ports are output.\n");
            ret = OMX_ErrorPortsNotCompatible;
            goto bail;
        }

        // limoi component default:
        // - read only, the input port that receives this structure cannot
        //   alter the contents of buffers supplied on the tunnel.
        pTunnelSetup->nTunnelFlags = 0;
        pTunnelSetup->eSupplier =
            LIMOI_BUF_SUPPLIER(limoi_port)->eBufferSupplier;

        limoi_port->tunnel_comp = hTunneledComp;
        limoi_port->tunnel_port = nTunneledPort;
	} else if (OMX_DirInput == port->eDir)
    {
        OMX_PARAM_PORTDEFINITIONTYPE tunneled_port;
        OMX_PARAM_BUFFERSUPPLIERTYPE tunneled_supplier;
        OMX_PARAM_BUFFERSUPPLIERTYPE * supplier;

        if (pTunnelSetup->eSupplier == OMX_BufferSupplyMax)
        {
            LIMERR("both ports are input.\n");
            ret = OMX_ErrorPortsNotCompatible;
            goto bail;
        }

        //we can safely ignore error check here, already done in prior steps.
        limoi_set_spec_version(&tunneled_port,
                               sizeof(OMX_PARAM_PORTDEFINITIONTYPE));
        tunneled_port.nPortIndex = nTunneledPort;

        lthread_mutex_unlock(&limoi_port->mutex);
        lthread_mutex_unlock(&base->mutex);
        OMX_GetParameter(hTunneledComp, OMX_IndexParamPortDefinition,
                         &tunneled_port);
        lthread_mutex_lock(&base->mutex);
        lthread_mutex_lock(&limoi_port->mutex);

        // now to check ports compatibility.
        //TODO: more robust compatibility check required.
        if (port->eDomain == tunneled_port.eDomain)
        {
            switch (port->eDomain)
            {
            case OMX_PortDomainAudio :
                if (tunneled_port.format.audio.eEncoding
                    == port->format.audio.eEncoding) goto negotiate;
                LIMERR("audio ports are not compatible: %d %d.\n",
                       tunneled_port.format.audio.eEncoding,
                       port->format.audio.eEncoding);
                break;

            case OMX_PortDomainVideo :
                if (tunneled_port.format.video.eColorFormat
                    == port->format.video.eColorFormat
                    && tunneled_port.format.video.eCompressionFormat
                    == port->format.video.eCompressionFormat) goto negotiate;

                LIMERR("color format: %d | %d, compression: %d | %d\n",
                       port->format.video.eColorFormat,
                       tunneled_port.format.video.eColorFormat,
                       port->format.video.eCompressionFormat,
                       tunneled_port.format.video.eCompressionFormat);
                break;

            case OMX_PortDomainImage :
                if (tunneled_port.format.image.eColorFormat
                    == port->format.image.eColorFormat) goto negotiate;
                LIMERR("image ports are not compatible: %d %d.\n",
                       tunneled_port.format.image.eColorFormat,
                       port->format.image.eColorFormat);
                break;

            case OMX_PortDomainOther :
                if(tunneled_port.format.other.eFormat
                   == port->format.other.eFormat) goto negotiate;
                LIMERR("other ports are not compatible: %d %d.\n",
                       tunneled_port.format.other.eFormat,
                       port->format.other.eFormat);
                break;

            default :
                break;
            }
        } else
        {
            LIMERR("ports domain are not compatible: %d %d.\n",
                   port->eDomain, tunneled_port.eDomain);
        }
        ret = OMX_ErrorPortsNotCompatible;
        goto bail;

    negotiate:
        // now to negotiate buffer supplier.
        supplier = LIMOI_BUF_SUPPLIER(limoi_port);

        switch (supplier->eBufferSupplier)
        {
        case OMX_BufferSupplyInput :
            pTunnelSetup->eSupplier = OMX_BufferSupplyInput;
            break;

        case OMX_BufferSupplyUnspecified :
        case OMX_BufferSupplyOutput :
        default :
            pTunnelSetup->eSupplier = OMX_BufferSupplyOutput;
            break;
        }
        LIMDBG("tunneled %s port %u,  supplier: %s\n",
               base->name, nPort,
               (pTunnelSetup->eSupplier == OMX_BufferSupplyInput)?
               "input" : "output");

        // negotiate buffer count.
        // NOTE: buffer count shall only be modified at LOAD time, while
        //       buffer size can be assigned at run time.
        if (port->nBufferCountMin > tunneled_port.nBufferCountMin)
        {
            tunneled_port.nBufferCountActual = port->nBufferCountMin;
            port->nBufferCountActual = port->nBufferCountMin;
        } else
        {
            tunneled_port.nBufferCountActual = tunneled_port.nBufferCountMin;
            port->nBufferCountActual = tunneled_port.nBufferCountMin;
        }

        // all look good, finalize tunnel setup.
        supplier->eBufferSupplier = pTunnelSetup->eSupplier;

        limoi_set_spec_version(&tunneled_supplier,
                               sizeof(OMX_PARAM_BUFFERSUPPLIERTYPE));
        tunneled_supplier.nPortIndex = nTunneledPort;
        tunneled_supplier.eBufferSupplier = pTunnelSetup->eSupplier;

        limoi_port->tunnel_comp = hTunneledComp;
        limoi_port->tunnel_port = nTunneledPort;

        lthread_mutex_unlock(&limoi_port->mutex);
        lthread_mutex_unlock(&base->mutex);

        OMX_SetParameter(hTunneledComp, OMX_IndexParamPortDefinition,
                         &tunneled_port);

        OMX_SetParameter(hTunneledComp, OMX_IndexParamCompBufferSupplier,
                         &tunneled_supplier);

        buffer_share_validate_policy(hComp, nPort);
        buffer_share_validate_policy(hTunneledComp, nTunneledPort);

        return OMX_ErrorNone;
    } else
    {
        LIMERR("port is neither output nor input.\n");
        ret = OMX_ErrorPortsNotCompatible;
    }

 bail:
    lthread_mutex_unlock(&limoi_port->mutex);
    lthread_mutex_unlock(&base->mutex);
    return ret;
}
