/*
  Copyright (C) 2010 - 2011 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX IL standard video decoder component.

  REVISION:

    1) initial creation. ----------------------------------------- 2010-02-17 MG
*/
#define _GNU_SOURCE
#include "limoi.h"
#include "video-priv.h"
#include <sys/stat.h>
#include <fcntl.h>

//#undef LIMDBG
#ifndef LIMDBG
#define LIMDBG(fmt, arg...)
#endif


//DEBUG: define this to write yuv to 'vdecoder-debug.yuv' while rendering.
//#define DEBUG_SAVE2FILE (1024 * 1024 * 4)

#ifdef DEBUG_SAVE2FILE
static void debug_save2file (void * buf, int len)
{
    static int fd = -1;
    static int bytes;

    if (-1 == fd)
    {
        fd = open("vdecoder-debug.yuv", O_WRONLY | O_CREAT | O_TRUNC,
                  S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
        if (fd < 0)
        {
            LIMERR("unable to create debug file.\n");
            fd = -2;
        }
    }
    if (fd > 0)
    {
        if (bytes < DEBUG_SAVE2FILE)
        {
            bytes += write(fd, buf, len);
        } else
        {
            LIMINFO("yuv debug file closed.\n");
            close(fd);
            fd = -2;
        }
    }
}
#else
#define debug_save2file(x, y) ;
#endif

static OMX_ERRORTYPE main_handle (
    OMX_HANDLETYPE hComp,
    limoi_port_t ** audio_port,
    limoi_port_t ** video_port,
    limoi_port_t ** other_port,
    int * preloaded)
{
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComp);
    limoi_video_decoder_t * decoder = LIMOI_VIDEO_DECODER(hComp);
    limoi_port_t * video_port_in = *video_port;
    limoi_port_t * video_port_out = *(video_port + 1);
    OMX_BUFFERHEADERTYPE ** in = NULL;
    OMX_BUFFERHEADERTYPE ** out = NULL;
    OMX_ERRORTYPE ret = OMX_ErrorNone;

    LIMDBG("%s decoder get input port buffer.\n", base->name);
    in = limoi_component_get_port_buffer(hComp, video_port_in, -1);
    if (NULL == in)
    {
        LIMDBG("%s unable to get input port buffer.\n", base->name);
        ret = OMX_ErrorNotReady;
        goto bail1;
    } else if (!(*in)->nFilledLen && !((*in)->nFlags & OMX_BUFFERFLAG_EOS))
    {
        LIMDBG("%s decoder input port buffer zero length.\n", base->name);
        if (-1 == limoi_component_put_port_buffer(hComp, video_port_in, in))
        {
            ret = OMX_ErrorNotReady;
        }
        goto bail1;
    }
    LIMDBG("%s input buffer len: %u\n", base->name, (*in)->nFilledLen);

    LIMDBG("%s decoder get output port buffer.\n", base->name);
    out = limoi_component_get_port_buffer(hComp, video_port_out, -1);
    if (NULL == out)
    {
        LIMDBG("%s unable to get output port buffer.\n", base->name);
        ret = OMX_ErrorNotReady;
        goto bail2;
    } else if ((*out)->nFilledLen == 0)
    {
        LIMDBG("%s output buffer len: %u\n", base->name, (*out)->nAllocLen);

        //decoder still gets called when input buffer contains zero data
        //so that decoder is able to output 'cached' picture if any.
        //if ((*in)->nFilledLen)
        {
            LIMDBG("%s decoding bytes in: %u\n", base->name,
                   (*in)->nFilledLen);
            ret = decoder->priv->callbacks.decode(hComp, *in, *out);
            LIMDBG("%s decoding bytes out: %u\n", base->name,
                   (*out)->nFilledLen);
        }

        if ((*in)->nFlags & OMX_BUFFERFLAG_EOS)
        {
            (*out)->nFlags |= OMX_BUFFERFLAG_EOS;
            limoi_component_deliver_event
                (hComp, OMX_EventBufferFlag, 0, (*in)->nFlags, NULL);
            limoi_component_deliver_event
                (hComp, OMX_EventBufferFlag, 1, (*out)->nFlags, NULL);
        }

        LIMDBG("%s put input port buffer back: %p.\n",
               base->name, (*in)->pBuffer);
        if (!(*in)->nFilledLen)
        {
            int tmp;

            tmp = limoi_component_put_port_buffer(hComp, video_port_in, in);
            if (tmp)
            {
                if (tmp == -1) ret = OMX_ErrorNotReady;
                else
                {
                    LIMDBG("%s unable to put back input port buffer.\n",
                           base->name);
                }
            }
            in = NULL;
        }
    }

    //put back output buffer if decode succeeds and output buffer
    //has data.
    if ((*out)->nFilledLen || (*out)->nFlags & OMX_BUFFERFLAG_EOS)
    {
        int tmp;

        debug_save2file((*out)->pBuffer + (*out)->nOffset,
                        (*out)->nFilledLen);
        (*preloaded)++;

        LIMDBG("%s put output port buffer back: %p %u.\n",
               base->name, (*out)->pBuffer, (*out)->nFilledLen);

        tmp = limoi_component_put_port_buffer(hComp, video_port_out, out);
        if (tmp)
        {
            if (tmp == -1)
            {
                LIMDBG("%s put output buffer, tunnel/client not ready.\n",
                       base->name);
                ret = OMX_ErrorNotReady;
            }
            else
            {
                LIMDBG("%s unable to put back output port buffer.\n",
                       base->name);
            }
        }
        out = NULL;
    }

 bail2:
    if (in) limoi_component_release_port_buffer(hComp, video_port_in, in);
    if (out) limoi_component_release_port_buffer(hComp, video_port_out, out);
 bail1:
    return ret;
}

static OMX_ERRORTYPE prep_cmd_handle (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    limoi_video_decoder_t * decoder = LIMOI_VIDEO_DECODER(cmd->hComponent);
    limoi_video_decoder_priv_t * pdecoder = decoder->priv;

    //decoder callback
    if (pdecoder->callbacks.command_prep)
    {
        ret = pdecoder->callbacks.command_prep(cmd);
        if (OMX_ErrorNone != ret)
        {
            LIMERR("%s command callback returned error %s\n",
                   LIMOI_COMPONENT_PRIV(cmd->hComponent)->name,
                   limoi_omx_error_string(ret));
        }
    }
    return ret;
}

static OMX_ERRORTYPE post_cmd_handle (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    limoi_video_decoder_t * decoder = LIMOI_VIDEO_DECODER(cmd->hComponent);
    limoi_video_decoder_priv_t * pdecoder = decoder->priv;

    if (pdecoder->callbacks.command_post)
    {
        ret = pdecoder->callbacks.command_post(cmd);
        if (OMX_ErrorNone != ret)
        {
            LIMERR("%s command callback returned error %s\n",
                   LIMOI_COMPONENT_PRIV(cmd->hComponent)->name,
                   limoi_omx_error_string(ret));
        }
    }
    return ret;
}

static int add_port (
    OMX_HANDLETYPE hComp,
    OMX_VIDEO_CODINGTYPE type)
{
	OMX_PARAM_PORTDEFINITIONTYPE port;
    OMX_PORT_PARAM_TYPE port_param;

    limoi_set_spec_version(&port_param, sizeof(OMX_PORT_PARAM_TYPE));
    port_param.nPorts = 2;
    port_param.nStartPortNumber = 0; //VPB = 0
    limoi_component_set_parameter(hComp, OMX_IndexParamVideoInit, &port_param);

    limoi_set_spec_version(&port, sizeof(OMX_PARAM_PORTDEFINITIONTYPE));

    port.nPortIndex = 0;
    port.eDir = OMX_DirInput;
    port.nBufferCountActual = LIMOI_VIDEO_DECODER_INPUT_BUFFER_NUM;
    port.nBufferCountMin = LIMOI_VIDEO_DECODER_INPUT_BUFFER_NUM;
    port.nBufferSize = LIMOI_VIDEO_DECODER_INPUT_BUFFER_BYTES;
    port.bEnabled = OMX_TRUE;
    port.bPopulated = OMX_FALSE;
    port.eDomain = OMX_PortDomainVideo;
    port.bBuffersContiguous = OMX_FALSE;
    port.nBufferAlignment = 0;
    port.format.video.cMIMEType = NULL;
    port.format.video.pNativeRender = 0;
    port.format.video.bFlagErrorConcealment = OMX_FALSE;
    port.format.video.eCompressionFormat = type;
    port.format.video.eColorFormat = OMX_COLOR_FormatUnused;
    limoi_component_add_port(hComp, &port);

    port.nPortIndex = 1;
    port.eDir = OMX_DirOutput;
    port.nBufferCountActual = LIMOI_VIDEO_DECODER_OUTPUT_BUFFER_NUM;
    port.nBufferCountMin = LIMOI_VIDEO_DECODER_OUTPUT_BUFFER_NUM;
    port.nBufferSize = LIMOI_VIDEO_DECODER_OUTPUT_BUFFER_BYTES;
    port.format.video.eCompressionFormat = OMX_VIDEO_CodingUnused;
    port.format.video.eColorFormat = OMX_COLOR_FormatYUV420Planar;
    port.format.video.nFrameWidth = 720;
    port.format.video.nFrameHeight = 480;
    limoi_component_add_port(hComp, &port);

    return 0;
}

int limoi_video_decoder_set_pixel_format (
    OMX_HANDLETYPE hComp,
    int w,
    int h,
    OMX_COLOR_FORMATTYPE color_format,
    int aspect_ratio_num,
    int aspect_ratio_den)
{
    OMX_PARAM_PORTDEFINITIONTYPE port;
    OMX_CONFIG_SCALEFACTORTYPE scale;
    int changed = 0;

    LIMDBG("%s set pixel format w x h: %d x %d\n",
           LIMOI_COMPONENT_PRIV(hComp)->name, w, h);

    limoi_set_spec_version(&scale, sizeof(OMX_CONFIG_SCALEFACTORTYPE));
    scale.nPortIndex = 1;
    if (aspect_ratio_den)
    {
        scale.xWidth = (uint64_t)65536 * aspect_ratio_num / aspect_ratio_den;
    } else
    {
        scale.xWidth = 65536;
    }
    scale.xHeight = 65536;
    limoi_component_set_parameter(hComp, OMX_IndexConfigCommonScale, &scale);

    limoi_set_spec_version(&port, sizeof(OMX_PARAM_PORTDEFINITIONTYPE));
    port.nPortIndex = 1;
    if (OMX_ErrorNone != limoi_component_get_parameter
        (hComp, OMX_IndexParamPortDefinition, &port))
    {
        changed = 1;
    } else
    {
        if (port.format.video.nFrameWidth != w
            || port.format.video.nFrameHeight != h
            || port.format.video.eColorFormat != color_format)
        {
            changed = 1;
        }
    }

    if (changed)
    {
        limoi_set_spec_version(&port, sizeof(OMX_PARAM_PORTDEFINITIONTYPE));
        port.nPortIndex = 1;

        port.format.video.nFrameWidth = w;
        port.format.video.nFrameHeight = h;
        port.format.video.eColorFormat = color_format;

        limoi_component_set_parameter
            (hComp, OMX_IndexParamPortDefinition, &port);
        limoi_component_deliver_event
            (hComp, OMX_EventPortSettingsChanged, 1, 0, NULL);
    }

    return 0;
}

int limoi_video_decoder_set_callbacks (
    OMX_HANDLETYPE hComp,
    limoi_video_decoder_callbacks_t * callbacks)
{
    limoi_video_decoder_t * decoder = LIMOI_VIDEO_DECODER(hComp);
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComp);

    lthread_mutex_lock(&base->mutex);
    decoder->priv->callbacks = *callbacks;
    lthread_mutex_unlock(&base->mutex);

    return 0;
}

OMX_ERRORTYPE limoi_video_decoder_deinit (
    OMX_HANDLETYPE hComponent)
{
    limoi_video_decoder_t * decoder = LIMOI_VIDEO_DECODER(hComponent);

    free(decoder->priv);

    // parent destructor.
    return limoi_component_deinit(hComponent);
}

static OMX_ERRORTYPE video_decoder_init (
    OMX_HANDLETYPE * hComp,
	int bytes,
	const char * comp_name,
	const char * comp_version,
    OMX_VIDEO_CODINGTYPE type,
    int synchronous)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    OMX_COMPONENTTYPE * hdl;
    limoi_component_callbacks_t callbacks;
    limoi_video_decoder_t * decoder;
    limoi_video_decoder_priv_t * priv = NULL;

    LIMDBG("init component %s, version %s.\n", comp_name, comp_version);

    priv = calloc(1, sizeof(limoi_video_decoder_priv_t));
    if (NULL == priv)
    {
        LIMERR("out of memory upon video decoder init.\n");
        ret = OMX_ErrorInsufficientResources;
        goto bail;
    }

    ret = synchronous?
        limoi_component_init_sync(hComp, bytes) :
        limoi_component_init(hComp, bytes);

    if (OMX_ErrorNone != ret) goto bail;
    hdl = (OMX_COMPONENTTYPE *)*hComp;

    decoder = LIMOI_VIDEO_DECODER(*hComp);
    decoder->priv = priv;    

    limoi_component_set_name_version(*hComp, comp_name, comp_version);
    add_port(*hComp, type);

    callbacks.command_prep = prep_cmd_handle;
    callbacks.command_post = post_cmd_handle;
    callbacks.main = main_handle;
    callbacks.sub = NULL;
    limoi_component_set_callbacks(*hComp, &callbacks);

    LIMDBG("component %s initialized.\n", comp_name);
    hdl->ComponentDeInit = limoi_video_decoder_deinit;

 bail:
    if (OMX_ErrorNone != ret && priv) free(priv);
    return ret;
}

OMX_ERRORTYPE limoi_video_decoder_init (
    OMX_HANDLETYPE * hComp,
	int bytes,
	const char * comp_name,
	const char * comp_version,
    OMX_VIDEO_CODINGTYPE type)
{
    return video_decoder_init(hComp, bytes, comp_name, comp_version, type, 0);
}

OMX_ERRORTYPE limoi_video_decoder_init_sync (
    OMX_HANDLETYPE * hComp,
	int bytes,
	const char * comp_name,
	const char * comp_version,
    OMX_VIDEO_CODINGTYPE type)
{
    return video_decoder_init(hComp, bytes, comp_name, comp_version, type, 1);
}
