/*
  Copyright (C) 2010 - 2011 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX IL standard video renderer component.

  REVISION:

    1) initial creation. ----------------------------------------- 2010-02-21 MG
*/
#define _GNU_SOURCE
#include "limoi.h"
#include "video-priv.h"
#include <sys/stat.h>
#include <fcntl.h>

//#undef LIMDBG
#ifndef LIMDBG
#define LIMDBG(fmt, arg...)
#endif


//DEBUG: define this to write yuv to 'vrenderer-debug.yuv' while rendering.
//#define DEBUG_SAVE2FILE (1024 * 1024 * 4)

#ifdef DEBUG_SAVE2FILE
static void debug_save2file (void * buf, int len)
{
    static int fd = -1;
    static int bytes;

    if (-1 == fd)
    {
        fd = open("vrenderer-debug.yuv", O_WRONLY | O_CREAT | O_TRUNC,
                  S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
        if (fd < 0)
        {
            LIMERR("unable to create debug file.\n");
            fd = -2;
        }
    }
    if (fd > 0)
    {
        if (bytes < DEBUG_SAVE2FILE)
        {
            bytes += write(fd, buf, len);
        } else
        {
            LIMINFO("yuv debug file closed.\n");
            close(fd);
            fd = -2;
        }
    }
}
#else
#define debug_save2file(x, y) ;
#endif

static OMX_ERRORTYPE main_handle (
    OMX_HANDLETYPE hComp,
    limoi_port_t ** audio_port,
    limoi_port_t ** video_port,
    limoi_port_t ** other_port,
    int * preloaded)
{
    limoi_video_renderer_t * renderer = LIMOI_VIDEO_RENDERER(hComp);
    limoi_port_t * yuv_port = *video_port;
    OMX_BUFFERHEADERTYPE ** buf;
    OMX_U32 flags;
    OMX_ERRORTYPE ret = OMX_ErrorNone;

    if (*preloaded == 0)
    {
        //do not requrie preload.
        *preloaded = LIMOI_PRELOAD_BUFFER_COUNT;
        goto bail;
    }

    LIMDBG("%s renderer get port buffer.\n",
           LIMOI_COMPONENT_PRIV(hComp)->name);
    buf = limoi_component_get_port_buffer(hComp, yuv_port, -1);
    if (NULL == buf)
    {
        LIMDBG("%s unable to get port buffer.\n",
               LIMOI_COMPONENT_PRIV(hComp)->name);
        ret = OMX_ErrorNotReady;
        goto bail;
    }

    if ((*buf)->nFilledLen || (*buf)->nFlags & OMX_BUFFERFLAG_EOS)
    {
        LIMDBG("%s renderer render buffer bytes: %u flag: %u.\n",
               LIMOI_COMPONENT_PRIV(hComp)->name, (*buf)->nFilledLen,
               (*buf)->nFlags);

        if ((*buf)->nFilledLen)
        {
            LIMDBG("%s renderer render buffer bytes: %u.\n",
                   LIMOI_COMPONENT_PRIV(hComp)->name, (*buf)->nFilledLen);
            debug_save2file((*buf)->pBuffer + (*buf)->nOffset,
                            (*buf)->nFilledLen);
            renderer->priv->callbacks.render(hComp, *buf);
        }

        if ((*buf)->nFlags & OMX_BUFFERFLAG_EOS)
        {
            limoi_component_deliver_event
                (hComp, OMX_EventBufferFlag, 0, (*buf)->nFlags, NULL);
        }

        if (!(*buf)->nFilledLen)
        {
            flags = (*buf)->nFlags;
            if (-1 == limoi_component_put_port_buffer(hComp, yuv_port, buf))
            {
                LIMDBG("%s unable to put back port buffer.\n",
                       LIMOI_COMPONENT_PRIV(hComp)->name);
                ret = OMX_ErrorNotReady;
                //not an error, tunnel port may be in the wrong state.
            }
            buf = NULL;

            if (flags & OMX_BUFFERFLAG_EOS)
            {
                limoi_component_deliver_event
                    (hComp, OMX_EventBufferFlag, yuv_port->index,
                     flags, NULL);
            }
        }
    }

    if (buf)
    {
        if ((*buf)->nFilledLen)
            limoi_component_release_port_buffer(hComp, yuv_port, buf);
        else
            limoi_component_put_port_buffer(hComp, yuv_port, buf);
    }

 bail:
    return ret;
}

static OMX_ERRORTYPE prep_cmd_handle (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    limoi_video_renderer_t * renderer = LIMOI_VIDEO_RENDERER(cmd->hComponent);
    limoi_video_renderer_priv_t * prenderer = renderer->priv;

    if (prenderer->callbacks.command_prep)
    {
        ret = prenderer->callbacks.command_prep(cmd);
        if (OMX_ErrorNone != ret)
        {
            LIMERR("%s command callback returned error %s\n",
                   LIMOI_COMPONENT_PRIV(cmd->hComponent)->name,
                   limoi_omx_error_string(ret));
        }
    }
    return ret;
}

static OMX_ERRORTYPE post_cmd_handle (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    limoi_video_renderer_t * renderer = LIMOI_VIDEO_RENDERER(cmd->hComponent);
    limoi_video_renderer_priv_t * prenderer = renderer->priv;

    if (prenderer->callbacks.command_post)
    {
        ret = prenderer->callbacks.command_post(cmd);
        if (OMX_ErrorNone != ret)
        {
            LIMERR("%s command callback returned error %s\n",
                   LIMOI_COMPONENT_PRIV(cmd->hComponent)->name,
                   limoi_omx_error_string(ret));
        }
    }
    return ret;
}

static int add_port (
    OMX_HANDLETYPE hComp)
{
	OMX_PARAM_PORTDEFINITIONTYPE port;
    OMX_PORT_PARAM_TYPE port_param;

    limoi_set_spec_version(&port_param, sizeof(OMX_PORT_PARAM_TYPE));
    port_param.nPorts = 1;
    port_param.nStartPortNumber = 0; //VPB = 0
    limoi_component_set_parameter(hComp, OMX_IndexParamVideoInit, &port_param);

    limoi_set_spec_version(&port, sizeof(OMX_PARAM_PORTDEFINITIONTYPE));

    port.nPortIndex = 0;
    port.eDir = OMX_DirInput;
    port.nBufferCountActual = 2;
    port.nBufferCountMin = 2;
    port.nBufferSize = 352 * 288 + 352 * 288 / 2;
    port.bEnabled = OMX_TRUE;
    port.bPopulated = OMX_FALSE;
    port.eDomain = OMX_PortDomainVideo;
    port.bBuffersContiguous = OMX_FALSE;
    port.nBufferAlignment = 0;
    port.format.video.cMIMEType = NULL;
    port.format.video.pNativeRender = 0;
    port.format.video.nFrameWidth = 0;
    port.format.video.nFrameHeight = 0;
    port.format.video.nStride = 0;
    port.format.video.nSliceHeight = 0;
    port.format.video.nBitrate = 0;
    port.format.video.bFlagErrorConcealment = OMX_FALSE;
    port.format.video.eCompressionFormat = OMX_VIDEO_CodingUnused;
    port.format.video.eColorFormat = OMX_COLOR_FormatYUV420Planar;
    port.format.video.pNativeWindow = NULL;
    limoi_component_add_port(hComp, &port);

    return 0;
}

int limoi_video_renderer_set_callbacks (
    OMX_HANDLETYPE hComp,
    limoi_video_renderer_callbacks_t * callbacks)
{
    limoi_video_renderer_t * renderer = LIMOI_VIDEO_RENDERER(hComp);
    limoi_component_priv_t * base = LIMOI_COMPONENT_PRIV(hComp);

    lthread_mutex_lock(&base->mutex);
    renderer->priv->callbacks = *callbacks;
    lthread_mutex_unlock(&base->mutex);

    return 0;
}

OMX_ERRORTYPE limoi_video_renderer_deinit (
    OMX_HANDLETYPE hComponent)
{
    limoi_video_renderer_t * renderer = LIMOI_VIDEO_RENDERER(hComponent);

    free(renderer->priv);

    // parent destructor.
    return limoi_component_deinit(hComponent);
}

static OMX_ERRORTYPE video_renderer_init (
    OMX_HANDLETYPE * hComp,
	int bytes,
	const char * comp_name,
	const char * comp_version,
    int synchronous)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    OMX_COMPONENTTYPE * hdl;
    limoi_component_callbacks_t callbacks;
    limoi_video_renderer_t * renderer;
    limoi_video_renderer_priv_t * priv = NULL;

    LIMDBG("init component %s, version %s.\n", comp_name, comp_version);

    priv = calloc(1, sizeof(limoi_video_renderer_priv_t));
    if (NULL == priv)
    {
        LIMERR("out of memory upon video renderer init.\n");
        ret = OMX_ErrorInsufficientResources;
        goto bail;
    }

    ret = synchronous?
        limoi_component_init_sync(hComp, bytes) :
        limoi_component_init(hComp, bytes);

    if (OMX_ErrorNone != ret) goto bail;
    hdl = (OMX_COMPONENTTYPE *)*hComp;

    renderer = LIMOI_VIDEO_RENDERER(*hComp);
    renderer->priv = priv;

    limoi_component_set_name_version(*hComp, comp_name, comp_version);
    limoi_component_add_role(*hComp, LIMOI_VIDEO_RENDERER_ROLE);
    add_port(*hComp);

    callbacks.command_prep = prep_cmd_handle;
    callbacks.command_post = post_cmd_handle;
    callbacks.main = main_handle;
    callbacks.sub = NULL;
    limoi_component_set_callbacks(*hComp, &callbacks);

    LIMDBG("component %s initialized.\n", comp_name);
    hdl->ComponentDeInit = limoi_video_renderer_deinit;

 bail:
    if (OMX_ErrorNone != ret && priv) free(priv);
    return ret;
}

OMX_ERRORTYPE limoi_video_renderer_init (
    OMX_HANDLETYPE * hComp,
	int bytes,
	const char * comp_name,
	const char * comp_version)
{
    return video_renderer_init(hComp, bytes, comp_name, comp_version, 0);
}

OMX_ERRORTYPE limoi_video_renderer_init_sync (
    OMX_HANDLETYPE * hComp,
	int bytes,
	const char * comp_name,
	const char * comp_version)
{
    return video_renderer_init(hComp, bytes, comp_name, comp_version, 1);
}
