/*
  Copyright (C) 2009 - 2011 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX IL ALSA sink component.

  REVISION:

    1) initial creation. ----------------------------------------- 2009-12-05 MG
*/

#include "limoi-alsa.h"

//#undef LIMDBG
#ifndef LIMDBG
#define LIMDBG(fmt, arg...)
#endif

#define COMPONENT_NAME "OMX.limoi.alsa_sink"

#define ALSA_PLAY_DEVICE "default"

static OMX_ERRORTYPE disable_port (limoi_component_cmd_t * cmd)
{
    alsa_private_t * alsa = LIMOI_ALSA_SINK(cmd->hComponent);

    LIMDBG("%s disable port 0\n", COMPONENT_NAME);
    if (alsa->handle)
    {
        snd_pcm_close(alsa->handle);
        alsa->handle = NULL;
    }
    return OMX_ErrorNone;
}

static OMX_ERRORTYPE enable_port (limoi_component_cmd_t * cmd)
{
    alsa_private_t * alsa = LIMOI_ALSA_SINK(cmd->hComponent);
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    int err;
    OMX_AUDIO_PARAM_PCMMODETYPE pcm;
    snd_pcm_format_t snd_pcm_format;

    LIMDBG("%s enable port 0\n", COMPONENT_NAME);
    if (alsa->handle) snd_pcm_close(alsa->handle);
    err = snd_pcm_open(&alsa->handle, ALSA_PLAY_DEVICE,
                       SND_PCM_STREAM_PLAYBACK, 0);
    if (err < 0)
    {
        alsa->handle = NULL;
        LIMERR("%s playback open error: %s\n",
               COMPONENT_NAME, snd_strerror(err));
        ret = OMX_ErrorUndefined;
        goto bail;
    }

    limoi_set_spec_version(&pcm, sizeof(OMX_AUDIO_PARAM_PCMMODETYPE));
    pcm.nPortIndex = 0;
    if (OMX_ErrorNone !=
        OMX_GetParameter(cmd->hComponent, OMX_IndexParamAudioPcm, &pcm))
    {
        LIMDBG("%s unable to get pcm param, using default.\n", COMPONENT_NAME);
        pcm.nChannels = 2;
        pcm.eNumData = OMX_NumericalDataSigned;
        pcm.eEndian = OMX_EndianMax; //don't care == host type
        pcm.bInterleaved = OMX_TRUE;
        pcm.nBitPerSample = 16;
        pcm.nSamplingRate = 44100;
    }
    LIMDBG("%s pcm parameters:\n"
            "\tnChannels     : %d\n"
            "\tnSampleRate   : %d\n"
            "\tnBitPerSample : %d\n", COMPONENT_NAME,
            pcm.nChannels, pcm.nSamplingRate, pcm.nBitPerSample);

    switch (pcm.nBitPerSample)
    {
        //TODO:
    default :
        snd_pcm_format = SND_PCM_FORMAT_S16;
    }

    alsa->sample_step = (pcm.nBitPerSample / 8) * pcm.nChannels;
    alsa->sample_rate = pcm.nSamplingRate;

    if ((err = snd_pcm_set_params
         (alsa->handle,
          snd_pcm_format,
          SND_PCM_ACCESS_RW_INTERLEAVED,
          pcm.nChannels,
          pcm.nSamplingRate,
          1, //soft_resample
          LIMOI_ALSA_LATENCY * 1000 //latency
          )) < 0)
    {
        LIMERR("set params error: %s\n", snd_strerror(err));
#if 1 //FIXME: for whatever reason, alsa doesn't like these parameters.
        snd_pcm_close(alsa->handle);
        alsa->handle = NULL;
#else
        ret = OMX_ErrorUndefined;
#endif
    }

    if (OMX_ErrorNone != ret)
    {
        if (alsa->handle)
        {
            snd_pcm_close(alsa->handle);
            alsa->handle = NULL;
        }
    }

 bail:
    return ret;
}

static OMX_ERRORTYPE set_state_prep (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    OMX_STATETYPE next_state = cmd->nParam1;
    OMX_STATETYPE state = limoi_component_state(cmd->hComponent);
    alsa_private_t * alsa = LIMOI_ALSA_SINK(cmd->hComponent);

    LIMDBG("%s callback prep set state from %s to %s.\n", COMPONENT_NAME,
           limoi_omx_state_string(state), limoi_omx_state_string(next_state));

    switch (state)
    {
    case OMX_StateIdle :
        if (OMX_StateExecuting == next_state)
        {
            //snd_pcm_pause(alsa->handle, 0);
        } else if (OMX_StatePause == next_state)
        {
            //snd_pcm_pause(alsa->handle, 1);
        } else if (OMX_StateLoaded == next_state)
        {
            disable_port(cmd);
        }
        break;

    case OMX_StateExecuting :
        if (OMX_StatePause == next_state)
        {
            alsa->prev_timestamp = 0;
            //snd_pcm_pause(alsa->handle, 1);
        }
        break;

    case OMX_StatePause :
        if (OMX_StateExecuting == next_state)
        {
            //snd_pcm_pause(alsa->handle, 0);
        }
        break;

    default :
        break;
    }

    return ret;
}

static OMX_ERRORTYPE set_state_post (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    OMX_STATETYPE next_state = cmd->nParam1;
    OMX_STATETYPE state = limoi_component_state(cmd->hComponent);
    alsa_private_t * alsa = LIMOI_ALSA_SINK(cmd->hComponent);

    LIMDBG("%s callback post set state from %s to %s.\n", COMPONENT_NAME,
           limoi_omx_state_string(state), limoi_omx_state_string(next_state));

    switch (state)
    {
    case OMX_StateLoaded :

        // transition to StateIdle
        ret = enable_port(cmd);
        alsa->prev_timestamp = 0;
        break;

    default :
        break;
    }

    return ret;
}

static OMX_ERRORTYPE prep_cmd_handle (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    OMX_HANDLETYPE hComp = cmd->hComponent;

    LIMDBG("%s prep command handler.\n", COMPONENT_NAME);
    limoi_component_lock(hComp);
    switch (cmd->Cmd)
    {
    case OMX_CommandStateSet :
        ret = set_state_prep(cmd);
        break;

    case OMX_CommandPortDisable :
        ret = disable_port(cmd);
        break;

    case OMX_CommandPortEnable :
        ret = enable_port(cmd);
        break;

    default :
        break;
    }
    limoi_component_unlock(hComp);
    return ret;
}

static OMX_ERRORTYPE post_cmd_handle (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    OMX_HANDLETYPE hComp = cmd->hComponent;

    LIMDBG("%s post command handler.\n", COMPONENT_NAME);
    limoi_component_lock(hComp);
    switch (cmd->Cmd)
    {
    case OMX_CommandStateSet :
        ret = set_state_post(cmd);
        break;

    default :
        break;
    }
    limoi_component_unlock(hComp);
    return ret;
}

static int clock_handle (
    OMX_HANDLETYPE hComp,
    OMX_HANDLETYPE clock_tunnel_comp,
    OMX_BUFFERHEADERTYPE * clock_buf)
{
    LIMDBG("%s: clock handle.\n", COMPONENT_NAME);
    clock_buf->nFilledLen = 0;

    return 0;
}

static int render_handle (
    OMX_HANDLETYPE hComp,
    OMX_BUFFERHEADERTYPE * buf,
    OMX_HANDLETYPE clock_tunnel_comp,
    int clock_tunnel_port,
    int clock_port)
{
    snd_pcm_sframes_t ret = 0;
    alsa_private_t * alsa = LIMOI_ALSA_SINK(hComp);
    int offset = 0;
    int samples;

    limoi_component_lock(hComp);
    if (!alsa->handle)
    {
        //upon error, discard pcm so video can still move forward.
        buf->nFilledLen = 0;
        limoi_component_unlock(hComp);
        return 0;
    }

    samples = buf->nFilledLen / alsa->sample_step;

    if (clock_tunnel_comp
        && (alsa->prev_timestamp == 0
            || alsa->prev_timestamp + 0 < buf->nTimeStamp))
    {
        OMX_TIME_CONFIG_TIMESTAMPTYPE ts;
        snd_pcm_sframes_t avail;
        int diff;

        alsa->prev_timestamp = buf->nTimeStamp;

        ASSERT(alsa->sample_rate);
        avail = snd_pcm_avail_update(alsa->handle);
        diff = LIMOI_ALSA_LATENCY - avail * 1000 / alsa->sample_rate;

        LIMDBG("%s available pcm samples: %ld %d.\n", COMPONENT_NAME,
               avail, diff);

        limoi_set_spec_version(&ts, sizeof(OMX_TIME_CONFIG_TIMESTAMPTYPE));
        ts.nPortIndex = clock_tunnel_port;
        ts.nTimestamp = buf->nTimeStamp - (OMX_TICKS)diff * 1000;
        if (OMX_ErrorNone != OMX_SetConfig
            (clock_tunnel_comp, OMX_IndexConfigTimeCurrentAudioReference,
             &ts))
        {
            LIMERR("%s unable to update reference clock.\n",
                   COMPONENT_NAME);
        }
        LIMDBG("%s audio pts at sink: %"PRId64" ref: %"PRId64"\n",
               COMPONENT_NAME, buf->nTimeStamp, ts.nTimestamp);
    }

    //LIMDBG("%s rendering %d.\n", COMPONENT_NAME, samples);
    while (samples)
    {
        ret = snd_pcm_writei(alsa->handle,
                             buf->pBuffer + buf->nOffset + offset,
                             samples);
        if (ret == -EAGAIN) continue;
        if (ret < 0)
        {
            LIMDBG("underrun?\n");
            ret = snd_pcm_recover(alsa->handle, ret, 0);
            if (ret < 0)
            {
                LIMERR("%s unrecoverable error: %s\n",
                       COMPONENT_NAME, snd_strerror(ret));
                break;
            }
        }
        /*
        if (ret != samples)
        {
            LIMDBG("Short write (expected %d, wrote %li)\n",
                   samples, ret);
        }
        */

        offset += ret * alsa->sample_step;
        samples -= ret;
    }
    buf->nFilledLen = 0;

    limoi_component_unlock(hComp);

    return (ret > 0)? 0 : -1;
}

static OMX_ERRORTYPE ComponentDeInit (
    OMX_HANDLETYPE hComponent)
{
    return limoi_audio_renderer_deinit(hComponent);
}

static OMX_ERRORTYPE ComponentInit (OMX_HANDLETYPE * hComponent, const char * name)
{
    OMX_COMPONENTTYPE * hdl;
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    limoi_audio_renderer_callbacks_t callbacks;
    alsa_private_t * alsa;

    LIMDBG("loading %s component.\n", COMPONENT_NAME);
    ret = limoi_audio_renderer_init(hComponent, sizeof(alsa_private_t),
                                    COMPONENT_NAME, VERSION);
    if (OMX_ErrorNone != ret) goto bail;
    hdl = (OMX_COMPONENTTYPE *)*hComponent;

    alsa = LIMOI_ALSA_SINK(*hComponent);

    hdl->ComponentDeInit = ComponentDeInit;

    callbacks.command_prep = prep_cmd_handle;
    callbacks.command_post = post_cmd_handle;
    callbacks.render = render_handle;
    callbacks.clock = clock_handle;
    limoi_audio_renderer_set_callbacks(*hComponent, &callbacks);

    LIMDBG("%s component loaded.\n", COMPONENT_NAME);

 bail:
    return ret;
}

lim_omx_component_registration_t ** OMX_ComponentRegistration (int * num)
{
    static char * roles[] = {LIMOI_AUDIO_RENDERER_ROLE, NULL};
    static lim_omx_component_registration_t comp_reg = {
        COMPONENT_NAME, roles, ComponentInit
    };
    static lim_omx_component_registration_t * regs[] = {&comp_reg};

    *num = 1;
    return regs;
}
