/*
  Copyright (C) 2009 - 2011 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX IL clock component.

  REVISION:

    1) initial creation. ----------------------------------------- 2009-11-23 MG
*/
#define _GNU_SOURCE
#include "limoi-clock.h"

//#undef LIMDBG
#ifndef LIMDBG
#define LIMDBG(fmt, arg...)
#endif

#define COMPONENT_NAME "OMX.limoi.clock"
#define ROLE_NAME "clock.binary"
#define MAX_CLOCK_PORTS 8

typedef struct {
    OMX_HANDLETYPE h_comp;
    OMX_U32 port_index;
    OMX_PTR client_priv;
    OMX_TICKS media_time;
    OMX_TICKS offset;
} media_time_request_t;

static int media_time_rqst_compare (
    const void * obj1,
    const void * obj2)
{
    media_time_request_t * rqst1 = (media_time_request_t *)obj1;
    media_time_request_t * rqst2 = (media_time_request_t *)obj2;
    OMX_TICKS t1, t2;
    clock_private_t * clock = LIMOI_CLOCK(rqst1->h_comp);

    t1 = rqst1->media_time - rqst1->offset;
    t2 = rqst2->media_time - rqst2->offset;
    if (t1 < t2) return (clock->mclock_xScale < 0)? -1 : 1;
    else if (t1 > t2) return (clock->mclock_xScale < 0)? 1 : -1;
    return 0;
}

static void broadcast_event (
    OMX_HANDLETYPE hComp,
    OMX_TIME_MEDIATIMETYPE * m_time)
{
    int ii;
    limoi_port_t * clock_port;
    OMX_BUFFERHEADERTYPE ** buf;
    OMX_PARAM_PORTDEFINITIONTYPE * port;
    OMX_STATETYPE state = limoi_component_state(hComp);

    if (state != OMX_StateExecuting && state != OMX_StatePause) return;

    for (ii = 0; ii < MAX_CLOCK_PORTS; ii++)
    {
        clock_port = limoi_component_get_port(hComp, ii);
        port = limoi_component_get_port_definition(clock_port);
        if (OMX_TRUE != port->bEnabled) continue;

        buf = limoi_component_get_port_buffer(hComp, clock_port, -1);
        if (NULL == buf)
        {
            LIMWARN("%s unable to get port %d buffer.\n", COMPONENT_NAME, ii);
        } else
        {
            memcpy((*buf)->pBuffer, m_time, sizeof(OMX_TIME_MEDIATIMETYPE));
            (*buf)->nFilledLen = sizeof(OMX_TIME_MEDIATIMETYPE);

            limoi_component_unlock(hComp);
            if (limoi_component_put_port_buffer(hComp, clock_port, buf))
            {
                LIMWARN("%s unable to put back port %d buffer.\n",
                        COMPONENT_NAME, ii);
            }
            limoi_component_lock(hComp);
        }
    }
}

#if 0
// upon state transition, clock port buffer no longer available.
static void broadcast_state_update (OMX_HANDLETYPE hComp)
{
    clock_private_t * clock = LIMOI_CLOCK(hComp);
    OMX_TIME_MEDIATIMETYPE m_time;
    ltime_val_t tv;
    OMX_TICKS wclock_now;
    OMX_TICKS mclock_now;

    ltime_gettimeofday(&tv);
    wclock_now = ltime_tv2us(&tv);
    mclock_now = clock->rclock_base
        + ((clock->mclock_xScale *
            (wclock_now - clock->wclock_base)) >> 16);

    //notify all client components.
    m_time.eUpdateType = OMX_TIME_UpdateClockStateChanged;
    m_time.pClientPrivate = NULL;
    m_time.nMediaTimestamp = mclock_now;
    m_time.nOffset = 0;
    m_time.nWallTimeAtMediaTime = wclock_now;
    m_time.xScale = clock->mclock_xScale;
    m_time.eState = clock->mclock_eState;

    broadcast_event(hComp, &m_time);
}
#endif

static OMX_ERRORTYPE main_handle (
    OMX_HANDLETYPE hComp,
    limoi_port_t ** audio_port,
    limoi_port_t ** video_port,
    limoi_port_t ** other_port,
    int * preloaded)
{
    clock_private_t * clock = LIMOI_CLOCK(hComp);
    media_time_request_t * next_rqst;

    limoi_component_lock(hComp);
    if (*preloaded == 0)
    {
        //do not requrie preload.
        *preloaded = LIMOI_PRELOAD_BUFFER_COUNT;
        goto bail;
    }

    //check to wait for client start time.
    if (clock->mclock_nWaitMask)
    {
        LIMDBG("%s waiting for client start time: %x\n", COMPONENT_NAME,
               clock->mclock_nWaitMask);
        limoi_component_cond_timedwait(hComp, 100);
        goto bail;
    }

    //check to fulfill media time request.
    while (1)
    {
        ltime_val_t tv;
        OMX_TICKS w_now;
        OMX_TICKS m_now;
        OMX_TICKS m_rqst;
        int wait_ms;
        limoi_port_t * clock_port;
        OMX_BUFFERHEADERTYPE ** buf;

        next_rqst = limq_dequeue_get(clock->media_time_rqst_q);
        if (NULL == next_rqst)
        {
            limoi_component_cond_timedwait(hComp, 100);
            goto bail;
        }

        LIMDBG("%s checking to fulfill media time request.\n", COMPONENT_NAME);
        wait_ms = 0;

        ltime_gettimeofday(&tv);
        w_now = ltime_tv2us(&tv);

        m_now = clock->rclock_base + ((clock->mclock_xScale *
                                       (w_now - clock->wclock_base)) >> 16);

        m_rqst = next_rqst->media_time
            - ((next_rqst->offset * clock->mclock_xScale) >> 16);

        if (clock->mclock_xScale > 0)
        {
            if (m_now < m_rqst)
                wait_ms = (((m_rqst - m_now) << 16) / clock->mclock_xScale)
                    / 1000;
        } else if (clock->mclock_xScale < 0)
        {
            if (m_now > m_rqst)
                wait_ms = (((m_rqst - m_now) << 16) / clock->mclock_xScale)
                    / 1000;
        } else wait_ms = 100;
        LIMDBG("%s scheduler wait status:\n"
               "\treference  : %"PRId64"\n"
               "\tm_now      : %"PRId64"\n"
               "\tm_rqst     : %"PRId64"\n"
               "\twait_ms    : %d\n", COMPONENT_NAME,
               clock->rclock_base, m_now, m_rqst, wait_ms);

        if (wait_ms)
        {
            limq_dequeue_release(clock->media_time_rqst_q, next_rqst);
            limoi_component_cond_timedwait(hComp, wait_ms);
            goto bail;
        }

        clock_port = *(other_port + next_rqst->port_index);
        buf = limoi_component_get_port_buffer(hComp, clock_port, -1);
        if (NULL == buf)
        {
            LIMWARN("%s unable to get port %u buffer.\n",
                    COMPONENT_NAME, next_rqst->port_index);
            limq_dequeue_release(clock->media_time_rqst_q, next_rqst);
            goto bail;
        } else
        {
            OMX_TIME_MEDIATIMETYPE * m_time;
            OMX_TICKS w_media;

            w_media = ((next_rqst->media_time - clock->rclock_base) << 16)
                / clock->mclock_xScale + clock->wclock_base;

            m_time = (OMX_TIME_MEDIATIMETYPE *)((*buf)->pBuffer);
            m_time->eUpdateType = OMX_TIME_UpdateRequestFulfillment;
            m_time->pClientPrivate = next_rqst->client_priv;
            m_time->nMediaTimestamp = next_rqst->media_time;
            m_time->nOffset = w_media - w_now;
            m_time->nWallTimeAtMediaTime = w_media;
            m_time->xScale = clock->mclock_xScale;
            m_time->eState = clock->mclock_eState;
            limq_dequeue_put(clock->media_time_rqst_q, next_rqst);

            LIMDBG("%s fulfilled port %u media time request:\n"
                   "\tw_media              : %lld\n"
                   "\tw_now                : %lld\n"
                   "\tnMediaTimeStamp      : %lld\n"
                   "\tnOffset              : %lld\n"
                   "\tnWallTimeAtMediaTime : %lld\n",
                   COMPONENT_NAME, next_rqst->port_index,
                   w_media, w_now,
                   m_time->nMediaTimestamp, m_time->nOffset,
                   m_time->nWallTimeAtMediaTime);

            (*buf)->nFilledLen = sizeof(OMX_TIME_MEDIATIMETYPE);
            limoi_component_unlock(hComp);
            if (limoi_component_put_port_buffer(hComp, clock_port, buf))
            {
                LIMDBG("%s unable to put back port %u buffer.\n",
                       COMPONENT_NAME, next_rqst->port_index);
                //not an error, tunnel port may be in the wrong state.
            }
            limoi_component_lock(hComp);
        }
    }

 bail:
    limoi_component_unlock(hComp);
    return OMX_ErrorNone;
}

static OMX_ERRORTYPE set_state_prep (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    OMX_HANDLETYPE hComp = cmd->hComponent;
    OMX_STATETYPE next_state = cmd->nParam1;
    OMX_STATETYPE state = limoi_component_state(hComp);
    clock_private_t * clock = LIMOI_CLOCK(hComp);

    LIMDBG("%s callback prep set state from %s to %s.\n", COMPONENT_NAME,
           limoi_omx_state_string(state), limoi_omx_state_string(next_state));

    switch (state)
    {
    case OMX_StateIdle :
        if (OMX_StateExecuting == next_state)
        {
        } else if (OMX_StatePause == next_state)
        {
        } else if (OMX_StateLoaded == next_state)
        {
            if (clock->media_time_rqst_q)
            {
                limq_destroy(clock->media_time_rqst_q);
                clock->media_time_rqst_q = NULL;
            }

            clock->mclock_eState = OMX_TIME_ClockStateStopped;
        } else
        {
        }
        break;

    case OMX_StateExecuting :
        if (OMX_StatePause == next_state)
        {
        } else if (OMX_StateIdle == next_state)
        {
            //invalidate all pending media time request
            limq_reset(clock->media_time_rqst_q);
            //clear client wait flags
            clock->mclock_nWaitMask = 0;
            limoi_component_cond_broadcast(hComp);

            clock->mclock_eState = OMX_TIME_ClockStateStopped;
            //broadcast_state_update(hComp);
        } else
        {
        }
        break;

    case OMX_StatePause :
        if (OMX_StateExecuting == next_state)
        {
        } else if (OMX_StateIdle == next_state)
        {
            //invalidate all pending media time request
            limq_reset(clock->media_time_rqst_q);
            //clear client wait flags
            clock->mclock_nWaitMask = 0;
            limoi_component_cond_broadcast(hComp);

            clock->mclock_eState = OMX_TIME_ClockStateStopped;
            //broadcast_state_update(hComp);
        } else
        {
        }
        break;

    default :
        break;
    }

    return ret;
}

static OMX_ERRORTYPE set_state_post (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    OMX_HANDLETYPE hComp = cmd->hComponent;
    OMX_STATETYPE next_state = cmd->nParam1;
    OMX_STATETYPE state = limoi_component_state(hComp);
    clock_private_t * clock = LIMOI_CLOCK(hComp);

    LIMDBG("%s callback post set state from %s to %s.\n", COMPONENT_NAME,
           limoi_omx_state_string(state), limoi_omx_state_string(next_state));

    switch (state)
    {
    case OMX_StateLoaded :

        // transition to StateIdle
        clock->media_time_rqst_q = limq_create_ext
            (sizeof(media_time_request_t), MAX_MEDIA_TIME_REQUESTS,
             "clock media time request", media_time_rqst_compare);
        if (clock->media_time_rqst_q == NULL)
        {
            LIMERR("%s unable to create request queue.\n", COMPONENT_NAME);
        }

        // reference clock base hasn't arrived yet.
        clock->rclock_base_set = 0;
        break;

    default :
        break;
    }

    return ret;
}

static OMX_ERRORTYPE prep_cmd_handle (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    OMX_HANDLETYPE hComp = cmd->hComponent;
    clock_private_t * clock = LIMOI_CLOCK(hComp);

    LIMDBG("%s prep command handler.\n", COMPONENT_NAME);

    limoi_component_lock(hComp);
    switch (cmd->Cmd)
    {
    case OMX_CommandStateSet :
        ret = set_state_prep(cmd);
        break;

    case OMX_CommandPortDisable :
        clock->mclock_nWaitMask &= ~(1 << cmd->nParam1);
        limoi_component_cond_broadcast(hComp);
        break;

    default :
        break;
    }

    limoi_component_unlock(hComp);
    return ret;
}

static OMX_ERRORTYPE post_cmd_handle (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    OMX_HANDLETYPE hComp = cmd->hComponent;

    LIMDBG("%s post command handler.\n", COMPONENT_NAME);

    limoi_component_lock(hComp);
    switch (cmd->Cmd)
    {
    case OMX_CommandStateSet :
        ret = set_state_post(cmd);
        break;

    default :
        break;
    }

    limoi_component_unlock(hComp);
    return ret;
}

static int add_port (OMX_HANDLETYPE hComp)
{
    OMX_PARAM_PORTDEFINITIONTYPE port;
    OMX_PORT_PARAM_TYPE port_param;
    int ii;

    limoi_set_spec_version(&port_param, sizeof(OMX_PORT_PARAM_TYPE));
    port_param.nPorts = MAX_CLOCK_PORTS;
    port_param.nStartPortNumber = 0; //OPB = 0
    limoi_component_set_parameter(hComp, OMX_IndexParamOtherInit, &port_param);

    limoi_set_spec_version(&port, sizeof(OMX_PARAM_PORTDEFINITIONTYPE));

    port.eDir = OMX_DirOutput;
    port.nBufferCountActual = 8;
    port.nBufferCountMin = 8;
    port.nBufferSize = sizeof(OMX_TIME_MEDIATIMETYPE);
    port.bEnabled = OMX_TRUE;
    port.bPopulated = OMX_FALSE;
    port.eDomain = OMX_PortDomainOther;
    port.bBuffersContiguous = OMX_FALSE;
    port.nBufferAlignment = 0;
    port.format.other.eFormat = OMX_OTHER_FormatTime;

    for (ii = 0; ii < MAX_CLOCK_PORTS; ii++)
    {
        port.nPortIndex = ii;
        limoi_component_add_port(hComp, &port);
    }

    return 0;
}

static OMX_ERRORTYPE GetConfig (
    OMX_HANDLETYPE hComponent,
    OMX_INDEXTYPE nIndex,
    OMX_PTR pCompConfigStruct)
{
    clock_private_t * clock = LIMOI_CLOCK(hComponent);

    limoi_component_lock(hComponent);
    switch (nIndex)
    {
    case OMX_IndexConfigTimeCurrentMediaTime :
        {
            OMX_TIME_CONFIG_TIMESTAMPTYPE * ts;
            ltime_val_t tv;
            OMX_TICKS wclock_now;

            ts = (OMX_TIME_CONFIG_TIMESTAMPTYPE *)pCompConfigStruct;
            //if (ts->nPortIndex != OMX_ALL)
            //{
            //    LIMWARN("invalid port index: %u, ignored.\n", ts->nPortIndex);
            //}

            ltime_gettimeofday(&tv);
            wclock_now = ltime_tv2us(&tv);
            ts->nTimestamp = clock->rclock_base
                + ((clock->mclock_xScale
                    * (wclock_now - clock->wclock_base)) >> 16);
            LIMDBG("current media time\n"
                   "\tts          : %"PRId64"\n"
                   "\trclock_base : %"PRId64"\n"
                   "\twclock_base : %"PRId64"\n"
                   "\twclock      : %"PRId64"\n",
                   ts->nTimestamp, clock->rclock_base,
                   clock->wclock_base, wclock_now);
        }
        break;

    case OMX_IndexConfigTimeScale :
        {
            OMX_TIME_CONFIG_SCALETYPE * scale;
            scale = (OMX_TIME_CONFIG_SCALETYPE *)pCompConfigStruct;
            scale->xScale = clock->mclock_xScale;
        }
        break;

    case OMX_IndexConfigTimeClockState :
        {
            OMX_TIME_CONFIG_CLOCKSTATETYPE * st;
            st = (OMX_TIME_CONFIG_CLOCKSTATETYPE *)pCompConfigStruct;

            st->eState = clock->mclock_eState;
            st->nStartTime = clock->mclock_nStartTime;
            st->nOffset = clock->mclock_nOffset;
            st->nWaitMask = clock->mclock_nWaitMask;
        }
        break;

    case OMX_IndexConfigTimeCurrentWallTime :
        {
            OMX_TIME_CONFIG_TIMESTAMPTYPE * ts;
            ltime_val_t tv;

            ts = (OMX_TIME_CONFIG_TIMESTAMPTYPE *)pCompConfigStruct;
            ltime_gettimeofday(&tv);
            ts->nTimestamp = ltime_tv2us(&tv);
            LIMDBG("%s current wall time: %lld\n",
                   COMPONENT_NAME, ts->nTimestamp);
        }
        break;

    case OMX_IndexConfigTimeActiveRefClock :
        {
            OMX_TIME_CONFIG_ACTIVEREFCLOCKTYPE * ref;
            ref = (OMX_TIME_CONFIG_ACTIVEREFCLOCKTYPE *)pCompConfigStruct;
            ref->eClock = clock->rclock_eClock;
        }
        break;

    default :
        limoi_component_get_parameter(hComponent, nIndex, pCompConfigStruct);
        break;
    }
    limoi_component_unlock(hComponent);
    return OMX_ErrorNone;
}

static OMX_ERRORTYPE SetConfig (
    OMX_HANDLETYPE hComponent,
    OMX_INDEXTYPE nIndex,
    OMX_PTR pCompConfigStruct)
{
    clock_private_t * clock = LIMOI_CLOCK(hComponent);
    OMX_ERRORTYPE ret = OMX_ErrorNone;

    limoi_component_lock(hComponent);
    switch (nIndex)
    {
    case OMX_IndexConfigTimeScale :
        {
            OMX_TIME_CONFIG_SCALETYPE * scale;
            ltime_val_t tv;
            OMX_TICKS wclock_now;
            OMX_TIME_MEDIATIMETYPE m_time;

            //rebase the time first.
            ltime_gettimeofday(&tv);
            wclock_now = ltime_tv2us(&tv);

            //Rbase = Mnow
            clock->rclock_base = clock->rclock_base
                + ((clock->mclock_xScale *
                   (wclock_now - clock->wclock_base)) >> 16);

            //Wbase = Wnow
            clock->wclock_base = wclock_now;

            //record scale change.
            scale = (OMX_TIME_CONFIG_SCALETYPE *)pCompConfigStruct;
            clock->mclock_xScale = scale->xScale;

            //invalidate all pending media time request
            limq_reset(clock->media_time_rqst_q);

            //notify all client components.
            m_time.eUpdateType = OMX_TIME_UpdateScaleChanged;
            m_time.pClientPrivate = NULL;
            m_time.nMediaTimestamp = clock->rclock_base;
            m_time.nOffset = 0;
            m_time.nWallTimeAtMediaTime = clock->wclock_base;
            m_time.xScale = clock->mclock_xScale;
            m_time.eState = clock->mclock_eState;

            broadcast_event(hComponent, &m_time);
        }
        break;

    case OMX_IndexConfigTimeClientStartTime :
        {
            OMX_TIME_CONFIG_TIMESTAMPTYPE * ts;
            ts = (OMX_TIME_CONFIG_TIMESTAMPTYPE *)pCompConfigStruct;
            if (ts->nPortIndex >= MAX_CLOCK_PORTS)
            {
                ret = OMX_ErrorBadPortIndex;
            } else
            {
                OMX_U32 prev_mask = clock->mclock_nWaitMask;

                clock->mclock_nWaitMask &= ~(1 << ts->nPortIndex);
                if (clock->mclock_nStartTime > ts->nTimestamp)
                {
                    clock->mclock_nStartTime = ts->nTimestamp;
                }
                limoi_component_cond_broadcast(hComponent);

                if (prev_mask && !clock->mclock_nWaitMask)
                {
                    ltime_val_t tv;
                    OMX_TICKS wclock_now;
                    OMX_TICKS mclock_now;
                    OMX_TIME_MEDIATIMETYPE m_time;

                    clock->mclock_eState = OMX_TIME_ClockStateRunning;

                    ltime_gettimeofday(&tv);
                    wclock_now = ltime_tv2us(&tv);
                    mclock_now = clock->rclock_base
                        + ((clock->mclock_xScale *
                           (wclock_now - clock->wclock_base)) >> 16);

                    //notify all client components.
                    m_time.eUpdateType = OMX_TIME_UpdateClockStateChanged;
                    m_time.pClientPrivate = NULL;
                    m_time.nMediaTimestamp = mclock_now;
                    m_time.nOffset = 0;
                    m_time.nWallTimeAtMediaTime = wclock_now;
                    m_time.xScale = clock->mclock_xScale;
                    m_time.eState = clock->mclock_eState;

                    broadcast_event(hComponent, &m_time);
                }
            }
        }
        break;

    case OMX_IndexConfigTimeClockState :
        {
            ltime_val_t tv;
            OMX_TICKS wclock_now;
            OMX_TICKS mclock_now;
            OMX_TIME_MEDIATIMETYPE m_time;
            OMX_TIME_CONFIG_CLOCKSTATETYPE * st;
            st = (OMX_TIME_CONFIG_CLOCKSTATETYPE *)pCompConfigStruct;

            ltime_gettimeofday(&tv);
            wclock_now = ltime_tv2us(&tv);
            mclock_now = clock->rclock_base
                + ((clock->mclock_xScale *
                   (wclock_now - clock->wclock_base)) >> 16);

            if (st->eState == OMX_TIME_ClockStateStopped)
            {
                clock->mclock_eState = OMX_TIME_ClockStateStopped;
                limq_reset(clock->media_time_rqst_q);
                limoi_component_cond_broadcast(hComponent);
            } else if (st->eState == OMX_TIME_ClockStateRunning)
            {
                clock->mclock_eState = st->eState;
                clock->mclock_nStartTime = st->nStartTime;
                clock->mclock_nOffset = st->nOffset;
                clock->mclock_nWaitMask = 0;
                limoi_component_cond_broadcast(hComponent);
            } else if (clock->mclock_eState != OMX_TIME_ClockStateStopped)
            {
                LIMWARN("%s invalid clock state transition from %d to %d\n",
                        COMPONENT_NAME, clock->mclock_eState, st->eState);
                ret = OMX_ErrorBadParameter;
                break;
            } else
            {
                ASSERT(st->eState == OMX_TIME_ClockStateWaitingForStartTime);

                clock->mclock_eState = st->eState;
                clock->mclock_nStartTime = st->nStartTime;
                clock->mclock_nOffset = st->nOffset;
                clock->mclock_nWaitMask = st->nWaitMask;
                limoi_component_cond_broadcast(hComponent);
            }

            //notify all client components.
            m_time.eUpdateType = OMX_TIME_UpdateClockStateChanged;
            m_time.pClientPrivate = NULL;
            m_time.nMediaTimestamp = mclock_now;
            m_time.nOffset = 0;
            m_time.nWallTimeAtMediaTime = wclock_now;
            m_time.xScale = clock->mclock_xScale;
            m_time.eState = clock->mclock_eState;

            broadcast_event(hComponent, &m_time);
        }
        break;

    case OMX_IndexConfigTimeActiveRefClock :
        {
            OMX_TIME_CONFIG_ACTIVEREFCLOCKTYPE * ref;
            ref = (OMX_TIME_CONFIG_ACTIVEREFCLOCKTYPE *)pCompConfigStruct;
            clock->rclock_eClock = ref->eClock;
        }
        break;

    case OMX_IndexConfigTimeCurrentAudioReference :
        LIMDBG("%s reference clock update from audio.\n", COMPONENT_NAME);
        if (clock->rclock_eClock == OMX_TIME_RefClockAudio)
        {
            OMX_TIME_CONFIG_TIMESTAMPTYPE * ts;
            ts = (OMX_TIME_CONFIG_TIMESTAMPTYPE *)pCompConfigStruct;

            if (ts->nPortIndex >= MAX_CLOCK_PORTS)
            {
                LIMWARN("%s bad port index: %u.\n",
                        COMPONENT_NAME, ts->nPortIndex);
                ret = OMX_ErrorBadPortIndex;
            } else
            {
                ltime_val_t tv;
                //static OMX_TICKS prv_wbase;
                //static OMX_TICKS prv_ts;

                ltime_gettimeofday(&tv);
                clock->wclock_base = ltime_tv2us(&tv);
                clock->wclock_base += clock->mclock_nOffset;
                clock->rclock_base = ts->nTimestamp;
                clock->rclock_base_set = 1;

                limoi_component_cond_broadcast(hComponent);
                LIMDBG("%s reference clock updated by audio as: %"PRId64"\n",
                       COMPONENT_NAME, ts->nTimestamp);
                /*
                if (ts->nTimestamp - prv_ts < clock->wclock_base - prv_wbase)
                {
                    LIMWARN("%s wallclock goes too fast?\n"
                            "\tnOffset        : %"PRId64"\n"
                            "\tcurrent wbase  : %"PRId64"\n"
                            "\tcurrent time   : %"PRId64"\n"
                            "\tprev wbase     : %"PRId64"\n"
                            "\tprev time      : %"PRId64"\n"
                            "\twbase diff     : %"PRId64"\n"
                            "\ttime diff      : %"PRId64"\n",
                            COMPONENT_NAME,
                            clock->mclock_nOffset,
                            clock->wclock_base, ts->nTimestamp,
                            prv_wbase, prv_ts,
                            clock->wclock_base - prv_wbase,
                            ts->nTimestamp - prv_ts);
                }
                prv_wbase = clock->wclock_base;
                prv_ts = ts->nTimestamp;
                */
            }
        }
        break;

    case OMX_IndexConfigTimeCurrentVideoReference :
        LIMDBG("%s reference clock update from video.\n", COMPONENT_NAME);
        if (clock->rclock_eClock == OMX_TIME_RefClockVideo)
        {
            OMX_TIME_CONFIG_TIMESTAMPTYPE * ts;
            ts = (OMX_TIME_CONFIG_TIMESTAMPTYPE *)pCompConfigStruct;

            if (ts->nPortIndex >= MAX_CLOCK_PORTS)
            {
                LIMWARN("%s bad port index: %u.\n",
                        COMPONENT_NAME, ts->nPortIndex);
                ret = OMX_ErrorBadPortIndex;
            } else
            {
                ltime_val_t tv;

                ltime_gettimeofday(&tv);
                clock->wclock_base = ltime_tv2us(&tv);
                clock->wclock_base += clock->mclock_nOffset;
                clock->rclock_base = ts->nTimestamp;
                clock->rclock_base_set = 1;

                limoi_component_cond_broadcast(hComponent);
                LIMDBG("%s reference clock updated by video as: %"PRId64"\n",
                       COMPONENT_NAME, ts->nTimestamp);
            }
        }
        break;

    case OMX_IndexConfigTimeMediaTimeRequest :
        {
            OMX_TIME_CONFIG_MEDIATIMEREQUESTTYPE * rqst;
            media_time_request_t * mrqst;

            if (clock->mclock_xScale == 0)
            {
                LIMWARN("%s current clock scale is zero, never fulfill.\n",
                        COMPONENT_NAME);
                ret = OMX_ErrorBadParameter;
                break;
            }

            LIMDBG("%s number of requests in queue: %d\n", COMPONENT_NAME,
                   limq_entries(clock->media_time_rqst_q));

            rqst = (OMX_TIME_CONFIG_MEDIATIMEREQUESTTYPE *)pCompConfigStruct;

            //validate request, if reference base has been set.
            if (clock->rclock_base_set)
            {
                OMX_TICKS w_now;
                OMX_TICKS m_now;
                ltime_val_t tv;

                ltime_gettimeofday(&tv);
                w_now = ltime_tv2us(&tv);
                m_now = clock->rclock_base
                    + ((clock->mclock_xScale *
                        (w_now - clock->wclock_base)) >> 16);
                if ((rqst->nMediaTimestamp - m_now < 0
                     && clock->mclock_xScale > 0)
                    || (rqst->nMediaTimestamp - m_now > 0
                        && clock->mclock_xScale < 0))
                {
                    LIMINFO("%s media time request to the past, ignored:\n"
                            "\tdistance       : %"PRId64"\n"
                            "\trequested time : %"PRId64"\n"
                            "\tcurrent time   : %"PRId64"\n"
                            "\tclock scale    : %d\n", COMPONENT_NAME,
                            rqst->nMediaTimestamp - m_now,
                            rqst->nMediaTimestamp, m_now,
                            clock->mclock_xScale);
                    ret = OMX_ErrorBadParameter;
                    break;
                }
            }

            mrqst = limq_enqueue_get(clock->media_time_rqst_q);
            if (NULL == mrqst)
            {
                LIMWARN("%s media time request queue is full.\n",
                        COMPONENT_NAME);
                ret = OMX_ErrorUndefined;
                break;
            }

            // is reference base set?
            if (!clock->rclock_base_set)
            {
                // NO, reset so the first request won't get stuck.
                // with this, the very first request will be fulfilled
                // in up to 100ms. In other words, if reference clock base
                // does not arrive in 100ms, reference clock base will be
                // reset to the first requested media time.
                ltime_val_t tv;

                ltime_gettimeofday(&tv);
                clock->wclock_base = ltime_tv2us(&tv);
                clock->wclock_base += clock->mclock_nOffset;
                clock->rclock_base = rqst->nMediaTimestamp - 100000;
                //LIMWARN("%s: reference clock reset upon first request.\n",
                //        COMPONENT_NAME);
                clock->rclock_base_set = 1;
            }

            mrqst->h_comp = hComponent;
            mrqst->port_index = rqst->nPortIndex;
            mrqst->client_priv = rqst->pClientPrivate;
            mrqst->media_time = rqst->nMediaTimestamp;
            mrqst->offset = rqst->nOffset;
            limq_enqueue_put(clock->media_time_rqst_q, mrqst);
            limoi_component_cond_broadcast(hComponent);
        }
        break;

    default :
        limoi_component_set_parameter(hComponent, nIndex, pCompConfigStruct);
        break;
    }

    limoi_component_unlock(hComponent);
    return ret;
}

static OMX_ERRORTYPE ComponentDeInit (OMX_HANDLETYPE hComponent)
{
    return limoi_component_deinit(hComponent);
}

static OMX_ERRORTYPE ComponentInit (OMX_HANDLETYPE * hComp, const char * name)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    OMX_COMPONENTTYPE * hdl;
    limoi_component_callbacks_t callbacks;
    clock_private_t * clock;

    LIMDBG("init component %s, version %s.\n", COMPONENT_NAME, VERSION);

    ret = limoi_component_init(hComp, sizeof(clock_private_t));
    if (OMX_ErrorNone != ret) goto bail;
    hdl = (OMX_COMPONENTTYPE *)(*hComp);

    clock = LIMOI_CLOCK(*hComp);

    limoi_component_set_name_version(*hComp, COMPONENT_NAME, VERSION);
    limoi_component_add_role(*hComp, ROLE_NAME);
    add_port(*hComp);

    callbacks.command_prep = prep_cmd_handle;
    callbacks.command_post = post_cmd_handle;
    callbacks.main = main_handle;
    callbacks.sub = NULL;
    limoi_component_set_callbacks(*hComp, &callbacks);

    LIMDBG("component %s initialized.\n", COMPONENT_NAME);
    hdl->ComponentDeInit = ComponentDeInit;
    hdl->GetConfig = GetConfig;
    hdl->SetConfig = SetConfig;

    //clock state initialization.
    clock->mclock_eState = OMX_TIME_ClockStateStopped;

    // set default wall clock base and media clock scale.
    {
        ltime_val_t tv;
        ltime_gettimeofday(&tv);
        clock->wclock_base = ltime_tv2us(&tv);
        clock->mclock_xScale = 1 << 16;
    }

 bail:
    return ret;
}

lim_omx_component_registration_t ** OMX_ComponentRegistration (int * num)
{
    static char * roles[] = {ROLE_NAME, NULL};
    static lim_omx_component_registration_t comp_reg = {
        COMPONENT_NAME, roles, ComponentInit
    };
    static lim_omx_component_registration_t * regs[] = {&comp_reg};

    *num = 1;
    return regs;
}
