/*
  Copyright (C) 2010 - 2011 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX IL ffmpeg audio decoder component.

  REVISION:

    1) initial creation. ----------------------------------------- 2010-02-18 MG
*/

#include <stdio.h>
#include <stdlib.h>
#include "OMX_Core.h"
#include "OMX_Types.h"
#include "OMX_Component.h"
#include "limoi-ffmpeg-decode-audio.h"

//#undef LIMDBG
#ifndef LIMDBG
#define LIMDBG(fmt, arg...)
#endif

#define COMPONENT_NAME "OMX.limoi.ffmpeg.decode.audio"
static const char * roles[] = {
    "audio_decoder.aac",
    "audio_decoder.amr",
    "audio_decoder.mp3",
    "audio_decoder.ogg",
    "audio_decoder.ra",
    "audio_decoder.wma",
    "audio_decoder.adpcm",
    "audio_decoder.pcm",
    NULL
};

static void find_decoder (
    OMX_HANDLETYPE hComp,
    AVStream * stream)
{
    ffmpeg_adecoder_private_t * ffmpeg_decoder = LIMOI_FFMPEG_ADECODER(hComp);
    AVCodecContext * avcodec = (stream)? stream->codec : NULL;

    if (NULL == ffmpeg_decoder->codec)
    {
        if (avcodec)
        {
            //REMOVEME: disable 5.1 output and force stereo audio output
            if (avcodec->codec_type == AVMEDIA_TYPE_AUDIO)
            {
                if (avcodec->channels > 0)
                {
                    avcodec->request_channels = FFMIN(2, avcodec->channels);
                } else
                {
                    avcodec->request_channels = 2;
                }
            }

            ffmpeg_decoder->stream = stream;
            ffmpeg_decoder->codec = avcodec_find_decoder(avcodec->codec_id);
        } else
        {
            //TODO:
        }

        if (NULL == ffmpeg_decoder->codec)
        {
            LIMWARN("%s unable to find codec.\n", COMPONENT_NAME);
        }
    }

    if (ffmpeg_decoder->codec && NULL == ffmpeg_decoder->avcodec)
    {
        if (avcodec)
        {
            ffmpeg_decoder->avcodec = avcodec;
        } else
        {
            ffmpeg_decoder->avcodec = avcodec_alloc_context();
            ffmpeg_decoder->avcodec_allocated = 1;
        }

        if (avcodec_open(ffmpeg_decoder->avcodec, ffmpeg_decoder->codec) < 0)
        {
            LIMERR("%s unable to open codec.\n", COMPONENT_NAME);
            ffmpeg_decoder->avcodec = NULL;
        } else
        {
            av_init_packet(&ffmpeg_decoder->avpkt);
        }
    }
}

static OMX_ERRORTYPE set_state (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    OMX_HANDLETYPE hComp = cmd->hComponent;
    OMX_STATETYPE next_state = cmd->nParam1;
    OMX_STATETYPE state = limoi_component_state(hComp);
    ffmpeg_adecoder_private_t * ffmpeg_decoder = LIMOI_FFMPEG_ADECODER(hComp);

    switch (state)
    {
    case OMX_StateLoaded :

        // transition to StateIdle
        omx_ffmpeg_codec_init();

        ffmpeg_decoder->avcodec = NULL;
        ffmpeg_decoder->avcodec_allocated = 0;
        ffmpeg_decoder->codec = NULL;

        ffmpeg_decoder->inited = 0;
        ffmpeg_decoder->pcm_format = SAMPLE_FMT_S16;
        ffmpeg_decoder->reformat = NULL;
        break;

    case OMX_StateIdle :
        if (OMX_StateLoaded == next_state)
        {
            if (ffmpeg_decoder->avcodec)
            {
                avcodec_close(ffmpeg_decoder->avcodec);
                if (ffmpeg_decoder->avcodec_allocated)
                {
                    av_free(ffmpeg_decoder->avcodec);
                    ffmpeg_decoder->avcodec_allocated = 0;
                }
                ffmpeg_decoder->avcodec = NULL;

                if (ffmpeg_decoder->reformat)
                {
                    av_audio_convert_free(ffmpeg_decoder->reformat);
                    ffmpeg_decoder->reformat = NULL;
                }
            }
        }
        break;

    default :
        break;
    }
    return ret;
}

static OMX_ERRORTYPE prep_cmd_handle (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    OMX_HANDLETYPE hComp = cmd->hComponent;
    ffmpeg_adecoder_private_t * ffmpeg_decoder = LIMOI_FFMPEG_ADECODER(hComp);

    LIMDBG("%s prep command handler.\n", COMPONENT_NAME);
    limoi_component_lock(hComp);

    switch (cmd->Cmd)
    {
    case OMX_CommandFlush :
        if (ffmpeg_decoder->avcodec && cmd->nParam1 == 0)
        {
            //flush codec input buffers upon input port flushing.
            LIMDBG("flushing audio decoder input.\n");
            avcodec_flush_buffers(ffmpeg_decoder->avcodec);
        }
        break;

    default :
        break;
    }
    limoi_component_unlock(hComp);

    return ret;
}

static OMX_ERRORTYPE post_cmd_handle (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    OMX_HANDLETYPE hComp = cmd->hComponent;

    LIMDBG("%s post command handler.\n", COMPONENT_NAME);
    limoi_component_lock(hComp);

    switch (cmd->Cmd)
    {
    case OMX_CommandStateSet :
        ret = set_state(cmd);
        break;

    default :
        break;
    }
    limoi_component_unlock(hComp);

    return ret;
}

static int decode_handle (
    OMX_HANDLETYPE hComp,
    OMX_BUFFERHEADERTYPE * in, OMX_BUFFERHEADERTYPE * out)
{
    ffmpeg_adecoder_private_t * ffmpeg_decoder;
    int out_size, len;
    OMX_U32 out_offset;

    ffmpeg_decoder = LIMOI_FFMPEG_ADECODER(hComp);

    //LIMDBG("%s decode handle.\n", COMPONENT_NAME);
    limoi_component_lock(hComp);

    if (NULL == ffmpeg_decoder->codec || NULL == ffmpeg_decoder->avcodec)
    {
        find_decoder(hComp, in->pOutputPortPrivate);
    }
    if (NULL == ffmpeg_decoder->codec || NULL == ffmpeg_decoder->avcodec)
    {
        in->nFilledLen = 0;
        goto bail;
    }

    LIMDBG("%s input data length: %u, offset %u\n", COMPONENT_NAME,
           in->nFilledLen, in->nOffset);
    ffmpeg_decoder->avpkt.data = in->pBuffer + in->nOffset;
    ffmpeg_decoder->avpkt.size = in->nFilledLen;
    in->nFilledLen = 0;
    ASSERT(out->nAllocLen >= AVCODEC_MAX_AUDIO_FRAME_SIZE);
    out_offset = out->nOffset;

    while (ffmpeg_decoder->avpkt.size > 0)
    {
        out_size = AVCODEC_MAX_AUDIO_FRAME_SIZE;

        //align the output buffer to 16 bytes boundary.
        out->nOffset = out_offset + 16 - (unsigned long)
            (out->pBuffer + out_offset + out->nFilledLen) % 16;

        len = avcodec_decode_audio3
            (ffmpeg_decoder->avcodec,
             (int16_t *)(out->pBuffer + out->nOffset + out->nFilledLen),
             &out_size,
             &ffmpeg_decoder->avpkt);

        if (len < 0)
        {
            //LIMERR("%s error while decoding\n", COMPONENT_NAME);
            //reset output and skip this frame.
            out->nOffset = out_offset;
            goto bail;
        }

        ffmpeg_decoder->avpkt.size -= len;
        ffmpeg_decoder->avpkt.data += len;
        in->nOffset += len;
        if (out_size <= 0) continue;

        if (ffmpeg_decoder->avcodec->sample_fmt != ffmpeg_decoder->pcm_format)
        {
            LIMDBG("%s allocate pcm reformat context.\n", COMPONENT_NAME);
            if (ffmpeg_decoder->reformat)
            {
                av_audio_convert_free(ffmpeg_decoder->reformat);
            }
            ffmpeg_decoder->reformat = av_audio_convert_alloc
                (SAMPLE_FMT_S16, 1, ffmpeg_decoder->avcodec->sample_fmt,
                 1, NULL, 0);
            if (NULL == ffmpeg_decoder->reformat)
            {
                LIMERR("%s unable to convert sample format from %s to %s\n",
                       COMPONENT_NAME,
                       av_get_sample_fmt_name
                       (ffmpeg_decoder->avcodec->sample_fmt),
                       av_get_sample_fmt_name
                       (SAMPLE_FMT_S16));
                //TODO: slient audio
            }
            ffmpeg_decoder->pcm_format = ffmpeg_decoder->avcodec->sample_fmt;
        }

        if (ffmpeg_decoder->reformat)
        {
            DECLARE_ALIGNED(16, uint8_t, buf)[AVCODEC_MAX_AUDIO_FRAME_SIZE];
            const void * in_buf[6] = {buf};
            void *out_buf[6] = {out->pBuffer + out->nOffset + out->nFilledLen};
            int in_stride[6] = {av_get_bytes_per_sample(ffmpeg_decoder->avcodec->sample_fmt)};
            int out_stride[6] = {2};
            int samples = out_size / in_stride[0];

            memcpy(buf, out->pBuffer + out->nOffset + out->nFilledLen, out_size);
            if (av_audio_convert(ffmpeg_decoder->reformat, out_buf, out_stride,
                                 in_buf, in_stride, samples) < 0)
            {
                LIMERR("%s av_audio_convert() failed\n", COMPONENT_NAME);
                //TODO: silent.
            }

            out_size = samples * 2;
        }

        if (out_size > 0)
        {
            AVRational * time_base = &ffmpeg_decoder->stream->time_base;

            out->nFilledLen += out_size;
            out->nTimeStamp = ((in->nTimeStamp * 1000 * time_base->num)
                               / time_base->den) * 1000;

            LIMDBG("audio pts at decoder: %"PRId64"\n", out->nTimeStamp);

            in->nFilledLen = ffmpeg_decoder->avpkt.size;
            if (in->nFilledLen)
            {
                OMX_TICKS pcm_ts;

                pcm_ts= ((OMX_TICKS)out_size * 1000000)
                    / ((OMX_TICKS)ffmpeg_decoder->avcodec->sample_rate
                       * ffmpeg_decoder->avcodec->channels * 2);
                in->nTimeStamp += ((pcm_ts * time_base->den) / time_base->num)
                    / 1000000;
                LIMDBG("%s jumbo audio frame? pcm_ts %"PRId64" %d/%d\n",
                       COMPONENT_NAME, pcm_ts, time_base->num, time_base->den);
            }
            LIMDBG("out size: %d %d\n", out_size, in->nFilledLen);

            if (ffmpeg_decoder->inited == 0
                || (ffmpeg_decoder->pcm_channels
                    != ffmpeg_decoder->avcodec->channels
                    || ffmpeg_decoder->pcm_sample_rate
                    != ffmpeg_decoder->avcodec->sample_rate))
            {
                OMX_AUDIO_PARAM_PCMMODETYPE pcm_param;

                LIMDBG("%s channels: %d sample_rate: %d\n", COMPONENT_NAME,
                       ffmpeg_decoder->avcodec->channels,
                       ffmpeg_decoder->avcodec->sample_rate);

                ffmpeg_decoder->pcm_channels = ffmpeg_decoder->avcodec->channels;
                ffmpeg_decoder->pcm_sample_rate = ffmpeg_decoder->avcodec->sample_rate;

                pcm_param.nChannels = ffmpeg_decoder->avcodec->channels;
                pcm_param.eNumData = OMX_NumericalDataSigned;
                pcm_param.eEndian = OMX_EndianMax; //don't care == host type
                pcm_param.bInterleaved = OMX_TRUE;
                pcm_param.nBitPerSample = 16;
                pcm_param.nSamplingRate = ffmpeg_decoder->avcodec->sample_rate;

                //NOTE:
                //if detected pcm mode here mismatches with the default or
                //what client supplies, client may choose to disable output
                //port in order to properly initialize the tunnel down the
                //stream, in which case the decoded pcm may get permanently
                //lost, we choose to do nothing with in the decoder and let
                //client deal with it, by either,
                // - copy and save the pcm, or
                // - seek to the beginning of stream and restart using info
                //   gained in previous run
                limoi_audio_decoder_set_pcm_mode(hComp, &pcm_param);

                ffmpeg_decoder->inited = 1;
            }

            break;
        }
    }

 bail:
    limoi_component_unlock(hComp);
    return 0;
}

static OMX_ERRORTYPE ComponentDeInit (OMX_HANDLETYPE hComp)
{
    return limoi_audio_decoder_deinit(hComp);
}

static OMX_ERRORTYPE ComponentInit (OMX_HANDLETYPE * hComponent, const char * name)
{
    OMX_COMPONENTTYPE * hdl;
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    limoi_audio_decoder_callbacks_t callbacks;
    ffmpeg_adecoder_private_t * ffmpeg_decoder;
    const char ** role;

    LIMDBG("loading %s component.\n", COMPONENT_NAME);
    ret = limoi_audio_decoder_init(hComponent,
                                   sizeof(ffmpeg_adecoder_private_t),
                                   COMPONENT_NAME, VERSION,
                                   OMX_AUDIO_CodingAutoDetect);
    if (OMX_ErrorNone != ret) goto bail;
    hdl = (OMX_COMPONENTTYPE *)*hComponent;

    ffmpeg_decoder  = LIMOI_FFMPEG_ADECODER(*hComponent);

    role = &roles[0];
    while (*role) limoi_component_add_role(*hComponent, *role++);

    hdl->ComponentDeInit = ComponentDeInit;

    callbacks.command_prep = prep_cmd_handle;
    callbacks.command_post = post_cmd_handle;
    callbacks.decode = decode_handle;
    limoi_audio_decoder_set_callbacks(*hComponent, &callbacks);

    LIMDBG("%s component loaded.\n", COMPONENT_NAME);

 bail:
    return ret;
}

lim_omx_component_registration_t ** OMX_ComponentRegistration (int * num)
{
    static lim_omx_component_registration_t comp_reg = {
        COMPONENT_NAME, (char **)roles, ComponentInit
    };
    static lim_omx_component_registration_t * regs[] = {&comp_reg};

    *num = 1;
    return regs;
}
