/*
  Copyright (C) 2010 - 2011 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX IL ffmpeg video decoder component.

  REVISION:

    1) initial creation. ----------------------------------------- 2010-02-18 MG
*/

#include <stdio.h>
#include <stdlib.h>
#include "OMX_Core.h"
#include "OMX_Types.h"
#include "OMX_Component.h"
#include "limoi-ffmpeg-decode-video.h"

//#undef LIMDBG
#ifndef LIMDBG
#define LIMDBG(fmt, arg...)
#endif

#define COMPONENT_NAME "OMX.limoi.ffmpeg.decode.video"
static const char * roles[] = {
    "video_decoder.avc",
    "video_decoder.mpeg2",
    "video_decoder.mpeg4",
    "video_decoder.h263",
    "video_decoder.wmv",
    "video_decoder.rv",
    "video_decoder.mjpeg",
    NULL
};

static void find_decoder (
    OMX_HANDLETYPE hComp,
    AVStream * stream)
{
    ffmpeg_vdecoder_private_t * ffmpeg_decoder = LIMOI_FFMPEG_VDECODER(hComp);
    AVCodecContext * avcodec = (stream)? stream->codec : NULL;

    if (NULL == ffmpeg_decoder->codec)
    {
        if (avcodec)
        {
            ffmpeg_decoder->stream = stream;
            ffmpeg_decoder->codec = avcodec_find_decoder(avcodec->codec_id);
        } else
        {
            //TODO:
            LIMWARN("%s component requires this context.\n", COMPONENT_NAME);
        }

        if (NULL == ffmpeg_decoder->codec)
        {
            LIMWARN("%s unable to find codec.\n", COMPONENT_NAME);
        }
    }

    if (ffmpeg_decoder->codec && NULL == ffmpeg_decoder->avcodec)
    {
        if (avcodec)
        {
            ffmpeg_decoder->avcodec = avcodec;
        } else
        {
            ffmpeg_decoder->avcodec = avcodec_alloc_context();
            ffmpeg_decoder->avcodec_allocated = 1;
        }

        if (avcodec_open(ffmpeg_decoder->avcodec, ffmpeg_decoder->codec) < 0)
        {
            LIMERR("%s unable to open codec.\n", COMPONENT_NAME);
            ffmpeg_decoder->avcodec = NULL;
        } else
        {
            av_init_packet(&ffmpeg_decoder->avpkt);
            ffmpeg_decoder->avpkt.size = 0;
            ffmpeg_decoder->avpkt.data = NULL;
        }
    }
}

static OMX_ERRORTYPE set_state (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    OMX_HANDLETYPE hComp = cmd->hComponent;
    OMX_STATETYPE next_state = cmd->nParam1;
    OMX_STATETYPE state = limoi_component_state(hComp);
    ffmpeg_vdecoder_private_t * ffmpeg_decoder = LIMOI_FFMPEG_VDECODER(hComp);

    LIMDBG("%s from %s to %s.\n", COMPONENT_NAME,
           limoi_omx_state_string(state), limoi_omx_state_string(next_state));

    switch (state)
    {
    case OMX_StateLoaded :

        // transition to StateIdle
        omx_ffmpeg_codec_init();

        avcodec_get_frame_defaults(&ffmpeg_decoder->frame);
        ffmpeg_decoder->avcodec = NULL;
        ffmpeg_decoder->avcodec_allocated = 0;
        ffmpeg_decoder->codec = NULL;
        ffmpeg_decoder->inited = 0;
        ffmpeg_decoder->has_frame = 0;
        ffmpeg_decoder->sws = NULL;
        break;

    case OMX_StateIdle :
        if (OMX_StateLoaded == next_state)
        {
            if (ffmpeg_decoder->avcodec)
            {
                avcodec_close(ffmpeg_decoder->avcodec);
                if (ffmpeg_decoder->avcodec_allocated)
                {
                    av_free(ffmpeg_decoder->avcodec);
                    ffmpeg_decoder->avcodec_allocated = 0;
                }
                ffmpeg_decoder->avcodec = NULL;
            }
        }
        break;

    default :
        break;
    }

    return ret;
}

static OMX_ERRORTYPE prep_cmd_handle (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    OMX_HANDLETYPE hComp = cmd->hComponent;
    ffmpeg_vdecoder_private_t * ffmpeg_decoder = LIMOI_FFMPEG_VDECODER(hComp);

    LIMDBG("%s prep command handler.\n", COMPONENT_NAME);
    limoi_component_lock(hComp);

    switch (cmd->Cmd)
    {
    case OMX_CommandFlush :
        if (ffmpeg_decoder->avcodec && cmd->nParam1 == 0)
        {
            LIMDBG("flushing video decoder input.\n");
            //flush codec input buffers upon input port flushing.
            avcodec_flush_buffers(ffmpeg_decoder->avcodec);
            ffmpeg_decoder->pts = 0;
        }
        break;

    default :
        break;
    }
    limoi_component_unlock(hComp);

    return ret;
}

static OMX_ERRORTYPE post_cmd_handle (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    OMX_HANDLETYPE hComp = cmd->hComponent;

    LIMDBG("%s post command handler.\n", COMPONENT_NAME);
    limoi_component_lock(hComp);

    switch (cmd->Cmd)
    {
    case OMX_CommandStateSet :
        ret = set_state(cmd);
        break;

    default :
        break;
    }
    limoi_component_unlock(hComp);

    return ret;
}

static OMX_TICKS get_pts (
    ffmpeg_vdecoder_private_t * ffmpeg_decoder,
    int jumbo_frame)
{
    OMX_TICKS pts = AV_NOPTS_VALUE;
    AVRational * time_base;

    //if it is jumbo frame, go calculate the pts
    if (!jumbo_frame)
    {
        //honor decoder reordered pts if dts is not present
        time_base = &ffmpeg_decoder->stream->time_base;
        if (ffmpeg_decoder->avpkt.dts == AV_NOPTS_VALUE)
        {
            pts = ffmpeg_decoder->frame.reordered_opaque;
        } else
        {
            pts = ffmpeg_decoder->avpkt.dts;
        }

        if (pts == AV_NOPTS_VALUE)
        {
            /*
            //try picture pts as the last fallback
            pts = ffmpeg_decoder->frame.pts;
            if (pts != AV_NOPTS_VALUE)
            {
                LIMDBG("%s using picture pts.\n", COMPONENT_NAME);
                time_base = &ffmpeg_decoder->avcodec->time_base;
            }
            */
        }

        //if by here pts is not available, we'll calculate one
        if (pts != AV_NOPTS_VALUE)
        {
            ASSERT(time_base->den);
            pts = ((pts * 1000 * time_base->num) / time_base->den) * 1000;
        }
    }

    if (pts == AV_NOPTS_VALUE || pts < ffmpeg_decoder->pts)
    {
        time_base = &ffmpeg_decoder->avcodec->time_base;
        LIMINFO("%s pts not available or going backward, calculated it\n"
                "\tprevious pts : %"PRId64"\n"
                "\tcurrent pts  : %"PRId64"\n"
                "\ttime base    : %d/%d.\n"
                "\tjumbo_frame  : %s\n",
                COMPONENT_NAME, ffmpeg_decoder->pts, pts,
                time_base->num, time_base->den,
                jumbo_frame? "YES" : "NO");
        ASSERT(time_base->den);
        pts = ffmpeg_decoder->pts
            + (OMX_TICKS)1000000 * time_base->num / time_base->den;
    }

    //save latest pts for future reference
    ffmpeg_decoder->pts = pts;

    return pts;
}

//calculate output picture dimension, this effectively enforces an upper
//limit to what gets out of decoder:
//- max width  : 1920
//- max height : 1080
static void calculate_output_dimension (int * w, int * h)
{
    int w1, h1;

    ASSERT(*w && *h);

    //align output width on 4-pixel boundary so that we have
    //max compatibility with underlying scaler/renderer.
    *w = (*w + 3) & ~3;
    //for the same reason, make sure height is an even number.
    *h = *h & ~1;

    if (*w <= 1920 && *h <= 1080) return;

    w1 = *w;
    h1 = *h;
    if (*w > 1920)
    {
        w1 = *w;
        *w = 720;
        *h = (*h * 720) / w1;
    }
    if (*h && *h <= 1080) return;

    if (*h)
    {
        *h = 480;
        *w = (w1 * 480) / h1;
    }
    if (*w && *h && *w <= 1920 && *h <= 1080) return;

    LIMINFO("%s don't know how to resize %d x %d, default to 720 x 480.\n",
            COMPONENT_NAME, w1, h1);
    *w = 720;
    *h = 480;
}

static int output_picture (
    ffmpeg_vdecoder_private_t * ffmpeg_decoder,
    OMX_BUFFERHEADERTYPE * in,
    OMX_BUFFERHEADERTYPE * out)
{
    AVPicture pict;
    int w, h;
    int pitch0, pitch1;
    int ret = 0;

    w = ffmpeg_decoder->avcodec->width;
    h = ffmpeg_decoder->avcodec->height;
    calculate_output_dimension(&w, &h);
    ASSERT(w && h);

    //align pitch on 16-pixel boundary.
    pitch0 = (w + 15) & ~15;
    pitch1 = pitch0 / 2;

    if ((pitch0 + pitch1) * h >
        out->nAllocLen - out->nFilledLen - out->nOffset)
    {
        LIMINFO("%s default output buffer too small, ignored.\n"
                "\twidth       : %d\n"
                "\theight      : %d\n"
                "\tnAllocLen   : %u\n"
                "\tnFilledLen  : %u\n"
                "\tnOffset     : %u\n",
                COMPONENT_NAME,
                w, h,
                out->nAllocLen,
                out->nFilledLen,
                out->nOffset);
        goto bail;
    }

    //scale to 420
    ffmpeg_decoder->sws = sws_getCachedContext
        (ffmpeg_decoder->sws,
         ffmpeg_decoder->avcodec->width,
         ffmpeg_decoder->avcodec->height,
         ffmpeg_decoder->avcodec->pix_fmt,
         w, h, PIX_FMT_YUV420P,
         SWS_BICUBIC, NULL, NULL, NULL);

    if (NULL == ffmpeg_decoder->sws)
    {
        LIMERR("%s unable to get swscaler context.\n",
               COMPONENT_NAME);
        ret = -1;
        goto bail;
    }

    pict.data[0] = out->pBuffer + out->nFilledLen;
    pict.data[1] = pict.data[0] + pitch0 * h + out->nFilledLen;
    pict.data[2] = pict.data[1] + pitch1 * h / 2 + out->nFilledLen;

    pict.linesize[0] = pitch0;
    pict.linesize[1] = pitch1;
    pict.linesize[2] = pitch1;

    sws_scale(ffmpeg_decoder->sws,
              (const uint8_t * const *)ffmpeg_decoder->frame.data,
              ffmpeg_decoder->frame.linesize,
              0, ffmpeg_decoder->avcodec->height,
              pict.data, pict.linesize);
    LIMDBG("%s scaling parameters:\n"
           "\tfrom %4d x %4d\n"
           "\tto   %4d x %4d\n", COMPONENT_NAME,
           ffmpeg_decoder->frame.linesize[0],
           ffmpeg_decoder->avcodec->height,
           pitch0, h);

    out->nFilledLen += (pitch0 + pitch1) * h;
    LIMDBG("%s filled len: %u w x h = %d x %d\n", COMPONENT_NAME,
           out->nFilledLen, w, h);
    ffmpeg_decoder->has_frame = 0;

 bail:
    return ret;
}

static int decode_handle (
    OMX_HANDLETYPE hComp,
    OMX_BUFFERHEADERTYPE * in,
    OMX_BUFFERHEADERTYPE * out)
{
    ffmpeg_vdecoder_private_t * ffmpeg_decoder;
    int got_picture, len;
    int jumbo_frame = 0;

    ffmpeg_decoder = LIMOI_FFMPEG_VDECODER(hComp);

    limoi_component_lock(hComp);

    if (NULL == ffmpeg_decoder->codec || NULL == ffmpeg_decoder->avcodec)
    {
        find_decoder(hComp, in->pOutputPortPrivate);
    }
    if (NULL == ffmpeg_decoder->codec || NULL == ffmpeg_decoder->avcodec)
    {
        LIMWARN("%s codec not found yet.\n", COMPONENT_NAME);
        in->nFilledLen = 0;
        goto bail;
    }

    if (ffmpeg_decoder->has_frame)
    {
        LIMDBG("%s output cached picture.\n", COMPONENT_NAME);
        out->nTimeStamp = get_pts(ffmpeg_decoder, 0);
        output_picture(ffmpeg_decoder, in, out);
        goto bail;
    }

    if (!ffmpeg_decoder->avpkt.size)
    {
        if (!in->nFilledLen) goto bail;

        memcpy(&ffmpeg_decoder->avpkt, in->pBuffer + in->nFilledLen,
               sizeof(AVPacket));

        ffmpeg_decoder->avpkt.data = in->pBuffer;
        ffmpeg_decoder->avpkt.size = in->nFilledLen;
        //ffmpeg_decoder->avpkt.pts = in->nTimeStamp;
        in->nFilledLen = 0;
        LIMDBG("%s decoder input pts: %"PRId64"\n",
               COMPONENT_NAME, ffmpeg_decoder->avpkt.pts);
    } else
    {
        LIMDBG("%s continue decoding left-over data.\n", COMPONENT_NAME);
        jumbo_frame = 1;
    }

    while (ffmpeg_decoder->avpkt.size > 0)
    {
        LIMDBG("%s pkt->size : %d %p\n", COMPONENT_NAME,
               ffmpeg_decoder->avpkt.size,
               ffmpeg_decoder->avpkt.data);
        if ((unsigned long)ffmpeg_decoder->avpkt.data & 3)
        {
            unsigned long addr = (unsigned long)ffmpeg_decoder->avpkt.data;
            LIMINFO("%s packet data buffer %p is not aligned.\n",
                    COMPONENT_NAME, ffmpeg_decoder->avpkt.data);
            addr = (addr + 3) & ~3;
            ffmpeg_decoder->avpkt.data = (void *)addr;
        }

        ffmpeg_decoder->avcodec->reordered_opaque = (jumbo_frame)?
            AV_NOPTS_VALUE : ffmpeg_decoder->avpkt.pts;
        len = avcodec_decode_video2(ffmpeg_decoder->avcodec,
                                    &ffmpeg_decoder->frame, &got_picture,
                                    &ffmpeg_decoder->avpkt);

        if (len < 0)
        {
            LIMINFO("%s error while decoding\n", COMPONENT_NAME);
            ffmpeg_decoder->avpkt.size = 0;
            goto bail;
        }

        if (got_picture)
        {
            if (ffmpeg_decoder->inited == 0)
            {
                int w, h;

                w = ffmpeg_decoder->avcodec->width;
                h = ffmpeg_decoder->avcodec->height;
                calculate_output_dimension(&w, &h);
                ASSERT(w && h);

                LIMDBG("aspect ratio info: %d %d %d %d\n", w, h,
                       ffmpeg_decoder->avcodec->sample_aspect_ratio.num,
                       ffmpeg_decoder->avcodec->sample_aspect_ratio.den);

                limoi_video_decoder_set_pixel_format
                    (hComp, w, h, OMX_COLOR_FormatYUV420Planar,
                     ffmpeg_decoder->avcodec->sample_aspect_ratio.num,
                     ffmpeg_decoder->avcodec->sample_aspect_ratio.den);
                ffmpeg_decoder->inited = 1;
                ffmpeg_decoder->pts = 0;

                //always delay outputing the very first frame, since component
                //down the stream may require the demension info to actually
                //allocate the buffers.
                ffmpeg_decoder->has_frame = 1;
            } else
            {
                //always fetch pts first before trying to output picture
                //so that pts history is preserved in case things go wrong
                //with particular picture output.
                out->nTimeStamp = get_pts(ffmpeg_decoder, jumbo_frame);
                LIMDBG("%s video pts at decoder: %"PRId64"\n",
                       COMPONENT_NAME, out->nTimeStamp);

                if (output_picture(ffmpeg_decoder, in, out))
                {
                    ffmpeg_decoder->avpkt.size = 0;
                    goto bail;
                }
            }
        }

        ffmpeg_decoder->avpkt.size -= len;
        ffmpeg_decoder->avpkt.data += len;

        if (ffmpeg_decoder->avpkt.size > 0)
        {
            LIMINFO("%s multiple picture in one packet?\n", COMPONENT_NAME);
        }

        if (got_picture) break;
    }

 bail:
    limoi_component_unlock(hComp);
    return 0;
}


static OMX_ERRORTYPE ComponentDeInit (OMX_HANDLETYPE hComp)
{
    return limoi_video_decoder_deinit(hComp);
}

static OMX_ERRORTYPE ComponentInit (OMX_HANDLETYPE * hComponent, const char * name)
{
    OMX_COMPONENTTYPE * hdl;
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    limoi_video_decoder_callbacks_t callbacks;
    ffmpeg_vdecoder_private_t * ffmpeg_decoder;
    const char ** role;

    LIMDBG("loading %s component.\n", COMPONENT_NAME);
    ret = limoi_video_decoder_init(hComponent,
                                   sizeof(ffmpeg_vdecoder_private_t),
                                   COMPONENT_NAME, VERSION,
                                   OMX_VIDEO_CodingAutoDetect);
    if (OMX_ErrorNone != ret) goto bail;
    hdl = (OMX_COMPONENTTYPE *)*hComponent;

    ffmpeg_decoder  = LIMOI_FFMPEG_VDECODER(*hComponent);

    role = &roles[0];
    while (*role) limoi_component_add_role(*hComponent, *role++);

    hdl->ComponentDeInit = ComponentDeInit;

    callbacks.command_prep = prep_cmd_handle;
    callbacks.command_post = post_cmd_handle;
    callbacks.decode = decode_handle;
    limoi_video_decoder_set_callbacks(*hComponent, &callbacks);

    LIMDBG("%s component loaded.\n", COMPONENT_NAME);

 bail:
    return ret;
}

lim_omx_component_registration_t ** OMX_ComponentRegistration (int * num)
{
    static lim_omx_component_registration_t comp_reg = {
        COMPONENT_NAME, (char **)roles, ComponentInit
    };
    static lim_omx_component_registration_t * regs[] = {&comp_reg};

    *num = 1;
    return regs;
}
