/*
  Copyright (C) 2010 - 2011 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX IL ffmpeg audio encoder component.

  REVISION:

    1) initial creation. ----------------------------------------- 2010-05-19 MG
*/

#include <stdio.h>
#include <stdlib.h>
#include "OMX_Core.h"
#include "OMX_Types.h"
#include "OMX_Component.h"
#include "limoi-ffmpeg-encode-audio.h"

//#undef LIMDBG
#ifndef LIMDBG
#define LIMDBG(fmt, arg...)
#endif

#define COMPONENT_NAME "OMX.limoi.ffmpeg.encode.audio"
static const char * roles[] = {
    "audio_encoder.aac",
    "audio_encoder.mp3",
    /* "audio_encoder.amr", */
    /* "audio_encoder.ogg", */
    /* "audio_encoder.ra", */
    /* "audio_encoder.wma", */
    /* "audio_encoder.adpcm", */
    /* "audio_encoder.pcm", */
    NULL
};

static void find_encoder (
    OMX_HANDLETYPE hComp)
{
    ffmpeg_aencoder_private_t * enc = LIMOI_FFMPEG_AENCODER(hComp);
    OMX_AUDIO_PARAM_PCMMODETYPE pcm;
    enum CodecID codec_id = CODEC_ID_NONE;
    int bitrate = 0;

    if (NULL == enc->codec)
    {
        char role[OMX_MAX_STRINGNAME_SIZE];
        if (OMX_ErrorNone != limoi_component_get_current_role(hComp, role))
        {
            LIMERR("%s unable to get current role.\n", COMPONENT_NAME);
            return;
        }

        if (!strcmp("audio_encoder.aac", role))
        {
            OMX_AUDIO_PARAM_AACPROFILETYPE aac;

            limoi_set_spec_version(&aac, sizeof(OMX_AUDIO_PARAM_AACPROFILETYPE));
            aac.nPortIndex = 1;
            if (OMX_ErrorNone != limoi_component_get_parameter
                (hComp, OMX_IndexParamAudioAac, &aac))
            {
                LIMWARN("unable to get encoder AAC parameter.\n");
            } else
            {
                bitrate = aac.nBitRate;
            }
            codec_id = CODEC_ID_AAC;
        } else if (!strcmp("audio_encoder.mp3", role))
        {
            OMX_AUDIO_PARAM_MP3TYPE mp3;

            limoi_set_spec_version(&mp3, sizeof(OMX_AUDIO_PARAM_MP3TYPE));
            mp3.nPortIndex = 1;
            if (OMX_ErrorNone != limoi_component_get_parameter
                (hComp, OMX_IndexParamAudioMp3, &mp3))
            {
                LIMWARN("unable to get encoder MP3 parameter.\n");
            } else
            {
                bitrate = mp3.nBitRate;
            }

            //TODO: mp3 support?
            codec_id = CODEC_ID_MP2;
        } else
        {
            LIMWARN("%s unsupported role: %s\n", COMPONENT_NAME, role);
            return;
        }

        enc->codec = avcodec_find_encoder(codec_id);
        if (NULL == enc->codec)
        {
            LIMWARN("%s unable to find codec.\n", COMPONENT_NAME);
            return;
        } else enc->codec->id = codec_id;
    }

    limoi_set_spec_version(&pcm, sizeof(OMX_AUDIO_PARAM_PCMMODETYPE));
    pcm.nPortIndex = 0;
    if (OMX_ErrorNone != limoi_component_get_parameter
        (hComp, OMX_IndexParamAudioPcm, &pcm))
    {
        LIMWARN("unable to get encoder PCM parameter.\n");
        return;
    }

    enc->avcodec = avcodec_alloc_context();

    avcodec_get_context_defaults2(enc->avcodec, enc->codec->type);
    enc->avcodec->codec_type = AVMEDIA_TYPE_AUDIO;
    enc->avcodec->channels = pcm.nChannels;
    enc->avcodec->sample_rate = pcm.nSamplingRate;
    if (bitrate) enc->avcodec->bit_rate = bitrate;
    enc->avcodec->codec_id = enc->codec->id;
    //TODO:
    enc->avcodec->sample_fmt = SAMPLE_FMT_S16;
    enc->avcodec->channel_layout = 0;

    enc->avcodec->time_base= (AVRational){1, enc->avcodec->sample_rate};

    if (avcodec_open(enc->avcodec, enc->codec) < 0)
    {
        LIMERR("%s unable to open codec.\n", COMPONENT_NAME);
        free(enc->avcodec);
        enc->avcodec = NULL;
        return;
    }

    enc->frame_size = enc->avcodec->frame_size * enc->avcodec->channels
        * av_get_bytes_per_sample(enc->avcodec->sample_fmt);
    enc->frame_bytes = 0;
    enc->frame_buffer = malloc(enc->frame_size);
    if (NULL == enc->frame_buffer)
    {
        LIMERR("%s out of memory.\n", COMPONENT_NAME);
        free(enc->avcodec);
        enc->avcodec = NULL;
        return;
    }

    enc->avcodec_allocated = 1;
}

static OMX_ERRORTYPE set_state (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    OMX_HANDLETYPE hComp = cmd->hComponent;
    OMX_STATETYPE next_state = cmd->nParam1;
    OMX_STATETYPE state = limoi_component_state(hComp);
    ffmpeg_aencoder_private_t * enc = LIMOI_FFMPEG_AENCODER(hComp);

    switch (state)
    {
    case OMX_StateLoaded :

        // transition to StateIdle
        omx_ffmpeg_codec_init();

        enc->avcodec = NULL;
        enc->avcodec_allocated = 0;
        enc->codec = NULL;

        enc->frame_buffer = NULL;
        break;

    case OMX_StateIdle :
        if (OMX_StateLoaded == next_state)
        {
            if (enc->avcodec)
            {
                avcodec_close(enc->avcodec);
                if (enc->avcodec_allocated)
                {
                    av_free(enc->avcodec);
                    enc->avcodec_allocated = 0;
                }
                enc->avcodec = NULL;

                if (enc->frame_buffer)
                {
                    free(enc->frame_buffer);
                    enc->frame_buffer = NULL;
                }
            }
        }
        break;

    default :
        break;
    }
    return ret;
}

static OMX_ERRORTYPE cmd_handle (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    OMX_HANDLETYPE hComp = cmd->hComponent;

    LIMDBG("%s command handler.\n", COMPONENT_NAME);
    limoi_component_lock(hComp);

    switch (cmd->Cmd)
    {
    case OMX_CommandStateSet :
        ret = set_state(cmd);
        break;

    default :
        break;
    }
    limoi_component_unlock(hComp);

    return ret;
}

static int encode_handle (
    OMX_HANDLETYPE hComp,
    OMX_BUFFERHEADERTYPE * in, OMX_BUFFERHEADERTYPE * out)
{
    ffmpeg_aencoder_private_t * enc;
    int len;

    enc = LIMOI_FFMPEG_AENCODER(hComp);

    //LIMDBG("%s encode handle.\n", COMPONENT_NAME);
    limoi_component_lock(hComp);

    if (NULL == enc->codec || NULL == enc->avcodec)
    {
        find_encoder(hComp);
    }
    if (NULL == enc->codec || NULL == enc->avcodec) goto bail;

    if (out->nAllocLen < FF_MIN_BUFFER_SIZE)
    {
        LIMWARN("%s output buffer too small: %d, required minimum: %d\n",
                COMPONENT_NAME, out->nAllocLen, FF_MIN_BUFFER_SIZE);
        goto bail;
    }

    //LIMDBG("%s input data length: %u, offset %u\n",
    //       COMPONENT_NAME, in->nFilledLen, in->nOffset);

    if (enc->frame_bytes)
    {
        int bytes = enc->frame_size - enc->frame_bytes;
        if (bytes > in->nFilledLen) bytes = in->nFilledLen;

        memcpy(enc->frame_buffer + enc->frame_bytes,
               in->pBuffer + in->nOffset, bytes);
        enc->frame_bytes += bytes;
        in->nOffset += bytes;
        in->nFilledLen -= bytes;

        if (enc->frame_bytes == enc->frame_size)
        {
            len = avcodec_encode_audio
                (enc->avcodec,
                 (uint8_t *)(out->pBuffer + out->nOffset),
                 out->nAllocLen - out->nFilledLen,
                 (int16_t *)enc->frame_buffer);
            if (len < 0)
            {
                LIMWARN("%s encoding error.\n", COMPONENT_NAME);
                goto bail;
            }
            enc->frame_bytes = 0;
            out->nFilledLen += len;
        }
    }

    while (in->nFilledLen >= enc->frame_size
           && out->nAllocLen - out->nFilledLen >= FF_MIN_BUFFER_SIZE)
    {
        len = avcodec_encode_audio
            (enc->avcodec,
             (uint8_t *)(out->pBuffer + out->nOffset + out->nFilledLen),
             out->nAllocLen - out->nFilledLen,
             (int16_t *)(in->pBuffer + in->nOffset));
        if (len < 0)
        {
            LIMWARN("%s encoding error.\n", COMPONENT_NAME);
            goto bail;
        } else
        {
            //LIMINFO("%s encoded audio bytes: %d\n", COMPONENT_NAME, len);
        }

        in->nFilledLen -= enc->frame_size;
        in->nOffset += enc->frame_size;
        out->nFilledLen += len;
    }

    if (in->nFilledLen && in->nFilledLen < enc->frame_size)
    {
        memcpy(enc->frame_buffer, in->pBuffer + in->nOffset, in->nFilledLen);
        in->nFilledLen = 0;
    }

 bail:
    limoi_component_unlock(hComp);
    return 0;
}

static OMX_ERRORTYPE ComponentDeInit (OMX_HANDLETYPE hComp)
{
    return limoi_audio_encoder_deinit(hComp);
}

static OMX_ERRORTYPE ComponentInit (OMX_HANDLETYPE * hComponent, const char * name)
{
    OMX_COMPONENTTYPE * hdl;
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    limoi_audio_encoder_callbacks_t callbacks;
    ffmpeg_aencoder_private_t * enc;
    const char ** role;

    LIMDBG("loading %s component.\n", COMPONENT_NAME);
    ret = limoi_audio_encoder_init(hComponent,
                                   sizeof(ffmpeg_aencoder_private_t),
                                   COMPONENT_NAME, VERSION,
                                   OMX_AUDIO_CodingAutoDetect);
    if (OMX_ErrorNone != ret) goto bail;
    hdl = (OMX_COMPONENTTYPE *)*hComponent;

    enc  = LIMOI_FFMPEG_AENCODER(*hComponent);

    role = &roles[0];
    while (*role) limoi_component_add_role(*hComponent, *role++);

    hdl->ComponentDeInit = ComponentDeInit;

    callbacks.command_prep = NULL;
    callbacks.command_post = cmd_handle;
    callbacks.encode = encode_handle;
    limoi_audio_encoder_set_callbacks(*hComponent, &callbacks);

    LIMDBG("%s component loaded.\n", COMPONENT_NAME);

 bail:
    return ret;
}

lim_omx_component_registration_t ** OMX_ComponentRegistration (int * num)
{
    static lim_omx_component_registration_t comp_reg = {
        COMPONENT_NAME, (char **)roles, ComponentInit
    };
    static lim_omx_component_registration_t * regs[] = {&comp_reg};

    *num = 1;
    return regs;
}
