/*
  Copyright (C) 2010 - 2011 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX IL ffmpeg video encoder component.

  REVISION:

    1) initial creation. ----------------------------------------- 2010-05-19 MG
*/

#include <stdio.h>
#include <stdlib.h>
#include "OMX_Core.h"
#include "OMX_Types.h"
#include "OMX_Component.h"
#include "limoi-ffmpeg-encode-video.h"

//#undef LIMDBG
#ifndef LIMDBG
#define LIMDBG(fmt, arg...)
#endif

#define COMPONENT_NAME "OMX.limoi.ffmpeg.encode.video"
static const char * roles[] = {
    "video_encoder.h263",
    "video_encoder.mpeg4",
    /* "video_encoder.mpeg2", */
    /* "video_encoder.avc", */
    /* "video_encoder.wmv", */
    /* "video_encoder.rv", */
    /* "video_encoder.mjpeg", */
    NULL
};

static void find_encoder (
    OMX_HANDLETYPE hComp)
{
    ffmpeg_vencoder_private_t * enc = LIMOI_FFMPEG_VENCODER(hComp);
    enum CodecID codec_id = CODEC_ID_NONE;
    OMX_PARAM_PORTDEFINITIONTYPE i_port_def;
    OMX_VIDEO_PORTDEFINITIONTYPE * i_video = &i_port_def.format.video;
    OMX_PARAM_PORTDEFINITIONTYPE o_port_def;
    OMX_VIDEO_PORTDEFINITIONTYPE * o_video = &o_port_def.format.video;

    if (NULL == enc->codec)
    {
        char role[OMX_MAX_STRINGNAME_SIZE];
        if (OMX_ErrorNone != limoi_component_get_current_role(hComp, role))
        {
            LIMERR("%s unable to get current role.\n", COMPONENT_NAME);
            return;
        }

        if (!strcmp("video_encoder.h263", role))
        {
            codec_id = CODEC_ID_H263P;
        } else if (!strcmp("video_encoder.mpeg4", role))
        {
            codec_id = CODEC_ID_MPEG4;
        } else
        {
            LIMWARN("%s unsupported role: %s\n", COMPONENT_NAME, role);
            return;
        }

        enc->codec = avcodec_find_encoder(codec_id);
        if (NULL == enc->codec)
        {
            LIMWARN("%s unable to find codec.\n", COMPONENT_NAME);
            return;
        }
    }

    limoi_set_spec_version(&i_port_def, sizeof(OMX_PARAM_PORTDEFINITIONTYPE));
    i_port_def.nPortIndex = 0;
    if (OMX_ErrorNone != limoi_component_get_parameter
        (hComp, OMX_IndexParamPortDefinition, &i_port_def))
    {
        LIMWARN("unable to get encoder input port parameter.\n");
        return;
    }

    limoi_set_spec_version(&o_port_def, sizeof(OMX_PARAM_PORTDEFINITIONTYPE));
    o_port_def.nPortIndex = 1;
    if (OMX_ErrorNone != limoi_component_get_parameter
        (hComp, OMX_IndexParamPortDefinition, &o_port_def))
    {
        LIMWARN("unable to get encoder output port parameter.\n");
        return;
    }

    if (i_video->nFrameWidth != o_video->nFrameWidth
        || i_video->nFrameHeight != o_video->nFrameHeight)
    {
        LIMWARN("%s requested encoding dimension does not match with input.\n"
                "\tinput wxh : %dx%d, requested output wxh: %dx%d\n",
                i_video->nFrameWidth, i_video->nFrameHeight,
                o_video->nFrameWidth, o_video->nFrameHeight);
    }

    enc->avcodec = avcodec_alloc_context();

    avcodec_get_context_defaults2(enc->avcodec, enc->codec->type);
    enc->avcodec->codec_type = AVMEDIA_TYPE_VIDEO;

    enc->avcodec->width = i_video->nFrameWidth;
    enc->avcodec->height = i_video->nFrameHeight;
    enc->avcodec->sample_aspect_ratio =
        av_d2q(1 * enc->avcodec->height/enc->avcodec->width, 255);

    //TODO: check to honor pixel format
    enc->avcodec->pix_fmt = PIX_FMT_YUV420P;

    //TODO: per-codec rate control
    enc->avcodec->bit_rate = o_video->nBitrate;

    enc->avcodec->codec_id = enc->codec->id;

    enc->avcodec->time_base.den = 65536;
    enc->avcodec->time_base.num = o_video->xFramerate >> 16;

    if (avcodec_open(enc->avcodec, enc->codec) < 0)
    {
        LIMERR("%s unable to open codec.\n", COMPONENT_NAME);
        free(enc->avcodec);
        enc->avcodec = NULL;
        return;
    }

    enc->avcodec_allocated = 1;
}

static OMX_ERRORTYPE set_state (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    OMX_HANDLETYPE hComp = cmd->hComponent;
    OMX_STATETYPE next_state = cmd->nParam1;
    OMX_STATETYPE state = limoi_component_state(hComp);
    ffmpeg_vencoder_private_t * enc = LIMOI_FFMPEG_VENCODER(hComp);

    switch (state)
    {
    case OMX_StateLoaded :

        // transition to StateIdle
        omx_ffmpeg_codec_init();

        enc->avcodec = NULL;
        enc->avcodec_allocated = 0;
        enc->codec = NULL;

        break;

    case OMX_StateIdle :
        if (OMX_StateLoaded == next_state)
        {
            if (enc->avcodec)
            {
                avcodec_close(enc->avcodec);
                if (enc->avcodec_allocated)
                {
                    av_free(enc->avcodec);
                    enc->avcodec_allocated = 0;
                }
                enc->avcodec = NULL;
            }
        }
        break;

    default :
        break;
    }

    return ret;
}

static OMX_ERRORTYPE cmd_handle (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    OMX_HANDLETYPE hComp = cmd->hComponent;

    LIMDBG("%s command handler.\n", COMPONENT_NAME);
    limoi_component_lock(hComp);

    switch (cmd->Cmd)
    {
    case OMX_CommandStateSet :
        ret = set_state(cmd);
        break;

    default :
        break;
    }

    limoi_component_unlock(hComp);
    return ret;
}

static int encode_handle (
    OMX_HANDLETYPE hComp,
    OMX_BUFFERHEADERTYPE * in,
    OMX_BUFFERHEADERTYPE * out)
{
    ffmpeg_vencoder_private_t * enc;
    AVFrame frame;
    int len;
    int planner_bytes;

    enc = LIMOI_FFMPEG_VENCODER(hComp);

    //LIMDBG("%s encode handle.\n", COMPONENT_NAME);
    limoi_component_lock(hComp);

    if (NULL == enc->codec || NULL == enc->avcodec)
    {
        find_encoder(hComp);
    }
    if (NULL == enc->codec || NULL == enc->avcodec) goto bail;

    LIMDBG("%s input data length: %u, offset %u\n",
           COMPONENT_NAME, in->nFilledLen, in->nOffset);
    avcodec_get_frame_defaults(&frame);

    //TODO: hard-coded to YUV420P
    planner_bytes = enc->avcodec->width * enc->avcodec->height;
    frame.data[0] = in->pBuffer + in->nOffset;
    frame.data[1] = frame.data[0] + planner_bytes;
    frame.data[2] = frame.data[1] + planner_bytes / 4;
    frame.linesize[0] = enc->avcodec->width;
    frame.linesize[1] = enc->avcodec->width / 2;
    frame.linesize[2] = enc->avcodec->width / 2;

    len = avcodec_encode_video(enc->avcodec,
                               (uint8_t *)(out->pBuffer + out->nOffset),
                               out->nAllocLen, &frame);
    if (len < 0)
    {
        LIMWARN("%s encoding error.\n", COMPONENT_NAME);
        goto bail;
    } else
    {
        //LIMINFO("%s encoded video bytes: %d\n", COMPONENT_NAME, len);
    }

    in->nFilledLen = 0;
    out->nFilledLen += len;

 bail:
    limoi_component_unlock(hComp);
    return 0;
}


static OMX_ERRORTYPE ComponentDeInit (OMX_HANDLETYPE hComp)
{
    return limoi_video_encoder_deinit(hComp);
}

static OMX_ERRORTYPE ComponentInit (OMX_HANDLETYPE * hComponent, const char * name)
{
    OMX_COMPONENTTYPE * hdl;
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    limoi_video_encoder_callbacks_t callbacks;
    ffmpeg_vencoder_private_t * enc;
    const char ** role;

    LIMDBG("loading %s component.\n", COMPONENT_NAME);
    ret = limoi_video_encoder_init(hComponent,
                                   sizeof(ffmpeg_vencoder_private_t),
                                   COMPONENT_NAME, VERSION,
                                   OMX_VIDEO_CodingAutoDetect);
    if (OMX_ErrorNone != ret) goto bail;
    hdl = (OMX_COMPONENTTYPE *)*hComponent;

    enc  = LIMOI_FFMPEG_VENCODER(*hComponent);

    role = &roles[0];
    while (*role) limoi_component_add_role(*hComponent, *role++);

    hdl->ComponentDeInit = ComponentDeInit;

    callbacks.command_prep = NULL;
    callbacks.command_post = cmd_handle;
    callbacks.encode = encode_handle;
    limoi_video_encoder_set_callbacks(*hComponent, &callbacks);

    LIMDBG("%s component loaded.\n", COMPONENT_NAME);

 bail:
    return ret;
}

lim_omx_component_registration_t ** OMX_ComponentRegistration (int * num)
{
    static lim_omx_component_registration_t comp_reg = {
        COMPONENT_NAME, (char **)roles, ComponentInit
    };
    static lim_omx_component_registration_t * regs[] = {&comp_reg};

    *num = 1;
    return regs;
}
