/*
  Copyright (C) 2010 - 2011 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX IL ffmpeg muxer component.

  REVISION:

    1) initial creation. ----------------------------------------- 2010-05-19 MG
*/

#include <stdio.h>
#include <stdlib.h>
#include "OMX_Core.h"
#include "OMX_Types.h"
#include "OMX_Component.h"
#include "limoi-ffmpeg-mux.h"

//#undef LIMDBG
#ifndef LIMDBG
#define LIMDBG(fmt, arg...)
#endif

#define COMPONENT_NAME "OMX.limoi.ffmpeg.mux"
#define ROLE_NAME "container_muxer.all"

static ffmpeg_cb_t * ffmpeg_cb;

static int get_content_uri (OMX_HANDLETYPE hComp)
{
    ffmpeg_muxer_private_t * muxer = LIMOI_FFMPEG_MUXER(hComp);
    void * buf[(sizeof(OMX_PARAM_CONTENTURITYPE) + LIMOI_URI_MAX_LENGTH) /
               sizeof(void *)];
    OMX_PARAM_CONTENTURITYPE * uri_param = (OMX_PARAM_CONTENTURITYPE *)&buf[0];

    limoi_set_spec_version(uri_param, sizeof(OMX_PARAM_CONTENTURITYPE)
                           + LIMOI_URI_MAX_LENGTH);
    if (OMX_ErrorNone == limoi_component_get_parameter
        (hComp, OMX_IndexParamContentURI, uri_param))
    {
        strncpy((char *)muxer->uri, (char *)uri_param->contentURI,
                LIMOI_URI_MAX_LENGTH - 1);
        muxer->uri[LIMOI_URI_MAX_LENGTH - 1] = 0;
    } else
    {
        muxer->uri[0] = 0;
        return -1;
    }
    return 0;
}

static OMX_ERRORTYPE set_state (
    limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    OMX_HANDLETYPE hComp = cmd->hComponent;
    OMX_STATETYPE next_state = cmd->nParam1;
    OMX_STATETYPE state = limoi_component_state(hComp);
    ffmpeg_muxer_private_t * muxer = LIMOI_FFMPEG_MUXER(hComp);

    LIMDBG("%s callback set state from %s to %s.\n", COMPONENT_NAME,
           limoi_omx_state_string(state), limoi_omx_state_string(next_state));

    switch (state)
    {
    case OMX_StateLoaded :

        // transition to StateIdle
        muxer->inited = 0;
        muxer->avformat = NULL;

        omx_ffmpeg_mux_init();
        break;

    case OMX_StateIdle :
        if (OMX_StateLoaded == next_state)
        {
            av_write_trailer(muxer->avformat);
            if (muxer->avformat)
            {
                int ii;
                for(ii = 0; ii < muxer->avformat->nb_streams; ii++)
                {
                    av_freep(&muxer->avformat->streams[ii]);
                }
                av_free(muxer->avformat);
            }
            if (muxer->bio)
            {
                av_free(muxer->bio);
                muxer->bio = NULL;
            }
        }
        break;

    default :
        break;
    }

    return ret;
}

static OMX_ERRORTYPE cmd_handle (
    limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    OMX_HANDLETYPE hComp = cmd->hComponent;

    LIMDBG("%s command handler.\n", COMPONENT_NAME);
    limoi_component_lock(hComp);

    switch (cmd->Cmd)
    {
    case OMX_CommandStateSet :
        ret = set_state(cmd);
        break;

    default :
        break;
    }
    limoi_component_unlock(hComp);

    return ret;
}

static int write_packet_cb (
    void * opaque,
    uint8_t * buf,
    int buf_size)
{
    ffmpeg_cb_t * cb = (ffmpeg_cb_t *)opaque;
    CPresult ret;

    //LIMDBG("%s read bytes: %d\n", COMPONENT_NAME, buf_size);
    ret = cb->pipe->Write(cb->content, (CPbyte *)buf, (CPuint)buf_size);
    if (ret)
    {
        LIMERR("%s content write error: %d\n", COMPONENT_NAME, ret);
    }

    return buf_size;
}

static AVStream * add_audio_stream (AVFormatContext *oc, enum CodecID codec_id)
{
    AVCodecContext *c;
    AVStream *st;

    st = av_new_stream(oc, 1);
    if (!st)
    {
        LIMERR("%s could not allocate audio stream\n", COMPONENT_NAME);
        goto bail;
    }

    c = st->codec;
    c->codec_id = codec_id;
    c->codec_type = AVMEDIA_TYPE_AUDIO;

    /* put sample parameters */
    c->sample_fmt = SAMPLE_FMT_S16;
    c->bit_rate = 64000;
    c->sample_rate = 44100;
    c->channels = 2;

    // some formats want stream headers to be separate
    if(oc->oformat->flags & AVFMT_GLOBALHEADER)
        c->flags |= CODEC_FLAG_GLOBAL_HEADER;

 bail:
    return st;
}

#define STREAM_FRAME_RATE 30
#define STREAM_PIX_FMT PIX_FMT_YUV420P /* default pix_fmt */
static AVStream * add_video_stream (AVFormatContext *oc, enum CodecID codec_id)
{
    AVCodecContext *c;
    AVStream *st;

    st = av_new_stream(oc, 0);
    if (!st)
    {
        LIMERR("%s could not allocate video stream\n", COMPONENT_NAME);
        goto bail;
    }

    c = st->codec;
    c->codec_id = codec_id;
    c->codec_type = AVMEDIA_TYPE_VIDEO;

    /* put sample parameters */
    c->bit_rate = 400000;
    /* resolution must be a multiple of two */
    c->width = 640; //352;
    c->height = 440; //288;
    /* time base: this is the fundamental unit of time (in seconds) in terms
       of which frame timestamps are represented. for fixed-fps content,
       timebase should be 1/framerate and timestamp increments should be
       identically 1. */
    c->time_base.den = STREAM_FRAME_RATE;
    c->time_base.num = 1;
    c->gop_size = 12; /* emit one intra frame every twelve frames at most */
    c->pix_fmt = STREAM_PIX_FMT;
    if (c->codec_id == CODEC_ID_MPEG2VIDEO) {
        /* just for testing, we also add B frames */
        c->max_b_frames = 2;
    }
    if (c->codec_id == CODEC_ID_MPEG1VIDEO){
        /* Needed to avoid using macroblocks in which some coeffs overflow.
           This does not happen with normal video, it just happens here as
           the motion of the chroma plane does not match the luma plane. */
        c->mb_decision=2;
    }
    // some formats want stream headers to be separate
    if(oc->oformat->flags & AVFMT_GLOBALHEADER)
        c->flags |= CODEC_FLAG_GLOBAL_HEADER;

 bail:
    return st;
}

static int mux_handle (
    OMX_HANDLETYPE hComp,
    CP_PIPETYPE * pipe,
    CPhandle content,
    OMX_BUFFERHEADERTYPE * audio_in,
    OMX_BUFFERHEADERTYPE * video_in)
{
    ffmpeg_muxer_private_t * muxer = LIMOI_FFMPEG_MUXER(hComp);
    OMX_PARAM_PORTDEFINITIONTYPE port_def;
    OMX_AUDIO_PORTDEFINITIONTYPE * audio = &port_def.format.audio;
    OMX_VIDEO_PORTDEFINITIONTYPE * video = &port_def.format.video;

    limoi_component_lock(hComp);
    ffmpeg_cb = &muxer->cb;

    if (muxer->inited == 0)
    {
        ffmpeg_cb = &muxer->cb;

        ffmpeg_cb->pipe = pipe;
        ffmpeg_cb->content = content;
        ffmpeg_cb->hComp = hComp;

        if (get_content_uri(hComp)) goto bail;
        else if (NULL == muxer->bio)
        {
            LIMDBG("%s output URI: %s\n", COMPONENT_NAME, muxer->uri);
            muxer->bio = avio_alloc_context
                (muxer->bio_buffer,
                 FFMPEG_BIO_BUFFER_SIZE,
                 1,
                 (void *)ffmpeg_cb,
                 NULL,
                 write_packet_cb,
                 NULL);
            if (NULL == muxer->bio)
            {
                LIMERR("%s unable to init byteIO.\n", COMPONENT_NAME);
                goto bail;
            }
        }

        if (NULL == muxer->avformat)
        {
            muxer->avformat = avformat_alloc_context();
            if (NULL == muxer->avformat)
            {
                LIMERR("%s out of memory.\n", COMPONENT_NAME);
                goto bail;
            }
            muxer->avformat->oformat = NULL;
            muxer->avformat->pb = muxer->bio;
        }

        if (NULL == muxer->avformat->oformat)
        {
            muxer->avformat->oformat =
                av_guess_format(NULL, (char *)muxer->uri, NULL);
            if (!muxer->avformat->oformat)
            {
                LIMERR("%s unable to find output format.\n", COMPONENT_NAME);
                goto bail;
            } else
            {
                snprintf(muxer->avformat->filename,
                         sizeof(muxer->avformat->filename), "%s",
                         (char *)muxer->uri);
            }
        }

        port_def.nSize = sizeof(OMX_PARAM_PORTDEFINITIONTYPE);
        port_def.nPortIndex = 0;
        if (OMX_ErrorNone != OMX_GetParameter
            (hComp, OMX_IndexParamPortDefinition, &port_def))
        {
            LIMWARN("%s unable to get audio type.\n", COMPONENT_NAME);
        }
        muxer->avformat->oformat->audio_codec =
            omx_ffmpeg_audio_codec_id_omx2ffmpeg(audio->eEncoding);

        port_def.nPortIndex = 1;
        if (OMX_ErrorNone != OMX_GetParameter
            (hComp, OMX_IndexParamPortDefinition, &port_def))
        {
            LIMWARN("%s unable to get video type.\n", COMPONENT_NAME);
        }
        muxer->avformat->oformat->video_codec =
            omx_ffmpeg_video_codec_id_omx2ffmpeg(video->eCompressionFormat);

        if (muxer->avformat->oformat->video_codec != CODEC_ID_NONE)
        {
            add_video_stream(muxer->avformat,
                             muxer->avformat->oformat->video_codec);
        }

        if (muxer->avformat->oformat->audio_codec != CODEC_ID_NONE)
        {
            add_audio_stream(muxer->avformat,
                             muxer->avformat->oformat->audio_codec);
        }

        //av_dump_format(muxer->avformat, 0, (char *)muxer->uri, 1);
        avformat_write_header(muxer->avformat, NULL);
        muxer->inited = 1;
    } else if (audio_in || video_in)
    {
        AVPacket pkt;

        if (audio_in)
        {
            LIMDBG("%s mux audio: %u\n", COMPONENT_NAME, audio_in->nFilledLen);
            av_init_packet(&pkt);
            pkt.size = audio_in->nFilledLen;
            pkt.pts = AV_NOPTS_VALUE; //TODO: assign the correct pts
            pkt.flags |= AV_PKT_FLAG_KEY;
            pkt.stream_index = 1; //TODO: do not hard-code stream index.
            pkt.data = audio_in->pBuffer + audio_in->nOffset;
            audio_in->nFilledLen = 0;
            if (av_interleaved_write_frame(muxer->avformat, &pkt))
            {
                LIMERR("%s: error writing audio frame.\n", COMPONENT_NAME);
            }
        }

        if (video_in)
        {
            LIMDBG("%s mux video: %u\n", COMPONENT_NAME, video_in->nFilledLen);
            av_init_packet(&pkt);
            pkt.size = video_in->nFilledLen;
            pkt.pts = AV_NOPTS_VALUE; //TODO: assign the correct pts
            pkt.flags |= AV_PKT_FLAG_KEY; //TODO: is coded frame key frame?
            pkt.stream_index = 0; //TODO: do not hard-code stream index.
            pkt.data = video_in->pBuffer + video_in->nOffset;
            video_in->nFilledLen = 0;

            if (av_interleaved_write_frame(muxer->avformat, &pkt))
            {
                LIMERR("%s: error writing video frame.\n", COMPONENT_NAME);
            }
        }
    }

 bail:
    limoi_component_unlock(hComp);
    return 0;
}

static OMX_ERRORTYPE ComponentDeInit (
    OMX_HANDLETYPE hComp)
{
    return limoi_muxer_deinit(hComp);
}

static OMX_ERRORTYPE ComponentInit (OMX_HANDLETYPE * hComp, const char * name)
{
    OMX_COMPONENTTYPE * hdl;
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    limoi_muxer_callbacks_t callbacks;
    ffmpeg_muxer_private_t * muxer;

    LIMDBG("loading %s component.\n", COMPONENT_NAME);
    ret = limoi_muxer_init(hComp, sizeof(ffmpeg_muxer_private_t),
                             COMPONENT_NAME, VERSION);
    if (OMX_ErrorNone != ret) goto bail;
    hdl = (OMX_COMPONENTTYPE *)*hComp;

    muxer = LIMOI_FFMPEG_MUXER(*hComp);

    limoi_component_add_role(*hComp, ROLE_NAME);

    hdl->ComponentDeInit = ComponentDeInit;

    callbacks.command_prep = NULL;
    callbacks.command_post = cmd_handle;
    callbacks.mux = mux_handle;
    limoi_muxer_set_callbacks(*hComp, &callbacks);

    LIMDBG("%s component loaded.\n", COMPONENT_NAME);

 bail:
    return ret;
}

lim_omx_component_registration_t ** OMX_ComponentRegistration (int * num)
{
    static char * roles[] = {ROLE_NAME, NULL};
    static lim_omx_component_registration_t comp_reg = {
        COMPONENT_NAME, roles, ComponentInit
    };
    static lim_omx_component_registration_t * regs[] = {&comp_reg};

    *num = 1;
    return regs;
}
