/*
  Copyright (C) 2009 - 2011 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX IL MAD decoder component.

  REVISION:

    1) initial creation. ----------------------------------------- 2009-12-22 MG
*/

#include "limoi-mad.h"

//#undef LIMDBG
#ifndef LIMDBG
#define LIMDBG(fmt, arg...)
#endif

#define COMPONENT_NAME "OMX.limoi.mad_dec"
#define ROLE_NAME "audio_decoder.mp3"

#define MIN(x, y) ((x) < (y))? (x) : (y)
#define MAX(x, y) ((x) < (y))? (y) : (x)

static OMX_ERRORTYPE set_state (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    OMX_HANDLETYPE hComp = cmd->hComponent;
    OMX_STATETYPE next_state = cmd->nParam1;
    OMX_STATETYPE state = limoi_component_state(hComp);
    mad_private_t * mad = LIMOI_MAD_DEC(hComp);

    limoi_component_lock(hComp);
    switch (state)
    {
    case OMX_StateLoaded :

        // transition to StateIdle
        mad->bitstream_buf = malloc(LIMOI_MAD_BITSTREAM_BUFFER_BYTES);
        if (NULL == mad->bitstream_buf)
        {
            ret = OMX_ErrorInsufficientResources;
            break;
        }
        mad->bitstream_bytes = 0;
        mad->decoder_inited = 0;

        mad->decoder.mode         = MAD_DECODER_MODE_SYNC;
        mad->decoder.options      = 0;
        mad->decoder.async.pid    = 0;
        mad->decoder.async.in     = -1;
        mad->decoder.async.out    = -1;
        mad->decoder.sync = malloc(sizeof(*(mad->decoder.sync)));
        if (NULL == mad->decoder.sync)
        {
            free(mad->bitstream_buf);
            ret = OMX_ErrorInsufficientResources;
            break;
        }

        mad_stream_init(&mad->decoder.sync->stream);
        mad_frame_init(&mad->decoder.sync->frame);
        mad_synth_init(&mad->decoder.sync->synth);

        mad_stream_options(&mad->decoder.sync->stream, mad->decoder.options);

        break;

    case OMX_StateIdle :
        if (OMX_StateLoaded == next_state)
        {
            mad_synth_finish(&mad->decoder.sync->synth);
            mad_frame_finish(&mad->decoder.sync->frame);
            mad_stream_finish(&mad->decoder.sync->stream);

            if (mad->decoder.sync) free(mad->decoder.sync);
            if (mad->bitstream_buf) free(mad->bitstream_buf);
            mad->decoder.sync = NULL;
            mad->bitstream_buf = NULL;
        }
        break;

    default :
        break;
    }
    limoi_component_unlock(hComp);

    return ret;
}

static OMX_ERRORTYPE cmd_handle (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    //mad_private_t * mad = LIMOI_MAD_DEC(cmd->hComponent);

    LIMDBG("%s command handler.\n", COMPONENT_NAME);

    switch (cmd->Cmd)
    {
    case OMX_CommandStateSet :
        ret = set_state(cmd);
        break;

    default :
        break;
    }

    return ret;
}

/*
 * The following utility routine performs simple rounding, clipping, and
 * scaling of MAD's high-resolution samples down to 16 bits. It does not
 * perform any dithering or noise shaping, which would be recommended to
 * obtain any exceptional audio quality. It is therefore not recommended to
 * use this routine if high-quality output is desired.
 */

static inline signed int scale (
    mad_fixed_t sample)
{
    /* round */
    sample += (1L << (MAD_F_FRACBITS - 16));

    /* clip */
    if (sample >= MAD_F_ONE)
        sample = MAD_F_ONE - 1;
    else if (sample < -MAD_F_ONE)
        sample = -MAD_F_ONE;

    /* quantize */
    return sample >> (MAD_F_FRACBITS + 1 - 16);
}

static enum mad_flow output_pcm (
    void * dptr,
    struct mad_header const * header,
    struct mad_pcm * pcm)
{
    unsigned int nchannels, nsamples;
    mad_fixed_t const *left_ch, *right_ch;
    char * data = (char *)dptr;

    /* pcm->samplerate contains the sampling frequency */

    nchannels = pcm->channels;
    nsamples  = pcm->length;
    left_ch   = pcm->samples[0];
    right_ch  = pcm->samples[1];

    while (nsamples--)
    {
        signed int sample;

        /* output sample(s) in 16-bit signed little-endian PCM */

        sample = scale(*left_ch++);
        //*data++ = (sample >> 0) & 0xff;
        //*data++ = (sample >> 8) & 0xff;
        //FIXME: full gain?
        *data++ = (sample >> 1) & 0xff;
        *data++ = (sample >> 9) & 0xff;

        if (nchannels == 2)
        {
            sample = scale(*right_ch++);
            //*data++ = (sample >> 0) & 0xff;
            //*data++ = (sample >> 8) & 0xff;
            *data++ = (sample >> 1) & 0xff;
            *data++ = (sample >> 9) & 0xff;
        }
    }

    return MAD_FLOW_CONTINUE;
}

static int decode_handle (
    OMX_HANDLETYPE hComp,
    OMX_BUFFERHEADERTYPE * in, OMX_BUFFERHEADERTYPE * out)
{
    mad_private_t * mad = LIMOI_MAD_DEC(hComp);
    struct mad_stream * stream;
    struct mad_frame * frame;
    struct mad_synth * synth;
    struct mad_decoder * decoder = &mad->decoder;
    int bytes;
    int left;

    limoi_component_lock(hComp);

    stream = &decoder->sync->stream;
    frame = &decoder->sync->frame;
    synth = &decoder->sync->synth;

    while (1)
    {
        if (stream->buffer == NULL || stream->error == MAD_ERROR_BUFLEN)
        {
            if (!in->nFilledLen) break;

            if (stream->next_frame)
            {
                left = mad->bitstream_bytes
                    - (stream->next_frame - mad->bitstream_buf);
                memmove(mad->bitstream_buf, stream->next_frame, left);
            } else left = 0;

            bytes = MIN(in->nFilledLen,
                        LIMOI_MAD_BITSTREAM_BUFFER_BYTES - left);

            memcpy(mad->bitstream_buf + left,
                   in->pBuffer + in->nOffset, bytes);
            in->nFilledLen -= bytes;
            in->nOffset += bytes;
            mad->bitstream_bytes = left + bytes;

            mad_stream_buffer(stream,
                              mad->bitstream_buf,
                              mad->bitstream_bytes);
            stream->error = 0;
        }

        if (mad_frame_decode(frame, stream))
        {
            if (MAD_RECOVERABLE(stream->error))
            {
                LIMWARN("%s recoverable error %s.\n", COMPONENT_NAME,
                        mad_stream_errorstr(stream));
                continue;
            }
            else
            {
                if(stream->error == MAD_ERROR_BUFLEN)
                    continue;
                else
                {
                    LIMERR("%s unrecoverable error %s.\n", COMPONENT_NAME,
                           mad_stream_errorstr(stream));
                    break;
                }
            }
        }

        mad_synth_frame(synth, frame);

        if (synth->pcm.length)
        {
            if (!mad->decoder_inited)
            {
                OMX_AUDIO_PARAM_PCMMODETYPE pcm_param;

                mad->decoder_inited = 1;

                pcm_param.nChannels = synth->pcm.channels;
                pcm_param.eNumData = OMX_NumericalDataSigned;
                pcm_param.eEndian = OMX_EndianMax; //don't care == host type
                pcm_param.bInterleaved = OMX_TRUE;
                pcm_param.nBitPerSample = 16;
                pcm_param.nSamplingRate = synth->pcm.samplerate;
                limoi_audio_decoder_set_pcm_mode(hComp, &pcm_param);
                printf("\n%s stream info:\n"
                       "\tlayer      : %d\n"
                       "\tmode       : %s\n"
                       "\tbitrate    : %lu\n"
                       "\tsamplerate : %u \n",
                       COMPONENT_NAME,
                       frame->header.layer,
                       (frame->header.mode)? "STEREO" : "MONO",
                       frame->header.bitrate,
                       frame->header.samplerate);
            }

            output_pcm(out->pBuffer + out->nOffset + out->nFilledLen,
                       &frame->header, &synth->pcm);
            out->nFilledLen += synth->pcm.length * 2 * synth->pcm.channels;

            break;
        }
    }
    limoi_component_unlock(hComp);

    return 0;
}

static OMX_ERRORTYPE ComponentDeInit (
    OMX_HANDLETYPE hComponent)
{
    return limoi_audio_decoder_deinit(hComponent);
}

static OMX_ERRORTYPE ComponentInit (OMX_HANDLETYPE * hComponent, const char * name)
{
    OMX_COMPONENTTYPE * hdl;
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    limoi_audio_decoder_callbacks_t callbacks;
    mad_private_t * mad;

    LIMDBG("loading %s component.\n", COMPONENT_NAME);
    ret = limoi_audio_decoder_init(hComponent, sizeof(mad_private_t),
                                   COMPONENT_NAME, VERSION,
                                   OMX_AUDIO_CodingMP3);
    if (OMX_ErrorNone != ret) goto bail;
    hdl = (OMX_COMPONENTTYPE *)*hComponent;

    mad = LIMOI_MAD_DEC(*hComponent);

    limoi_component_add_role(*hComponent, ROLE_NAME);

    hdl->ComponentDeInit = ComponentDeInit;

    callbacks.command_prep = NULL;
    callbacks.command_post = cmd_handle;
    callbacks.decode = decode_handle;
    limoi_audio_decoder_set_callbacks(*hComponent, &callbacks);

    LIMDBG("%s component loaded.\n", COMPONENT_NAME);

 bail:
    return ret;
}

lim_omx_component_registration_t ** OMX_ComponentRegistration (int * num)
{
    static char * roles[] = {ROLE_NAME, NULL};
    static lim_omx_component_registration_t comp_reg = {
        COMPONENT_NAME, roles, ComponentInit
    };
    static lim_omx_component_registration_t * regs[] = {&comp_reg};

    *num = 1;
    return regs;
}
