/*
  Copyright (C) 2009 - 2011 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX IL OGG decoder component.

  REVISION:

    1) initial creation. ----------------------------------------- 2009-12-22 MG
*/

#include "limoi-ogg.h"

//#undef LIMDBG
#ifndef LIMDBG
#define LIMDBG(fmt, arg...)
#endif

#define COMPONENT_NAME "OMX.limoi.ogg_dec"
#define ROLE_NAME "audio_decoder_with_framing.ogg"

static OMX_ERRORTYPE set_state (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    OMX_HANDLETYPE hComp = cmd->hComponent;
    OMX_STATETYPE next_state = cmd->nParam1;
    OMX_STATETYPE state = limoi_component_state(hComp);
    ogg_private_t * ogg = LIMOI_OGG_DEC(cmd->hComponent);

    limoi_component_lock(hComp);
    switch (state)
    {
    case OMX_StateLoaded :

        // transition to StateIdle
        ogg->decoder_inited = 0;
        ogg->header_counts = 0;
        ogg_sync_init(&ogg->oy);
        break;

    case OMX_StateIdle :
        if (OMX_StateLoaded == next_state)
        {
            ogg_sync_clear(&ogg->oy);
            if (ogg->header_counts)
            {
                ogg_stream_clear(&ogg->os);
                vorbis_info_clear(&ogg->vi);
                vorbis_comment_clear(&ogg->vc);
            }
            if (ogg->decoder_inited)
            {
                vorbis_block_clear(&ogg->vb);
                vorbis_dsp_clear(&ogg->vd);
            }
        }
        break;

    default :
        break;
    }
    limoi_component_unlock(hComp);

    return ret;
}

static OMX_ERRORTYPE cmd_handle (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    //ogg_private_t * ogg = LIMOI_OGG_DEC(cmd->hComponent);

    LIMDBG("%s command handler.\n", COMPONENT_NAME);

    switch (cmd->Cmd)
    {
    case OMX_CommandStateSet :
        ret = set_state(cmd);
        break;

    default :
        break;
    }

    return ret;
}

static int decode_handle (
    OMX_HANDLETYPE hComp,
    OMX_BUFFERHEADERTYPE * in, OMX_BUFFERHEADERTYPE * out)
{
    ogg_private_t * ogg = LIMOI_OGG_DEC(hComp);
    char * oggbuf;
    int ret;
    OMX_AUDIO_PARAM_PCMMODETYPE pcm_param;

    //LIMDBG("%s decode handle.\n", COMPONENT_NAME);
    limoi_component_lock(hComp);

    if (ogg->decoder_inited && ogg->decoder_pcm_ready) goto decode_pcm_out;
    if (ogg->decoder_inited && ogg->decoder_first_time)
    {
        ogg->decoder_first_time = 0;
        goto decode_start;
    }

    LIMDBG("%s filled length: %u offset: %u\n", COMPONENT_NAME,
           in->nFilledLen, in->nOffset);
    oggbuf = ogg_sync_buffer(&ogg->oy, in->nFilledLen);
    memcpy(oggbuf, in->pBuffer + in->nOffset, in->nFilledLen);
    ogg_sync_wrote(&ogg->oy, in->nFilledLen);
    in->nFilledLen = 0;

 decode_start:
    while (ogg->decoder_inited)
    {
        ret = ogg_sync_pageout(&ogg->oy, &ogg->og);
        if (0 == ret)
        {
            if (in->nFilledLen)
            {
                oggbuf = ogg_sync_buffer(&ogg->oy, in->nFilledLen);
                memcpy(oggbuf, in->pBuffer + in->nOffset, in->nFilledLen);
                ogg_sync_wrote(&ogg->oy, in->nFilledLen);
                in->nFilledLen = 0;
                continue;
            }
            LIMDBG("no more data.\n");
            ogg->decoder_pcm_ready = 0;

            limoi_component_unlock(hComp);
            return 0;
        }
        if (0 > ret) continue;

        ogg_stream_pagein(&ogg->os, &ogg->og);
        while (1)
        {
            float ** pcm;
            int samples;

            ret = ogg_stream_packetout(&ogg->os, &ogg->op);
            if (0 == ret) break;
            if (0 > ret) continue;

            if (vorbis_synthesis(&ogg->vb, &ogg->op) == 0)
                vorbis_synthesis_blockin(&ogg->vd, &ogg->vb);

        decode_pcm_out:
            while ((samples = vorbis_synthesis_pcmout(&ogg->vd, &pcm)) > 0)
            {
                int ii, jj;
                int clipflag = 0;
                int convsize = (out->nAllocLen - out->nFilledLen)
                    / (2 * ogg->vi.channels);
                int bout = (samples < convsize)? samples : convsize;

                /* convert floats to 16 bit signed ints (host order) and
                   interleave */
                for (ii = 0; ii < ogg->vi.channels; ii++)
                {
                    void * buf = out->pBuffer + out->nFilledLen + out->nOffset;
                    ogg_int16_t * ptr = (ogg_int16_t *)buf + ii;
                    float  * mono = pcm[ii];

                    for (jj = 0; jj < bout; jj++)
                    {
                        int val = floor(mono[jj] * 32767.f + .5f);

                        /* might as well guard against clipping */
                        if (val > 32767)
                        {
                            val = 32767;
                            clipflag = 1;
                        }
                        if (val < -32768)
                        {
                            val = -32768;
                            clipflag = 1;
                        }
                        *ptr = val >> 1;
                        ptr += ogg->vi.channels;
                    }
                }

                /*
                if (clipflag)
                {
                    LIMWARN("clipping in frame %ld\n",
                            (long)(ogg->vd.sequence));
                }
                */

                vorbis_synthesis_read(&ogg->vd, bout);

                out->nFilledLen += 2 * ogg->vi.channels * bout;
                LIMDBG("filled len: %u, samples: %d, convsize: %d\n",
                       out->nFilledLen, samples, convsize);
                if (samples >= convsize)
                {
                    ogg->decoder_pcm_ready = 1;

                    limoi_component_unlock(hComp);
                    return 0;
                }
            }
        }
        if (ogg_page_eos(&ogg->og))
        {
            LIMDBG("EOS reached.\n");

            limoi_component_unlock(hComp);
            return 0;
        }
    }

    //initialize stream and decoder.
    if (ogg->header_counts == 0)
    {
        //get first page out.
        if (ogg_sync_pageout(&ogg->oy, &ogg->og) != 1)
        {
            LIMERR("unable to get first page, unknown stream?\n");
            goto bail;
        }
        ogg_stream_init(&ogg->os, ogg_page_serialno(&ogg->og));
        vorbis_info_init(&ogg->vi);
        vorbis_comment_init(&ogg->vc);

        if (ogg_stream_pagein(&ogg->os, &ogg->og) == -1)
        {
            LIMERR("reading first page.\n");
            goto bail2;
        }

        if (ogg_stream_packetout(&ogg->os, &ogg->op) != 1)
        {
            LIMERR("reading intial header.\n");
            goto bail2;
        }

        if (vorbis_synthesis_headerin(&ogg->vi, &ogg->vc,&ogg->op) < 0)
        {
            LIMERR("stream does not contain vorbis.\n");
            goto bail2;
        }

        //we are vorbis, now looking for comments and codebook headers.
        ogg->header_counts++;
    }

    while (ogg->header_counts < 3)
    {
        ret = ogg_sync_pageout(&ogg->oy, &ogg->og);
        if (0 == ret)
        {
            LIMDBG("need more data.\n");

            limoi_component_unlock(hComp);
            return 0;
        }

        if (1 == ret)
        {
            ogg_stream_pagein(&ogg->os, &ogg->og);
            while (ogg->header_counts < 3)
            {
                ret = ogg_stream_packetout(&ogg->os, &ogg->op);
                if (0 == ret) break;
                if (0 > ret)
                {
                    LIMERR("oops, corrupted stream?\n");
                    goto bail2;
                }

                ret = vorbis_synthesis_headerin(&ogg->vi, &ogg->vc,&ogg->op);
                if (0 > ret)
                {
                    LIMERR("corrupted secondary header.\n");
                    goto bail2;
                }
                ogg->header_counts++;
            }
        }
    }

    //throw away user comment.
    {
        char ** ptr = ogg->vc.user_comments;
        while (*ptr)
        {
            printf("\t%s\n", *ptr);
            ptr++;
        }
        printf("\n\tbitstream is %d channel, %ldHz\n",
               ogg->vi.channels, ogg->vi.rate);
        printf("\tencoded by: %s\n\n", ogg->vc.vendor);
    }

    if (vorbis_synthesis_init(&ogg->vd, &ogg->vi))
    {
        LIMERR("unable to initialize vorbis.\n");
        goto bail2;
    }
    vorbis_block_init(&ogg->vd, &ogg->vb);

    pcm_param.nChannels = ogg->vi.channels;
    pcm_param.eNumData = OMX_NumericalDataSigned;
    pcm_param.eEndian = OMX_EndianMax; //don't care == host type
    pcm_param.bInterleaved = OMX_TRUE;
    pcm_param.nBitPerSample = 16;
    pcm_param.nSamplingRate = ogg->vi.rate;
    limoi_audio_decoder_set_pcm_mode(hComp, &pcm_param);

    ogg->decoder_inited = 1;
    ogg->decoder_pcm_ready = 0;
    ogg->decoder_first_time = 1;

    limoi_component_unlock(hComp);
    return 0;

 bail2:
    ogg_stream_clear(&ogg->os);
    vorbis_info_clear(&ogg->vi);
    vorbis_comment_clear(&ogg->vc);

 bail:
    limoi_component_unlock(hComp);
    return -1;
}

static OMX_ERRORTYPE ComponentDeInit (
    OMX_HANDLETYPE hComponent)
{
    return limoi_audio_decoder_deinit(hComponent);
}

static OMX_ERRORTYPE ComponentInit (OMX_HANDLETYPE * hComponent, const char * name)
{
    OMX_COMPONENTTYPE * hdl;
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    limoi_audio_decoder_callbacks_t callbacks;
    ogg_private_t * ogg;

    LIMDBG("loading %s component.\n", COMPONENT_NAME);
    ret = limoi_audio_decoder_init(hComponent, sizeof(ogg_private_t),
                                   COMPONENT_NAME, VERSION,
                                   OMX_AUDIO_CodingVORBIS);
    if (OMX_ErrorNone != ret) goto bail;
    hdl = (OMX_COMPONENTTYPE *)*hComponent;

    ogg  = LIMOI_OGG_DEC(*hComponent);

    limoi_component_add_role(*hComponent, ROLE_NAME);

    hdl->ComponentDeInit = ComponentDeInit;

    callbacks.command_prep = NULL;
    callbacks.command_post = cmd_handle;
    callbacks.decode = decode_handle;
    limoi_audio_decoder_set_callbacks(*hComponent, &callbacks);

    LIMDBG("%s component loaded.\n", COMPONENT_NAME);

 bail:
    return ret;
}

lim_omx_component_registration_t ** OMX_ComponentRegistration (int * num)
{
    static char * roles[] = {ROLE_NAME, NULL};
    static lim_omx_component_registration_t comp_reg = {
        COMPONENT_NAME, roles, ComponentInit
    };
    static lim_omx_component_registration_t * regs[] = {&comp_reg};

    *num = 1;
    return regs;
}
