/*
  Copyright (C) 2010 - 2011 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX IL SDL video sink component.

  REVISION:

    1) initial creation. ----------------------------------------- 2010-02-22 MG
*/

#include <stdio.h>
#include <stdlib.h>
#include "limoi-sdl-video-sink.h"

//#undef LIMDBG
#ifndef LIMDBG
#define LIMDBG(fmt, arg...)
#endif

#define COMPONENT_NAME "OMX.limoi.sdl.renderer.video"

static void redraw (sdl_video_renderer_private_t * video_renderer)
{
    SDL_Rect rect;

    ASSERT(video_renderer && video_renderer->yuv);
    rect.x = 0;
    rect.y = 0;
    rect.w = video_renderer->disp_w;
    rect.h = video_renderer->disp_h;
    SDL_DisplayYUVOverlay(video_renderer->yuv, &rect);
}

static void * event_thread (void * arg)
{
    sdl_video_renderer_private_t * video_renderer;
    SDL_Event event;
    //double incr, pos, frac;
    int toggle_full_screen = 0;

    video_renderer = (sdl_video_renderer_private_t *)arg;

    lthread_mutex_lock(&video_renderer->mutex);
    video_renderer->event_state = SE_IDLE;
    lthread_cond_broadcast(&video_renderer->cond);
    lthread_mutex_unlock(&video_renderer->mutex);

    while (1)
    {
        lthread_mutex_lock(&video_renderer->mutex);
        if (video_renderer->event_state == SE_STOP)
        {
            lthread_mutex_unlock(&video_renderer->mutex);
            break;
        }
        lthread_mutex_unlock(&video_renderer->mutex);

        if (!SDL_PollEvent(&event))
        {
            usleep(100000);
            continue;
        }

        lthread_mutex_lock(&video_renderer->mutex);
        switch(event.type)
        {
        case SDL_KEYDOWN:
            switch(event.key.keysym.sym)
            {
            case SDLK_f:
                toggle_full_screen ^= 1;
                if (toggle_full_screen)
                {
                    video_renderer->screen = SDL_SetVideoMode
                        (video_renderer->full_screen_w,
                         video_renderer->full_screen_h, 0,
                         SDL_HWSURFACE | SDL_ASYNCBLIT | SDL_HWACCEL
                         | SDL_FULLSCREEN);
                    video_renderer->disp_w = video_renderer->full_screen_w;
                    video_renderer->disp_h = video_renderer->full_screen_h;
                    break;
                }// else fall through to restore display

            case SDLK_ESCAPE:
                video_renderer->disp_w = video_renderer->prev_disp_w;
                video_renderer->disp_h = video_renderer->prev_disp_h;
                video_renderer->screen = SDL_SetVideoMode
                    (video_renderer->disp_w, video_renderer->disp_h, 0,
                     SDL_HWSURFACE | SDL_ASYNCBLIT | SDL_HWACCEL
                     | SDL_RESIZABLE);
                break;

            default:
                break;
            }
            break;

        case SDL_MOUSEBUTTONDOWN:
            break;

        case SDL_VIDEORESIZE:
            LIMDBG("video resized to %d x %d.\n",
                   event.resize.w, event.resize.h);
            video_renderer->screen = SDL_SetVideoMode
                (event.resize.w, event.resize.h, 0,
                 SDL_HWSURFACE | SDL_ASYNCBLIT | SDL_HWACCEL | SDL_RESIZABLE);
            video_renderer->prev_disp_w =
                video_renderer->disp_w = event.resize.w;
            video_renderer->prev_disp_h =
                video_renderer->disp_h = event.resize.h;
            break;

        case SDL_QUIT:
            //TODO: gracefully shut down.
            exit(0);
            break;

        default:
            break;
        }
        redraw(video_renderer);
        lthread_mutex_unlock(&video_renderer->mutex);
    }

    lthread_mutex_lock(&video_renderer->mutex);
    video_renderer->event_state = SE_INVALID;
    lthread_cond_broadcast(&video_renderer->cond);
    lthread_mutex_unlock(&video_renderer->mutex);

    return NULL;
}

static OMX_ERRORTYPE disable_port (limoi_component_cmd_t * cmd)
{
    sdl_video_renderer_private_t * video_renderer;
    OMX_ERRORTYPE ret = OMX_ErrorNone;

    video_renderer = LIMOI_SDL_VIDEO_RENDERER(cmd->hComponent);

    LIMDBG("%s disable port 0\n", COMPONENT_NAME);
    lthread_mutex_lock(&video_renderer->mutex);

    if (video_renderer->yuv)
    {
        SDL_FreeYUVOverlay(video_renderer->yuv);
        SDL_Quit();

        video_renderer->screen = NULL;
        video_renderer->yuv = NULL;

        while (video_renderer->event_state != SE_INVALID)
        {
            video_renderer->event_state = SE_STOP;
            lthread_cond_broadcast(&video_renderer->cond);
            if (lthread_cond_timedwait
                (&video_renderer->cond, &video_renderer->mutex, 5000))
            {
                LIMERR("sync thread state transition timeout.\n");
                break;
            }
        }
        lthread_mutex_unlock(&video_renderer->mutex);
        lthread_join(video_renderer->event_thread);
        goto bail;
    }

    lthread_mutex_unlock(&video_renderer->mutex);

 bail:
    return ret;
}

static OMX_ERRORTYPE enable_port (limoi_component_cmd_t * cmd)
{
    OMX_PARAM_PORTDEFINITIONTYPE port;
    sdl_video_renderer_private_t * video_renderer;
    OMX_ERRORTYPE ret = OMX_ErrorNone;

    video_renderer = LIMOI_SDL_VIDEO_RENDERER(cmd->hComponent);

    LIMDBG("%s enable port 0\n", COMPONENT_NAME);

    lthread_mutex_lock(&video_renderer->mutex);

    limoi_set_spec_version(&port, sizeof(OMX_PARAM_PORTDEFINITIONTYPE));
    port.nPortIndex = 0;
    limoi_component_get_parameter(cmd->hComponent,
                                  OMX_IndexParamPortDefinition,
                                  &port);
    if (port.format.video.nFrameWidth
        && port.format.video.nFrameHeight
        && video_renderer->yuv == NULL)
    {
        OMX_CONFIG_SCALEFACTORTYPE scale;
        int disp_w = 0;
        int disp_h = 0;

        video_renderer->pic_w = port.format.video.nFrameWidth;
        video_renderer->pic_h = port.format.video.nFrameHeight;
        video_renderer->color_format = port.format.video.eColorFormat;

        limoi_set_spec_version(&scale, sizeof(OMX_CONFIG_SCALEFACTORTYPE));
        scale.nPortIndex = 0;
        if (OMX_ErrorNone == limoi_component_get_parameter
            (cmd->hComponent, OMX_IndexConfigCommonScale, &scale))
        {
            LIMDBG("scale.xWidth = %d xHeight = %d\n",
                   scale.xWidth, scale.xHeight);
            disp_w = port.format.video.nFrameWidth * scale.xWidth / 65536;
            disp_h = port.format.video.nFrameHeight * scale.xHeight / 65536;
        }

        //safe check to undo invalid aspect ratio
        if (!disp_w || !disp_h)
        {
            disp_w = port.format.video.nFrameWidth;
            disp_h = port.format.video.nFrameHeight;
        }

        video_renderer->prev_disp_w =
            video_renderer->disp_w = disp_w;
        video_renderer->prev_disp_h =
            video_renderer->disp_h = disp_h;

        if (SDL_Init(SDL_INIT_VIDEO))
        {
            LIMERR("unable to initialize SDL - %s\n", SDL_GetError());
            ret = OMX_ErrorUndefined;
            goto bail;
        }
        SDL_WM_SetCaption("LIM OpenMAX Player", "LIM");

        {
            const SDL_VideoInfo *vi = SDL_GetVideoInfo();
            LIMDBG("full screen: %d x %d\n", vi->current_w, vi->current_h);
            video_renderer->full_screen_w = vi->current_w;
            video_renderer->full_screen_h = vi->current_h;
        }

        SDL_EventState(SDL_ACTIVEEVENT, SDL_IGNORE);
        SDL_EventState(SDL_MOUSEMOTION, SDL_IGNORE);
        SDL_EventState(SDL_SYSWMEVENT, SDL_IGNORE);
        SDL_EventState(SDL_USEREVENT, SDL_IGNORE);

        if (video_renderer->screen == NULL)
        {
            video_renderer->screen = SDL_SetVideoMode
                (disp_w, disp_h, 0,
                 SDL_HWSURFACE | SDL_ASYNCBLIT | SDL_HWACCEL | SDL_RESIZABLE);
            if (NULL == video_renderer->screen)
            {
                LIMWARN("%s unable to set video mode: %d x %d\n",
                        COMPONENT_NAME, disp_w, disp_h);
            }
        }
        if (video_renderer->screen && NULL == video_renderer->yuv)
        {
            if (video_renderer->color_format != OMX_COLOR_FormatYUV420Planar)
            {
                LIMWARN("%s format must be 420 plannr for now.\n",
                        COMPONENT_NAME);
                goto bail;
            }

            video_renderer->yuv = SDL_CreateYUVOverlay
                (video_renderer->pic_w, video_renderer->pic_h,
                 SDL_YV12_OVERLAY, video_renderer->screen);

            LIMDBG("%s SDL setup:\n"
                   "\tinput  w x h  : %d x %d\n"
                   "\tyuv    w x h  : %d x %d\n"
                   "\tplanes        : %d\n"
                   "\tformat        : %x\n"
                   "\tpitch[0]      : %d\n"
                   "\tpitch[1]      : %d\n"
                   "\tpitch[2]      : %d\n", COMPONENT_NAME,
                   video_renderer->pic_w, video_renderer->pic_h,
                   video_renderer->yuv->w, video_renderer->yuv->h,
                   video_renderer->yuv->planes,
                   video_renderer->yuv->format,
                   video_renderer->yuv->pitches[0],
                   video_renderer->yuv->pitches[1],
                   video_renderer->yuv->pitches[2]);

            if (NULL == video_renderer->yuv)
            {
                LIMWARN("%s unable to create overlay: %d x %d\n",
                        COMPONENT_NAME,
                        video_renderer->pic_w, video_renderer->pic_h);
            }

            video_renderer->event_state = SE_INVALID;
            if (lthread_create
                (&video_renderer->event_thread,
                 event_thread, video_renderer, 1))
            {
                LIMERR("unable to create SDL event thread.\n");
                ret = OMX_ErrorUndefined;
                goto bail;
            }

            while (video_renderer->event_state != SE_IDLE)
            {
                if (lthread_cond_timedwait
                    (&video_renderer->cond, &video_renderer->mutex, 5000))
                {
                    LIMERR("SDL event thread state transition timeout.\n");
                    goto bail;
                }
            }
        }
    }

 bail:
    lthread_mutex_unlock(&video_renderer->mutex);
    return ret;
}

static OMX_ERRORTYPE set_state_post (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    OMX_HANDLETYPE hComp = cmd->hComponent;
    OMX_STATETYPE next_state = cmd->nParam1;
    OMX_STATETYPE state = limoi_component_state(hComp);
    sdl_video_renderer_private_t * video_renderer =
        LIMOI_SDL_VIDEO_RENDERER(hComp);

    LIMDBG("%s callback post set state from %s to %s.\n", COMPONENT_NAME,
           limoi_omx_state_string(state), limoi_omx_state_string(next_state));

    switch (state)
    {
    case OMX_StateLoaded :

        // transition to StateIdle
        video_renderer->screen = NULL;
        video_renderer->yuv = NULL;

        {
            OMX_PARAM_PORTDEFINITIONTYPE port;
            limoi_set_spec_version(&port, sizeof(OMX_PARAM_PORTDEFINITIONTYPE));
            port.nPortIndex = 0;
            limoi_component_get_parameter(hComp,
                                          OMX_IndexParamPortDefinition,
                                          &port);
            if (port.bEnabled)
            {
                ret = enable_port(cmd);
            }
        }
        break;

    case OMX_StateIdle :
        if (OMX_StateLoaded == next_state)
        {
            disable_port(cmd);
        }
        break;

    default :
        break;
    }

    return ret;
}

static OMX_ERRORTYPE prep_cmd_handle (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    //sdl_video_renderer_private_t * video_renderer;
    //video_renderer = LIMOI_SDL_VIDEO_RENDERER(cmd->hComponent);

    LIMDBG("%s prep command handler.\n", COMPONENT_NAME);

    switch (cmd->Cmd)
    {
    case OMX_CommandPortDisable :
        ret = disable_port(cmd);
        break;

    case OMX_CommandPortEnable :
        ret = enable_port(cmd);
        break;

    default :
        break;
    }

    return ret;
}

static OMX_ERRORTYPE post_cmd_handle (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    //sdl_video_renderer_private_t * video_renderer;
    //video_renderer = LIMOI_SDL_VIDEO_RENDERER(cmd->hComponent);

    LIMDBG("%s post command handler.\n", COMPONENT_NAME);

    switch (cmd->Cmd)
    {
    case OMX_CommandStateSet :
        ret = set_state_post(cmd);
        break;

    default :
        break;
    }

    return ret;
}

static int render_handle (OMX_HANDLETYPE hComp, OMX_BUFFERHEADERTYPE * buf)
{
    sdl_video_renderer_private_t * video_renderer;

    video_renderer = LIMOI_SDL_VIDEO_RENDERER(hComp);

    lthread_mutex_lock(&video_renderer->mutex);

    if (video_renderer->yuv)
    {
        //AVPicture pict;
        SDL_Rect rect;
        uint8_t * y;
        uint8_t * u;
        uint8_t * v;
        unsigned int bytes;
        int pitch0, pitch1;

        //align pitch on 16-pixel boundary.
        pitch0 = (video_renderer->pic_w + 15) & ~15;
        pitch1 = pitch0 / 2;

        //FIXME: hard-coded to be YUV420 plannar
        y = buf->pBuffer;
        u = y + pitch0 * video_renderer->pic_h;
        v = u + pitch1 * video_renderer->pic_h / 2;

        SDL_LockYUVOverlay(video_renderer->yuv);

        if (video_renderer->yuv->pitches[0] != pitch0
            || video_renderer->yuv->pitches[1] != pitch1
            || video_renderer->yuv->pitches[2] != pitch1)
        {
            int hh;
            uint8_t * y2;
            uint8_t * u2;
            uint8_t * v2;

            LIMINFO("%s unexpected pitch mismatch, line copying.\n"
                    "\tinput  w x h  : %d x %d\n"
                    "\tpitch[0]      : %d\n"
                    "\tpitch[1]      : %d\n"
                    "\tpitch[2]      : %d\n", COMPONENT_NAME,
                    video_renderer->pic_w, video_renderer->pic_h,
                    video_renderer->yuv->pitches[0],
                    video_renderer->yuv->pitches[1],
                    video_renderer->yuv->pitches[2]);
            y2 = video_renderer->yuv->pixels[0];
            u2 = video_renderer->yuv->pixels[2];
            v2 = video_renderer->yuv->pixels[1];

            for (hh = 0; hh < video_renderer->pic_h; hh++)
            {
                memcpy(y2, y, video_renderer->yuv->pitches[0]);
                y2 += video_renderer->yuv->pitches[0];
                y += pitch0;
            }
            for (hh = 0; hh < video_renderer->pic_h / 2; hh++)
            {
                memcpy(u2, u, video_renderer->yuv->pitches[2]);
                u2 += video_renderer->yuv->pitches[2];
                u += pitch1;
            }
            for (hh = 0; hh < video_renderer->pic_h / 2; hh++)
            {
                memcpy(v2, v, video_renderer->yuv->pitches[1]);
                v2 += video_renderer->yuv->pitches[1];
                v += pitch1;
            }
        } else
        {
            bytes = pitch0 * video_renderer->pic_h;
            memcpy(video_renderer->yuv->pixels[0], y, bytes);

            bytes = pitch1 * video_renderer->pic_h / 2;
            memcpy(video_renderer->yuv->pixels[2], u, bytes);

            bytes = pitch1 * video_renderer->pic_h / 2;
            memcpy(video_renderer->yuv->pixels[1], v, bytes);
        }
        SDL_UnlockYUVOverlay(video_renderer->yuv);

        LIMDBG("%s displayed frame pts: %"PRId64"\n",
               COMPONENT_NAME, buf->nTimeStamp);
        rect.x = 0;
        rect.y = 0;
        rect.w = video_renderer->disp_w;
        rect.h = video_renderer->disp_h;
        SDL_DisplayYUVOverlay(video_renderer->yuv, &rect);
    }

    lthread_mutex_unlock(&video_renderer->mutex);
    buf->nFilledLen = 0;

    return 0;
}

static OMX_ERRORTYPE ComponentDeInit (
    OMX_HANDLETYPE hComponent)
{
    sdl_video_renderer_private_t * video_renderer;

    video_renderer = LIMOI_SDL_VIDEO_RENDERER(hComponent);

    // sdl_video_renderer destructor.
    lthread_mutex_destroy(&video_renderer->mutex);
    lthread_cond_destroy(&video_renderer->cond);

    // parent destructor.
    return limoi_video_renderer_deinit(hComponent);
}

static OMX_ERRORTYPE ComponentInit (OMX_HANDLETYPE * hComponent, const char * name)
{
    OMX_COMPONENTTYPE * hdl;
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    limoi_video_renderer_callbacks_t callbacks;
    sdl_video_renderer_private_t * video_renderer;

    LIMDBG("loading %s component.\n", COMPONENT_NAME);
    ret = limoi_video_renderer_init
        (hComponent, sizeof(sdl_video_renderer_private_t),
         COMPONENT_NAME, VERSION);
    if (OMX_ErrorNone != ret) goto bail;
    hdl = (OMX_COMPONENTTYPE *)*hComponent;

    video_renderer = LIMOI_SDL_VIDEO_RENDERER(*hComponent);
    lthread_mutex_init(&video_renderer->mutex);
    lthread_cond_init(&video_renderer->cond);

    hdl->ComponentDeInit = ComponentDeInit;

    callbacks.command_prep = prep_cmd_handle;
    callbacks.command_post = post_cmd_handle;
    callbacks.render = render_handle;
    limoi_video_renderer_set_callbacks(*hComponent, &callbacks);

    LIMDBG("%s component loaded.\n", COMPONENT_NAME);

 bail:
    return ret;
}

lim_omx_component_registration_t ** OMX_ComponentRegistration (int * num)
{
    static char * roles[] = {LIMOI_VIDEO_RENDERER_ROLE, NULL};
    static lim_omx_component_registration_t comp_reg = {
        COMPONENT_NAME, roles, ComponentInit
    };
    static lim_omx_component_registration_t * regs[] = {&comp_reg};

    *num = 1;
    return regs;
}
