/*
  Copyright (C) 2010 - 2011 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX IL video scheduler component.

  REVISION:

    1) initial creation. ----------------------------------------- 2010-02-27 MG
*/

#include <stdio.h>
#include <stdlib.h>
#include "OMX_Core.h"
#include "OMX_Types.h"
#include "OMX_Component.h"
#include "limoi-vscheduler.h"

//#undef LIMDBG
#ifndef LIMDBG
#define LIMDBG(fmt, arg...)
#endif

//define the following to enable buffer sharing support.
#define BUFFER_SHARE_ENABLED

#define COMPONENT_NAME "OMX.limoi.video_scheduler"
#define ROLE_NAME "video_scheduler.binary"

static void reset_pool (OMX_HANDLETYPE hComp)
{
    video_scheduler_private_t * video_scheduler = VIDEO_SCHEDULER(hComp);
    OMX_BUFFERHEADERTYPE ** in;
    int ii;

    for (ii = 0; ii < LIMOI_MAX_CLOCK_PORT_BUFFERS; ii++)
    {
        if (video_scheduler->hold_buf_pool[ii].priv)
        {
            in = video_scheduler->hold_buf_pool[ii].priv;
            LIMDBG("%s release port %u buffer %p\n", COMPONENT_NAME,
                   (*in)->nInputPortIndex, in);
            limoi_component_unlock(hComp);
            limoi_component_release_port_buffer
                (hComp, limoi_component_get_port
                 (hComp, (*in)->nInputPortIndex), in);
            limoi_component_lock(hComp);
            video_scheduler->hold_buf_pool[ii].priv = NULL;
        }
    }
    video_scheduler->hold_buf_count = 0;
}

static void * add_to_pool (
    OMX_HANDLETYPE hComp,
    void * buf)
{
    video_scheduler_private_t * video_scheduler = VIDEO_SCHEDULER(hComp);
    int ii;
    static uint32_t running_id;

    if (video_scheduler->hold_buf_count >= LIMOI_MAX_CLOCK_PORT_BUFFERS)
        return NULL;

    for (ii = 0; ii < LIMOI_MAX_CLOCK_PORT_BUFFERS; ii++)
    {
        if (!video_scheduler->hold_buf_pool[ii].priv)
        {
            video_scheduler->hold_buf_pool[ii].priv = buf;
            video_scheduler->hold_buf_pool[ii].id = running_id;
            video_scheduler->buf_signature[ii] = running_id;
            video_scheduler->hold_buf_count++;
            running_id++;
            return &video_scheduler->buf_signature[ii];
        }
    }
    return NULL;
}

static int remove_from_pool (
    OMX_HANDLETYPE hComp,
    vsm_rqst_priv_t * rqst)
{
    video_scheduler_private_t * video_scheduler = VIDEO_SCHEDULER(hComp);

    if (!video_scheduler->hold_buf_count) return -1;
    rqst->priv = NULL;
    video_scheduler->hold_buf_count--;
    return 0;
}

static OMX_ERRORTYPE set_state_prep (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    OMX_HANDLETYPE hComp = cmd->hComponent;
    OMX_STATETYPE next_state = cmd->nParam1;
    OMX_STATETYPE state = limoi_component_state(hComp);

    LIMDBG("%s callback prep set state from %s to %s.\n", COMPONENT_NAME,
           limoi_omx_state_string(state), limoi_omx_state_string(next_state));

    switch (state)
    {
    case OMX_StateExecuting :
        if (OMX_StateIdle == next_state)
        {
            reset_pool(hComp);
        }
        break;

    case OMX_StatePause :
        if (OMX_StateIdle == next_state)
        {
            reset_pool(hComp);
        }
        break;

    default :
        break;
    }

    return ret;
}

static OMX_ERRORTYPE set_state_post (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    OMX_HANDLETYPE hComp = cmd->hComponent;
    OMX_STATETYPE next_state = cmd->nParam1;
    OMX_STATETYPE state = limoi_component_state(hComp);
    video_scheduler_private_t * video_scheduler = VIDEO_SCHEDULER(hComp);

    LIMDBG("%s callback post set state from %s to %s.\n", COMPONENT_NAME,
           limoi_omx_state_string(state), limoi_omx_state_string(next_state));

    switch (state)
    {
    case OMX_StateLoaded :

        // transition to StateIdle
        video_scheduler->reference_clock = 0;

        video_scheduler->clock_scale_set = 0;
        //setting this value to -1 triggers a mannual clock state query.
        video_scheduler->clock_running = -1;
        break;

    default :
        break;
    }

    return ret;
}

static OMX_ERRORTYPE prep_cmd_handle (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    OMX_HANDLETYPE hComp = cmd->hComponent;

    LIMDBG("%s prep command handler.\n", COMPONENT_NAME);
    limoi_component_lock(hComp);
    switch (cmd->Cmd)
    {
    case OMX_CommandStateSet :
        ret = set_state_prep(cmd);
        break;

    case OMX_CommandFlush :
        //reset buffer pool upon input port flushing.
        if (cmd->nParam1 == 0) reset_pool(hComp);
        break;

    default :
        break;
    }
    limoi_component_unlock(hComp);

    return ret;
}

static OMX_ERRORTYPE post_cmd_handle (limoi_component_cmd_t * cmd)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    OMX_HANDLETYPE hComp = cmd->hComponent;

    LIMDBG("%s post command handler.\n", COMPONENT_NAME);
    limoi_component_lock(hComp);
    switch (cmd->Cmd)
    {
    case OMX_CommandStateSet :
        ret = set_state_post(cmd);
        break;

    default :
        break;
    }
    limoi_component_unlock(hComp);

    return ret;
}

static int clock_handle (
    OMX_HANDLETYPE hComp,
    limoi_port_t * in_port,
    limoi_port_t * out_port,
    OMX_BUFFERHEADERTYPE * out_buf,
    OMX_HANDLETYPE clock_tunnel_comp,
    OMX_BUFFERHEADERTYPE * clock_buf)
{
    video_scheduler_private_t * video_scheduler = VIDEO_SCHEDULER(hComp);
    OMX_TIME_MEDIATIMETYPE * mtype;
    OMX_BUFFERHEADERTYPE ** in;
    vsm_rqst_priv_t * rqst = NULL;

    LIMDBG("%s: clock handle.\n", COMPONENT_NAME);

    mtype = (OMX_TIME_MEDIATIMETYPE *)clock_buf->pBuffer;

    LIMDBG("%s clock update : %d\n"
           "\tprivate       : %p\n"
           "\ttimestamp     : %"PRId64"\n"
           "\toffset        : %"PRId64"\n"
           "\twall time     : %"PRId64"\n"
           "\tscale         : %d\n"
           "\tstate         : %d\n",
           COMPONENT_NAME,
           mtype->eUpdateType,
           mtype->pClientPrivate,
           mtype->nMediaTimestamp,
           mtype->nOffset,
           mtype->nWallTimeAtMediaTime,
           mtype->xScale,
           mtype->eState);

    limoi_component_lock(hComp);

    switch (mtype->eUpdateType)
    {
    case OMX_TIME_UpdateScaleChanged :
        video_scheduler->clock_scale = mtype->xScale;
        clock_buf->nFilledLen = 0;
        break;

    case OMX_TIME_UpdateClockStateChanged :
        {
            OMX_TIME_CONFIG_ACTIVEREFCLOCKTYPE ref;

            LIMDBG("%s clock state change.\n", COMPONENT_NAME);
            limoi_set_spec_version
                (&ref, sizeof(OMX_TIME_CONFIG_ACTIVEREFCLOCKTYPE));

            limoi_component_unlock(hComp);
            if (OMX_ErrorNone != OMX_GetConfig
                (clock_tunnel_comp,
                 OMX_IndexConfigTimeActiveRefClock, &ref))
            {
                ref.eClock = OMX_TIME_RefClockNone;
            }
            limoi_component_lock(hComp);

            if (OMX_TIME_RefClockVideo == ref.eClock)
            {
                video_scheduler->reference_clock = 1;
            }
        }
        video_scheduler->clock_running
            = (mtype->eState == OMX_TIME_ClockStateRunning)? 1 : 0;
        clock_buf->nFilledLen = 0;
        break;

    default :
        LIMWARN("%s unhandled clock event: %d.\n", COMPONENT_NAME,
                mtype->eUpdateType);
        clock_buf->nFilledLen = 0;
        break;

    case OMX_TIME_UpdateRequestFulfillment :

        LIMDBG("%s request fulfillment: %p %"PRId64".\n", COMPONENT_NAME,
               mtype->pClientPrivate, mtype->nMediaTimestamp);
        if (mtype->pClientPrivate >=
            (void *)&video_scheduler->buf_signature[0]
            && mtype->pClientPrivate <
            (void *)&video_scheduler->buf_signature
            [LIMOI_MAX_CLOCK_PORT_BUFFERS])
        {
            rqst = &video_scheduler->hold_buf_pool[0]
                + (mtype->pClientPrivate
                   - (void *)&video_scheduler->buf_signature[0])
                / sizeof(uint32_t);
            in = rqst->priv;
            if (rqst->id != *(uint32_t *)mtype->pClientPrivate || !in || !(*in)
                || (*in)->nTimeStamp != mtype->nMediaTimestamp)
            {
                rqst = NULL;
                LIMINFO("%s request seems no longer valid.\n", COMPONENT_NAME);
            }
        }

        if (!rqst)
        {
            clock_buf->nFilledLen = 0;
            LIMINFO("%s request %p is no longer in pool.\n",
                    COMPONENT_NAME, mtype->pClientPrivate);
        } else if (out_buf)
        {
            clock_buf->nFilledLen = 0;
            ASSERT(out_buf->nAllocLen >= (*in)->nFilledLen);
            if ((*in)->nFilledLen)
            {
#ifdef BUFFER_SHARE_ENABLED
                if (!limoi_component_port_buffer_share_enabled(out_port))
#endif
                memcpy(out_buf->pBuffer, (*in)->pBuffer, (*in)->nFilledLen);

                out_buf->nTimeStamp = (*in)->nTimeStamp;
            }

            if ((*in)->nFlags & OMX_BUFFERFLAG_EOS)
            {
                LIMDBG("%s EOS delivering at %"PRId64"\n", COMPONENT_NAME,
                       (*in)->nTimeStamp);

                out_buf->nFlags |= OMX_BUFFERFLAG_EOS;
                limoi_component_deliver_event
                    (hComp, OMX_EventBufferFlag, 0, (*in)->nFlags, NULL);
                limoi_component_deliver_event
                    (hComp, OMX_EventBufferFlag, 1, out_buf->nFlags, NULL);
            }

            out_buf->nFilledLen = (*in)->nFilledLen;

            if (video_scheduler->reference_clock)
            {
                OMX_TIME_CONFIG_TIMESTAMPTYPE ts;
                limoi_set_spec_version
                    (&ts, sizeof(OMX_TIME_CONFIG_TIMESTAMPTYPE));
                ts.nPortIndex = clock_buf->nOutputPortIndex;
                ts.nTimestamp = (*in)->nTimeStamp;

                LIMDBG("%s set reference clock as %"PRId64"\n",
                       COMPONENT_NAME, ts.nTimestamp);

                limoi_component_unlock(hComp);
                OMX_SetConfig(clock_tunnel_comp,
                              OMX_IndexConfigTimeCurrentVideoReference, &ts);
                limoi_component_lock(hComp);
            }

            (*in)->nFilledLen = 0;

#ifdef BUFFER_SHARE_ENABLED
            limoi_component_share_port_buffer
                (hComp, in_port, out_port, in, out_buf);
#endif
            remove_from_pool(hComp, rqst);
            limoi_component_unlock(hComp);

            LIMDBG("%s put back input buffer: %p.\n", COMPONENT_NAME, in);
            limoi_component_put_port_buffer(hComp, in_port, in);
            LIMDBG("video pts scheduled: %"PRId64"\n", (*in)->nTimeStamp);

            limoi_component_lock(hComp);
        }
        LIMDBG("%s request fulfillment complete.\n", COMPONENT_NAME);
        break;
    }
    limoi_component_unlock(hComp);

    return 0;
}

//manually query clock state and set current clock scale.
//caller must hold video_scheduler->mutex
//required in case component misses the clock broadcasting.
static inline void query_clock_state (
    OMX_HANDLETYPE hComp,
    OMX_HANDLETYPE clock_tunnel_comp)
{
    OMX_TIME_CONFIG_ACTIVEREFCLOCKTYPE ref;
    OMX_TIME_CONFIG_SCALETYPE scale;
    OMX_TIME_CONFIG_CLOCKSTATETYPE state;
    video_scheduler_private_t * video_scheduler = VIDEO_SCHEDULER(hComp);

    limoi_set_spec_version(&ref, sizeof(OMX_TIME_CONFIG_ACTIVEREFCLOCKTYPE));
    limoi_set_spec_version(&scale, sizeof(OMX_TIME_CONFIG_SCALETYPE));
    limoi_set_spec_version(&state, sizeof(OMX_TIME_CONFIG_CLOCKSTATETYPE));

    limoi_component_unlock(hComp);

    if (OMX_ErrorNone != OMX_GetConfig
        (clock_tunnel_comp,
         OMX_IndexConfigTimeActiveRefClock, &ref))
    {
        ref.eClock = OMX_TIME_RefClockNone;
    }

    if (OMX_ErrorNone != OMX_GetConfig
        (clock_tunnel_comp, OMX_IndexConfigTimeScale, &scale))
    {
        scale.xScale = 0;
    }

    if (OMX_ErrorNone != OMX_GetConfig
        (clock_tunnel_comp, OMX_IndexConfigTimeClockState, &state))
    {
        state.eState = OMX_TIME_ClockStateStopped;
    } else
    {
        LIMDBG("%s current clock state\n"
               "\teState     : %d\n"
               "\tnStartTime : %"PRId64"\n"
               "\tnOffset    : %"PRId64"\n"
               "\tnWaitMask  : %x\n", COMPONENT_NAME,
               state.eState,
               state.nStartTime,
               state.nOffset,
               state.nWaitMask);
    }

    limoi_component_lock(hComp);

    if (OMX_TIME_RefClockVideo == ref.eClock)
    {
        video_scheduler->reference_clock = 1;
    }
    video_scheduler->clock_scale = scale.xScale;
    video_scheduler->clock_running
        = (state.eState == OMX_TIME_ClockStateRunning)? 1 : 0;
    video_scheduler->clock_scale_set = 1;
}

static int schedule_handle (
    OMX_HANDLETYPE hComp,
    limoi_port_t * in_port,
    OMX_BUFFERHEADERTYPE ** in_buf,
    OMX_BUFFERHEADERTYPE * out_buf,
    OMX_HANDLETYPE clock_tunnel_comp,
    int clock_tunnel_port,
    int clock_port)
{
    video_scheduler_private_t * video_scheduler = VIDEO_SCHEDULER(hComp);
    int ret = 0;

    LIMDBG("%s scheduling buffer %p\n", COMPONENT_NAME, *in_buf);
    limoi_component_lock(hComp);

    if (!video_scheduler->clock_scale_set
        || video_scheduler->clock_running == -1)
    {
        query_clock_state(hComp, clock_tunnel_comp);
    }

    if (video_scheduler->clock_scale && video_scheduler->clock_running)
    {
        //in this case, client relies on scheduler to internally
        //schedule buffer, thus the clock port must be tunneled.
        OMX_TIME_CONFIG_MEDIATIMEREQUESTTYPE rqst;
        void * pool_rqst;
        int eos_retry_count = 10;

        LIMDBG("%s scheduling pts %"PRId64".\n", COMPONENT_NAME,
               (*in_buf)->nTimeStamp);
        ASSERT(clock_tunnel_comp && clock_tunnel_port != 0x7fffffff);

        //hold port buffer before adding buffer to pool.
        limoi_component_unlock(hComp);
        limoi_component_hold_port_buffer(hComp, in_port, in_buf);
        limoi_component_lock(hComp);

        pool_rqst = add_to_pool(hComp, in_buf);
        limoi_set_spec_version
            (&rqst, sizeof(OMX_TIME_CONFIG_MEDIATIMEREQUESTTYPE));
        rqst.nPortIndex = clock_tunnel_port;
        rqst.pClientPrivate = pool_rqst;
        rqst.nMediaTimestamp = (*in_buf)->nTimeStamp;
        rqst.nOffset = 1000; //schedule 1ms ahead.

    eos_retry:
        if ((*in_buf)->nFlags & OMX_BUFFERFLAG_EOS)
        {
            OMX_TIME_CONFIG_TIMESTAMPTYPE ts;

            limoi_set_spec_version
                (&ts, sizeof(OMX_TIME_CONFIG_TIMESTAMPTYPE));
            ts.nPortIndex = OMX_ALL;

            LIMDBG("%s EOS detected at %"PRId64"\n", COMPONENT_NAME,
                   (*in_buf)->nTimeStamp);

            if (OMX_ErrorNone !=
                OMX_GetConfig(clock_tunnel_comp,
                              OMX_IndexConfigTimeCurrentMediaTime, &ts))
            {
                LIMINFO("%s unable to get current media time.\n",
                        COMPONENT_NAME);
            } else
            {
                LIMDBG("%s current media time: %"PRId64"\n",
                       COMPONENT_NAME, ts.nTimestamp);
                //EOS delivery attempt in 10 miliseconds.
                rqst.nMediaTimestamp =
                (*in_buf)->nTimeStamp = ts.nTimestamp + 10000;
            }
        }

        limoi_component_unlock(hComp);
        if (OMX_ErrorNone != OMX_SetConfig
            (clock_tunnel_comp, OMX_IndexConfigTimeMediaTimeRequest, &rqst))
        {
            vsm_rqst_priv_t * priv;

            if ((*in_buf)->nFlags & OMX_BUFFERFLAG_EOS)
            {
                //retry the scheduling and delivery of EOS.
                if (eos_retry_count--)
                {
                    LIMINFO("%s missing EOS delivery, retry.\n",
                            COMPONENT_NAME);
                    limoi_component_lock(hComp);
                    goto eos_retry;
                } else
                {
                    LIMERR("%s missing EOS delivery, retry failed.\n",
                           COMPONENT_NAME);
                }
            }

            LIMINFO("%s media time request failed.\n", COMPONENT_NAME);

            //upon failure, dereference from pool first.
            limoi_component_lock(hComp);
            priv = &video_scheduler->hold_buf_pool[0]
                + (pool_rqst - (void *)&video_scheduler->buf_signature[0])
                / sizeof(uint32_t);

            remove_from_pool(hComp, priv);
            limoi_component_unlock(hComp);

            //now it is safe to release this buffer.
            limoi_component_release_port_buffer(hComp, in_port, in_buf);
        }
        limoi_component_lock(hComp);
    } else
    {
        // input is not consumed.
        LIMDBG("%s input is not consumed.\n", COMPONENT_NAME);
        ret = 1;
    }
    LIMDBG("%s scheduling buffer %p done\n", COMPONENT_NAME, *in_buf);
    limoi_component_unlock(hComp);

    return ret;
}

static OMX_ERRORTYPE ComponentDeInit (
    OMX_HANDLETYPE hComp)
{
    return limoi_video_scheduler_deinit(hComp);
}

static OMX_ERRORTYPE ComponentInit (OMX_HANDLETYPE * hComp, const char * name)
{
    OMX_COMPONENTTYPE * hdl;
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    limoi_video_scheduler_callbacks_t callbacks;
    video_scheduler_private_t * video_scheduler;
#ifdef BUFFER_SHARE_ENABLED
    limoi_buffer_share_policy_t policy;
#endif

    LIMDBG("loading %s component.\n", COMPONENT_NAME);
    ret = limoi_video_scheduler_init(hComp, sizeof(video_scheduler_private_t),
                             COMPONENT_NAME, VERSION);
    if (OMX_ErrorNone != ret) goto bail;
    hdl = (OMX_COMPONENTTYPE *)*hComp;

    video_scheduler = VIDEO_SCHEDULER(*hComp);

    limoi_component_add_role(*hComp, ROLE_NAME);

    hdl->ComponentDeInit = ComponentDeInit;

    callbacks.command_prep = prep_cmd_handle;
    callbacks.command_post = post_cmd_handle;
    callbacks.schedule = schedule_handle;
    callbacks.clock = clock_handle;
    limoi_video_scheduler_set_callbacks(*hComp, &callbacks);

#ifdef BUFFER_SHARE_ENABLED
    //output port shares buffer from input port.
    policy.num_share_pairs = 1;
    policy.ports[0] = 0;
    policy.ports[1] = 1;
    limoi_component_set_buffer_share_policy(*hComp, &policy);
#endif

    LIMDBG("%s component loaded.\n", COMPONENT_NAME);

 bail:
    return ret;
}

lim_omx_component_registration_t ** OMX_ComponentRegistration (int * num)
{
    static char * roles[] = {ROLE_NAME, NULL};
    static lim_omx_component_registration_t comp_reg = {
        COMPONENT_NAME, roles, ComponentInit
    };
    static lim_omx_component_registration_t * regs[] = {&comp_reg};

    *num = 1;
    return regs;
}
