/*
  Copyright (C) 2009 - 2011 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX IL core content pipe module.

  REVISION:

    2) pipe read-write access support. --------------------------- 2011-01-31 MG
    1) initial creation. ----------------------------------------- 2009-11-24 MG
*/

#include "contentpipe.h"

static content_access_itf_t * CONTENT_ACCESS_ITFS[] = {
    &content_access_itf_file,
    NULL
};

#define LIMS_LOOP_RUNNING  1
#define LIMS_LOOP_STOPPING 2
#define LIMS_LOOP_STOPPED  3

/* +----+------+--------------------+-----------------------+------------+
   |    |======|####################|=======================|            |
   +----^------^--------------------^-----------------------^------------+
        |      |                    |                       |
        tail   wr_tail              mark                    head
        |      |<--bytes_writable-->|<---bytes_readable---->|
        |<----------------bytes_available------------------>|
*/
typedef struct {
    int loop;
    int eof;
    int bytes_requested; /* bytes requested by client. */
    int bytes_available; /* bytes available within pipe buffer. */
    int bytes_readable; /* bytes available for reading. */
    int bytes_writable; /* bytes available for reading. */
    int head; /* head within pipe buffer. */
    int tail; /* tail within pipe buffer. */
    int mark; /* current access point within pipe buffer. */
    int wr_tail; /* writing tail within pipe buffer. */
    off_t content_offset; /* actual content physical offset. */
    off_t offset; /* absolute stream offset in pipe. */

    lthread_mutex_t mutex;
    lthread_cond_t cond;
    lthread_t thread;
    void * buf; /* internal pipe buffer. */
    void * rd_buf; /* read buffer returned by ReadBuffer. */
    int rd_buf_bytes; /* read buffer bytes. */
    void * rd_buf_a; /* allocated read buffer. */
    int rd_buf_a_bytes; /* allocated read buffer size in bytes. */
    void * wr_buf; /* write buffer returned by GetWriteBuffer. */
    int wr_buf_bytes; /* write buffer bytes. */
    void * wr_buf_a; /* allocated write buffer. */
    int wr_buf_a_bytes; /* allocated write buffer size in bytes. */
    CP_ACCESSTYPE access_type;
    CPresult (*callback) (CP_EVENTTYPE eEvent, CPuint iParam);

    content_access_itf_t * itf;
    content_handle_t content;
    content_info_t info;
} lim_stream_t;

static void * lims_thread_read (void * arg)
{
    lim_stream_t * lims = (lim_stream_t *)arg;
    CPuint bytes;
    CP_EVENTTYPE event = CP_EventMax;
    CPuint param;

    while (1)
    {
        lthread_mutex_lock(&lims->mutex);
        while ((lims->bytes_readable > lims->info.circular_buffer_bytes / 2
                || lims->eof) && lims->loop == LIMS_LOOP_RUNNING)
        {
            /*
            LIMDBG("file stream read controls:\n"
                   "\toffset    : %lu\n"
                   "\thead      : %d\n"
                   "\ttail      : %d\n"
                   "\tmark      : %d\n"
                   "\tavailable : %d\n"
                   "\ttotal     : %d\n",
                   lims->offset, lims->head, lims->tail,
                   lims->mark, lims->bytes_available,
                   lims->info.circular_buffer_bytes);
            */
            if (lims->eof && lims->bytes_requested)
            {
                if (lims->callback)
                {
                    lims->bytes_requested = 0;
                    param = lims->bytes_readable;

                    lthread_mutex_unlock(&lims->mutex);
                    lims->callback(CP_BytesAvailable, param);
                    lthread_mutex_lock(&lims->mutex);
                    //mutex has been relinquished above, do not go to wait
                    //before rechecking conditions first.
                    continue;
                }
            }

            lthread_cond_wait(&lims->cond, &lims->mutex);
        }

        if (lims->loop != LIMS_LOOP_RUNNING)
        {
            lthread_mutex_unlock(&lims->mutex);
            event = CP_EventMax;
            break;
        }

        if (lims->bytes_available == lims->info.circular_buffer_bytes)
        {
            lims->bytes_available -= lims->info.commit_step_bytes;
            lims->tail += lims->info.commit_step_bytes;
            if (lims->tail >= lims->info.circular_buffer_bytes)
                lims->tail -= lims->info.circular_buffer_bytes;
        }

        bytes = lims->info.circular_buffer_bytes - lims->bytes_available;
        bytes = (bytes > lims->info.commit_step_bytes)?
            lims->info.commit_step_bytes : bytes;
        bytes = (bytes > lims->info.circular_buffer_bytes - lims->head)?
            lims->info.circular_buffer_bytes - lims->head : bytes;

        if (lims->itf->read(lims->content, lims->buf + lims->head, &bytes))
        {
            //error, callback to client.
            lthread_mutex_unlock(&lims->mutex);
            if (lims->callback)
            {
                event = CP_PipeDisconnected;
                param = 0;
            }
            break;
        } else if (bytes == 0)
        {
            lims->eof = 1;
        } else
        {
            lims->bytes_available += bytes;
            lims->bytes_readable += bytes;
            lims->head += bytes;
            ASSERT(lims->head <= lims->info.circular_buffer_bytes);

            if (lims->head == lims->info.circular_buffer_bytes)
                lims->head = 0;
        }

        if ((lims->bytes_requested <= lims->bytes_readable)
            && lims->bytes_requested)
        {
            if (lims->callback)
            {
                lims->bytes_requested = 0;
                param = lims->bytes_readable;

                lthread_mutex_unlock(&lims->mutex);
                lims->callback(CP_BytesAvailable, param);
                continue;
            }
        }
        lthread_mutex_unlock(&lims->mutex);
    }

    if (CP_EventMax != event) lims->callback(event, param);

    lthread_mutex_lock(&lims->mutex);
    lims->loop = LIMS_LOOP_STOPPED;
    lthread_cond_broadcast(&lims->cond);
    lthread_mutex_unlock(&lims->mutex);
    return NULL;
}

static int flush_writable_bytes (lim_stream_t * lims, int one_shot)
{
    CPuint bytes;
    int contiguous_bytes;
    int write_bytes;
    void * ptr;

    while (lims->bytes_writable)
    {
        contiguous_bytes = (lims->mark <= lims->wr_tail)?
            lims->info.circular_buffer_bytes - lims->wr_tail:
            lims->mark - lims->wr_tail;

        ptr = lims->buf + lims->wr_tail;
        write_bytes = (contiguous_bytes > lims->info.commit_step_bytes)?
            lims->info.commit_step_bytes : contiguous_bytes;
        contiguous_bytes = write_bytes;

        if (lims->offset - lims->bytes_writable != lims->content_offset)
        {
            lims->itf->seek(lims->content,
                            lims->offset - lims->bytes_writable
                            - lims->content_offset,
                            SEEK_CUR);
            lims->content_offset = lims->offset - lims->bytes_writable;
        }

        while (contiguous_bytes)
        {
            bytes = contiguous_bytes;

            if (lims->itf->write(lims->content, ptr, &bytes))
            {
                LIMERR("unable to write to content.\n");
                //error, callback to client.
                return -1;
            } else if(bytes == contiguous_bytes) break;
            ptr += bytes;
            contiguous_bytes -= bytes;
        }

        lims->content_offset += write_bytes;
        lims->wr_tail += write_bytes;
        if (lims->wr_tail == lims->info.circular_buffer_bytes)
            lims->wr_tail = 0;
        lims->bytes_writable -= write_bytes;

        if (one_shot) break;
    }

    ASSERT(!lims->wr_tail || lims->wr_tail == lims->mark || one_shot);
    ASSERT(lims->bytes_writable == 0 || one_shot);
    return 0;
}

static void * lims_thread_write (void * arg)
{
    lim_stream_t * lims = (lim_stream_t *)arg;
    CP_EVENTTYPE event = CP_EventMax;
    CPuint param = 0;

    while (1)
    {
        lthread_mutex_lock(&lims->mutex);

        while (lims->bytes_writable < lims->info.commit_step_bytes
               && lims->loop == LIMS_LOOP_RUNNING)
        {
            /*
            LIMDBG("file stream write controls:\n"
                   "\toffset    : %lu\n"
                   "\thead      : %d\n"
                   "\ttail      : %d\n"
                   "\tmark      : %d\n"
                   "\tavailable : %d\n"
                   "\ttotal     : %d\n",
                   lims->offset, lims->head, lims->tail,
                   lims->mark, lims->bytes_available,
                   lims->info.circular_buffer_bytes);
            */
            lthread_cond_wait(&lims->cond, &lims->mutex);
        }

        //upon shutting down, wait till all bytes are flushed.
        if (lims->loop != LIMS_LOOP_RUNNING && !lims->bytes_writable)
        {
            lthread_mutex_unlock(&lims->mutex);
            break;
        }

        if (flush_writable_bytes(lims, 1))
        {
            if (lims->callback)
            {
                event = CP_PipeDisconnected;
                param = 0;
            }
            lthread_mutex_unlock(&lims->mutex);
            goto bail;
        }

        if ((lims->bytes_requested <=
             lims->info.circular_buffer_bytes - lims->bytes_writable)
            && lims->bytes_requested)
        {
            if (lims->callback)
            {
                lims->bytes_requested = 0;
                event = CP_BytesAvailable;
                param = lims->info.circular_buffer_bytes
                    - lims->bytes_writable;
            }
        }
        lthread_mutex_unlock(&lims->mutex);

        if (CP_EventMax != event)
        {
            lims->callback(event, param);
            event = CP_EventMax;
        }
    }

bail:
    if (CP_EventMax != event) lims->callback(event, param);

    lthread_mutex_lock(&lims->mutex);
    lims->loop = LIMS_LOOP_STOPPED;
    lthread_cond_broadcast(&lims->cond);
    lthread_mutex_unlock(&lims->mutex);
    return NULL;
}

static void * lims_thread_read_write (void * arg)
{
    lim_stream_t * lims = (lim_stream_t *)arg;
    CPuint bytes;
    CP_EVENTTYPE event = CP_EventMax;
    CPuint param = CP_EventMax;

    while (1)
    {
        lthread_mutex_lock(&lims->mutex);
        while ((lims->bytes_readable > lims->info.circular_buffer_bytes / 2
                || lims->eof)
               && lims->bytes_writable < lims->info.commit_step_bytes
               && lims->loop == LIMS_LOOP_RUNNING)
        {
            /*
            LIMINFO("file stream write controls:\n"
                    "\toffset         : %lu\n"
                    "\thead           : %d\n"
                    "\ttail           : %d\n"
                    "\tmark           : %d\n"
                    "\twr_tail        : %d\n"
                    "\tcontent offset : %d\n"
                    "\twritable       : %d\n"
                    "\treadable       : %d\n"
                    "\tavailable      : %d\n"
                    "\ttotal          : %d\n"
                    "\teof            : %d\n",
                    lims->offset, lims->head, lims->tail,
                    lims->mark, lims->wr_tail, lims->content_offset,
                    lims->bytes_writable, lims->bytes_readable,
                    lims->bytes_available,
                    lims->info.circular_buffer_bytes, lims->eof);
            */
            if (lims->eof && lims->bytes_requested)
            {
                if (lims->callback)
                {
                    lims->bytes_requested = 0;
                    param = lims->bytes_readable;

                    lthread_mutex_unlock(&lims->mutex);
                    lims->callback(CP_BytesAvailable, param);
                    lthread_mutex_lock(&lims->mutex);
                    //mutex has been relinquished above, do not go to wait
                    //before rechecking conditions first.
                    continue;
                }
            }

            lthread_cond_wait(&lims->cond, &lims->mutex);
        }

        if (lims->loop != LIMS_LOOP_RUNNING && !lims->bytes_writable)
        {
            lthread_mutex_unlock(&lims->mutex);
            event = CP_EventMax;
            break;
        }

        if (lims->bytes_writable >= lims->info.commit_step_bytes
            || lims->loop != LIMS_LOOP_RUNNING)
        {
            if (flush_writable_bytes(lims, 1))
            {
                if (lims->callback)
                {
                    event = CP_PipeDisconnected;
                    param = 0;
                }
                lthread_mutex_unlock(&lims->mutex);
                goto bail;
            }
            lthread_mutex_unlock(&lims->mutex);
            continue;
        }

        if (lims->bytes_available == lims->info.circular_buffer_bytes)
        {
            lims->bytes_available -= lims->info.commit_step_bytes;
            lims->tail += lims->info.commit_step_bytes;
            if (lims->tail >= lims->info.circular_buffer_bytes)
                lims->tail -= lims->info.circular_buffer_bytes;
        }

        bytes = lims->info.circular_buffer_bytes - lims->bytes_available;
        bytes = (bytes > lims->info.commit_step_bytes)?
            lims->info.commit_step_bytes : bytes;
        bytes = (bytes > lims->info.circular_buffer_bytes - lims->head)?
            lims->info.circular_buffer_bytes - lims->head : bytes;

        if (lims->offset + lims->bytes_readable != lims->content_offset)
        {
            lims->itf->seek(lims->content,
                            lims->offset + lims->bytes_readable
                            - lims->content_offset,
                            SEEK_CUR);
            lims->content_offset = lims->offset + lims->bytes_readable;
        }

        if (lims->itf->read(lims->content, lims->buf + lims->head, &bytes))
        {
            LIMERR("content read error.\n");
            //error, callback to client.
            lthread_mutex_unlock(&lims->mutex);
            if (lims->callback)
            {
                event = CP_PipeDisconnected;
                param = 0;
            }
            break;
        } else if (bytes == 0)
        {
            lims->eof = 1;
        } else
        {
            lims->content_offset += bytes;
            lims->bytes_available += bytes;
            lims->bytes_readable += bytes;
            lims->head += bytes;
            ASSERT(lims->head <= lims->info.circular_buffer_bytes);

            if (lims->head == lims->info.circular_buffer_bytes)
                lims->head = 0;
        }

        if ((lims->bytes_requested <= lims->bytes_readable)
            && lims->bytes_requested)
        {
            if (lims->callback)
            {
                lims->bytes_requested = 0;
                param = lims->bytes_readable;

                lthread_mutex_unlock(&lims->mutex);
                lims->callback(CP_BytesAvailable, param);
                continue;
            }
        }
        lthread_mutex_unlock(&lims->mutex);
    }

 bail:
    if (CP_EventMax != event) lims->callback(event, param);

    lthread_mutex_lock(&lims->mutex);
    lims->loop = LIMS_LOOP_STOPPED;
    lthread_cond_broadcast(&lims->cond);
    lthread_mutex_unlock(&lims->mutex);
    return NULL;
}
  
static CPresult Open (
    CPhandle* hContent,
    CPstring szURI,
    CP_ACCESSTYPE eAccess)
{
    CPresult ret;
    content_access_itf_t * itf;
    lim_stream_t * lims;
    void * (*thread_func) (void * arg);

    if (NULL == szURI)
    {
        LIMERR("NULL szURI.\n");
        return KD_EINVAL;
    }

    itf = CONTENT_ACCESS_ITFS[0];
    while (itf)
    {
        if (itf->can_access(szURI)) break;
    }

    if (!itf)
    {
        LIMERR("unexpected error.\n");
        return KD_EINVAL;
    }

    lims = calloc(1, sizeof(lim_stream_t));
    if (NULL == lims)
    {
        LIMERR("out of memory.\n");
        return KD_ENOMEM;
    }

    ret = itf->open(&lims->content, szURI, eAccess);
    if (ret) goto bail;
    else lims->itf = itf;

    itf->info(lims->content, &lims->info);
    lims->buf = malloc(lims->info.circular_buffer_bytes);
    if (NULL == lims->buf)
    {
        ret = KD_ENOMEM;
        goto bail;
    }

    switch (eAccess)
    {
    case CP_AccessRead :
        thread_func = lims_thread_read;
        break;
    case CP_AccessWrite :
        thread_func = lims_thread_write;
        break;
    case CP_AccessReadWrite :
        thread_func = lims_thread_read_write;
        break;
    default :
        LIMERR("unhandled content pipe access type: %d\n", eAccess);
        ret = KD_EINVAL;
        goto bail;
    }

    lthread_mutex_init(&lims->mutex);
    lthread_cond_init(&lims->cond);

    lims->loop = LIMS_LOOP_RUNNING;
    lims->access_type = eAccess;

    if (lthread_create(&lims->thread, thread_func, lims, 1))
    {
        ret = KD_ENOMEM;
        lthread_mutex_destroy(&lims->mutex);
        lthread_cond_destroy(&lims->cond);
        goto bail;
    }

    *hContent = lims;
    return 0;

 bail:
    if (lims)
    {
        if (lims->content) itf->close(lims->content);
        if (lims->buf) free(lims->buf);
        free(lims);
    }
    return ret;
}

static CPresult Close (CPhandle hContent)
{
    lim_stream_t * lims;

    if (NULL == hContent) return KD_EINVAL;

    lims = (lim_stream_t *)hContent;

    lthread_mutex_lock(&lims->mutex);
    //wait till thread shuts down, if not already.
    while (lims->loop != LIMS_LOOP_STOPPED)
    {
        lims->loop = LIMS_LOOP_STOPPING;
        lthread_cond_broadcast(&lims->cond);
        lthread_cond_wait(&lims->cond, &lims->mutex);
    }
    lthread_mutex_unlock(&lims->mutex);

    lims->itf->close(lims->content);

    lthread_mutex_destroy(&lims->mutex);
    lthread_cond_destroy(&lims->cond);

    if (lims->buf) free(lims->buf);
    if (lims->rd_buf_a) free(lims->rd_buf_a);
    if (lims->wr_buf_a) free(lims->wr_buf_a);

    free(lims);
    return 0;
}

static CPresult Create (CPhandle *hContent, CPstring szURI)
{
    return Open(hContent, szURI, CP_AccessWrite);
}

static CPresult CheckAvailableBytes (
    CPhandle hContent,
    CPuint nBytesRequested,
    CP_CHECKBYTESRESULTTYPE *eResult)
{
    lim_stream_t * lims = (lim_stream_t *)hContent;
    CPuint pipe_capacity;

    lthread_mutex_lock(&lims->mutex);

    pipe_capacity = (CP_AccessWrite == lims->access_type)?
        lims->info.circular_buffer_bytes :
        lims->info.circular_buffer_bytes / 2;
    if (nBytesRequested > pipe_capacity)
    {
        LIMWARN("requested bytes %lu beyond pipe capacity: %lu\n",
                nBytesRequested, pipe_capacity);
        *eResult = CP_CheckBytesInsufficientBytes;
        goto bail;
    }

    if (CP_AccessRead == lims->access_type
        || CP_AccessReadWrite == lims->access_type)
    {
        if (nBytesRequested <= lims->bytes_readable)
        {
            *eResult = CP_CheckBytesOk;
            goto bail;
        } else if (!lims->eof)
        {
            *eResult = CP_CheckBytesNotReady;
            lims->bytes_requested = nBytesRequested;
        } else if (lims->bytes_readable)
        {
            *eResult = CP_CheckBytesInsufficientBytes;
            lims->bytes_requested = nBytesRequested;
        } else
        {
            *eResult = CP_CheckBytesAtEndOfStream;
        }
    } else if (CP_AccessWrite == lims->access_type)
    {
        if (nBytesRequested <=
            lims->info.circular_buffer_bytes - lims->bytes_available)
        {
            *eResult = CP_CheckBytesOk;
            goto bail;
        } else if (lims->loop == LIMS_LOOP_STOPPING
                   || lims->loop == LIMS_LOOP_STOPPED)
        {
            *eResult = CP_CheckBytesAtEndOfStream;
        } else
        {
            *eResult = CP_CheckBytesOutOfBuffers;
            lims->bytes_requested = nBytesRequested;
        }
    } else
    {
        LIMWARN("unknown content access type: %d\n", lims->access_type);
    }

 bail:
    lthread_cond_broadcast(&lims->cond);
    lthread_mutex_unlock(&lims->mutex);
    return 0;
}

static CPresult SetPosition (
    CPhandle hContent,
    CPint nOffset,
    CP_ORIGINTYPE eOrigin)
{
    CPresult ret = 0;
    lim_stream_t * lims = (lim_stream_t *)hContent;

    if (CP_AccessWrite == lims->access_type) return KD_EINVAL;

    lthread_mutex_lock(&lims->mutex);
    if (lims->access_type == CP_AccessRead
        || lims->access_type == CP_AccessReadWrite)
    {
        // flush all writable bytes, if any
        if (flush_writable_bytes(lims, 0))
        {
            lthread_mutex_unlock(&lims->mutex);
            if (lims->callback) lims->callback(CP_PipeDisconnected, 0);
            ret = KD_EIO;
            goto bail;
        }

        switch (eOrigin)
        {
        case CP_OriginEnd :
            nOffset = lims->info.content_length_bytes + nOffset;
            // fall through

        case CP_OriginBegin :
            nOffset -= lims->offset;
            // fall through

        case CP_OriginCur :
            if (nOffset < lims->bytes_readable - lims->bytes_available
                || nOffset > lims->bytes_readable)
            {
                // seek in file directly.
                if (lims->itf->seek(lims->content,
                                    nOffset - lims->bytes_readable,
                                    SEEK_CUR))
                {
                    LIMWARN("stream position lost.\n");
                    ret = KD_EINVAL;
                    break;
                } else
                {
                    lims->head = lims->tail = lims->wr_tail = lims->mark = 0;
                    lims->bytes_available = lims->bytes_readable = 0;
                    lims->offset += nOffset;
                    lims->eof = 0;
                    lims->content_offset = lims->offset;
                }
            } else
            {
                // seek within pipe buffer.
                lims->mark += nOffset;
                if (lims->mark >= lims->info.circular_buffer_bytes)
                {
                    lims->mark -= lims->info.circular_buffer_bytes;
                } else if (lims->mark < 0)
                {
                    lims->mark += lims->info.circular_buffer_bytes;
                }

                lims->offset += nOffset;
                lims->bytes_readable -= nOffset;
            }
            lthread_cond_broadcast(&lims->cond);
            break;

        default :
            ret = KD_EINVAL;
            break;
        }
    }else
    {
        LIMWARN("unknown content access type: %d\n", lims->access_type);
    }

    lthread_mutex_unlock(&lims->mutex);

 bail:
    return ret;
}

static CPresult GetPosition (CPhandle hContent, CPuint *pPosition)
{
    lim_stream_t * lims = (lim_stream_t *)hContent;

    lthread_mutex_lock(&lims->mutex);
    *pPosition = lims->offset;
    lthread_mutex_unlock(&lims->mutex);

    return 0;
}

static CPresult Read (CPhandle hContent, CPbyte *pData, CPuint nSize)
{
    CPresult ret = 0;
    lim_stream_t * lims = (lim_stream_t *)hContent;
    CPuint bytes;
    CPuint bytes_read;

    if (CP_AccessWrite == lims->access_type) return KD_EINVAL;

    lthread_mutex_lock(&lims->mutex);
    if (lims->access_type == CP_AccessRead
        || lims->access_type == CP_AccessReadWrite)
    {
        bytes = (lims->bytes_readable >= nSize)? nSize : lims->bytes_readable;
        if (lims->eof && bytes < nSize)
        {
            // client can query what's left in pipe by CheckAvailableBytes
            ret = KD_ENOMEM;
            goto bail;
        }

        if (flush_writable_bytes(lims, 0))
        {
            lthread_cond_broadcast(&lims->cond);
            lthread_mutex_unlock(&lims->mutex);
            if (lims->callback) lims->callback(CP_PipeDisconnected, 0);
            ret = KD_EIO;
            goto bail2;
        }

        if (bytes)
        {
            int contiguous_bytes;

            contiguous_bytes = (lims->head > lims->mark)?
                lims->head - lims->mark :
                lims->info.circular_buffer_bytes - lims->mark;

            if (contiguous_bytes >= bytes)
            {
                memcpy(pData, lims->buf + lims->mark, bytes);
                lims->mark += bytes;
            } else
            {
                memcpy(pData, lims->buf + lims->mark, contiguous_bytes);
                memcpy(pData + contiguous_bytes, lims->buf,
                       bytes - contiguous_bytes);
                lims->mark = bytes - contiguous_bytes;
            }
            lims->bytes_readable -= bytes;
            lims->offset += bytes;
            if (lims->mark == lims->info.circular_buffer_bytes)
                lims->mark = 0;
        }

        if (bytes == nSize) goto bail;

        // directly read from content.
        pData += bytes;
        bytes_read = bytes = nSize - bytes;
        ret = lims->itf->read(lims->content, pData, &bytes_read);
        if (ret)
        {
            LIMERR("unable to read from content, ret = %d.\n", ret);
            if (ret == 0) lims->eof = 1;
            goto bail;
        } else  if (bytes != bytes_read)
        {
            //LIMWARN("partial reading detected, eof?\n");
            lims->eof = 1;
        }
        // reset pipe controls.
        lims->head = lims->tail = lims->wr_tail = lims->mark = 0;
        lims->bytes_available = lims->bytes_readable = 0;
        lims->offset += bytes_read;
        lims->content_offset = lims->offset;
    } else
    {
        LIMWARN("unknown content access type: %d\n", lims->access_type);
    }

 bail:
    lthread_cond_broadcast(&lims->cond);
    lthread_mutex_unlock(&lims->mutex);
 bail2:
    return ret;
}

static CPresult ReadBuffer (
    CPhandle hContent,
    CPbyte **ppBuffer,
    CPuint *nSize,
    CPbool bForbidCopy)
{
    CPresult ret = 0;
    lim_stream_t * lims = (lim_stream_t *)hContent;
    CPuint bytes;
    CPuint bytes_read;

    if (CP_AccessWrite == lims->access_type) return KD_EINVAL;
    if (*nSize == 0 || NULL == ppBuffer) return KD_EINVAL;
    *ppBuffer = NULL;

    lthread_mutex_lock(&lims->mutex);
    if (lims->access_type == CP_AccessRead
        || CP_AccessReadWrite == lims->access_type)
    {
        int contiguous_bytes = 0;

        if (lims->rd_buf)
        {
            LIMERR("read buffer hasn't been released yet.\n");
            ret = KD_ENOMEM;
            goto bail2;
        }

        if (flush_writable_bytes(lims, 0))
        {
            lthread_cond_broadcast(&lims->cond);
            lthread_mutex_unlock(&lims->mutex);
            if (lims->callback) lims->callback(CP_PipeDisconnected, 0);
            ret = KD_EIO;
            goto bail3;
        }

        if (lims->bytes_readable)
        {
            contiguous_bytes = (lims->head > lims->mark)?
                lims->head - lims->mark :
                lims->info.circular_buffer_bytes - lims->mark;
        }

        if (contiguous_bytes >= *nSize
            || (lims->bytes_readable && OMX_TRUE == bForbidCopy))
        {
            *nSize = (*nSize > contiguous_bytes)?
                contiguous_bytes : *nSize;
            *ppBuffer = lims->buf + lims->mark;
        } else
        {
            if (lims->rd_buf_a_bytes < *nSize)
            {
                lims->rd_buf_a = realloc(lims->rd_buf_a, *nSize);
                if (NULL == lims->rd_buf_a)
                {
                    LIMERR("out of memory.\n");
                    goto bail;
                } else lims->rd_buf_a_bytes = *nSize;
            }

            *ppBuffer = lims->rd_buf_a;

            bytes = (*nSize > lims->bytes_readable)?
                      lims->bytes_readable : *nSize;
            if (bytes)
            {
                memcpy(*ppBuffer, lims->buf + lims->mark, contiguous_bytes);
                memcpy(*ppBuffer + contiguous_bytes, lims->buf,
                       bytes - contiguous_bytes);
            }

            if (bytes < *nSize)
            {
                // directly read from contents
                bytes_read = *nSize - bytes;
                ret = lims->itf->read
                    (lims->content, *ppBuffer + bytes, &bytes_read);
                if (ret)
                {
                    LIMERR("unable to read from content.\n");
                    *nSize = bytes;
                } else
                {
                    if (bytes_read != *nSize - bytes) lims->eof = 1;
                    *nSize = bytes + bytes_read;

                    lims->rd_buf = *ppBuffer;
                    lims->rd_buf_bytes = *nSize;

                    // reset pipe controls.
                    lims->head = lims->tail = lims->wr_tail = lims->mark = 0;
                    lims->bytes_available = lims->bytes_readable = 0;
                    lims->offset += bytes;
                    lims->content_offset = lims->offset;
                    goto bail;
                }
            }
        }

        lims->rd_buf = *ppBuffer;
        lims->rd_buf_bytes = *nSize;

        lims->offset += lims->rd_buf_bytes;
        lims->bytes_readable -= lims->rd_buf_bytes;
        lims->mark += lims->rd_buf_bytes;
        if (lims->mark >= lims->info.circular_buffer_bytes)
            lims->mark -= lims->info.circular_buffer_bytes;
    } else
    {
        LIMWARN("unknown content access type: %d\n", lims->access_type);
    }

 bail:
    if (0 == *nSize || NULL == lims->rd_buf)
    {
        lims->rd_buf = NULL;
        *nSize = 0;
        ret = KD_ENOMEM;
    }
    lthread_cond_broadcast(&lims->cond);

 bail2:
    lthread_mutex_unlock(&lims->mutex);

 bail3:
    return ret;
}

static CPresult ReleaseReadBuffer (CPhandle hContent, CPbyte *pBuffer)
{
    CPresult ret = 0;
    lim_stream_t * lims = (lim_stream_t *)hContent;

    if (CP_AccessWrite == lims->access_type) return KD_EINVAL;

    lthread_mutex_lock(&lims->mutex);
    if (CP_AccessRead == lims->access_type
        || CP_AccessReadWrite == lims->access_type)
    {
        if (lims->rd_buf != pBuffer)
        {
            LIMDBG("invalid buffer: buf: %p bytes: %d\n",
                   lims->rd_buf, lims->rd_buf_bytes);
        } else
        {
            lims->rd_buf = NULL;
            lims->rd_buf_bytes = 0;
        }
    } else
    {
        LIMWARN("unknown content access type: %d\n", lims->access_type);
    }

    lthread_cond_broadcast(&lims->cond);
    lthread_mutex_unlock(&lims->mutex);

    return ret;
}

static CPresult Write (CPhandle hContent, CPbyte *pData, CPuint nSize)
{
    CPresult ret = 0;
    lim_stream_t * lims = (lim_stream_t *)hContent;
    int contiguous_bytes;
    int retried = 0;

    if (CP_AccessRead == lims->access_type) return KD_EINVAL;

 retry:
    lthread_mutex_lock(&lims->mutex);
    if (lims->access_type == CP_AccessWrite
        || lims->access_type == CP_AccessReadWrite)
    {
        if (lims->wr_buf || lims->wr_buf_bytes)
        {
            /* if client interleaves Write and WriteBuffer, Write API
               must wait till fetched buffer is written back. Since in
               most cases, buffer is directly fetched from internal buffer
               to save memory copying. */
            if (retried)
            {
                LIMWARN("interleaved write and writeBuffer API.\n");
                ret = KD_EAGAIN;
                goto bail2;
            } else
            {
                lthread_cond_broadcast(&lims->cond);
                lthread_mutex_unlock(&lims->mutex);
                // sleep 15ms, so we still comply with IL spec < 20ms
                usleep(15000);
                retried = 1;
                goto retry;
            }
        } else if (lims->info.circular_buffer_bytes - lims->bytes_writable
                   >= nSize)
        {
            contiguous_bytes = (lims->mark >= lims->wr_tail)?
                lims->info.circular_buffer_bytes - lims->mark :
                lims->wr_tail - lims->mark;

            if (contiguous_bytes >= nSize)
            {
                memcpy(lims->buf + lims->mark, pData, nSize);
                lims->mark += nSize;
                if (lims->mark == lims->info.circular_buffer_bytes)
                    lims->mark = 0;
            } else
            {
                memcpy(lims->buf + lims->mark, pData, contiguous_bytes);
                memcpy(lims->buf, pData + contiguous_bytes,
                       nSize - contiguous_bytes);
                lims->mark = nSize - contiguous_bytes;
            }
            lims->bytes_writable += nSize;
            lims->offset += nSize;

            if (nSize > lims->bytes_readable)
            {
                lims->bytes_available += nSize - lims->bytes_readable;
                lims->bytes_readable = 0;
                lims->head = lims->mark;

                if (lims->bytes_available > lims->info.circular_buffer_bytes)
                {
                    lims->tail += lims->bytes_available;
                    lims->tail -= lims->info.circular_buffer_bytes;
                    if (lims->tail == lims->info.circular_buffer_bytes)
                        lims->tail = 0;
                }
            } else lims->bytes_readable -= nSize;
        } else
        {
            // flush and directly write to content
            if (flush_writable_bytes(lims, 0))
            {
                LIMERR("unable to flush write buffer.\n");
                lthread_mutex_unlock(&lims->mutex);
                if (lims->callback) lims->callback(CP_PipeDisconnected, 0);
                ret = KD_EIO;
                goto bail;
            }

            ret = lims->itf->write(lims->content, pData, &nSize);
            if (ret)
            {
                LIMERR("unable to write to content.\n");
                lthread_mutex_unlock(&lims->mutex);
                if (lims->callback) lims->callback(CP_PipeDisconnected, 0);
                ret = KD_EIO;
                goto bail;
            } else
            {
                lims->head = lims->tail = lims->mark = lims->wr_tail = 0;
                lims->bytes_available = 0;
                lims->bytes_readable = lims->bytes_writable = 0;
                lims->offset += nSize;
                lims->content_offset = lims->offset;
            }
        }
    } else
    {
        LIMWARN("unknown content access type: %d\n", lims->access_type);
    }
 bail2:
    lthread_cond_broadcast(&lims->cond);
    lthread_mutex_unlock(&lims->mutex);
 bail:
    return ret;
}

static CPresult GetWriteBuffer (
    CPhandle hContent,
    CPbyte **ppBuffer,
    CPuint *nSize)
{
    CPresult ret = 0;
    lim_stream_t * lims = (lim_stream_t *)hContent;
    int contiguous_bytes;

    if (CP_AccessRead == lims->access_type) return KD_EINVAL;

    *ppBuffer = NULL;
    lthread_mutex_lock(&lims->mutex);
    if (CP_AccessWrite == lims->access_type
        || CP_AccessReadWrite == lims->access_type)
    {
        if (lims->wr_buf || lims->wr_buf_bytes)
        {
            LIMERR("write buffer hasn't been written back yet.\n");
            ret = KD_ENOMEM;
            *nSize = 0;
            goto bail;
        }

        contiguous_bytes =
            ((lims->mark == lims->wr_tail && lims->bytes_writable)
             || lims->mark < lims->wr_tail)?
            lims->wr_tail - lims->mark :
            lims->info.circular_buffer_bytes - lims->mark;

        if (contiguous_bytes >= *nSize)
        {
            *ppBuffer = lims->buf + lims->mark;
        } else
        {
            if (lims->wr_buf_a_bytes < *nSize)
            {
                lims->wr_buf_a = realloc(lims->wr_buf_a, *nSize);
                if (NULL == lims->wr_buf_a)
                {
                    LIMERR("out of memory.\n");
                    *nSize = 0;
                    ret = KD_ENOMEM;
                    goto bail;
                } else lims->wr_buf_a_bytes = *nSize;
            }

            *ppBuffer = lims->wr_buf_a;
        }
        lims->wr_buf = *ppBuffer;
        lims->wr_buf_bytes = *nSize;
    } else
    {
        LIMWARN("unknown content access type: %d\n", lims->access_type);
    }

bail:
    lthread_mutex_unlock(&lims->mutex);
    return ret;
}

static CPresult WriteBuffer (
    CPhandle hContent,
    CPbyte *pBuffer,
    CPuint nFilledSize)
{
    CPresult ret = 0;
    lim_stream_t * lims = (lim_stream_t *)hContent;

    if (CP_AccessRead == lims->access_type) return KD_EINVAL;

    lthread_mutex_lock(&lims->mutex);
    if (CP_AccessWrite == lims->access_type
        || CP_AccessReadWrite == lims->access_type)
    {
        if (lims->wr_buf != pBuffer || lims->wr_buf_bytes == 0
            || lims->wr_buf_bytes < nFilledSize)
        {
            LIMERR("invalid write buffer: buf: %p bytes: %d, filled: %d\n",
                   lims->wr_buf, lims->wr_buf_bytes, nFilledSize);
            ret = KD_EINVAL;
            goto bail;
        }

        if (!BUFFER_CONTAINS
            (lims->wr_buf, lims->buf, lims->info.circular_buffer_bytes))
        {
            lims->wr_buf = NULL;
            lims->wr_buf_bytes = 0;
            lthread_mutex_unlock(&lims->mutex);

            return Write(hContent, pBuffer, nFilledSize);
        } else
        {
            lims->mark += nFilledSize;
            if (lims->mark == lims->info.circular_buffer_bytes)
                lims->mark = 0;
        }
        lims->bytes_writable += nFilledSize;
        lims->offset += nFilledSize;

        if (nFilledSize > lims->bytes_readable)
        {
            lims->bytes_available += nFilledSize - lims->bytes_readable;
            lims->bytes_readable = 0;
            lims->head = lims->mark;

            if (lims->bytes_available > lims->info.circular_buffer_bytes)
            {
                lims->tail += lims->bytes_available;
                lims->tail -= lims->info.circular_buffer_bytes;
                if (lims->tail == lims->info.circular_buffer_bytes)
                    lims->tail = 0;
            }
        } else lims->bytes_readable -= nFilledSize;

        lims->wr_buf = NULL;
        lims->wr_buf_bytes = 0;
    } else
    {
        LIMWARN("unknown content access type: %d\n", lims->access_type);
    }

bail:
    lthread_cond_broadcast(&lims->cond);
    lthread_mutex_unlock(&lims->mutex);
    return ret;
}

static CPresult RegisterCallback (
    CPhandle hContent,
    CPresult (*ClientCallback)(CP_EVENTTYPE eEvent, CPuint iParam))
{
    lim_stream_t * lims = (lim_stream_t *)hContent;

    lthread_mutex_lock(&lims->mutex);
    lims->callback = ClientCallback;
    lthread_mutex_unlock(&lims->mutex);

    return 0;
}

static CP_PIPETYPE lim_content_pipe = {
    .Open = Open,
    .Close = Close,
    .Create = Create,
    .CheckAvailableBytes = CheckAvailableBytes,
    .SetPosition = SetPosition,
    .GetPosition = GetPosition,
    .Read = Read,
    .ReadBuffer = ReadBuffer,
    .ReleaseReadBuffer = ReleaseReadBuffer,
    .Write = Write,
    .GetWriteBuffer = GetWriteBuffer,
    .WriteBuffer = WriteBuffer,
    .RegisterCallback = RegisterCallback
};

OMX_ERRORTYPE OMX_GetContentPipe (
    OMX_HANDLETYPE *hPipe,
    OMX_STRING szURI)
{
    ASSERT(hPipe);
    if (NULL == szURI)
    {
        *hPipe = NULL;
        return OMX_ErrorBadParameter;
    }

    *hPipe = &lim_content_pipe;
    return OMX_ErrorNone;
}

