/*
  Copyright (C) 2009 - 2011 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX IL core content pipe: file support.

  REVISION:

    1) initial creation. ----------------------------------------- 2009-12-06 MG
*/

#include "contentpipe.h"

#define FILE_STREAM_READ_STEP_BYTES       (1024 * 8)
#define FILE_STREAM_CIRCULAR_READ_BUFFER_BYTES \
    (FILE_STREAM_READ_STEP_BYTES * 512)

#define FILE_STREAM_WRITE_STEP_BYTES      (1024 * 8)
#define FILE_STREAM_CIRCULAR_WRITE_BUFFER_BYTES \
  (FILE_STREAM_WRITE_STEP_BYTES * 512)

typedef struct
{
    int fd;
    content_info_t info;
} limoi_file_stream_t;

static int file_can_access (CPstring szURI)
{
    //TODO
    return 1;
}

static CPresult file_open (
    content_handle_t * content,
    CPstring szURI,
    CP_ACCESSTYPE eAccess)
{
    CPresult ret = KD_EACCES;
    limoi_file_stream_t * fs = NULL;
    char * fname;

    ASSERT(szURI);

    fs = calloc(1, sizeof(limoi_file_stream_t));
    if (NULL == fs)
    {
        LIMERR("out of memory.\n");
        return KD_ENOMEM;
    }

    // strip file:// if present
    fname = (strstr(szURI, "file://") == szURI)?
        szURI + strlen("file://") : szURI;

    switch (eAccess)
    {
    case CP_AccessRead :
        fs->fd = open(fname, O_RDONLY);
        fs->info.circular_buffer_bytes =
            FILE_STREAM_CIRCULAR_READ_BUFFER_BYTES;
        fs->info.commit_step_bytes = FILE_STREAM_READ_STEP_BYTES;
        break;

    case CP_AccessWrite :
        fs->fd = open(fname, O_WRONLY | O_CREAT | O_TRUNC,
                      S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
        fs->info.circular_buffer_bytes =
            FILE_STREAM_CIRCULAR_WRITE_BUFFER_BYTES;
        fs->info.commit_step_bytes = FILE_STREAM_WRITE_STEP_BYTES;
        break;

    case CP_AccessReadWrite :
        fs->fd = open(fname, O_RDWR | O_CREAT,
                      S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
        fs->info.circular_buffer_bytes =
            FILE_STREAM_CIRCULAR_WRITE_BUFFER_BYTES;
        fs->info.commit_step_bytes = FILE_STREAM_WRITE_STEP_BYTES;
        break;

    default : ASSERT(0);
    }

    if (-1 == fs->fd)
    {
        LIMERR("unable to open file: %s\n", fname);
        goto bail;
    } else
    {
        struct stat st;
        if (fstat(fs->fd, &st))
        {
            LIMWARN("unable to get file status, file size is hard-coded.\n");
            //something has gone wrong, or maybe the
            //_LARGEFILE_SOURCE support is not enabled.
            fs->info.content_length_bytes = 0x7fffffff;
        } else
        {
            fs->info.content_length_bytes = st.st_size;
        }

        *content = fs;
        return 0;
    }

 bail:
    if (fs)
    {
        if (fs->fd != -1) close(fs->fd);
        free(fs);
    }
    return ret;
}

static CPresult file_close (content_handle_t content)
{
    limoi_file_stream_t * fs;

    ASSERT(content);
    fs = (limoi_file_stream_t *)content;
    ASSERT(fs->fd != -1);
    close(fs->fd);

    return 0;
}

static CPresult file_seek (content_handle_t content,
                           CPint nOffset,
                           CP_ORIGINTYPE eOrigin)
{
    limoi_file_stream_t * fs;
    int whence;
    CPresult ret = 0;

    LIMDBG("content %p seek to %ld from %d\n", content, nOffset, eOrigin);

    ASSERT(content);
    fs = (limoi_file_stream_t *)content;
    ASSERT(fs->fd != -1);

    switch (eOrigin)
    {
    case CP_OriginBegin :
        whence = SEEK_SET;
        break;
    case CP_OriginEnd :
        whence = SEEK_END;
        break;
    case CP_OriginCur :
        whence = SEEK_CUR;
        break;
    default: ASSERT(0);
    }

    if (-1 == lseek(fs->fd, nOffset, SEEK_CUR))
    {
        LIMWARN("stream position lost.\n");
        ret = KD_EINVAL;
    }
    return ret;
}

static CPresult file_read (content_handle_t content,
                           CPbyte * pData,
                           CPuint * nSize)
{
    limoi_file_stream_t * fs;
    ssize_t bytes;

    LIMDBG("content %p read %lu\n", content, *nSize);

    ASSERT(content);
    fs = (limoi_file_stream_t *)content;
    ASSERT(fs->fd != -1);

    bytes = read(fs->fd, pData, *nSize);
    if (bytes < 0) return KD_EIO;
    *nSize = bytes;
    return 0;
}

static CPresult file_write (content_handle_t content,
                            CPbyte * pData,
                            CPuint * nSize)
{
    limoi_file_stream_t * fs;
    ssize_t bytes;

    LIMDBG("content %p write %lu\n", content, *nSize);

    ASSERT(content);
    fs = (limoi_file_stream_t *)content;
    ASSERT(fs->fd != -1);

    bytes = write(fs->fd, pData, *nSize);
    if (bytes < 0) return KD_EIO;
    *nSize = bytes;
    return 0;
}

static CPresult file_info (content_handle_t content,
                           content_info_t * info)
{
    limoi_file_stream_t * fs;

    ASSERT(content);
    fs = (limoi_file_stream_t *)content;
    ASSERT(fs->fd != -1);

    *info = fs->info;

    return 0;
}

content_access_itf_t content_access_itf_file = {
    .can_access = file_can_access,
    .open = file_open,
    .close = file_close,
    .seek = file_seek,
    .read = file_read,
    .write = file_write,
    .info = file_info,
};
