/*
  Copyright (C) 2009 - 2011 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX IL core component handle module.

  REVISION:

    2) component multi-instantiation support. -------------------- 2011-05-22 MG
    1) initial creation. ----------------------------------------- 2009-11-22 MG
*/

#define _GNU_SOURCE
#include <dlfcn.h>
#include <sys/types.h>
#include "core.h"

#undef LIMDBG
#ifndef LIMDBG
#define LIMDBG(fmt, arg...)
#endif

component_reg_t * get_comp_registration_by_name (const char * name)
{
    component_reg_t * reg = NULL;
    list_t * list = limoi_omx.components_list_head.next;

    while (list != &limoi_omx.components_list_head)
    {
        int res;
        reg = (component_reg_t *)list;
        res = strncmp(reg->name, name, OMX_MAX_STRINGNAME_SIZE);
        if (!res) return reg;
        if (res > 0) break;
        list = list->next;
    }

    return NULL;
}

component_reg_t * get_comp_registration_by_handle (void * comp_handle)
{
    component_reg_t * reg = NULL;
    list_t * list = limoi_omx.components_list_head.next;

    while (list != &limoi_omx.components_list_head)
    {
        list_t * list2;
        instance_reg_t * comp;

        reg = (component_reg_t *)list;

        list2 = reg->obj->comps_list_head.next;
        while (list2 != &reg->obj->comps_list_head)
        {
            comp = (instance_reg_t *)list2;
            if (comp_handle == comp->handle) return reg;
            list2 = list2->next;
        }

        list = list->next;
    }

    return NULL;
}

OMX_ERRORTYPE OMX_GetHandle (
    OMX_HANDLETYPE* pHandle,
    OMX_STRING cComponentName,
    OMX_PTR pAppData,
    OMX_CALLBACKTYPE* pCallBacks)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    component_reg_t * reg = NULL;

    *pHandle = NULL;
    lthread_mutex_lock(&limoi_omx.mutex);

    if (strlen(cComponentName) >= OMX_MAX_STRINGNAME_SIZE)
    {
        LIMERR("invalid component name: %s.\n", cComponentName);
        ret = OMX_ErrorInvalidComponentName;
        goto bail;
    }

    reg = get_comp_registration_by_name(cComponentName);
    if (NULL == reg)
    {
        LIMERR("unable to find component: %s\n", cComponentName);
        ret = OMX_ErrorInvalidComponentName;
        goto bail;
    }

    if (reg->obj->init)
    {
        lthread_mutex_unlock(&limoi_omx.mutex);
        ret = reg->obj->init(pHandle, reg->name);
        lthread_mutex_lock(&limoi_omx.mutex);
    } else if (reg->obj->so)
    {
        if (!reg->obj->so->lib)
        {
            reg->obj->so->lib = dlopen(reg->obj->so->name, RTLD_LAZY);
            if (NULL == reg->obj->so->lib)
            {
                LIMERR("unexpected error: %s\n", dlerror());
                ret = OMX_ErrorInsufficientResources;
                goto bail;
            }
        }

        if (!reg->obj->so->registration)
        {
            reg->obj->so->registration =
                dlsym(reg->obj->so->lib, "OMX_ComponentRegistration");
        }

        if (NULL == reg->obj->so->registration)
        {
            LIMERR("insufficient resources.\n");
            if (!reg->obj->so->ref_count)
            {dlclose(reg->obj->so->lib);
                reg->obj->so->lib = NULL;
            }
            ret = OMX_ErrorInsufficientResources;
            goto bail;
        } else
        {
            int num;
            lim_omx_component_registration_t ** comp_reg;
            comp_reg = reg->obj->so->registration(&num);
            while (num--)
            {
                if (!strcmp((*comp_reg)->name, cComponentName)) break;
                comp_reg++;
            }

            reg->obj->so->ref_count++;

            lthread_mutex_unlock(&limoi_omx.mutex);
            ret = (*comp_reg)->init(pHandle, cComponentName);
            lthread_mutex_lock(&limoi_omx.mutex);
        }
    } else
    {
        lthread_mutex_unlock(&limoi_omx.mutex);
        ret = reg->obj->core->OMX_GetHandle
            (pHandle, cComponentName, pAppData, pCallBacks);
        lthread_mutex_lock(&limoi_omx.mutex);
    }

    if (OMX_ErrorNone == ret)
    {
        if (reg->obj->init || reg->obj->so)
        {
            LIMDBG("%s event handler callback: %p\n",
                   cComponentName, pCallBacks->EventHandler);
            ((OMX_COMPONENTTYPE *)*pHandle)->SetCallbacks
                (*pHandle, pCallBacks, pAppData);
        }

        instance_reg_t * comp = malloc(sizeof(instance_reg_t));
        if (NULL == comp)
        {
            LIMERR("out of memory.\n");
            ret = OMX_ErrorInsufficientResources;
            goto bail;
        }
        comp->handle = *pHandle;
        list_init(&comp->list);
        list_add(&comp->list, &reg->obj->comps_list_head);

        lthread_mutex_unlock(&limoi_omx.mutex);
        return OMX_ErrorNone;
    }

 bail:
    lthread_mutex_unlock(&limoi_omx.mutex);
    if (*pHandle) free(*pHandle);
    *pHandle =  NULL;

    return ret;
}

OMX_ERRORTYPE OMX_FreeHandle (OMX_HANDLETYPE hComponent)
{
    component_reg_t * reg = NULL;
    list_t * list;

    lthread_mutex_lock(&limoi_omx.mutex);

    list = limoi_omx.components_list_head.next;
    while (list != &limoi_omx.components_list_head)
    {
        list_t * list2;
        instance_reg_t * comp;

        reg = (component_reg_t *)list;

        list2 = reg->obj->comps_list_head.next;
        while (list2 != &reg->obj->comps_list_head)
        {
            comp = (instance_reg_t *)list2;
            if (hComponent == comp->handle)
            {
                lthread_mutex_unlock(&limoi_omx.mutex);
                if (reg->obj->init || reg->obj->so)
                {
                    if (OMX_ErrorNone !=
                        ((OMX_COMPONENTTYPE *)hComponent)->
                        ComponentDeInit(hComponent))
                    {
                        LIMWARN("unable to deinit component?\n");
                    }
                } else
                {
                    reg->obj->core->OMX_FreeHandle(hComponent);
                }
                lthread_mutex_lock(&limoi_omx.mutex);

                list_del(list2);
                free(comp);

                if (!list_entries(&reg->obj->comps_list_head))
                {
                    if (reg->obj->so)
                    {
                        //if SO is no longer reference by any components,
                        //close it to free up memory.
                        reg->obj->so->ref_count--;
                        ASSERT(reg->obj->so->ref_count >= 0);
                        if (0 == reg->obj->so->ref_count && reg->obj->so->lib)
                        {
                            dlclose(reg->obj->so->lib);
                            reg->obj->so->lib = NULL;
                            reg->obj->so->registration = NULL;
                        }
                    }
                }
                lthread_mutex_unlock(&limoi_omx.mutex);
                return OMX_ErrorNone;
            }
            list2 = list2->next;
        }

        list = list->next;
    }

    lthread_mutex_unlock(&limoi_omx.mutex);
    LIMERR("component not found.\n");
    return OMX_ErrorComponentNotFound;
}
