/*
  Copyright (C) 2009 - 2011 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX IL core init/deinit module.

  REVISION:

    2) static and dynamic loading support. ----------------------- 2011-05-24 MG
    1) initial creation. ----------------------------------------- 2009-11-22 MG
*/
#define _GNU_SOURCE
#include <dlfcn.h>
#include <sys/types.h>
#include <dirent.h>
#include "core.h"
#include "limdir.h"

#ifdef RESOURCE_MANAGER_ENABLED
    #include "resource-manager.h"
#else
    #define lim_resource_manager_init() ;
    #define lim_resource_manager_deinit() ;
#endif

#undef LIMDBG
#ifndef LIMDBG
#define LIMDBG(fmt, arg...)
#endif

// global LIMOI OMX core object
limoi_omx_t limoi_omx;

static int register_role (char ** roles, list_t * list_head)
{
    list_t * list;
    role_reg_t * role;

    while (*roles)
    {
        list = list_head->next;
        while (list != list_head)
        {
            role_reg_t * role0 = (role_reg_t *)list;
            int res = strcmp(role0->name, *roles);
            if (!res)
            {
                LIMDBG("duplicated role name: %s\n", *roles);
                goto bail;
            } else if (res > 0) break;
            else list = list->next;
        }

        role = malloc(sizeof(role_reg_t));
        if (NULL == role)
        {
            LIMERR("out of memory.\n");
            goto bail;
        }
        list_init(&role->list);
        role->name = strdup(*roles);
        if (NULL == role->name)
        {
            LIMERR("out of memory.\n");
            free(role);
            goto bail;
        }

        list_add_tail(&role->list, list);
        roles++;
    }

    return 0;

 bail:
    list = list_head->next;
    while (list != list_head)
    {
        role_reg_t * role0 = (role_reg_t *)list;
        list = list->next;

        list_del(&role0->list);
        free((void *)role0->name);
        free(role0);
    }
    return -1;
}

static int register_component (component_reg_t * comp)
{
    list_t * list = limoi_omx.components_list_head.next;

    while (list != &limoi_omx.components_list_head)
    {
        component_reg_t * comp0 = (component_reg_t *)list;
        int res = strcmp(comp0->name, comp->name);
        if (!res)
        {
            LIMDBG("duplicated component name: %s\n", comp->name);
            return -1;
        } else if (res > 0) break;
        else list = list->next;
    }

    list_add_tail(&comp->list, list);

    return 0;
}

//return 1: duplicate
//       0: success
//      -1: error
static int add_static_component (lim_omx_component_registration_t * reg)
{
    component_reg_t * comp = NULL;
    int ret = -1;

    ASSERT(reg && reg->name && reg->roles && *reg->roles && reg->init);

    comp = calloc(1, sizeof(component_reg_t));
    if (NULL == comp)
    {
        LIMERR("out of memory.\n");
        return -1;
    }

    list_init(&comp->roles_list_head);
    comp->name = strdup(reg->name);
    if (NULL == comp->name)
    {
        LIMERR("out of memory.\n");
        goto bail;
    }

    comp->obj = malloc(sizeof(component_obj_t));
    if (NULL == comp->obj)
    {
        LIMERR("out of memory.\n");
        goto bail;
    }

    list_init(&comp->obj->comps_list_head);
    comp->obj->so = NULL;
    comp->obj->core = NULL;
    comp->obj->init = reg->init;

    if (register_role(reg->roles, &comp->roles_list_head)) goto bail;
    if (!register_component(comp)) return 0;
    else ret = 1;

 bail:
    if (comp)
    {
        if (comp->name) free((void *)comp->name);
        if (comp->obj) free(comp->obj);
        free(comp);
    }

    return ret;
}

static int load_components_static (void)
{
    int ii = 0;
    int count = 0;
    int ret;

    while (LIM_OMX_ComponentRegistrationTable[ii])
    {
        ret = add_static_component(LIM_OMX_ComponentRegistrationTable[ii]);
        if (ret < 0)
        {
            LIMERR("unable to register static component: %s\n",
                   LIM_OMX_ComponentRegistrationTable[ii]->name);
            break;
        } else if (ret == 0) count++;
        ii++;
    }
    return count;
}

// return >= 1: num of component detected and loaded
//         = 0: not a valid component
//         =-1: unknown error detected, bail
static int add_dynamic_component (
    const char * cname,
    comp_registration_entry_t registration,
    comp_deregistration_entry_t deregistration)
{
    component_reg_t * comp = NULL;
    int num;
    lim_omx_component_registration_t ** preg = registration(&num);
    lim_omx_component_registration_t * reg;
    int num_comp_registered = 0;

    so_reg_t * so = malloc(sizeof(so_reg_t));
    if (NULL == so)
    {
        LIMERR("out of memory.\n");
        return -1;
    }
    so->name = strdup(cname);
    if (NULL == so->name)
    {
        LIMERR("out of memory.\n");
        free(so);
        return -1;
    }

    so->lib = NULL;
    so->registration = NULL;
    so->ref_count = 0;
    so->deregistration = deregistration;
    list_init(&so->list);
    list_add(&so->list, &limoi_omx.so_list_head);

    while (num--)
    {
        reg = *preg++;

        comp = calloc(1, sizeof(component_reg_t));
        if (NULL == comp)
        {
            LIMERR("out of memory.\n");
            goto bail;
        }

        list_init(&comp->roles_list_head);
        comp->name = strdup(reg->name);
        if (NULL == comp->name)
        {
            LIMERR("out of memory.\n");
            goto bail;
        }

        comp->obj = malloc(sizeof(component_obj_t));
        if (NULL == comp->obj)
        {
            LIMERR("out of memory.\n");
            goto bail;
        }

        list_init(&comp->obj->comps_list_head);
        comp->obj->so = so;
        comp->obj->init = NULL;
        comp->obj->core = NULL;

        if (register_role(reg->roles, &comp->roles_list_head)) goto bail;
        if (!register_component(comp)) num_comp_registered++;
        else
        {
            free((void *)comp->name);
            free(comp->obj);
            free(comp);
        }
        continue;

    bail:
        if (comp)
        {
            if (comp->name) free((void *)comp->name);
            if (comp->obj) free(comp->obj);
            free(comp);
        }
        break;
    }

    if (!num_comp_registered)
    {
        list_del(&so->list);
        free(so);
    }

    return num_comp_registered;
}

static void load_component (const char * cname, void * arg)
{
	void * so;
	void * registration;
    void * deregistration;

    LIMDBG("cname: %s\n", cname);
	so = dlopen(cname, RTLD_LAZY);
    if (so)
    {
        registration = dlsym(so, "OMX_ComponentRegistration");
        if (registration)
        {
            int ret;

            deregistration = dlsym(so, "OMX_ComponentDeregistration");
            ret = add_dynamic_component(cname, registration, deregistration);
            if (ret > 0)
            {
                *(int *)arg = *(int *)arg + ret;
            }
        }
        dlclose(so);
    } else
    {
        LIMDBG("%s\n", dlerror());
    }
}

// recursively load components from given path.
static int load_components_recursive (const char * com_path)
{
	int num_components = 0;
    return (limdir_recursive_scan(com_path, load_component, &num_components))?
        0: num_components;
}

// return >= 1: num of component detected and loaded
//         = 0: not a valid component
//         =-1: unknown error detected, bail
static int add_core_component (void * lib)
{
    int num_comp_registered = 0;
    int idx = 0;
    component_reg_t * comp = NULL;
    core_reg_t * core = malloc(sizeof(core_reg_t));
    if (NULL == core)
    {
        LIMERR("out of memory.\n");
        return -1;
    }
    core->lib = lib;

    core->OMX_Init = dlsym(lib, "OMX_Init");
    core->OMX_Deinit = dlsym(lib, "OMX_Deinit");
    core->OMX_ComponentNameEnum = dlsym(lib, "OMX_ComponentNameEnum");
    core->OMX_GetHandle = dlsym(lib, "OMX_GetHandle");
    core->OMX_FreeHandle = dlsym(lib, "OMX_FreeHandle");
    core->OMX_GetRolesOfComponent = dlsym(lib, "OMX_GetRolesOfComponent");

    if (NULL == core->OMX_Init
        || NULL == core->OMX_Deinit
        || NULL == OMX_ComponentNameEnum
        || NULL == core->OMX_GetHandle
        || NULL == core->OMX_FreeHandle
        || NULL == core->OMX_GetRolesOfComponent)
    {
        LIMERR("non-compliant OpenMAX IL core or imcomplete implementation.\n");
        free(core);
        return -1;
    } else if (OMX_ErrorNone != core->OMX_Init())
    {
        LIMERR("unable to initialize core.\n");
        free(core);
        return -1;
    }

    list_init(&core->list);
    list_add(&core->list, &limoi_omx.core_list_head);

    while (1)
    {
        char name[OMX_MAX_STRINGNAME_SIZE];
        OMX_ERRORTYPE ret;
        OMX_U32 roles_num;
        OMX_U8 ** roles;
        OMX_U8 * names;
        int ii;

        ret = core->OMX_ComponentNameEnum(name, OMX_MAX_STRINGNAME_SIZE, idx++);
        if (OMX_ErrorNoMore == ret)
        {
            LIMDBG("no more component.\n");
            break;
        } else if (OMX_ErrorNone != ret)
        {
            LIMERR("component enumeration error: %x.\n", ret);
            break;
        }

        comp = calloc(1, sizeof(component_reg_t));
        if (NULL == comp)
        {
            LIMERR("out of memory.\n");
            goto bail;
        }

        list_init(&comp->roles_list_head);
        comp->name = strdup(name);
        if (NULL == comp->name)
        {
            LIMERR("out of memory.\n");
            goto bail;
        }

        comp->obj = malloc(sizeof(component_obj_t));
        if (NULL == comp->obj)
        {
            LIMERR("out of memory.\n");
            goto bail;
        }

        list_init(&comp->obj->comps_list_head);
        comp->obj->so = NULL;
        comp->obj->init = NULL;
        comp->obj->core = core;

        if (OMX_ErrorNone !=
            core->OMX_GetRolesOfComponent(name, &roles_num, NULL))
        {
            LIMERR("unable to fetch roles of component: %s\n", name);
            goto bail;
        }
        if (!roles_num)
        {
            LIMERR("no roles found for: %s\n", name);
            goto bail;
        }
        names = malloc(roles_num * OMX_MAX_STRINGNAME_SIZE);
        if (NULL == names)
        {
            LIMERR("out of memory.\n");
            goto bail;
        }
        roles = malloc((roles_num + 1)* sizeof(OMX_U8 *));
        if (NULL == roles)
        {
            LIMERR("out of memory.\n");
            free(names);
            goto bail;
        }
        for (ii = 0; ii < roles_num; ii++)
        {
            roles[ii] = names + ii * OMX_MAX_STRINGNAME_SIZE;
        }
        roles[ii] = NULL;

        core->OMX_GetRolesOfComponent(name, &roles_num, roles);
        ii = register_role((char **)roles, &comp->roles_list_head);
        free(roles);
        free(names);
        if (ii)
        {
            LIMERR("unable to register role.\n");
            goto bail;
        }
        if (!register_component(comp)) num_comp_registered++;
        else
        {
            LIMDBG("unable to register component.\n");
            free((void *)comp->name);
            free(comp->obj);
            free(comp);
        }
        continue;

    bail:
        if (comp)
        {
            if (comp->name) free((void *)comp->name);
            if (comp->obj) free(comp->obj);
            free(comp);
        }
        break;
    }

    if (num_comp_registered <= 0)
    {
        list_del(&core->list);
        free(core);
    }

    return num_comp_registered;
}

static void load_core (const char * cname, void * arg)
{
	void * so;

    LIMDBG("cname: %s\n", cname);
	so = dlopen(cname, RTLD_LAZY | RTLD_GLOBAL);
    if (so)
    {
        int ret = add_core_component(so);
        if (ret > 0)
        {
            *(int *)arg = *(int *)arg + ret;
        } else
        {
            LIMDBG("no component added from: %s\n", cname);
            dlclose(so);
        }
    } else
    {
        LIMDBG("libopen error: %s\n", dlerror());
    }
}

// recursively load external cores from given path.
static int load_cores_recursive (const char * com_path)
{
	int num_components = 0;
    return (limdir_recursive_scan(com_path, load_core, &num_components))?
        0: num_components;
}

OMX_ERRORTYPE OMX_Init (void)
{
    int components;

    lim_resource_manager_init();

    LIMDBG("searching components at: %s\n", componentsdir);
    list_init(&limoi_omx.components_list_head);
    list_init(&limoi_omx.so_list_head);
    list_init(&limoi_omx.core_list_head);
    lthread_mutex_init(&limoi_omx.mutex);

    components = load_components_static();
    components += load_components_recursive(loadersdir);
    components += load_components_recursive(componentsdir);
    components += load_cores_recursive(othercoresdir);

    LIMDBG("total components found: %d\n", components);

    return (components <= 0)? OMX_ErrorUndefined : OMX_ErrorNone;
}

static void release_roles (component_reg_t * comp_reg)
{
    list_t * list = comp_reg->roles_list_head.next;
    while (list != &comp_reg->roles_list_head)
    {
        role_reg_t * reg = (role_reg_t *)list;

        LIMDBG("release role: %s\n", reg->name);

        free((void *)reg->name);

        list = list->next;

        list_del(&reg->list);
        free(reg);
    }
}

static void release_obj (component_reg_t * comp_reg)
{
    list_t * list = comp_reg->obj->comps_list_head.next;
    while (list != &comp_reg->obj->comps_list_head)
    {
        instance_reg_t * reg = (instance_reg_t *)list;

        if (reg->handle) free(reg->handle);

        list = list->next;

        list_del(&reg->list);
        free(reg);
    }

    free(comp_reg->obj);
}

/* The use of OMX_Deinit requires that all component handles acquired by the
   IL client in the system have been released, implying that all resources
   associated with components have been freed.
*/
OMX_ERRORTYPE OMX_Deinit (void)
{
    list_t * list = limoi_omx.components_list_head.next;
    while (list != &limoi_omx.components_list_head)
    {
        component_reg_t * reg = (component_reg_t *)list;

        LIMDBG("deinit: %s\n", reg->name);
        release_roles(reg);
        LIMDBG("deinit: release component object.\n");
        release_obj(reg);
        free((void *)reg->name);

        list = list->next;

        LIMDBG("deinit: remove component registry.\n");
        list_del(&reg->list);
        free(reg);
    }

    LIMDBG("deinit: shutdown all plugins.\n");
    list = limoi_omx.so_list_head.next;
    while (list != &limoi_omx.so_list_head)
    {
        so_reg_t * reg = (so_reg_t *)list;

        if (reg->deregistration)
        {
            LIMDBG("deinit: deregister dynamic components: %s.\n",
                   reg->name);
            if (!reg->lib)
            {
                reg->lib = dlopen(reg->name, RTLD_LAZY);
            }
            if (reg->lib)
            {
                reg->deregistration =
                    dlsym(reg->lib, "OMX_ComponentDeregistration");
                if (reg->deregistration) reg->deregistration();
            }
        }
        if (reg->lib) dlclose(reg->lib);
        free((void *)reg->name);

        list = list->next;

        list_del(&reg->list);
        free(reg);
    }

    LIMDBG("deinit: deinit all external cores.\n");
    list = limoi_omx.core_list_head.next;
    while (list != &limoi_omx.core_list_head)
    {
        core_reg_t * reg = (core_reg_t *)list;

        reg->OMX_Deinit();
        dlclose(reg->lib);

        list = list->next;

        list_del(&reg->list);
        free(reg);
    }
    LIMDBG("deinit: deinit resource manager.\n");
    lim_resource_manager_deinit();

    lthread_mutex_destroy(&limoi_omx.mutex);

    return OMX_ErrorNone;
}

OMX_ERRORTYPE OMX_ComponentNameEnum (
    OMX_STRING cComponentName,
    OMX_U32 nNameLength,
    OMX_U32 nIndex)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;
    list_t * list = limoi_omx.components_list_head.next;
    int idx = -1;

    if (nNameLength < OMX_MAX_STRINGNAME_SIZE || NULL == cComponentName)
    {
        ret = OMX_ErrorBadParameter;
        goto bail;
    }

    //TODO: support runtime enumeration.

    // iterate through registered components
    while (list != &limoi_omx.components_list_head)
    {
        idx++;
        if (idx == nIndex) break;
        list = list->next;
    }

    if (idx == nIndex)
    {
        strcpy((char *)cComponentName, ((component_reg_t *)list)->name);
    } else
    {
        ret = OMX_ErrorNoMore;
    }

 bail:
    return ret;
}
