/*
  Copyright (C) 2010 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX IL clock component test code.

  REVISION:

    1) initial creation. ----------------------------------------- 2010-03-04 MG
*/
#include <sys/stat.h>
#include <fcntl.h>
#include "stdio.h"
#include "stdlib.h"
#include <getopt.h>
#include "OMX_Core.h"
#include "OMX_Types.h"
#include "OMX_ContentPipe.h"
#include "OMX_Component.h"
#include "limlog.h"

#undef LIMDBG
#ifndef LIMDBG
#define LIMDBG(fmt, arg...)
#endif

static OMX_BUFFERHEADERTYPE * bufs[64];

static OMX_ERRORTYPE clock_event_handler (
    OMX_HANDLETYPE hComponent,
    OMX_PTR pAppData,
    OMX_EVENTTYPE eEvent,
    OMX_U32 Data1,
    OMX_U32 Data2,
    OMX_PTR pEventData)
{
    if(eEvent == OMX_EventCmdComplete)
    {
        if (Data1 == OMX_CommandStateSet)
        {
            LIMDBG("clock state set to %u\n", Data2);
        } else if (Data1 == OMX_CommandPortEnable)
        {
            LIMDBG("port %u enabled.\n", Data2);
        } else if (Data1 == OMX_CommandPortDisable)
        {
            LIMDBG("port %u disabled.\n", Data2);
        }
    } else if (eEvent == OMX_EventPortFormatDetected)
    {
        LIMDBG("port %u format detected.\n", Data1);
    } else if (eEvent == OMX_EventPortSettingsChanged)
    {
          LIMDBG("port %u settings changed.\n", Data1);
    } else if (eEvent == OMX_EventBufferFlag)
    {
        printf("\nEOS detected\n");
    } else if (eEvent == OMX_EventError)
    {
    } else
    {
        LIMDBG("param1 = %d param2 = %d\n", (int)Data1, (int)Data2);
    }
    return OMX_ErrorNone;
}

static OMX_ERRORTYPE clock_fill_buffer_done (
    OMX_HANDLETYPE hComponent,
    OMX_PTR pAppData,
    OMX_BUFFERHEADERTYPE* pBuffer)
{
    OMX_TIME_MEDIATIMETYPE * mtype;

    ASSERT(pBuffer && pBuffer->pBuffer);

    mtype = (OMX_TIME_MEDIATIMETYPE *)pBuffer->pBuffer;
    switch (mtype->eUpdateType)
    {
    case OMX_TIME_UpdateRequestFulfillment :
        printf("media time fulfilled as:\n");
        break;

    case OMX_TIME_UpdateScaleChanged :
        printf("clock scale changed as:\n");
        break;

    case OMX_TIME_UpdateClockStateChanged :
        printf("clock state changed as:\n");
        break;

    default :
        LIMERR("bogus udpate event detected: %d\n", mtype->eUpdateType);
        break;
    }

    printf("\tport index    : %u\n"
           "\tprivate       : %p\n"
           "\ttimestamp     : %"PRId64"\n"
           "\toffset        : %"PRId64"\n"
           "\twall time     : %"PRId64"\n"
           "\tscale         : %d\n"
           "\tstate         : %d\n",
           pBuffer->nOutputPortIndex,
           mtype->pClientPrivate,
           mtype->nMediaTimestamp,
           mtype->nOffset,
           mtype->nWallTimeAtMediaTime,
           mtype->xScale,
           mtype->eState);

    OMX_FillThisBuffer(hComponent, pBuffer);
    return OMX_ErrorNone;
}

OMX_CALLBACKTYPE clock_callbacks = {
    .EventHandler    = clock_event_handler,
    .EmptyBufferDone = NULL,
    .FillBufferDone  = clock_fill_buffer_done
};


int main (int argc, char ** argv)
{
    OMX_HANDLETYPE clock;
    char clock_component_name[OMX_MAX_STRINGNAME_SIZE];
    OMX_U32 num_comps = 1;
    OMX_U8 * names[1];
    OMX_PORT_PARAM_TYPE port_param;
    OMX_TIME_CONFIG_TIMESTAMPTYPE ts;
    OMX_TIME_CONFIG_CLOCKSTATETYPE state;
    OMX_TIME_CONFIG_MEDIATIMEREQUESTTYPE rqst;
    OMX_TIME_CONFIG_SCALETYPE scale;
    int ii;

    OMX_Init();

    names[0] = (OMX_U8 *)clock_component_name;
    if (OMX_ErrorNone !=
        OMX_GetComponentsOfRole((OMX_STRING)"clock.binary",
                                &num_comps, names))
    {
        LIMERR("unable to find clock component.\n");
        goto bail;
    }

    if (OMX_ErrorNone !=
        OMX_GetHandle(&clock, clock_component_name,
                      NULL, &clock_callbacks))
    {
        LIMERR("unable to get clock handle.\n");
        goto bail;
    }

    port_param.nSize = sizeof(OMX_PORT_PARAM_TYPE);
    OMX_GetParameter(clock, OMX_IndexParamOtherInit, &port_param);

    if (OMX_ErrorNone !=
        OMX_SendCommand(clock, OMX_CommandStateSet, OMX_StateIdle, NULL))
    {
        LIMERR("unable to set clock to idle.\n");
        goto bail;
    }

    LIMDBG("clock component supports %u ports.\n", port_param.nPorts);
    for (ii = 0; ii < port_param.nPorts * 8; ii++)
    {
        OMX_AllocateBuffer(clock, &bufs[ii], ii / 8, NULL,
                           sizeof(OMX_TIME_MEDIATIMETYPE));
    }

    if (OMX_ErrorNone !=
        OMX_SendCommand(clock,
                        OMX_CommandStateSet, OMX_StateExecuting, NULL))
    {
        LIMERR("unable to execute clock.\n");
        goto bail;
    }

    for (ii = 0; ii < port_param.nPorts * 8; ii++)
    {
        if (OMX_ErrorNone != OMX_FillThisBuffer(clock, bufs[ii]))
        {
            LIMERR("unable to fill buffer %d.\n", ii);
        }
    }

    //query wall time
    ts.nSize = sizeof(OMX_TIME_CONFIG_TIMESTAMPTYPE);
    OMX_GetConfig(clock, OMX_IndexConfigTimeCurrentWallTime, &ts);
    printf("current wall time: %"PRId64"\n", ts.nTimestamp);

    //test waiting for start time.
    state.nSize = sizeof(OMX_TIME_CONFIG_CLOCKSTATETYPE);
    OMX_GetConfig(clock, OMX_IndexConfigTimeClockState, &state);
    state.nWaitMask = 0x1;
    state.eState = OMX_TIME_ClockStateWaitingForStartTime;
    OMX_SetConfig(clock, OMX_IndexConfigTimeClockState, &state);

    ts.nPortIndex = 0;
    OMX_SetConfig(clock, OMX_IndexConfigTimeClientStartTime, &ts);

    //let the clock run a bit
    printf("running for a bit ...\n");
    sleep(3);

    //test scale change.
    scale.nSize = sizeof(OMX_TIME_CONFIG_SCALETYPE);
    scale.xScale = 1 << 16;
    OMX_SetConfig(clock, OMX_IndexConfigTimeScale, &scale);

    //test media time request.
    OMX_GetConfig(clock, OMX_IndexConfigTimeCurrentMediaTime, &ts);
    printf("current media time: %"PRId64"\n", ts.nTimestamp);

    OMX_GetConfig(clock, OMX_IndexConfigTimeCurrentWallTime, &ts);
    printf("current wall time: %"PRId64"\n", ts.nTimestamp);
    OMX_GetConfig(clock, OMX_IndexConfigTimeCurrentMediaTime, &ts);
    printf("current media time: %"PRId64"\n", ts.nTimestamp);

    rqst.nSize = sizeof(OMX_TIME_CONFIG_MEDIATIMEREQUESTTYPE);
    rqst.nPortIndex = 5;
    rqst.pClientPrivate = (void *)0xbeefbeef;
    rqst.nMediaTimestamp = ts.nTimestamp
		+ (((int64_t)2000000 * scale.xScale) >> 16);
    rqst.nOffset = 0;
    printf("current media time: %"PRId64", requesting time: %"PRId64"\n",
           ts.nTimestamp, rqst.nMediaTimestamp);
    OMX_SetConfig(clock, OMX_IndexConfigTimeMediaTimeRequest, &rqst);

    rqst.nSize = sizeof(OMX_TIME_CONFIG_MEDIATIMEREQUESTTYPE);
    rqst.nPortIndex = 7;
    rqst.pClientPrivate = (void *)0xdeaddead;
    rqst.nMediaTimestamp = ts.nTimestamp
		+ (((int64_t)1000000 * scale.xScale) >> 16);
    rqst.nOffset = 0;
    printf("current media time: %"PRId64", requesting time: %"PRId64"\n",
           ts.nTimestamp, rqst.nMediaTimestamp);
    OMX_SetConfig(clock, OMX_IndexConfigTimeMediaTimeRequest, &rqst);

    rqst.nSize = sizeof(OMX_TIME_CONFIG_MEDIATIMEREQUESTTYPE);
    rqst.nPortIndex = 2;
    rqst.pClientPrivate = (void *)0xdeadbeef;
    rqst.nMediaTimestamp = ts.nTimestamp
		+ (((int64_t)3000000 * scale.xScale) >> 16);
    rqst.nOffset = 0;
    printf("current media time: %"PRId64", requesting time: %"PRId64"\n",
           ts.nTimestamp, rqst.nMediaTimestamp);
    OMX_SetConfig(clock, OMX_IndexConfigTimeMediaTimeRequest, &rqst);

    //goodbye
    sleep(4);
    OMX_SendCommand(clock, OMX_CommandStateSet, OMX_StateIdle, NULL);
    OMX_SendCommand(clock, OMX_CommandStateSet, OMX_StateLoaded, NULL);

bail:
    OMX_Deinit();
    return 0;
}
