/*
  Copyright (C) 2009 - 2011 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX IL contentpipe test code.

  REVISION:

    1) initial creation. ----------------------------------------- 2009-12-08 MG
*/

#include "stdio.h"
#include "stdlib.h"
#include "OMX_Core.h"
#include "OMX_Types.h"
#include "OMX_ContentPipe.h"
#include "limlog.h"

#define READ_WRITE_ACCESS_TEST 1
#define WRITE_BUFFER_TEST 1

static int volatile preload = 0;
static int broken = 0;
static void read_test (const char * path);
static void write_test (const char * i_path, const char * o_path);

static void my_log_print (const void * arg, int level,
                          const void * log, int bytes)
{
    printf("%s", (char *)log);
}

int main (int argc, char ** argv)
{
    if (argc != 2 && argc != 3)
    {
        printf("USAGE:\n"
               "\tcontentpipe <in-url>\n"
               "\t\tread test to print contents\n"
               "\tcontentpipe <in-url> <out-url>\n"
               "\t\twrite test, copy contents from in to out.\n");
        exit(0);
    }

    //lim_log_set_level(LIM_LOG_DEBUG);
    lim_log_set_listener(my_log_print, NULL);

    if (argc == 2) read_test(argv[1]);
    else write_test(argv[1], argv[2]);

    return 0;
}

static  CPresult callback (CP_EVENTTYPE eEvent, CPuint iParam)
{
    switch (eEvent)
    {
    case CP_BytesAvailable :
        preload = 1;
        break;

    case CP_PipeDisconnected :
        broken = 1;
        break;

    default :
        break;
    }

    return 0;
}

static void read_test (const char * path)
{
    OMX_HANDLETYPE hdl;
    CP_PIPETYPE * cp;
    CPhandle c_hdl;
    unsigned char buf[32];
    CP_CHECKBYTESRESULTTYPE res;
    char *pBuf;
    CPuint bytes;

    if (OMX_ErrorNone != OMX_GetContentPipe(&hdl, (OMX_STRING)path))
    {
        LIMERR("unable to get contentpipe for %s\n", path);
        exit(-1);
    }
    cp = (CP_PIPETYPE *)hdl;

    if (cp->Open(&c_hdl, (OMX_STRING)path, CP_AccessRead))
    {
        LIMERR("unable to access content: %s\n", path);
        exit(-1);
    }
    cp->RegisterCallback(c_hdl, callback);

    cp->CheckAvailableBytes(c_hdl, 32, &res);
    while (!preload && res == CP_CheckBytesNotReady) ;

    if (1 != preload && res != CP_CheckBytesOk)
    {
        LIMERR("preload failed.\n");
        goto bail;
    }

    LIMINFO("Read interface test.\n");
    while (!cp->Read(c_hdl, (CPbyte *)buf, 32) && !broken)
    {
        int ii;
        for (ii = 0; ii < 32; ii++)
        {
            printf("%02x", buf[ii]);
        }
        printf("\n");
    }

    LIMINFO("ReadBuffer/ReleaseReadBuffer interface test.\n");
    cp->SetPosition(c_hdl, 0, CP_OriginBegin);
    bytes = 32;
    while (!cp->ReadBuffer(c_hdl, &pBuf, &bytes, OMX_TRUE) && !broken)
    {
        while (bytes)
        {
            int tmp = (bytes > 32)? 32 : bytes;
            int ii;
            for (ii = 0; ii < tmp; ii++)
            {
                printf("%02x", pBuf[ii]);
            }
            printf("\n");
            bytes -= tmp;
        }
        cp->ReleaseReadBuffer(c_hdl, pBuf);
        bytes = 32;
    }

 bail:
    cp->Close(c_hdl);
}

static void write_test (const char * i_path, const char * o_path)
{
    CP_PIPETYPE * i_cp;
    CP_PIPETYPE * o_cp;
    CPhandle ic_hdl;
    CPhandle oc_hdl;
    char *pBuf;
    CPuint bytes;

    if (OMX_ErrorNone !=
        OMX_GetContentPipe((OMX_HANDLETYPE*)&i_cp, (OMX_STRING)i_path))
    {
        LIMERR("unable to get contentpipe for %s\n", i_path);
        exit(-1);
    }
    if (OMX_ErrorNone !=
        OMX_GetContentPipe((OMX_HANDLETYPE*)&o_cp, (OMX_STRING)o_path))
    {
        LIMERR("unable to get contentpipe for %s\n", o_path);
        exit(-1);
    }

    if (i_cp->Open(&ic_hdl, (OMX_STRING)i_path, CP_AccessRead))
    {
        LIMERR("unable to access content: %s\n", i_path);
        exit(-1);
    }
#ifdef READ_WRITE_ACCESS_TEST
    if (o_cp->Open(&oc_hdl, (OMX_STRING)o_path, CP_AccessReadWrite))
#else
    if (o_cp->Open(&oc_hdl, (OMX_STRING)o_path, CP_AccessWrite))
#endif
    {
        LIMERR("unable to open path: %s\n", o_path);
        i_cp->Close(ic_hdl);
        exit(-1);
    }

    //FIXOMX: Read API does not return bytes read, thus the use of ReadBuffer.
    while (1)
    {
        CPresult ret;

        bytes = 1024 * 1024;
        ret = i_cp->ReadBuffer(ic_hdl, &pBuf, &bytes, OMX_TRUE);
        if (ret == KD_EAGAIN) {usleep(50000); continue;}
        if (ret) break;
        LIMDBG("bytes read: %lu\n", bytes);

#ifdef WRITE_BUFFER_TEST
        {
            CPbyte * wr_buf;
            CPbyte * buf = pBuf;
            CPuint bytes_write;
            CPuint bytes_filled = bytes;
            while (bytes)
            {
                bytes_write = bytes;
                if (o_cp->GetWriteBuffer(oc_hdl, &wr_buf, &bytes_write))
                {
                    LIMERR("unable to get write buffer.\n");
                    goto bail;
                }
                memcpy(wr_buf, buf, bytes_write);
                bytes -= bytes_write;
                buf += bytes_write;
            }
            if (o_cp->WriteBuffer(oc_hdl, wr_buf, bytes_filled))
            {
                LIMERR("unable to write buffer back.\n");
                goto bail;
            }
        }
#else
    retry:
        ret = o_cp->Write(oc_hdl, pBuf, bytes);
        if (ret == KD_EAGAIN) {usleep(50000); goto retry;}
#endif

        i_cp->ReleaseReadBuffer(ic_hdl, pBuf);

        if (ret)
        {
            LIMERR("unable to write: %lu bytes\n", bytes);
            break;
        }
    }
#ifdef WRITE_BUFFER_TEST
 bail:
#endif
    i_cp->Close(ic_hdl);
    o_cp->Close(oc_hdl);

    LIMINFO("content pipe write test done.\n");
}
