/*
  Copyright (C) 2010 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX IL demuxer test code.

  REVISION:

    1) initial creation. ----------------------------------------- 2010-01-29 MG
*/

#include "stdio.h"
#include "stdlib.h"
#include "OMX_Core.h"
#include "OMX_Types.h"
#include "OMX_ContentPipe.h"
#include "OMX_Component.h"
#include "limlog.h"

#define AUDIO_OUT_BUF_BYTES (1024 * 8)
#define VIDEO_OUT_BUF_BYTES (1024 * 512)

static OMX_BUFFERHEADERTYPE * a_buf1;
static OMX_BUFFERHEADERTYPE * a_buf2;
static OMX_BUFFERHEADERTYPE * v_buf1;
static OMX_BUFFERHEADERTYPE * v_buf2;
static int eof;

//simple test code, no locking/sync mechanism implemented
static int port0disable;
static int port1disable;

static void dump_port_formats (
    OMX_HANDLETYPE hComponent,
    OMX_U32 port)
{
    OMX_PARAM_U32TYPE num_streams;
    OMX_AUDIO_PARAM_PORTFORMATTYPE audio_port_format;
    OMX_VIDEO_PARAM_PORTFORMATTYPE video_port_format;
    int ii;

    //FIXME: specify spec version.
    num_streams.nSize = sizeof(OMX_PARAM_U32TYPE);
    num_streams.nPortIndex = port;
    if (OMX_ErrorNone !=
        OMX_GetParameter(hComponent, OMX_IndexParamNumAvailableStreams,
                         &num_streams))
    {
        LIMERR("unable to get available streams.\n");
        exit(-1);
    }
    LIMDBG("port %u available streams: %u\n", port, num_streams.nU32);
    if (!num_streams.nU32)
    {
        LIMWARN("port %u has zero stream.\n", port);
        exit(-1);
    }

    if (port == 0) // audio port
    {
        audio_port_format.nSize = sizeof(OMX_AUDIO_PARAM_PORTFORMATTYPE);
        audio_port_format.nPortIndex = 0;
        for (ii = 0; ii < num_streams.nU32; ii++)
        {
            audio_port_format.nIndex = ii;
            if (OMX_ErrorNone !=
                OMX_GetParameter(hComponent, OMX_IndexParamAudioPortFormat,
                                 &audio_port_format))
            {
                LIMERR("unable to get port 0 format with index %d\n", ii);
                exit(-1);
            }
            switch (audio_port_format.eEncoding)
            {
            case OMX_AUDIO_CodingVORBIS :
                printf("vorbis audio detected on port 0.\n");
                break;

            case OMX_AUDIO_CodingMP3 :
                printf("mp3 audio detected on port 0.\n");
                break;

            case OMX_AUDIO_CodingAAC :
                printf("aac audio detected on port 0.\n");
                break;

            default :
                printf("%d audio detected on port 0.\n",
                       audio_port_format.eEncoding);
                break;
            }
        }
    } else if (port == 1) //video port
    {
        video_port_format.nSize = sizeof(OMX_VIDEO_PARAM_PORTFORMATTYPE);
        video_port_format.nPortIndex = 1;
        for (ii = 0; ii < num_streams.nU32; ii++)
        {
            video_port_format.nIndex = ii;
            if (OMX_ErrorNone !=
                OMX_GetParameter(hComponent, OMX_IndexParamVideoPortFormat,
                                 &video_port_format))
            {
                LIMERR("unable to get port 1 format with index %d\n", ii);
                exit(-1);
            }
            switch (video_port_format.eCompressionFormat)
            {
            case OMX_VIDEO_CodingMPEG2 :
                printf("mpeg2 video detected on port 1.\n");
                break;

            case OMX_VIDEO_CodingMPEG4 :
                printf("mpeg4 video detected on port 1.\n");
                break;

            case OMX_VIDEO_CodingH263 :
                printf("h263 video detected on port 1.\n");
                break;

            case OMX_VIDEO_CodingAVC :
                printf("h264 video detected on port 1.\n");
                break;

            case OMX_VIDEO_CodingWMV :
                printf("wmv video detected on port 1.\n");
                break;

            default :
                printf("%d video detected on port 1.\n",
                       video_port_format.eCompressionFormat);
                break;
            }
        }
    }
}

static void activate_port (
    OMX_HANDLETYPE hComponent,
    OMX_U32 port)
{
    OMX_PARAM_U32TYPE active_stream;

    active_stream.nSize = sizeof(OMX_PARAM_U32TYPE);
    active_stream.nPortIndex = port;
    active_stream.nU32 = 0;
    if (OMX_ErrorNone !=
        OMX_SetParameter(hComponent, OMX_IndexParamActiveStream,
                         &active_stream))
    {
        LIMERR("unable to set active stream to %u.\n", active_stream.nU32);
        exit(-1);
    }

    OMX_SendCommand(hComponent,
                    OMX_CommandPortEnable, port, NULL);

    if (port == 0)
    {
        OMX_AllocateBuffer(hComponent, &a_buf1, 0, NULL,
                           AUDIO_OUT_BUF_BYTES);
        OMX_AllocateBuffer(hComponent, &a_buf2, 0, NULL,
                           AUDIO_OUT_BUF_BYTES);
        OMX_FillThisBuffer(hComponent, a_buf1);
        OMX_FillThisBuffer(hComponent, a_buf2);
    } else if (port == 1)
    {
        OMX_AllocateBuffer(hComponent, &v_buf1, 1, NULL,
                           VIDEO_OUT_BUF_BYTES);
        OMX_AllocateBuffer(hComponent, &v_buf2, 1, NULL,
                           VIDEO_OUT_BUF_BYTES);
        OMX_FillThisBuffer(hComponent, v_buf1);
        OMX_FillThisBuffer(hComponent, v_buf2);
    }
}

static OMX_ERRORTYPE demuxer_event_handler (
    OMX_HANDLETYPE hComponent,
    OMX_PTR pAppData,
    OMX_EVENTTYPE eEvent,
    OMX_U32 Data1,
    OMX_U32 Data2,
    OMX_PTR pEventData)
{
    if(eEvent == OMX_EventCmdComplete)
    {
        if (Data1 == OMX_CommandStateSet)
        {
            LIMDBG("demuxer state set to %u\n", Data2);
        } else if (Data1 == OMX_CommandPortEnable)
        {
            LIMDBG("port %u enabled.\n", Data2);
        } else if (Data1 == OMX_CommandPortDisable)
        {
            LIMDBG("port %u disabled.\n", Data2);
            if (Data2 != 2) activate_port(hComponent, Data2);
        }
    } else if (eEvent == OMX_EventPortFormatDetected)
    {
        LIMDBG("port %u format detected.\n", Data1);
    } else if (eEvent == OMX_EventPortSettingsChanged)
    {
          LIMDBG("port %u settings changed.\n", Data1);
          dump_port_formats(hComponent, Data1);

          if (Data1 == 0) port0disable = 2;
          else if (Data1 == 1) port1disable = 2;
          OMX_SendCommand(hComponent,
                          OMX_CommandPortDisable, Data1, NULL);
    } else if (eEvent == OMX_EventBufferFlag)
    {
        printf("\nEOS detected\n");
        eof = 1;
    } else if (eEvent == OMX_EventError)
    {
    } else
    {
        LIMDBG("param1 = %d param2 = %d\n", (int)Data1, (int)Data2);
    }
    return OMX_ErrorNone;
}

static OMX_ERRORTYPE demuxer_empty_buffer_done (
    OMX_HANDLETYPE hComponent,
    OMX_PTR pAppData,
    OMX_BUFFERHEADERTYPE* pBuffer)
{
    return OMX_ErrorNone;
}

static OMX_ERRORTYPE demuxer_fill_buffer_done (
    OMX_HANDLETYPE hComponent,
    OMX_PTR pAppData,
    OMX_BUFFERHEADERTYPE* pBuffer)
{
    static unsigned int dbg_count;

    if (port0disable && pBuffer->nOutputPortIndex == 0)
    {
        OMX_FreeBuffer(hComponent, 0, pBuffer);
        port0disable--;
    } else if (port1disable && pBuffer->nOutputPortIndex == 1)
    {
        OMX_FreeBuffer(hComponent, 1, pBuffer);
        port1disable--;
    } else
    {
        if (dbg_count++ < 32)
        {
            printf("port %u buffer filled with %u bytes.\n",
                   pBuffer->nOutputPortIndex, pBuffer->nFilledLen);
        } else if (!eof)
        {
            printf("packets count: %d\r", dbg_count);
        }

        if (!eof)
        {
            pBuffer->nFilledLen = 0;
            OMX_FillThisBuffer(hComponent, pBuffer);
        } else
        {
            OMX_FreeBuffer(hComponent, pBuffer->nOutputPortIndex, pBuffer);
        }
    }

    return OMX_ErrorNone;
}

OMX_CALLBACKTYPE demuxer_callbacks = {
    .EventHandler    = demuxer_event_handler,
    .EmptyBufferDone = demuxer_empty_buffer_done,
    .FillBufferDone  = demuxer_fill_buffer_done
};

//reset port buffer numbers to minimum for non-tunnel demo purpose.
static void reconfig_component_port (OMX_HANDLETYPE comp, int port, int bytes)
{
    OMX_PARAM_PORTDEFINITIONTYPE port_def;

    port_def.nSize = sizeof(OMX_PARAM_PORTDEFINITIONTYPE);
    port_def.nPortIndex = port;
    if (OMX_ErrorNone != OMX_GetParameter
        (comp, OMX_IndexParamPortDefinition, &port_def))
    {
        LIMERR("unable to get port %d definition.\n", port);
        exit(-1);
    } else
    {
        port_def.nBufferCountActual = 2;
        port_def.nBufferCountMin = 2;
        port_def.nBufferSize = bytes;
        OMX_SetParameter(comp, OMX_IndexParamPortDefinition, &port_def);
    }
}

static void my_log_print (const void * arg, int level,
                          const void * log, int bytes)
{
    printf("%s", (char *)log);
}

int main (int argc, char ** argv)
{
    OMX_HANDLETYPE demuxer;
    char demux_component_name[OMX_MAX_STRINGNAME_SIZE];
    OMX_U32 num_comps = 1;
    OMX_U8 * names[1];
    OMX_PARAM_CONTENTURITYPE * uri_param;
    int bytes;

    //lim_log_set_level(LIM_LOG_DEBUG);
    lim_log_set_listener(my_log_print, NULL);
    if (argc != 2)
    {
        printf("USAGE: demuxer <file-name>\n");
        exit(0);
    } else
    {
        LIMDBG("demuxing %s\n", argv[1]);
    }

    OMX_Init();

    names[0] = (OMX_U8 *)demux_component_name;
    if (OMX_ErrorNone !=
        OMX_GetComponentsOfRole((OMX_STRING)"container_demuxer.all",
                                &num_comps, names))
    {
        LIMERR("unable to find demuxer component.\n");
        goto bail;
    }

    if (OMX_ErrorNone !=
        OMX_GetHandle(&demuxer, demux_component_name,
                      NULL, &demuxer_callbacks))
    {
        LIMERR("unable to get demuxer handle.\n");
        goto bail;
    }
    //disable clock port.
    OMX_SendCommand(demuxer,
                    OMX_CommandPortDisable, 2, NULL);

    bytes = strlen(argv[1]);
    uri_param = (OMX_PARAM_CONTENTURITYPE *)malloc
        (sizeof(OMX_PARAM_CONTENTURITYPE) + bytes);
    uri_param->nSize = sizeof(OMX_PARAM_CONTENTURITYPE) + bytes;
    strcpy((char *)uri_param->contentURI, argv[1]);
    OMX_SetParameter(demuxer, OMX_IndexParamContentURI, uri_param);
    free(uri_param);

    reconfig_component_port(demuxer, 0, AUDIO_OUT_BUF_BYTES);
    reconfig_component_port(demuxer, 1, VIDEO_OUT_BUF_BYTES);
    if (OMX_ErrorNone !=
        OMX_SendCommand(demuxer, OMX_CommandStateSet, OMX_StateIdle, NULL))
    {
        LIMERR("unable to set demuxer to idle.\n");
        goto bail;
    }

    if (OMX_ErrorNone !=
        OMX_AllocateBuffer(demuxer,
                           &a_buf1, 0, NULL,
                           AUDIO_OUT_BUF_BYTES))
    {
        LIMERR("Unable to allocate audio output buffer\n");
        goto bail;
    }

    if (OMX_ErrorNone !=
        OMX_AllocateBuffer(demuxer,
                           &a_buf2, 0, NULL,
                           AUDIO_OUT_BUF_BYTES))
    {
        LIMERR("Unable to allocate audio output buffer2\n");
        goto bail;
    }

    if (OMX_ErrorNone !=
        OMX_AllocateBuffer(demuxer,
                           &v_buf1, 1, NULL,
                           VIDEO_OUT_BUF_BYTES))
    {
        LIMERR("Unable to allocate audio output buffer\n");
        goto bail;
    }

    if (OMX_ErrorNone !=
        OMX_AllocateBuffer(demuxer,
                           &v_buf2, 1, NULL,
                           VIDEO_OUT_BUF_BYTES))
    {
        LIMERR("Unable to allocate audio output buffer2\n");
        goto bail;
    }

    if (OMX_ErrorNone !=
        OMX_SendCommand(demuxer,
                        OMX_CommandStateSet, OMX_StateExecuting, NULL))
    {
        LIMERR("unable to execute demuxer.\n");
        goto bail;
    }

    OMX_FillThisBuffer(demuxer, a_buf1);
    OMX_FillThisBuffer(demuxer, a_buf2);
    OMX_FillThisBuffer(demuxer, v_buf1);
    OMX_FillThisBuffer(demuxer, v_buf2);

    while (!eof) sleep(1);

bail:
    OMX_Deinit();
    return 0;
}
