/*
  Copyright (C) 2010 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX IL encoder test code.

  REVISION:

    1) initial creation. ----------------------------------------- 2010-07-10 MG
*/

#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include "OMX_Core.h"
#include "OMX_Types.h"
#include "OMX_ContentPipe.h"
#include "OMX_Component.h"
#include "limlog.h"

#define AUDIO_INPUT_BUF_BYTES (1024 * 192)
#define AUDIO_OUTPUT_BUF_BYTES (1024 * 32)
#define VIDEO_INPUT_BUF_BYTES (1024 * 512)
#define VIDEO_OUTPUT_BUF_BYTES (1024 * 64)

static volatile int eof;
static OMX_BUFFERHEADERTYPE * i_buf1;
static OMX_BUFFERHEADERTYPE * i_buf2;
static OMX_BUFFERHEADERTYPE * o_buf1;
static OMX_BUFFERHEADERTYPE * o_buf2;
static int in_buf_bytes;
static int out_buf_bytes;

typedef struct
{
    int a_samplerate; /* audio samplerate. */
    int a_channels; /* audio channels. */
    int a_bitrate; /* audio output bitrate. */
    int v_width; /* video input width. */
    int v_height; /* video input height. */
    int v_framerate; /* video input framerate. */
    int v_bitrate; /* vidoe output bitrate. */
    int codec; /* codec type, dictates test case either for audio or video. */
    char * in_fname;  /* input file name. */
    char * out_fname; /* output file name. */
} options_t;

typedef struct
{
    options_t options;
    int in_fd;
    int out_fd;
    OMX_HANDLETYPE encoder;
} main_context_t;

enum {
    OPT_SAMPLERATE = 's',
    OPT_CHANNELS = 'c',
    OPT_BITRATE = 'b',
    OPT_WIDTH = 'w',
    OPT_HEIGHT = 'h',
    OPT_FRAME_RATE = 'r',
    OPT_CODEC = 'x',
    OPT_IN_FNAME = 'i',
    OPT_OUT_FNAME = 'o',
    OPT_HELP = '?',
};

static struct option cmd_opts_long[] = {
    {"help", 0, 0, '?'},
    { NULL, 0, NULL, 0 }
};
#define cmd_opts_short "?s:c:b:w:h:r:x:i:o:"

static void print_help(void)
{
    printf("to encode audio:\n"
           "   encode -s PCM_SAMPLERATE -c PCM_CHANNELS -b OUTPUT_BITRATE\n"
           "          -x AUDIO_CODEC [mp3]\n"
           "          -i PCM_INPUT_FILE -o OUTPUT_FILE\n"
           "to encode video:\n"
           "   encode -w WIDTH -h HEIGHT -r FRAMERATE -b OUTPUT_BITRATE\n"
           "          -x VIDEO_CODEC [h263 mpeg4\n"
           "          -i YUV_INPUT_FILE -o OUTPUT_FILE\n");
}

static int get_options (
    options_t * opts,
    int argc,
    char ** argv)
{
    int opt, index;
    char * codec = NULL;

    while ((opt = getopt_long(argc, argv,
                              cmd_opts_short,
                              cmd_opts_long,
                              &index)) != -1)
    {
        switch (opt)
        {
        case OPT_SAMPLERATE :
            opts->a_samplerate = atoi(optarg);
            break;

        case OPT_CHANNELS :
            opts->a_channels = atoi(optarg);
            break;

        case OPT_BITRATE :
            opts->a_bitrate = opts->v_bitrate = atoi(optarg);
            break;

        case OPT_WIDTH :
            opts->v_width = atoi(optarg);
            break;

        case OPT_HEIGHT :
            opts->v_height = atoi(optarg);
            break;

        case OPT_FRAME_RATE :
            opts->v_framerate = atoi(optarg);
            break;

        case OPT_CODEC :
            if (!strcmp(optarg, "mp3"))
                opts->codec = OMX_AUDIO_CodingMP3;
            else if (!strcmp(optarg, "aac"))
                opts->codec = OMX_AUDIO_CodingAAC;
            else if (!strcmp(optarg, "h263"))
                opts->codec = OMX_VIDEO_CodingH263;
            else if (!strcmp(optarg, "mpeg4"))
                opts->codec = OMX_VIDEO_CodingMPEG4;
            else
            {
                LIMERR("invalid codec: %s\n", optarg);
                goto bail;
            }
            codec = strdup(optarg);
            break;

        case OPT_IN_FNAME :
            opts->in_fname = optarg;
            break;

        case OPT_OUT_FNAME :
            opts->out_fname = optarg;
            break;

        default :
            LIMERR("invalid command line option : %s\n", optarg);
            //run through

        case OPT_HELP :
            goto bail;
        }
    }

    if (-1 == opts->codec || NULL == opts->in_fname)
    {
        LIMERR("mandatory codec and input file name parameter missing.\n");
        goto bail;
    }

    if (opts->v_width < 32 || opts->v_height < 32)
    {
        LIMERR("invalid parameter w = %d or h = %d\n",
               opts->v_width, opts->v_height);
        goto bail;
    }

    if (NULL == opts->out_fname)
    {
        opts->out_fname = malloc(64);
        strcpy(opts->out_fname, "output.");
        strcat(opts->out_fname, codec);
    }
    free(codec);

    return 0;

 bail:
    if (codec) free(codec);
    print_help();
    return -1;
}

static int get_input (main_context_t * ctx, OMX_BUFFERHEADERTYPE* buf)
{
    if (in_buf_bytes != read(ctx->in_fd, buf->pBuffer, in_buf_bytes))
    {
        return -1;
    } else
    {
        buf->nOffset = 0;
        buf->nFilledLen = in_buf_bytes;
    }
    return 0;
}

static OMX_ERRORTYPE encoder_event_handler (
    OMX_HANDLETYPE hComponent,
    OMX_PTR pAppData,
    OMX_EVENTTYPE eEvent,
    OMX_U32 Data1,
    OMX_U32 Data2,
    OMX_PTR pEventData)
{
    if(eEvent == OMX_EventCmdComplete)
    {
        if (Data1 == OMX_CommandStateSet)
        {
            LIMDBG("encoder state set to %u\n", Data2);
        } else if (Data1 == OMX_CommandPortEnable)
        {
            LIMDBG("port %u enabled.\n", Data2);
        } else if (Data1 == OMX_CommandPortDisable)
        {
            LIMDBG("port %u disabled.\n", Data2);
        }
    } else if (eEvent == OMX_EventPortFormatDetected)
    {
        LIMDBG("port %u format detected.\n", Data1);
    } else if (eEvent == OMX_EventPortSettingsChanged)
    {
        LIMDBG("port %u settings changed.\n", Data1);
    } else if (eEvent == OMX_EventBufferFlag)
    {
        printf("\nEOS detected\n");
    } else if (eEvent == OMX_EventError)
    {
        LIMERR("error event detected.\n");
    } else
    {
        LIMDBG("param1 = %d param2 = %d\n", (int)Data1, (int)Data2);
    }
    return OMX_ErrorNone;
}

static OMX_ERRORTYPE encoder_empty_buffer_done (
    OMX_HANDLETYPE hComponent,
    OMX_PTR pAppData,
    OMX_BUFFERHEADERTYPE* pBuffer)
{
    main_context_t * ctx = (main_context_t *)pAppData;

    LIMDBG("encoder empty buffer done: %u.\n", pBuffer->nFilledLen);
    if (!get_input(ctx, pBuffer))
    {
        OMX_EmptyThisBuffer(hComponent, pBuffer);
    } else eof = 1;

    return OMX_ErrorNone;
}

static OMX_ERRORTYPE encoder_fill_buffer_done (
    OMX_HANDLETYPE hComponent,
    OMX_PTR pAppData,
    OMX_BUFFERHEADERTYPE* pBuffer)
{
    main_context_t * ctx = (main_context_t *)pAppData;

    LIMDBG("encoder fill buffer done: %u.\n", pBuffer->nFilledLen);
    if (pBuffer->nFilledLen !=
        write(ctx->out_fd, pBuffer->pBuffer, pBuffer->nFilledLen))
    {
        LIMERR("unable to write to output file.\n");
        eof = 1;
    } else
    {
        pBuffer->nFilledLen = 0;
        OMX_FillThisBuffer(hComponent, pBuffer);
    }
    return OMX_ErrorNone;
}

OMX_CALLBACKTYPE encoder_callbacks = {
    .EventHandler    = encoder_event_handler,
    .EmptyBufferDone = encoder_empty_buffer_done,
    .FillBufferDone  = encoder_fill_buffer_done
};

typedef struct {
    int codec;
    char * role;
} codec_role_map_t;
static codec_role_map_t CODEC_ROLE_MAP[] = {
    {OMX_AUDIO_CodingMP3, "audio_encoder.mp3"},
    {OMX_AUDIO_CodingAAC, "audio_encoder.aac"},
    {OMX_VIDEO_CodingMPEG4, "video_encoder.mpeg4"},
    {OMX_VIDEO_CodingH263, "video_encoder.h263"},
    {-1, NULL}};

static int get_component (main_context_t * ctx)
{
    codec_role_map_t * role_map = &CODEC_ROLE_MAP[0];
    char component_name[OMX_MAX_STRINGNAME_SIZE];
    OMX_U32 num_comps = 1;
    OMX_U8 * names[1];

    while (role_map->role)
    {
        if (role_map->codec == ctx->options.codec) break;
        role_map++;
    }

    if (NULL == role_map->role)
    {
        LIMERR("unsupported codec: %d\n", ctx->options.codec);
        goto bail;
    }

    names[0] = (OMX_U8 *)component_name;
    if (OMX_ErrorNone !=
        OMX_GetComponentsOfRole((OMX_STRING)role_map->role,
                                &num_comps, names))
    {
        LIMERR("unable to find encoder component for role: %s.\n",
               role_map->role);
        goto bail;
    }
    LIMDBG("encoder component %s found.\n", component_name);

    if (OMX_ErrorNone !=
        OMX_GetHandle(&ctx->encoder, component_name,
                      ctx, &encoder_callbacks))
    {
        LIMERR("unable to get %s handle.\n", component_name);
        goto bail;
    }
    LIMDBG("encoder component handle ready.\n");
    return 0;

 bail:
    return -1;
}

//reset port buffer numbers to minimum for non-tunnel demo purpose.
static void reconfig_component_port (OMX_HANDLETYPE comp, int port, int bytes)
{
    OMX_PARAM_PORTDEFINITIONTYPE port_def;

    port_def.nSize = sizeof(OMX_PARAM_PORTDEFINITIONTYPE);
    port_def.nPortIndex = port;
    if (OMX_ErrorNone != OMX_GetParameter
        (comp, OMX_IndexParamPortDefinition, &port_def))
    {
        LIMERR("unable to get port %d definition.\n", port);
        exit(-1);
    } else
    {
        port_def.nBufferCountActual = 2;
        port_def.nBufferCountMin = 2;
        port_def.nBufferSize = bytes;
        OMX_SetParameter(comp, OMX_IndexParamPortDefinition, &port_def);
    }
}

static int reconfig_audio_ports (main_context_t * ctx)
{
    OMX_AUDIO_PARAM_PCMMODETYPE pcm;
    OMX_PARAM_COMPONENTROLETYPE role;

    /* set component role. */
    role.nSize = sizeof(OMX_PARAM_COMPONENTROLETYPE);
    strcpy((char *)role.cRole, (ctx->options.codec == OMX_AUDIO_CodingMP3)?
           "audio_encoder.mp3" : "audio_encoder.aac");
    if (OMX_ErrorNone != OMX_SetParameter
        (ctx->encoder, OMX_IndexParamStandardComponentRole, &role))
    {
        LIMERR("unable to set encoder role.\n");
        return -1;
    }

    reconfig_component_port(ctx->encoder, 0, AUDIO_INPUT_BUF_BYTES);
    reconfig_component_port(ctx->encoder, 1, AUDIO_OUTPUT_BUF_BYTES);

    pcm.nSize = sizeof(OMX_AUDIO_PARAM_PCMMODETYPE);
    pcm.nPortIndex = 0;
    if (OMX_ErrorNone != OMX_GetParameter
        (ctx->encoder, OMX_IndexParamAudioPcm, &pcm))
    {
        LIMWARN("unable to get encoder PCM parameter.\n");
    }
    pcm.nChannels = ctx->options.a_channels;
    pcm.eNumData = OMX_NumericalDataSigned;
    pcm.eEndian = OMX_EndianMax;
    pcm.bInterleaved = OMX_TRUE;
    pcm.nBitPerSample = 16;
    pcm.nSamplingRate = ctx->options.a_samplerate;
#if 0
    pcm.ePCMMODE = OMX_AUDIO_PCMModeLinear;
    pcm.eChannelMapping[0] = OMX_AUDIO_ChannelLF;
    pcm.eChannelMapping[1] = OMX_AUDIO_ChannelRF;
#endif
    if (OMX_ErrorNone != OMX_SetParameter
        (ctx->encoder, OMX_IndexParamAudioPcm, &pcm))
    {
        LIMERR("unable to set encoder PCM parameter.\n");
        return -1;
    }

    if (ctx->options.codec == OMX_AUDIO_CodingMP3)
    {
        OMX_AUDIO_PARAM_MP3TYPE mp3;

        mp3.nSize = sizeof(OMX_AUDIO_PARAM_MP3TYPE);
        mp3.nPortIndex = 1;
        if (OMX_ErrorNone != OMX_GetParameter
            (ctx->encoder, OMX_IndexParamAudioMp3, &mp3))
        {
            LIMWARN("unable to get encoder MP3 parameter.\n");
        }

        mp3.nChannels = ctx->options.a_channels;
        mp3.nBitRate = ctx->options.a_bitrate;
        mp3.nSampleRate = ctx->options.a_samplerate;

        if (OMX_ErrorNone != OMX_SetParameter
            (ctx->encoder, OMX_IndexParamAudioMp3, &mp3))
        {
            LIMERR("unable to set encoder MP3 parameter.\n");
            return -1;
        }
    } else
    {
        //TODO: AAC
    }

    return 0;
}

static int reconfig_video_ports (main_context_t * ctx)
{
    OMX_PARAM_PORTDEFINITIONTYPE port_def;
    OMX_VIDEO_PORTDEFINITIONTYPE * video = &port_def.format.video;
    OMX_PARAM_COMPONENTROLETYPE role;

    /* set component role. */
    role.nSize = sizeof(OMX_PARAM_COMPONENTROLETYPE);
    strcpy((char *)role.cRole, (ctx->options.codec == OMX_VIDEO_CodingH263)?
           "video_encoder.h263" : "video_encoder.mpeg4");
    if (OMX_ErrorNone != OMX_SetParameter
        (ctx->encoder, OMX_IndexParamStandardComponentRole, &role))
    {
        LIMERR("unable to set encoder role.\n");
        return -1;
    }

    reconfig_component_port(ctx->encoder, 0, VIDEO_INPUT_BUF_BYTES);
    reconfig_component_port(ctx->encoder, 1, VIDEO_OUTPUT_BUF_BYTES);

    // configure input parameters
    port_def.nSize = sizeof(OMX_PARAM_PORTDEFINITIONTYPE);
    port_def.nPortIndex = 0;
    OMX_GetParameter(ctx->encoder, OMX_IndexParamPortDefinition, &port_def);

    video->nFrameWidth = ctx->options.v_width;
    video->nFrameHeight = ctx->options.v_height;
    video->nStride = ctx->options.v_width;
    video->xFramerate = ctx->options.v_framerate << 16;
    // default values, can optionally ignore
    //video->eColorFormat = OMX_COLOR_FormatYUV420Planar;
    //video->eCompressionFormat = OMX_VIDEO_CodingUnused;
    OMX_SetParameter(ctx->encoder, OMX_IndexParamPortDefinition, &port_def);

    // configure output parameters
    port_def.nSize = sizeof(OMX_PARAM_PORTDEFINITIONTYPE);
    port_def.nPortIndex = 1;
    OMX_GetParameter(ctx->encoder, OMX_IndexParamPortDefinition, &port_def);

    video->nFrameWidth = ctx->options.v_width;
    video->nFrameHeight = ctx->options.v_height;
    video->nBitrate = ctx->options.v_bitrate;
    video->xFramerate = ctx->options.v_framerate << 16;
    video->eCompressionFormat = ctx->options.codec;
    OMX_SetParameter(ctx->encoder, OMX_IndexParamPortDefinition, &port_def);

    return 0;
}

static void my_log_print (const void * arg, int level,
                          const void * log, int bytes)
{
    printf("%s", (char *)log);
}

int main (int argc, char ** argv)
{
    main_context_t main_ctx = {
        .options.a_samplerate = 44100,
        .options.a_channels = 2,
        .options.a_bitrate = 128000,
        .options.v_width = 720,
        .options.v_height = 480,
        .options.v_framerate = 30,
        .options.v_bitrate = 1000000,
        .options.codec = -1,
        .options.in_fname = NULL,
        .options.out_fname = NULL,
        .in_fd = -1,
        .out_fd = -1,
    };
    lim_log_set_level(LIM_LOG_DEBUG);
    lim_log_set_level(LIM_LOG_INFO);
    lim_log_set_listener(my_log_print, NULL);
    if (get_options(&main_ctx.options, argc, argv))
    {
        exit(-1);
    }

    main_ctx.in_fd = open(main_ctx.options.in_fname, O_RDONLY, 0);
    if (main_ctx.in_fd < 0)
    {
        LIMERR("unable to open input: %s\n", main_ctx.options.in_fname);
        exit(-1);
    }
    main_ctx.out_fd = open(main_ctx.options.out_fname,
                           O_WRONLY | O_CREAT | O_TRUNC,
                           S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
    if (main_ctx.out_fd < 0)
    {
        LIMERR("unable to create output: %s.\n", main_ctx.options.out_fname);
        close(main_ctx.in_fd);
        exit(-1);
    }

    OMX_Init();

    if (get_component(&main_ctx))
    {
        exit(-1);
    }

    if (main_ctx.options.codec == OMX_AUDIO_CodingMP3
        || main_ctx.options.codec == OMX_AUDIO_CodingAAC)
    {
        if (reconfig_audio_ports(&main_ctx)) goto bail;
        in_buf_bytes = AUDIO_INPUT_BUF_BYTES;
        out_buf_bytes = AUDIO_OUTPUT_BUF_BYTES;
    } else
    {
        if (reconfig_video_ports(&main_ctx)) goto bail;

        // hard-coded to YUV420 planner
        in_buf_bytes = main_ctx.options.v_width
            * main_ctx.options.v_height * 3 / 2;
        out_buf_bytes = VIDEO_OUTPUT_BUF_BYTES;
    }

    OMX_SendCommand(main_ctx.encoder, OMX_CommandStateSet,
                    OMX_StateIdle, NULL);
    OMX_AllocateBuffer(main_ctx.encoder, &i_buf1, 0, NULL, in_buf_bytes);
    OMX_AllocateBuffer(main_ctx.encoder, &i_buf2, 0, NULL, in_buf_bytes);
    OMX_AllocateBuffer(main_ctx.encoder, &o_buf1, 1, NULL, out_buf_bytes);
    OMX_AllocateBuffer(main_ctx.encoder, &o_buf2, 1, NULL, out_buf_bytes);

    OMX_SendCommand(main_ctx.encoder, OMX_CommandStateSet,
                    OMX_StateExecuting, NULL);

    get_input(&main_ctx, i_buf1);
    get_input(&main_ctx, i_buf2);

    OMX_EmptyThisBuffer(main_ctx.encoder, i_buf1);
    OMX_EmptyThisBuffer(main_ctx.encoder, i_buf2);
    OMX_FillThisBuffer(main_ctx.encoder, o_buf1);
    OMX_FillThisBuffer(main_ctx.encoder, o_buf2);

    while(!eof) sleep(1);

 bail:
    LIMDBG("done encoding.\n");
    OMX_Deinit();
    return 0;
}
