/*
  Copyright (C) 2011 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX IL parser metadata test code.

  REVISION:

    1) initial creation. ----------------------------------------- 2011-03-27 MG
*/

#include "stdio.h"
#include "stdlib.h"
#include "OMX_Core.h"
#include "OMX_Types.h"
#include "OMX_ContentPipe.h"
#include "OMX_Component.h"
#include "limlog.h"

#define AUDIO_OUT_BUF_BYTES (1024 * 8)
#define VIDEO_OUT_BUF_BYTES (1024 * 512)

static OMX_BUFFERHEADERTYPE * a_buf1;
static OMX_BUFFERHEADERTYPE * a_buf2;
static OMX_BUFFERHEADERTYPE * v_buf1;
static OMX_BUFFERHEADERTYPE * v_buf2;

//simple test code, no locking/sync mechanism implemented
static volatile int format_detected;
static volatile int error_detected;
static volatile int eos;

static void print_metadata (
    OMX_HANDLETYPE demuxer,
    OMX_METADATASCOPETYPE scope,
    OMX_U32 specifier)
{
    OMX_CONFIG_METADATAITEMCOUNTTYPE count;
    void * buf[(sizeof(OMX_CONFIG_METADATAITEMTYPE) + 128) / sizeof(void*)];
    OMX_CONFIG_METADATAITEMTYPE * item = (OMX_CONFIG_METADATAITEMTYPE *)buf;

    //FIXME: specify spec version.
    count.nSize = sizeof(OMX_CONFIG_METADATAITEMCOUNTTYPE);
    count.eScopeMode = scope;
    count.nScopeSpecifier = specifier;

    if (OMX_ErrorNone != OMX_GetConfig
        (demuxer, OMX_IndexConfigMetadataItemCount, &count))
    {
        printf("unable to get metadata item count.\n");
        return;
    }

    if (count.nMetadataItemCount < 1)
    {
        printf("no meta data available.\n");
        return;
    }

    item->nSize = sizeof(OMX_CONFIG_METADATAITEMTYPE);
    item->eScopeMode = scope;
    item->nScopeSpecifier = specifier;
    item->eSearchMode = OMX_MetadataSearchItemByIndex;
    item->nMetadataItemIndex = 0;
    item->nValueMaxSize = 128;
    while (OMX_ErrorNone == OMX_GetConfig
           (demuxer, OMX_IndexConfigMetadataItem, item))
    {
        printf("%s : %s\n", item->nKey, item->nValue);
        item->nMetadataItemIndex++;
    }
}

static void display_metadata (OMX_HANDLETYPE demuxer)
{
    printf("all level metadata\n");
    print_metadata(demuxer, OMX_MetadataScopeAllLevels, 0);

    printf("\n\ntop level metadata\n");
    print_metadata(demuxer, OMX_MetadataScopeTopLevel, 0);

    printf("\n\naudio port metadata\n");
    print_metadata(demuxer, OMX_MetadataScopePortLevel, 0);

    printf("\n\nvideo port metadata\n");
    print_metadata(demuxer, OMX_MetadataScopePortLevel, 1);
}

static OMX_ERRORTYPE demuxer_event_handler (
    OMX_HANDLETYPE hComponent,
    OMX_PTR pAppData,
    OMX_EVENTTYPE eEvent,
    OMX_U32 Data1,
    OMX_U32 Data2,
    OMX_PTR pEventData)
{
    if (eEvent == OMX_EventPortFormatDetected)
    {
        LIMDBG("port %u format detected.\n", Data1);
        OMX_SendCommand(hComponent,
                        OMX_CommandPortDisable, Data1, NULL);
        format_detected = 1;
    } else if (eEvent == OMX_EventBufferFlag)
    {
        printf("\nEOS detected\n");
        eos = 1;
    } else if (eEvent == OMX_EventError)
    {
        printf("\nError %x detected.\n", Data1);
        error_detected = 1;
    }
    return OMX_ErrorNone;
}

OMX_CALLBACKTYPE demuxer_callbacks = {
    .EventHandler    = demuxer_event_handler,
    .EmptyBufferDone = NULL,
    .FillBufferDone  = NULL,
};

//reset port buffer numbers to minimum for non-tunnel demo purpose.
static void reconfig_component_port (OMX_HANDLETYPE comp, int port, int bytes)
{
    OMX_PARAM_PORTDEFINITIONTYPE port_def;

    port_def.nSize = sizeof(OMX_PARAM_PORTDEFINITIONTYPE);
    port_def.nPortIndex = port;
    if (OMX_ErrorNone != OMX_GetParameter
        (comp, OMX_IndexParamPortDefinition, &port_def))
    {
        LIMERR("unable to get port %d definition.\n", port);
        exit(-1);
    } else
    {
        port_def.nBufferCountActual = 2;
        port_def.nBufferCountMin = 2;
        port_def.nBufferSize = bytes;
        OMX_SetParameter(comp, OMX_IndexParamPortDefinition, &port_def);
    }
}

static void my_log_print (const void * arg, int level,
                          const void * log, int bytes)
{
    printf("%s", (char *)log);
}

int main (int argc, char ** argv)
{
    OMX_HANDLETYPE demuxer;
    char demux_component_name[OMX_MAX_STRINGNAME_SIZE];
    OMX_U32 num_comps = 1;
    OMX_U8 * names[1];
    OMX_PARAM_CONTENTURITYPE * uri_param;
    int bytes;

    //lim_log_set_level(LIM_LOG_DEBUG);
    lim_log_set_listener(my_log_print, NULL);
    if (argc != 2)
    {
        printf("USAGE: metadata <file-name>\n");
        exit(0);
    } else
    {
        LIMDBG("extracting metadata of %s\n", argv[1]);
    }

    OMX_Init();

    names[0] = (OMX_U8 *)demux_component_name;
    if (OMX_ErrorNone !=
        OMX_GetComponentsOfRole((OMX_STRING)"container_demuxer.all",
                                &num_comps, names))
    {
        LIMERR("unable to find demuxer component.\n");
        goto bail;
    }

    if (OMX_ErrorNone !=
        OMX_GetHandle(&demuxer, demux_component_name,
                      NULL, &demuxer_callbacks))
    {
        LIMERR("unable to get demuxer handle.\n");
        goto bail;
    }
    //disable clock port.
    OMX_SendCommand(demuxer,
                    OMX_CommandPortDisable, 2, NULL);

    bytes = strlen(argv[1]);
    uri_param = (OMX_PARAM_CONTENTURITYPE *)malloc
        (sizeof(OMX_PARAM_CONTENTURITYPE) + bytes);
    uri_param->nSize = sizeof(OMX_PARAM_CONTENTURITYPE) + bytes;
    strcpy((char *)uri_param->contentURI, argv[1]);
    OMX_SetParameter(demuxer, OMX_IndexParamContentURI, uri_param);
    free(uri_param);

    reconfig_component_port(demuxer, 0, AUDIO_OUT_BUF_BYTES);
    reconfig_component_port(demuxer, 1, VIDEO_OUT_BUF_BYTES);
    if (OMX_ErrorNone !=
        OMX_SendCommand(demuxer, OMX_CommandStateSet, OMX_StateIdle, NULL))
    {
        LIMERR("unable to set demuxer to idle.\n");
        goto bail;
    }

    if (OMX_ErrorNone !=
        OMX_AllocateBuffer(demuxer,
                           &a_buf1, 0, NULL,
                           AUDIO_OUT_BUF_BYTES))
    {
        LIMERR("Unable to allocate audio output buffer\n");
        goto bail;
    }

    if (OMX_ErrorNone !=
        OMX_AllocateBuffer(demuxer,
                           &a_buf2, 0, NULL,
                           AUDIO_OUT_BUF_BYTES))
    {
        LIMERR("Unable to allocate audio output buffer2\n");
        goto bail;
    }

    if (OMX_ErrorNone !=
        OMX_AllocateBuffer(demuxer,
                           &v_buf1, 1, NULL,
                           VIDEO_OUT_BUF_BYTES))
    {
        LIMERR("Unable to allocate audio output buffer\n");
        goto bail;
    }

    if (OMX_ErrorNone !=
        OMX_AllocateBuffer(demuxer,
                           &v_buf2, 1, NULL,
                           VIDEO_OUT_BUF_BYTES))
    {
        LIMERR("Unable to allocate audio output buffer2\n");
        goto bail;
    }

    if (OMX_ErrorNone !=
        OMX_SendCommand(demuxer,
                        OMX_CommandStateSet, OMX_StateExecuting, NULL))
    {
        LIMERR("unable to execute demuxer.\n");
        goto bail;
    }

    OMX_FillThisBuffer(demuxer, a_buf1);
    OMX_FillThisBuffer(demuxer, a_buf2);
    OMX_FillThisBuffer(demuxer, v_buf1);
    OMX_FillThisBuffer(demuxer, v_buf2);


    while (!format_detected && !error_detected && !eos) sleep(1);

    if (format_detected)
    {
        LIMDBG("displaying metadata.\n");
        display_metadata(demuxer);
    }

bail:
    OMX_Deinit();
    return 0;
}
