/*
  Copyright (C) 2011 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX IL muxer test code.

  REVISION:

    1) initial creation. ----------------------------------------- 2011-01-05 MG
*/

#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include "OMX_Core.h"
#include "OMX_Types.h"
#include "OMX_ContentPipe.h"
#include "OMX_Component.h"
#include "limlog.h"

#define AUDIO_INPUT_BUF_BYTES 5888 /* roughly 33ms for 44.1KHz stero input. */
#define AUDIO_OUTPUT_BUF_BYTES (1024 * 32)
#define VIDEO_INPUT_BUF_BYTES (1024 * 512)
#define VIDEO_OUTPUT_BUF_BYTES (1024 * 64)

static volatile int eof;
static OMX_BUFFERHEADERTYPE * ia_buf1;
static OMX_BUFFERHEADERTYPE * ia_buf2;
static OMX_BUFFERHEADERTYPE * iv_buf1;
static OMX_BUFFERHEADERTYPE * iv_buf2;
static int a_in_buf_bytes;
static int v_in_buf_bytes;

typedef struct
{
    int a_codec; /* audio codec type. */
    int a_samplerate; /* audio samplerate. */
    int a_channels; /* audio channels. */
    int v_codec; /* video codec type. */
    int v_width; /* video input width. */
    int v_height; /* video input height. */
    int v_framerate; /* video input framerate. */
    char * pcm_in_fname;  /* PCM input file name. */
    char * yuv_in_fname;  /* YUV input file name. */
    char * out_fname; /* output file name. */
} options_t;

typedef struct
{
    options_t options;
    int pcm_in_fd;
    int yuv_in_fd;
    OMX_HANDLETYPE a_encoder;
    OMX_HANDLETYPE v_encoder;
    OMX_HANDLETYPE muxer;
} main_context_t;

enum {
    OPT_AUDIO_CODEC = 'a',
    OPT_SAMPLERATE = 's',
    OPT_CHANNELS = 'c',
    OPT_VIDEO_CODEC = 'v',
    OPT_WIDTH = 'w',
    OPT_HEIGHT = 'h',
    OPT_FRAME_RATE = 'r',
    OPT_PCM_IN_FNAME = 'i',
    OPT_YUV_IN_FNAME = 'y',
    OPT_OUT_FNAME = 'o',
    OPT_HELP = '?',
};

static struct option cmd_opts_long[] = {
    {"help", 0, 0, '?'},
    { NULL, 0, NULL, 0 }
};
#define cmd_opts_short "?a:s:c:v:w:h:r:i:y:o:"

static void print_help(void)
{
    printf("encode and mux audio video inputs:\n"
           "   muxer -a AUDIO_CODEC -s PCM_SAMPLERATE -c PCM_CHANNELS\n"
           "         -v VIDEO_CODEC -w WIDTH -h HEIGHT -r FRAMERATE\n"
           "         -i PCM_INPUT_FILE -y YUV_INPUT_FILE -o OUTPUT_FILE\n");
}

static int get_options (
    options_t * opts,
    int argc,
    char ** argv)
{
    int opt, index;

    while ((opt = getopt_long(argc, argv,
                              cmd_opts_short,
                              cmd_opts_long,
                              &index)) != -1)
    {
        switch (opt)
        {
        case OPT_AUDIO_CODEC :
            if (!strcmp(optarg, "mp3"))
                opts->a_codec = OMX_AUDIO_CodingMP3;
            else if (!strcmp(optarg, "aac"))
                opts->a_codec = OMX_AUDIO_CodingAAC;
            else
            {
                LIMERR("invalid audio codec: %s\n", optarg);
                goto bail;
            }
            break;

        case OPT_SAMPLERATE :
            opts->a_samplerate = atoi(optarg);
            break;

        case OPT_CHANNELS :
            opts->a_channels = atoi(optarg);
            break;

        case OPT_VIDEO_CODEC :
            if (!strcmp(optarg, "h263"))
                opts->v_codec = OMX_VIDEO_CodingH263;
            else if (!strcmp(optarg, "mpeg4"))
                opts->v_codec = OMX_VIDEO_CodingMPEG4;
            else
            {
                LIMERR("invalid video codec: %s\n", optarg);
                goto bail;
            }
            break;

        case OPT_WIDTH :
            opts->v_width = atoi(optarg);
            break;

        case OPT_HEIGHT :
            opts->v_height = atoi(optarg);
            break;

        case OPT_FRAME_RATE :
            opts->v_framerate = atoi(optarg);
            break;

        case OPT_PCM_IN_FNAME :
            opts->pcm_in_fname = optarg;
            break;

        case OPT_YUV_IN_FNAME :
            opts->yuv_in_fname = optarg;
            break;

        case OPT_OUT_FNAME :
            opts->out_fname = optarg;
            break;

        default :
            LIMERR("invalid command line option : %s\n", optarg);
            //run through

        case OPT_HELP :
            goto bail;
        }
    }

    if (NULL == opts->pcm_in_fname || NULL == opts->yuv_in_fname
        || NULL == opts->out_fname)
    {
        LIMERR("mandatory input output file name parameter missing.\n");
        goto bail;
    }

    if (opts->v_width < 32 || opts->v_height < 32)
    {
        LIMERR("invalid parameter w = %d or h = %d\n",
               opts->v_width, opts->v_height);
        goto bail;
    }

    return 0;

 bail:
    print_help();
    return -1;
}

static int get_input (main_context_t * ctx, OMX_BUFFERHEADERTYPE* buf, int av)
{
    int in_buf_bytes = av? v_in_buf_bytes : a_in_buf_bytes;

    if (in_buf_bytes != read(av? ctx->yuv_in_fd : ctx->pcm_in_fd, buf->pBuffer,
                             in_buf_bytes))
    {
        return -1;
    } else
    {
        LIMDBG("get %s bytes: %d\n", av? "video" : "audio", in_buf_bytes);
        buf->nOffset = 0;
        buf->nFilledLen = in_buf_bytes;
    }
    return 0;
}

static OMX_ERRORTYPE a_encoder_event_handler (
    OMX_HANDLETYPE hComponent,
    OMX_PTR pAppData,
    OMX_EVENTTYPE eEvent,
    OMX_U32 Data1,
    OMX_U32 Data2,
    OMX_PTR pEventData)
{
    if(eEvent == OMX_EventCmdComplete)
    {
        if (Data1 == OMX_CommandStateSet)
        {
            LIMDBG("audio encoder state set to %u\n", Data2);
        } else if (Data1 == OMX_CommandPortEnable)
        {
            LIMDBG("port %u enabled.\n", Data2);
        } else if (Data1 == OMX_CommandPortDisable)
        {
            LIMDBG("port %u disabled.\n", Data2);
        }
    } else if (eEvent == OMX_EventPortFormatDetected)
    {
        LIMDBG("port %u format detected.\n", Data1);
    } else if (eEvent == OMX_EventPortSettingsChanged)
    {
        LIMDBG("port %u settings changed.\n", Data1);
    } else if (eEvent == OMX_EventBufferFlag)
    {
        printf("\nEOS detected\n");
    } else if (eEvent == OMX_EventError)
    {
        LIMERR("error event detected.\n");
    } else
    {
        LIMDBG("param1 = %d param2 = %d\n", (int)Data1, (int)Data2);
    }
    return OMX_ErrorNone;
}

static OMX_ERRORTYPE a_encoder_empty_buffer_done (
    OMX_HANDLETYPE hComponent,
    OMX_PTR pAppData,
    OMX_BUFFERHEADERTYPE* pBuffer)
{
    main_context_t * ctx = (main_context_t *)pAppData;

    LIMDBG("audio encoder empty buffer done: %u.\n", pBuffer->nFilledLen);
    if (!get_input(ctx, pBuffer, 0))
    {
        OMX_EmptyThisBuffer(hComponent, pBuffer);
    } else
    {
        eof = 1;
    }

    return OMX_ErrorNone;
}

OMX_CALLBACKTYPE a_encoder_callbacks = {
    .EventHandler    = a_encoder_event_handler,
    .EmptyBufferDone = a_encoder_empty_buffer_done,
    .FillBufferDone  = NULL
};

static OMX_ERRORTYPE v_encoder_event_handler (
    OMX_HANDLETYPE hComponent,
    OMX_PTR pAppData,
    OMX_EVENTTYPE eEvent,
    OMX_U32 Data1,
    OMX_U32 Data2,
    OMX_PTR pEventData)
{
    if(eEvent == OMX_EventCmdComplete)
    {
        if (Data1 == OMX_CommandStateSet)
        {
            LIMDBG("video encoder state set to %u\n", Data2);
        } else if (Data1 == OMX_CommandPortEnable)
        {
            LIMDBG("port %u enabled.\n", Data2);
        } else if (Data1 == OMX_CommandPortDisable)
        {
            LIMDBG("port %u disabled.\n", Data2);
        }
    } else if (eEvent == OMX_EventPortFormatDetected)
    {
        LIMDBG("port %u format detected.\n", Data1);
    } else if (eEvent == OMX_EventPortSettingsChanged)
    {
        LIMDBG("port %u settings changed.\n", Data1);
    } else if (eEvent == OMX_EventBufferFlag)
    {
        printf("\nEOS detected\n");
    } else if (eEvent == OMX_EventError)
    {
        LIMERR("error event detected.\n");
    } else
    {
        LIMDBG("param1 = %d param2 = %d\n", (int)Data1, (int)Data2);
    }
    return OMX_ErrorNone;
}

static OMX_ERRORTYPE v_encoder_empty_buffer_done (
    OMX_HANDLETYPE hComponent,
    OMX_PTR pAppData,
    OMX_BUFFERHEADERTYPE* pBuffer)
{
    main_context_t * ctx = (main_context_t *)pAppData;

    LIMDBG("video encoder empty buffer done: %u.\n", pBuffer->nFilledLen);
    if (!get_input(ctx, pBuffer, 1))
    {
        OMX_EmptyThisBuffer(hComponent, pBuffer);
    } else
    {
        eof = 1;
    }

    return OMX_ErrorNone;
}

OMX_CALLBACKTYPE v_encoder_callbacks = {
    .EventHandler    = v_encoder_event_handler,
    .EmptyBufferDone = v_encoder_empty_buffer_done,
    .FillBufferDone  = NULL
};

static OMX_ERRORTYPE muxer_event_handler (
    OMX_HANDLETYPE hComponent,
    OMX_PTR pAppData,
    OMX_EVENTTYPE eEvent,
    OMX_U32 Data1,
    OMX_U32 Data2,
    OMX_PTR pEventData)
{
    if(eEvent == OMX_EventCmdComplete)
    {
        if (Data1 == OMX_CommandStateSet)
        {
            LIMDBG("muxer state set to %u\n", Data2);
        } else if (Data1 == OMX_CommandPortEnable)
        {
            LIMDBG("port %u enabled.\n", Data2);
        } else if (Data1 == OMX_CommandPortDisable)
        {
            LIMDBG("port %u disabled.\n", Data2);
        }
    } else if (eEvent == OMX_EventPortFormatDetected)
    {
        LIMDBG("port %u format detected.\n", Data1);
    } else if (eEvent == OMX_EventPortSettingsChanged)
    {
        LIMDBG("port %u settings changed.\n", Data1);
    } else if (eEvent == OMX_EventBufferFlag)
    {
        printf("\nEOS detected\n");
    } else if (eEvent == OMX_EventError)
    {
        LIMERR("error event detected.\n");
    } else
    {
        LIMDBG("param1 = %d param2 = %d\n", (int)Data1, (int)Data2);
    }
    return OMX_ErrorNone;
}

OMX_CALLBACKTYPE muxer_callbacks = {
    .EventHandler    = muxer_event_handler,
    .EmptyBufferDone = NULL,
    .FillBufferDone  = NULL
};

typedef struct {
    int codec;
    char * role;
} codec_role_map_t;
static codec_role_map_t CODEC_ROLE_MAP[] = {
    {OMX_AUDIO_CodingMP3, "audio_encoder.mp3"},
    {OMX_AUDIO_CodingAAC, "audio_encoder.aac"},
    {OMX_VIDEO_CodingMPEG4, "video_encoder.mpeg4"},
    {OMX_VIDEO_CodingH263, "video_encoder.h263"},
    {-1, NULL}};

static int get_component (main_context_t * ctx, int type)
{
    codec_role_map_t * role_map = &CODEC_ROLE_MAP[0];
    char component_name[OMX_MAX_STRINGNAME_SIZE];
    OMX_U32 num_comps = 1;
    OMX_U8 * names[1];
    int codec = -1;
    OMX_HANDLETYPE * handle = NULL;
    OMX_CALLBACKTYPE * callbacks = NULL;
    const char * role;

    switch (type)
    {
    case 0 :
        codec = ctx->options.a_codec;
        handle = &ctx->a_encoder;
        callbacks = &a_encoder_callbacks;
        break;

    case 1 :
        codec = ctx->options.v_codec;
        handle = &ctx->v_encoder;
        callbacks = &v_encoder_callbacks;
        break;

    case 2 :
        role = "container_muxer.all";
        handle = &ctx->muxer;
        callbacks = &muxer_callbacks;
        goto find_component;

    default:
        LIMERR("unexpected component type: %d\n", type);
        goto bail;
    }

    while (role_map->role)
    {
        if (role_map->codec == codec) break;
        role_map++;
    }

    if (NULL == role_map->role)
    {
        LIMERR("unsupported codec: %d\n", codec);
        goto bail;
    } else role = role_map->role;

 find_component:
    names[0] = (OMX_U8 *)component_name;
    if (OMX_ErrorNone !=
        OMX_GetComponentsOfRole((OMX_STRING)role, &num_comps, names))
    {
        LIMERR("unable to find encoder component for role: %s.\n", role);
        goto bail;
    }
    LIMDBG("encoder component %s found.\n", component_name);

    if (OMX_ErrorNone != OMX_GetHandle(handle, component_name, ctx, callbacks))
    {
        LIMERR("unable to get %s handle.\n", component_name);
        goto bail;
    }
    return 0;

 bail:
    return -1;
}

//reset port buffer numbers to minimum for non-tunnel demo purpose.
static void reconfig_component_port (OMX_HANDLETYPE comp, int port, int bytes)
{
    OMX_PARAM_PORTDEFINITIONTYPE port_def;

    port_def.nSize = sizeof(OMX_PARAM_PORTDEFINITIONTYPE);
    port_def.nPortIndex = port;
    if (OMX_ErrorNone != OMX_GetParameter
        (comp, OMX_IndexParamPortDefinition, &port_def))
    {
        LIMERR("unable to get port %d definition.\n", port);
        exit(-1);
    } else
    {
        port_def.nBufferCountActual = 2;
        port_def.nBufferCountMin = 2;
        port_def.nBufferSize = bytes;
        OMX_SetParameter(comp, OMX_IndexParamPortDefinition, &port_def);
    }
}

static int reconfig_audio_ports (main_context_t * ctx)
{
    OMX_PARAM_PORTDEFINITIONTYPE port_def;
    OMX_AUDIO_PORTDEFINITIONTYPE * audio = &port_def.format.audio;
    OMX_AUDIO_PARAM_PCMMODETYPE pcm;
    OMX_PARAM_COMPONENTROLETYPE role;

    /* set component role. */
    role.nSize = sizeof(OMX_PARAM_COMPONENTROLETYPE);
    strcpy((char *)role.cRole, (ctx->options.a_codec == OMX_AUDIO_CodingMP3)?
           "audio_encoder.mp3" : "audio_encoder.aac");
    if (OMX_ErrorNone != OMX_SetParameter
        (ctx->a_encoder, OMX_IndexParamStandardComponentRole, &role))
    {
        LIMERR("unable to set encoder role.\n");
        return -1;
    }

    reconfig_component_port(ctx->a_encoder, 0, AUDIO_INPUT_BUF_BYTES);
    reconfig_component_port(ctx->a_encoder, 1, AUDIO_OUTPUT_BUF_BYTES);

    pcm.nSize = sizeof(OMX_AUDIO_PARAM_PCMMODETYPE);
    pcm.nPortIndex = 0;
    if (OMX_ErrorNone != OMX_GetParameter
        (ctx->a_encoder, OMX_IndexParamAudioPcm, &pcm))
    {
        LIMWARN("unable to get encoder PCM parameter.\n");
    }
    pcm.nChannels = ctx->options.a_channels;
    pcm.eNumData = OMX_NumericalDataSigned;
    pcm.eEndian = OMX_EndianMax;
    pcm.bInterleaved = OMX_TRUE;
    pcm.nBitPerSample = 16;
    pcm.nSamplingRate = ctx->options.a_samplerate;
#if 0
    pcm.ePCMMODE = OMX_AUDIO_PCMModeLinear;
    pcm.eChannelMapping[0] = OMX_AUDIO_ChannelLF;
    pcm.eChannelMapping[1] = OMX_AUDIO_ChannelRF;
#endif
    if (OMX_ErrorNone != OMX_SetParameter
        (ctx->a_encoder, OMX_IndexParamAudioPcm, &pcm))
    {
        LIMERR("unable to set encoder PCM parameter.\n");
        return -1;
    }

    if (ctx->options.a_codec == OMX_AUDIO_CodingMP3)
    {
        OMX_AUDIO_PARAM_MP3TYPE mp3;

        mp3.nSize = sizeof(OMX_AUDIO_PARAM_MP3TYPE);
        mp3.nPortIndex = 1;
        if (OMX_ErrorNone != OMX_GetParameter
            (ctx->a_encoder, OMX_IndexParamAudioMp3, &mp3))
        {
            LIMWARN("unable to get encoder MP3 parameter.\n");
        }

        mp3.nChannels = ctx->options.a_channels;
        //mp3.nBitRate = ctx->options.a_bitrate;
        mp3.nSampleRate = ctx->options.a_samplerate;

        if (OMX_ErrorNone != OMX_SetParameter
            (ctx->a_encoder, OMX_IndexParamAudioMp3, &mp3))
        {
            LIMERR("unable to set encoder MP3 parameter.\n");
            return -1;
        }

        port_def.nSize = sizeof(OMX_PARAM_PORTDEFINITIONTYPE);
        port_def.nPortIndex = 1;
        OMX_GetParameter(ctx->a_encoder, OMX_IndexParamPortDefinition, &port_def);
        audio->eEncoding = OMX_AUDIO_CodingMP3;
        OMX_SetParameter(ctx->a_encoder, OMX_IndexParamPortDefinition, &port_def);

        port_def.nSize = sizeof(OMX_PARAM_PORTDEFINITIONTYPE);
        port_def.nPortIndex = 0;
        OMX_GetParameter(ctx->muxer, OMX_IndexParamPortDefinition, &port_def);
        audio->eEncoding = OMX_AUDIO_CodingMP3;
        OMX_SetParameter(ctx->muxer, OMX_IndexParamPortDefinition, &port_def);
    } else
    {
        //TODO: AAC
    }

    return 0;
}

static int reconfig_video_ports (main_context_t * ctx)
{
    OMX_PARAM_PORTDEFINITIONTYPE port_def;
    OMX_VIDEO_PORTDEFINITIONTYPE * video = &port_def.format.video;
    OMX_PARAM_COMPONENTROLETYPE role;

    /* set component role. */
    role.nSize = sizeof(OMX_PARAM_COMPONENTROLETYPE);
    strcpy((char *)role.cRole, (ctx->options.v_codec == OMX_VIDEO_CodingH263)?
           "video_encoder.h263" : "video_encoder.mpeg4");
    if (OMX_ErrorNone != OMX_SetParameter
        (ctx->v_encoder, OMX_IndexParamStandardComponentRole, &role))
    {
        LIMERR("unable to set encoder role.\n");
        return -1;
    }

    reconfig_component_port(ctx->v_encoder, 0, VIDEO_INPUT_BUF_BYTES);
    reconfig_component_port(ctx->v_encoder, 1, VIDEO_OUTPUT_BUF_BYTES);

    // configure input parameters
    port_def.nSize = sizeof(OMX_PARAM_PORTDEFINITIONTYPE);
    port_def.nPortIndex = 0;
    OMX_GetParameter(ctx->v_encoder, OMX_IndexParamPortDefinition, &port_def);

    video->nFrameWidth = ctx->options.v_width;
    video->nFrameHeight = ctx->options.v_height;
    video->nStride = ctx->options.v_width;
    video->xFramerate = ctx->options.v_framerate << 16;
    // default values, can optionally ignore
    //video->eColorFormat = OMX_COLOR_FormatYUV420Planar;
    //video->eCompressionFormat = OMX_VIDEO_CodingUnused;
    OMX_SetParameter(ctx->v_encoder, OMX_IndexParamPortDefinition, &port_def);

    // configure output parameters
    port_def.nSize = sizeof(OMX_PARAM_PORTDEFINITIONTYPE);
    port_def.nPortIndex = 1;
    OMX_GetParameter(ctx->v_encoder, OMX_IndexParamPortDefinition, &port_def);

    video->nFrameWidth = ctx->options.v_width;
    video->nFrameHeight = ctx->options.v_height;
    video->nBitrate = 1000000;
    video->xFramerate = ctx->options.v_framerate << 16;
    video->eCompressionFormat = ctx->options.v_codec;
    OMX_SetParameter(ctx->v_encoder, OMX_IndexParamPortDefinition, &port_def);

    port_def.nSize = sizeof(OMX_PARAM_PORTDEFINITIONTYPE);
    port_def.nPortIndex = 1;
    OMX_GetParameter(ctx->muxer, OMX_IndexParamPortDefinition, &port_def);
    video->nFrameWidth = ctx->options.v_width;
    video->nFrameHeight = ctx->options.v_height;
    video->nBitrate = 1000000;
    video->xFramerate = ctx->options.v_framerate << 16;
    video->eCompressionFormat = ctx->options.v_codec;
    OMX_SetParameter(ctx->muxer, OMX_IndexParamPortDefinition, &port_def);

    return 0;
}

static void my_log_print (const void * arg, int level,
                          const void * log, int bytes)
{
    printf("%s", (char *)log);
}

int main (int argc, char ** argv)
{
    OMX_PARAM_CONTENTURITYPE * uri_param;
    int bytes;
    main_context_t main_ctx = {
        .options.a_codec = OMX_AUDIO_CodingAAC,
        .options.a_samplerate = 44100,
        .options.a_channels = 2,
        .options.v_codec = OMX_VIDEO_CodingMPEG4,
        .options.v_width = 720,
        .options.v_height = 480,
        .options.v_framerate = 30,
        .options.pcm_in_fname = NULL,
        .options.yuv_in_fname = NULL,
        .options.out_fname = NULL,
        .pcm_in_fd = -1,
        .yuv_in_fd = -1,
    };
    lim_log_set_level(LIM_LOG_DEBUG);
    lim_log_set_level(LIM_LOG_INFO);
    lim_log_set_listener(my_log_print, NULL);
    if (get_options(&main_ctx.options, argc, argv))
    {
        exit(-1);
    }

    main_ctx.pcm_in_fd = open(main_ctx.options.pcm_in_fname, O_RDONLY, 0);
    if (main_ctx.pcm_in_fd < 0)
    {
        LIMERR("unable to open pcm input: %s\n", main_ctx.options.pcm_in_fname);
        exit(-1);
    }

    main_ctx.yuv_in_fd = open(main_ctx.options.yuv_in_fname, O_RDONLY, 0);
    if (main_ctx.yuv_in_fd < 0)
    {
        LIMERR("unable to open yuv input: %s\n", main_ctx.options.yuv_in_fname);
        close(main_ctx.pcm_in_fd);
        exit(-1);
    }

    OMX_Init();

    if (get_component(&main_ctx, 0) || get_component(&main_ctx, 1)
        || get_component(&main_ctx, 2))
    {
        exit(-1);
    }

    if (reconfig_audio_ports(&main_ctx)) goto bail;
    a_in_buf_bytes = AUDIO_INPUT_BUF_BYTES;

    if (reconfig_video_ports(&main_ctx)) goto bail;

    // hard-coded to YUV420 planner
    v_in_buf_bytes = main_ctx.options.v_width
        * main_ctx.options.v_height * 3 / 2;

    // content URI
    bytes = strlen(main_ctx.options.out_fname);
    uri_param = (OMX_PARAM_CONTENTURITYPE *)malloc
        (sizeof(OMX_PARAM_CONTENTURITYPE) + bytes);
    uri_param->nSize = sizeof(OMX_PARAM_CONTENTURITYPE) + bytes;
    strcpy((char *)uri_param->contentURI, main_ctx.options.out_fname);
    OMX_SetParameter(main_ctx.muxer, OMX_IndexParamContentURI, uri_param);
    free(uri_param);

    if (OMX_ErrorNone !=
        OMX_SetupTunnel(main_ctx.a_encoder, 1, main_ctx.muxer, 0))
    {
        LIMERR("unable to tunnel audio encoder and muxer component.\n");
        goto bail;
    }

    if (OMX_ErrorNone !=
        OMX_SetupTunnel(main_ctx.v_encoder, 1, main_ctx.muxer, 1))
    {
        LIMERR("unable to tunnel video encoder and muxer component.\n");
        goto bail;
    }

    // audio encoder setup
    OMX_SendCommand(main_ctx.a_encoder, OMX_CommandStateSet,
                    OMX_StateIdle, NULL);
    OMX_AllocateBuffer(main_ctx.a_encoder, &ia_buf1, 0, NULL, a_in_buf_bytes);
    OMX_AllocateBuffer(main_ctx.a_encoder, &ia_buf2, 0, NULL, a_in_buf_bytes);

    OMX_SendCommand(main_ctx.a_encoder, OMX_CommandStateSet,
                    OMX_StateExecuting, NULL);

    get_input(&main_ctx, ia_buf1, 0);
    get_input(&main_ctx, ia_buf2, 0);

    OMX_EmptyThisBuffer(main_ctx.a_encoder, ia_buf1);
    OMX_EmptyThisBuffer(main_ctx.a_encoder, ia_buf2);

    // video encoder setup
    OMX_SendCommand(main_ctx.v_encoder, OMX_CommandStateSet,
                    OMX_StateIdle, NULL);
    OMX_AllocateBuffer(main_ctx.v_encoder, &iv_buf1, 0, NULL, v_in_buf_bytes);
    OMX_AllocateBuffer(main_ctx.v_encoder, &iv_buf2, 0, NULL, v_in_buf_bytes);

    OMX_SendCommand(main_ctx.v_encoder, OMX_CommandStateSet,
                    OMX_StateExecuting, NULL);

    get_input(&main_ctx, iv_buf1, 1);
    get_input(&main_ctx, iv_buf2, 1);

    OMX_EmptyThisBuffer(main_ctx.v_encoder, iv_buf1);
    OMX_EmptyThisBuffer(main_ctx.v_encoder, iv_buf2);

    // muxer set up
    OMX_SendCommand(main_ctx.muxer, OMX_CommandStateSet,
                    OMX_StateIdle, NULL);
    OMX_SendCommand(main_ctx.muxer, OMX_CommandStateSet,
                    OMX_StateExecuting, NULL);

    while(!eof) sleep(1);

    OMX_SendCommand(main_ctx.muxer, OMX_CommandStateSet,
                    OMX_StateIdle, NULL);
    OMX_SendCommand(main_ctx.muxer, OMX_CommandStateSet,
                    OMX_StateLoaded, NULL);

    //TODO: wait for muxer component shutdown event.
    sleep(3);

 bail:
    LIMDBG("done encoding.\n");
    OMX_Deinit();
    return 0;
}
