/*
  Copyright (C) 2010 SWOAG Technology <www.swoag.com>

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA
  02110-1301  USA

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  OpenMAX IL video sink component test code.

  REVISION:

    1) initial creation. ----------------------------------------- 2010-02-23 MG
*/
#include <sys/stat.h>
#include <fcntl.h>
#include "stdio.h"
#include "stdlib.h"
#include <getopt.h>
#include "OMX_Core.h"
#include "OMX_Types.h"
#include "OMX_ContentPipe.h"
#include "OMX_Component.h"
#include "limlog.h"

#undef LIMDBG
#ifndef LIMDBG
#define LIMDBG(fmt, arg...)
#endif

typedef struct
{
    int w;
    int h;
    int framerate;
    char * fname;
    int color_pattern;
} options_t;

static options_t options = {
    .w = 720,
    .h = 480,
    .framerate = 30,
    .fname = NULL,
    .color_pattern = 1
};

static OMX_BUFFERHEADERTYPE * buf1;
static OMX_BUFFERHEADERTYPE * buf2;
static int done;
static int yuv_buf_bytes;

//return:  0 success
//         1 success, last entry
//        -1 fail, no more.
static int get_test_pattern (OMX_BUFFERHEADERTYPE * buf)
{
    static int count = 0;
    void * yuv = buf->pBuffer;
    int ret = 0;
    int ii;

    if (!options.color_pattern)
    {
        static int fd = -1;
        if (fd < 0)
        {
            fd = open(options.fname, O_RDONLY, 0);
            if (fd < 0)
            {
                LIMERR("unable to open file: %s.\n", options.fname);
                printf("displaying color pattern with %d x %d\n",
                       options.w, options.h);
                options.color_pattern = 1;
                goto color_pattern;
            }
        }
        if (yuv_buf_bytes != read(fd, yuv, yuv_buf_bytes))
        {
            close(fd);
            exit(0);
        } else
        {
            usleep(1000000/options.framerate);
        }
    }
 color_pattern:
    if (options.color_pattern)
    {
        switch (count++)
        {
        case 17 :
            ret = -1;
            break;

        case 16 :
            ret = 1;

        case 0 :
            memset(yuv, 0x80, yuv_buf_bytes);
            break;

        default :
            memset(yuv, 0x80 - (count << 2), options.w * options.h);
            yuv += options.w * options.h;
            for (ii = 0; ii < options.h / 2; ii++)
            {
                memset(yuv, 0x80 + ii / 4 + (count << 2), options.w / 2);
                yuv += options.w / 2;
            }
            for (ii = 0; ii < options.h / 2; ii++)
            {
                memset(yuv, 0x80 - ii / 4 + (count << 2), options.w / 2);
                yuv += options.w / 2;
            }
            break;
        }
        usleep(400000);
    }
    buf->nFilledLen = yuv_buf_bytes;
    LIMDBG("%d buffer filled with %u / %u bytes\n",
           count, buf->nFilledLen, buf->nAllocLen);
    return ret;
}

static OMX_ERRORTYPE video_sink_event_handler (
    OMX_HANDLETYPE hComponent,
    OMX_PTR pAppData,
    OMX_EVENTTYPE eEvent,
    OMX_U32 Data1,
    OMX_U32 Data2,
    OMX_PTR pEventData)
{
    if(eEvent == OMX_EventCmdComplete)
    {
        if (Data1 == OMX_CommandStateSet)
        {
            LIMDBG("video_sink state set to %u\n", Data2);
        } else if (Data1 == OMX_CommandPortEnable)
        {
            LIMDBG("port %u enabled.\n", Data2);
        } else if (Data1 == OMX_CommandPortDisable)
        {
            LIMDBG("port %u disabled.\n", Data2);
        }
    } else if (eEvent == OMX_EventPortFormatDetected)
    {
        LIMDBG("port %u format detected.\n", Data1);
    } else if (eEvent == OMX_EventPortSettingsChanged)
    {
          LIMDBG("port %u settings changed.\n", Data1);
    } else if (eEvent == OMX_EventBufferFlag)
    {
        printf("\nEOS detected\n");
        done = 1;
    } else if (eEvent == OMX_EventError)
    {
    } else
    {
        LIMDBG("param1 = %d param2 = %d\n", (int)Data1, (int)Data2);
    }
    return OMX_ErrorNone;
}

static OMX_ERRORTYPE video_sink_empty_buffer_done (
    OMX_HANDLETYPE hComponent,
    OMX_PTR pAppData,
    OMX_BUFFERHEADERTYPE* pBuffer)
{
    int ret = get_test_pattern(pBuffer);

    LIMDBG("empty buffer done.\n");
    if (ret > 0)
    {
        pBuffer->nFlags |= OMX_BUFFERFLAG_EOS ;
    }
    if (ret >= 0)
    {
        OMX_EmptyThisBuffer(hComponent, pBuffer);
    } else
    {
        OMX_FreeBuffer(hComponent, 0, pBuffer);
    }

    return OMX_ErrorNone;
}

static OMX_ERRORTYPE video_sink_fill_buffer_done (
    OMX_HANDLETYPE hComponent,
    OMX_PTR pAppData,
    OMX_BUFFERHEADERTYPE* pBuffer)
{
    return OMX_ErrorNone;
}

OMX_CALLBACKTYPE video_sink_callbacks = {
    .EventHandler    = video_sink_event_handler,
    .EmptyBufferDone = video_sink_empty_buffer_done,
    .FillBufferDone  = video_sink_fill_buffer_done
};

enum {
    OPT_WIDTH = 'w',
    OPT_HEIGHT = 'h',
    OPT_FRAME_RATE = 'r',
    OPT_HELP = '?',
};

static struct option cmd_opts_long[] = {
    {"help", 0, 0, '?'},
    { NULL,              0, NULL, 0 }
};
#define cmd_opts_short "?w:h:r:"

static void print_help(void)
{
    printf("video_sink [-w width -h height -r fps yuv-input-file-name]\n");
}

static int get_options (
    options_t * opts,
    int argc,
    char ** argv)
{
    int opt, index;

    while ((opt = getopt_long(argc, argv,
                              cmd_opts_short,
                              cmd_opts_long,
                              &index)) != -1)
    {
        switch (opt)
        {
        case OPT_WIDTH :
            opts->w = atoi(optarg);
            break;

        case OPT_HEIGHT :
            opts->h = atoi(optarg);
            break;

        case OPT_FRAME_RATE :
            opts->framerate = atoi(optarg);
            break;

        default :
            LIMERR("invalid command line option : %s\n", optarg);
            //run through

        case OPT_HELP :
            print_help();
            return -1;
        }
    }
    if (optind < argc)
    {
        opts->fname = argv[optind];
    }

    if (opts->w < 32 || opts->h < 32)
    {
        opts->w = 720;
        opts->h = 480;
        LIMWARN("invalid w or h, using default 720 x 480.\n");
    }

    if (opts->fname && opts->w && opts->h) options.color_pattern = 0;
    if (opts->w < 32 || opts->h < 32)
    {
        opts->w = 720;
        opts->h = 480;
        options.color_pattern = 1;
    }
    if (!opts->framerate) opts->framerate = 30;
    return 0;
}

int main (int argc, char ** argv)
{
    OMX_HANDLETYPE video_sink;
    char video_sink_component_name[OMX_MAX_STRINGNAME_SIZE];
    OMX_U32 num_comps = 1;
    OMX_U8 * names[1];
    OMX_PARAM_PORTDEFINITIONTYPE port;

    if (get_options(&options, argc, argv))
    {
        exit(-1);
    }

    if (options.color_pattern)
    {
        printf("displaying color pattern with %d x %d\n",
               options.w, options.h);
    } else
    {
        printf("displaying yuv [%d x %d] sequence @ %dfps from %s\n",
               options.w, options.h, options.framerate, options.fname);
    }

    OMX_Init();

    names[0] = (OMX_U8 *)video_sink_component_name;
    if (OMX_ErrorNone !=
        OMX_GetComponentsOfRole((OMX_STRING)"iv_renderer.yuv.overlay",
                                &num_comps, names))
    {
        LIMERR("unable to find video_sink component.\n");
        goto bail;
    }

    if (OMX_ErrorNone !=
        OMX_GetHandle(&video_sink, video_sink_component_name,
                      NULL, &video_sink_callbacks))
    {
        LIMERR("unable to get video_sink handle.\n");
        goto bail;
    }

    port.nSize = sizeof(OMX_PARAM_PORTDEFINITIONTYPE);
    port.nPortIndex = 0;
    OMX_GetParameter(video_sink, OMX_IndexParamPortDefinition, &port);

    port.format.video.nFrameWidth = options.w;
    port.format.video.nFrameHeight = options.h;
    port.format.video.eColorFormat = OMX_COLOR_FormatYUV420Planar;
    OMX_SetParameter(video_sink, OMX_IndexParamPortDefinition, &port);

    yuv_buf_bytes = options.w * options.h + options.w * options.h / 2;

    if (OMX_ErrorNone !=
        OMX_SendCommand(video_sink, OMX_CommandStateSet, OMX_StateIdle, NULL))
    {
        LIMERR("unable to set video_sink to idle.\n");
        goto bail;
    }

    if (OMX_ErrorNone !=
        OMX_AllocateBuffer(video_sink,
                           &buf1, 0, NULL,
                           yuv_buf_bytes))
    {
        LIMERR("Unable to allocate yuv buffer\n");
        goto bail;
    }

    if (OMX_ErrorNone !=
        OMX_AllocateBuffer(video_sink,
                           &buf2, 0, NULL,
                           yuv_buf_bytes))
    {
        LIMERR("Unable to allocate yuv buffer2\n");
        goto bail;
    }

    if (OMX_ErrorNone !=
        OMX_SendCommand(video_sink,
                        OMX_CommandStateSet, OMX_StateExecuting, NULL))
    {
        LIMERR("unable to execute video_sink.\n");
        goto bail;
    }
    get_test_pattern(buf1);
    OMX_EmptyThisBuffer(video_sink, buf1);
    get_test_pattern(buf2);
    OMX_EmptyThisBuffer(video_sink, buf2);

    while (!done) sleep(1);

bail:
    OMX_Deinit();
    return 0;
}
